/*
 * File:      check_brocade_fan.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int index;  /* sensor index */
    int v2;     /* found swSensorType and it is fan (2) */
    int v3;     /* found swSensorStatus and it is not absent (6) */
    int v4;     /* found swSensorValue */
    int v5;     /* found swSensorInfo */
    char* info; /* information about the sensor */
} tScanRes;

/* data type for storing data */
typedef struct
{
    int   type;  /* type of the sensor */
    int   stat;  /* state of the sensor */
    int   speed; /* current speed in rounds per minute */
} tDataSet;


/*
 * SNMP base OID
 * SW-MIB::swSensorType.1   = .1.3.6.1.4.1.1588.2.1.1.1.1.22.1.2.1
 * SW-MIB::swSensorStatus.1 = .1.3.6.1.4.1.1588.2.1.1.1.1.22.1.3.1
 * SW-MIB::swSensorValue.1  = .1.3.6.1.4.1.1588.2.1.1.1.1.22.1.4.1
 * SW-MIB::swSensorInfo.1   = .1.3.6.1.4.1.1588.2.1.1.1.1.22.1.5.1
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 1588, 2, 1, 1, 1, 1, 22, 1 };
size_t baseOIDLen = 14;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_brocade_fan";
    progVersion = "0.2";
    helpIndexP = "-i INDEX";
    helpThresP = NULL;
    helpPurpose =
            "Checks the status of a fan of a Brocade SAN Switch.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 5 * 60;
    rrdHBeat = 11 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = TRUE;
    hasAddInfParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-i, --index INTEGER\n"
    "    Index of the fan which should be checked\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin checks if the SAN switch considers the speed of the fan as\n"
    "within the nominal range. If the sensor is outside this range a\n"
    "critical will be generated. If the fan is unknown, faulty or\n"
    "absent, a warning will be generated.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<addInf>;<speed>\n"
    "where\n"
    "    addInf ... additional information.\n"
    "    speed  ... current speed in rounds per minute.\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'current speed'=<speed>\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    return FALSE;
}

int ScanResIndex ( int aIndex )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 16 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [14] )
            {
                case 2: /* swSensorType */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].index = aVar->name [15];
                    if ( ( aVar->type == ASN_INTEGER ) )
                    {
                        scanRes [index].v2 = ( *(aVar->val.integer) == 2 );
                    }
                    else
                    {
                        scanRes [index].v2 = FALSE;
                    }
                    scanRes [index].v3 = FALSE;
                    scanRes [index].v4 = FALSE;
                    scanRes [index].v5 = FALSE;
                    scanRes [index].info = NULL;
                break;
                case 3: /* swSensorStatus */
                    index = ScanResIndex ( aVar->name [15] );
                    if ( ( aVar->type == ASN_INTEGER ) )
                    {
                        scanRes [index].v3 = ( *(aVar->val.integer) != 6 );
                    } /* else nothing to do */
                break;
                case 4: /* swSensorValue */
                {
                    index = ScanResIndex ( aVar->name [15] );
                    scanRes [index].v4 = ( aVar->type == ASN_INTEGER );
                }
                break;
                case 5: /* swSensorInfo */
                    index = ScanResIndex ( aVar->name [15] );
                    scanRes [index].v5 = ( aVar->type == ASN_OCTET_STR );
                    scanRes [index].info = VarListStrDup ( aVar );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v2 && scanRes [i].v3 && scanRes [i].v4 &&
                scanRes [i].v5 )
        {
            retVal = STATE_OK;
            printf ( "OK index: %i ; info: %s\n",
                    scanRes [i].index, scanRes [i].info );
            free ( scanRes [i].info );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_type  [16];
    oid o_stat  [16];
    oid o_speed [16];
    size_t oidLen;

    OidCpy ( o_type,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_stat,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_speed, &oidLen, baseOID, &baseOIDLen );
    o_type  [14] = 2; o_type  [15] = idx;
    o_stat  [14] = 3; o_stat  [15] = idx;
    o_speed [14] = 4; o_speed [15] = idx;
    oidLen = 16;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.type  = ReadInteger ( sess2, o_type,  oidLen );
    v.stat  = ReadInteger ( sess2, o_stat,  oidLen );
    v.speed = ReadInteger ( sess2, o_speed, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "speed",  GAUGE, rrdHBeat, "0", "30000" );
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MIN_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MIN_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MIN_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    params [i++] = RRA_MIN_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    char buf[1024];

    chk = sscanf ( aData, "%1023s;%i", buf, &(v.speed) );
    if ( chk != 2 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
    addInf = StrDup ( buf );
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* speed  = IntToStr    ( v.speed );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "speed" );
    params [i++] = StrConcat ( 3, ts, ":", speed );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( speed );
    free ( ts );
}

void PlotDataFan ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[rpm]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, NULL, aTiming );
    PPAddPlot ( pp, "speed", "rpm", FALSE );
    title = StrConcat ( 4, hostName, " fan ", addInf, " speed" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 7000, FALSE );
    params [i++] = CreateDEF ( "speedM", aFN, "speed", MAX );
    params [i++] = CreateDEF ( "speedA", aFN, "speed", AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "speedM", C_GAUGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "speedA", C_GAUGE1_AVG, pp, 0, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "speedA", "speedA", "speedM", "speedM", "%4.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataFan ( fn, PLOT_24H_5M,  now );
    PlotDataFan ( fn, PLOT_7D_15M,  now );
    PlotDataFan ( fn, PLOT_31D_60M, now );
    PlotDataFan ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    int retVal = STATE_OK;
    char* stat = NULL;

    ReadVals ();
    if ( v.type == 2 )
    {
        switch ( v.stat )
        {
            case  1: retVal = STATE_WARNING;  stat = "'unknown'";       break;
            case  2: retVal = STATE_WARNING;  stat = "'faulty'";        break;
            case  3: retVal = STATE_CRITICAL; stat = "'below minimum'"; break;
            case  4: retVal = STATE_OK;       stat = "'nominal'";       break;
            case  5: retVal = STATE_CRITICAL; stat = "'above maximum'"; break;
            case  6: retVal = STATE_WARNING;  stat = "'absent'";        break;
            default: retVal = STATE_UNKNOWN;  stat = "invalid";         break;
        }
        PrintExit ( retVal );
        printf ( " - speed is %s, %i rpm", stat, v.speed );
        if ( npPerfData )
        {
            printf ( "|'current speed'=%i\n", v.speed );
        }
        else
        {
            printf ( "|%s;%i\n", addInf, v.speed );
        }
    }
    else
    {
        retVal = STATE_UNKNOWN;
        v.speed = -1;
        PrintExit ( retVal );
        printf ( " - not a fan\n" );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ();
        if ( plotPerfData ) PlotPerfData ();
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
