/*
 * File:      check_netapp_overall.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include "globals.h"
#include "nagiosif.h"
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int v7;  /* swOperStatus */
    int v20; /* swDiagResult */
} tScanRes;

/* data type for storing data */
typedef struct
{
    int oStat; /* operational status */
    int diag;  /* result of the POST diagnostics */
} tDataSet;

/*
 * SNMP base OID
 * SW-MIB::swOperStatus.0 = .1.3.6.1.4.1.1588.2.1.1.1.1.7.0
 * SW-MIB::swDiagResult.0 = .1.3.6.1.4.1.1588.2.1.1.1.1.20.0
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 1588, 2, 1, 1, 1, 1 };
size_t baseOIDLen = 12;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_brocade_overall";
    progVersion = "0.2";
    helpIndexP = NULL;
    helpThresP = NULL;
    helpPurpose = "Checks the overall status of a Brocade SAN Switch.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
    /* rrdStep is unused */
    /* rrdHBeat is unused */
    supportsPerfLog = FALSE;
    /* Ugly hack to keep those d*ned parts from timing out */
    snmpTimeOut *= 3;
}

void HelpSpecificParams ()
{
    /* No specific parameters */
}

void HelpExplain ()
{
    printf (
    "The plugin checks the overall status of a Brocade SAN Switch.\n"
    "\n"
    "If the operational status is online, then OK will be returned. If the\n"
    "operational status is offline or testing, then a warning will be\n"
    "returned. If the operational status is faulty or the POST diagnostics\n"
    "reports anything but 'OK', then a critical state is returned.\n"
    "\n"
    );
    printf (
    "Note: there is an ugly hack inside: The SNMP timeout is set higher\n"
    "because those devices tend to time out otherwise. And once they timed\n"
    "out they take some time before recovering.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "There is no performance data provided.\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    return FALSE;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 14 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            switch ( aVar->name [12] )
            {
                case 7: /* swOperStatus */
                    scanResLen = 1;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [0].v7 = ( aVar->type == ASN_INTEGER );
                    scanRes [0].v20 = FALSE;
                break;
                case 20: /* swDiagResult */
                    scanRes [0].v20 = ( aVar->type == ASN_INTEGER );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } // no else
    /* usually that would be:
     *     else Log ( LOG_CRIT, "Wrong length while scanning services" );
     * but in the very case we are comming across some tables we don't
     * care about which are nevertheless right.
     */
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if ( scanRes [i].v7  && scanRes [i].v20 )
        {
            retVal = STATE_OK;
            printf ( "OK\n" );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_oStat [14];
    oid o_diag  [14];
    size_t oidLen;

    OidCpy ( o_oStat, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_diag,  &oidLen, baseOID, &baseOIDLen );
    o_oStat [12] =  7; o_oStat [13] = 0;
    o_diag  [12] = 20; o_diag  [13] = 0;
    oidLen = 14;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.oStat = ReadInteger ( sess2, o_oStat, oidLen );
    v.diag  = ReadInteger ( sess2, o_diag,  oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

void CreateDataFile ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

void ParseData ( const char* aData )
{
    /* performance data is not supported */
}

void WriteData ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

int CheckData ()
{
    int retVal = STATE_UNKNOWN;
    char* oStat = NULL;
    char* diag = NULL;

    ReadVals ();
    switch ( v.oStat )
    {
        case  1: retVal = STATE_OK;       oStat = "'online'";  break;
        case  2: retVal = STATE_WARNING;  oStat = "'offline'"; break;
        case  3: retVal = STATE_WARNING;  oStat = "'testing'"; break;
        case  4: retVal = STATE_CRITICAL; oStat = "'faulty'";  break;
        default: retVal = STATE_UNKNOWN;  oStat = "invalid";   break;
    }
    switch ( v.diag )
    {
        case  1:
            retVal = MAXIMUM ( retVal, STATE_OK);
            diag = "'ok'";
        break;
        case  2:
            retVal = MAXIMUM ( retVal, STATE_CRITICAL);
            diag = "'central memory fault'";
        break;
        case  3:
            retVal = MAXIMUM ( retVal, STATE_CRITICAL);
            diag = "'embedded port fault'";
        break;
        default:
            retVal = MAXIMUM ( retVal, STATE_UNKNOWN );
            diag = "invalid";
        break;
    }
    PrintExit ( retVal );
    printf ( " - operational status: %s, post: %s\n", oStat, diag );
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
