/*
 * File:      check_brocade_port.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int index;   /* interface index */
    int v3;      /* found swFCPortPhyState */
    int v4;      /* found swFCPortOpStatus */
    int v5;      /* found swFCPortAdmStatus */
    int v11;     /* found swFCPortTxWords */
    int v12;     /* found swFCPortRxWords */
    int v13;     /* found swFCPortTxFrames */
    int v14;     /* found swFCPortRxFrames */
    int pStat;   /* physical state */
    int aStat;   /* administrative status */
} tScanRes;

/* data type for storing data */
typedef struct
{
    int  pStat;
    int  oStat;
    int  aStat;
    uint iWrds;
    uint oWrds;
    uint iFrms;
    uint oFrms;
} tDataSet;

/*
 * SNMP base OID
 * SW-MIB::swFCPortIndex.IDX     = .1.3.6.1.4.1.1588.2.1.1.1.6.2.1.1.IDX
 * SW-MIB::swFCPortPhyState.IDX  = .1.3.6.1.4.1.1588.2.1.1.1.6.2.1.3.IDX
 * SW-MIB::swFCPortOpStatus.IDX  = .1.3.6.1.4.1.1588.2.1.1.1.6.2.1.4.IDX
 * SW-MIB::swFCPortAdmStatus.IDX = .1.3.6.1.4.1.1588.2.1.1.1.6.2.1.5.IDX
 * SW-MIB::swFCPortTxWords.IDX   = .1.3.6.1.4.1.1588.2.1.1.1.6.2.1.11.IDX
 * SW-MIB::swFCPortRxWords.IDX   = .1.3.6.1.4.1.1588.2.1.1.1.6.2.1.12.IDX
 * SW-MIB::swFCPortTxFrames.IDX  = .1.3.6.1.4.1.1588.2.1.1.1.6.2.1.13.IDX
 * SW-MIB::swFCPortRxFrames.IDX  = .1.3.6.1.4.1.1588.2.1.1.1.6.2.1.14.IDX
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 1588, 2, 1, 1, 1, 6, 2, 1 };
size_t baseOIDLen = 14;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_brocade_port";
    progVersion = "0.2";
    helpIndexP = "-i INDEX";
    helpThresP = NULL;
    helpPurpose =
            "Checks the status of a FC interface of a Broacade SAN Switch.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 5 * 60;
    rrdHBeat = 11 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = TRUE;
    hasAddInfParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-i, --index INTEGER\n"
    "    Index of the Fibre Channel interface which should be checked.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "If the interface is up, then the plugin generates an OK. If it is\n"
    "administratively down, then the plugin generates a warning. If it is\n"
    "down and administratively enabled, then the plugin returns critical.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<addInf>;<pStat>;<oStat>;<iWrds>;<oWrds>;<iFrms>;<oFrms>\n"
    "where\n"
    );
    printf (
    "    addInf ... additional information.\n"
    "    pStat  ... physical status of the the interface, c.f. SW-MIB.txt,\n"
    "               swFCPortPhyState\n"
    "    oStat  ... operational state, c.f. SW-MIB.txt, swFCPortOpStatus\n"
    "    iWrds  ... Input Fibre Channel word counter\n"
    "    oWrds  ... Output Fibre Channel word counter\n"
    "    iFrms  ... Input Fibre Channel frames counter\n"
    "    oFrms  ... Output Fibre Channel frames counter\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'input words'=<iWrds>c "
    "'output words'=<oWrds>c "
    "'input frames'=<iFrms>c "
    "'output frames'=<oFrms>c\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    return FALSE;
}

int ScanResIndex ( int aIndex )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 16 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [14] )
            {
                case  1: /* swFCPortIndex */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].index = aVar->name [15];
                    scanRes [index].v3 = FALSE;
                    scanRes [index].v4 = FALSE;
                    scanRes [index].v5 = FALSE;
                    scanRes [index].v11 = FALSE;
                    scanRes [index].v12 = FALSE;
                    scanRes [index].v13 = FALSE;
                    scanRes [index].v14 = FALSE;
                    scanRes [index].pStat = -1;
                    scanRes [index].aStat = -1;
                break;
                case  3: /* swFCPortPhyState */
                    index = ScanResIndex ( aVar->name [15] );
                    scanRes [index].v3 = ( aVar->type == ASN_INTEGER );
                    scanRes [index].pStat = *(aVar->val.integer);
                break;
                case  4: /* swFCPortOpStatus */
                    index = ScanResIndex ( aVar->name [15] );
                    scanRes [index].v4 = ( aVar->type == ASN_INTEGER );
                break;
                case  5: /* swFCPortAdmStatus */
                    index = ScanResIndex ( aVar->name [15] );
                    scanRes [index].v5 = ( aVar->type == ASN_INTEGER );
                    scanRes [index].aStat = *(aVar->val.integer);
                break;
                case 11: /* swFCPortTxWords */
                    index = ScanResIndex ( aVar->name [15] );
                    scanRes [index].v11 = ( aVar->type == ASN_COUNTER );
                break;
                case 12: /* swFCPortRxWords */
                    index = ScanResIndex ( aVar->name [15] );
                    scanRes [index].v12 = ( aVar->type == ASN_COUNTER );
                break;
                case 13: /* swFCPortTxFrames */
                    index = ScanResIndex ( aVar->name [15] );
                    scanRes [index].v13 = ( aVar->type == ASN_COUNTER );
                break;
                case 14: /* swFCPortRxFrames */
                    index = ScanResIndex ( aVar->name [15] );
                    scanRes [index].v14 = ( aVar->type == ASN_COUNTER );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v3  && scanRes [i].v4  && scanRes [i].v5  &&
                scanRes [i].v11 && scanRes [i].v12 && scanRes [i].v13 &&
                scanRes [i].v14 )
        {
            retVal = STATE_OK;
            printf ( "OK index: %i ; pStat: %i ; aStat: %i\n",
                    scanRes [i].index, scanRes [i].pStat, scanRes [i].aStat );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_pStat [16];
    oid o_oStat [16];
    oid o_aStat [16];
    oid o_iWrds [16];
    oid o_oWrds [16];
    oid o_iFrms [16];
    oid o_oFrms [16];
    size_t oidLen;

    OidCpy ( o_pStat, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_oStat, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_aStat, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_iWrds, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_oWrds, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_iFrms, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_oFrms, &oidLen, baseOID, &baseOIDLen );
    o_pStat [14] =  3; o_pStat [15] = idx;
    o_oStat [14] =  4; o_oStat [15] = idx;
    o_aStat [14] =  5; o_aStat [15] = idx;
    o_iWrds [14] = 11; o_iWrds [15] = idx;
    o_oWrds [14] = 12; o_oWrds [15] = idx;
    o_iFrms [14] = 13; o_iFrms [15] = idx;
    o_oFrms [14] = 14; o_oFrms [15] = idx;
    oidLen = 16;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.pStat = ReadInteger   ( sess2, o_pStat, oidLen );
    v.oStat = ReadInteger   ( sess2, o_oStat, oidLen );
    v.aStat = ReadInteger   ( sess2, o_aStat, oidLen );
    v.iWrds = ReadCounter32 ( sess2, o_iWrds, oidLen );
    v.oWrds = ReadCounter32 ( sess2, o_oWrds, oidLen );
    v.iFrms = ReadCounter32 ( sess2, o_iFrms, oidLen );
    v.oFrms = ReadCounter32 ( sess2, o_oFrms, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "iWrds",  COUNTER, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "oWrds",  COUNTER, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "iFrms",  COUNTER, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "oFrms",  COUNTER, rrdHBeat, "0", "U" );
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    char buf[1024];

    chk = sscanf ( aData, "%1023s;%i;%i;%u;%u;%u;%u",
            buf, &(v.pStat), &(v.oStat),
            &(v.iWrds), &(v.oWrds), &(v.iFrms), &(v.oFrms) );
    if ( chk != 7 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
    addInf = StrDup ( buf );
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* iWrds  = UIntToStr   ( v.iWrds );
    char* oWrds  = UIntToStr   ( v.oWrds );
    char* iFrms  = UIntToStr   ( v.iFrms );
    char* oFrms  = UIntToStr   ( v.oFrms );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "iWrds:oWrds:iFrms:oFrms" );
    params [i++] = StrConcat ( 9, ts, ":", iWrds, ":", oWrds, ":", iFrms,
            ":", oFrms );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( oFrms );
    free ( iFrms );
    free ( oWrds );
    free ( iWrds );
    free ( ts );
}

void PlotDataWrds ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[FCW/s]  <-- out | in -->";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "wrds", aTiming );
    PPAddPlot ( pp, "input",  "FCW/s", FALSE );
    PPAddPlot ( pp, "output", "FCW/s", FALSE );
    title = StrConcat ( 4, hostName, " interface ", addInf,
            " transfer (fiber channel words)" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_AUTO, 0, 0, TRUE );
    params [i++] = CreateDEF ( "iWrdsMR", aFN, "iWrds", MAX );
    params [i++] = CreateDEF ( "iWrdsAR", aFN, "iWrds", AVG );
    params [i++] = CreateDEF ( "oWrdsMR", aFN, "oWrds", MAX );
    params [i++] = CreateDEF ( "oWrdsAR", aFN, "oWrds", AVG );
    params [i++] = StrDup ( "CDEF:iWrdsM=iWrdsMR" );
    params [i++] = StrDup ( "CDEF:iWrdsA=iWrdsAR" );
    params [i++] = StrDup ( "CDEF:oWrdsM=oWrdsMR,-1,*" );
    params [i++] = StrDup ( "CDEF:oWrdsA=oWrdsAR,-1,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "iWrdsM", C_THRU_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "iWrdsA", C_THRU_AVG, pp, 0, AVG, TRUE );
    params [i++] = CreatePlot ( PLOT_AREA,  "oWrdsM", C_THRU_MAX, pp, 1, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "oWrdsA", C_THRU_AVG, pp, 1, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "iWrdsAR", "iWrdsAR", "iWrdsMR", "iWrdsMR", "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "oWrdsAR", "oWrdsAR", "oWrdsMR", "oWrdsMR", "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataFrms ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[FCF/s]  <-- out | in -->";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "frms", aTiming );
    PPAddPlot ( pp, "input",  "FCF/s", FALSE );
    PPAddPlot ( pp, "output", "FCF/s", FALSE );
    title = StrConcat ( 4, hostName, " interface ", addInf,
            " transfer (fiber channel frames)" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_AUTO, 0, 0, TRUE );
    params [i++] = CreateDEF ( "iFrmsMR", aFN, "iFrms", MAX );
    params [i++] = CreateDEF ( "iFrmsAR", aFN, "iFrms", AVG );
    params [i++] = CreateDEF ( "oFrmsMR", aFN, "oFrms", MAX );
    params [i++] = CreateDEF ( "oFrmsAR", aFN, "oFrms", AVG );
    params [i++] = StrDup ( "CDEF:iFrmsM=iFrmsMR" );
    params [i++] = StrDup ( "CDEF:iFrmsA=iFrmsAR" );
    params [i++] = StrDup ( "CDEF:oFrmsM=oFrmsMR,-1,*" );
    params [i++] = StrDup ( "CDEF:oFrmsA=oFrmsAR,-1,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "iFrmsM", C_THRU_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "iFrmsA", C_THRU_AVG, pp, 0, AVG, TRUE );
    params [i++] = CreatePlot ( PLOT_AREA,  "oFrmsM", C_THRU_MAX, pp, 1, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "oFrmsA", C_THRU_AVG, pp, 1, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "iFrmsAR", "iFrmsAR", "iFrmsMR", "iFrmsMR", "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "oFrmsAR", "oFrmsAR", "oFrmsMR", "oFrmsMR", "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataWrds ( fn, PLOT_24H_5M,  now );
    PlotDataFrms ( fn, PLOT_24H_5M,  now );
    PlotDataWrds ( fn, PLOT_7D_15M,  now );
    PlotDataFrms ( fn, PLOT_7D_15M,  now );
    PlotDataWrds ( fn, PLOT_31D_60M, now );
    PlotDataFrms ( fn, PLOT_31D_60M, now );
    PlotDataWrds ( fn, PLOT_366D_4H, now );
    PlotDataFrms ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    int retVal = STATE_OK;

    ReadVals ();
    if ( v.aStat == 1 ) /* that's "online" */
    {
        if ( v.oStat == 1 ) /* that's "online" */
        {
            printf ( "OK - Port %i: I: %u kFCW, %u FCF; O: %u kFCW, %u FCF",
                    idx, v.iWrds / 1024, v.iFrms, v.oWrds / 1024, v.oFrms );
        }
        else
        {
            char* stat = NULL;

            switch ( v.oStat )
            {
                case  0: stat = "'unknown'"; break;
                case  1: /* done that */     break;
                case  2: stat = "'offline'"; break;
                case  3: stat = "'testing'"; break;
                case  4: stat = "'faulty'";  break;
                default: stat = "invalid";   break;
            }
            printf ( "CRITICAL - Port %i operational status is %s",
                    idx, stat );
            retVal = STATE_CRITICAL;
        }
    }
    else
    {
        printf ( "WARNING - Port %i is administratively down", idx );
        retVal = STATE_WARNING;
    }
    if ( npPerfData )
    {
        printf ( "|'input words'=%uc", v.iWrds );
        printf ( " 'output words'=%uc", v.oWrds );
        printf ( " 'input frames'=%uc", v.iFrms );
        printf ( " 'output frames'=%uc\n", v.oFrms );
    }
    else
    {
        printf ( "|%s;%i;%i;%u;%u;%u;%u\n",
                addInf, v.pStat, v.oStat,
                v.iWrds, v.oWrds, v.iFrms, v.oFrms );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ();
        if ( plotPerfData ) PlotPerfData ();
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
