/*
 * File:      check_brocade_psu.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include "globals.h"
#include "nagiosif.h"
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int index;  /* sensor index */
    int v2;     /* found swSensorType and it is power-supply (3) */
    int v3;     /* found swSensorStatus and it is not absent (6) */
    int v5;     /* found swSensorInfo */
    char* info; /* information about the sensor */
} tScanRes;

/* data type for storing data */
typedef struct
{
    int   type;  /* type of the sensor */
    int   stat;  /* state of the sensor */
} tDataSet;


/*
 * SNMP base OID
 * SW-MIB::swSensorType.1   = .1.3.6.1.4.1.1588.2.1.1.1.1.22.1.2.1
 * SW-MIB::swSensorStatus.1 = .1.3.6.1.4.1.1588.2.1.1.1.1.22.1.3.1
 * SW-MIB::swSensorInfo.1   = .1.3.6.1.4.1.1588.2.1.1.1.1.22.1.5.1
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 1588, 2, 1, 1, 1, 1, 22, 1 };
size_t baseOIDLen = 14;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_brocade_psu";
    progVersion = "0.2";
    helpIndexP = "-i INDEX";
    helpThresP = NULL;
    helpPurpose =
            "Checks the status of a power supply of a Brocade SAN Switch.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
    /* rrdStep is unused */
    /* rrdHBeat is unused */
    supportsPerfLog = FALSE;
    hasIndexParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-i, --index INTEGER\n"
    "    Index of the power supply which should be checked\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin checks if the SAN switch considers the power supply as\n"
    "within the nominal range. If the psu is unknown or absent a warning\n"
    "will be generated. Otherwise, If the psu is not nominal, a critical\n"
    "will be generated.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "There is no performance data provided.\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    /* no specific parameters except --index which is handled elsewhere */
    return FALSE;
}

int ScanResIndex ( int aIndex )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 16 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [14] )
            {
                case 2: /* swSensorType */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].index = aVar->name [15];
                    if ( ( aVar->type == ASN_INTEGER ) )
                    {
                        scanRes [index].v2 = ( *(aVar->val.integer) == 3 );
                    }
                    else
                    {
                        scanRes [index].v2 = FALSE;
                    }
                    scanRes [index].v3 = FALSE;
                    scanRes [index].v5 = FALSE;
                    scanRes [index].info = NULL;
                break;
                case 3: /* swSensorStatus */
                    index = ScanResIndex ( aVar->name [15] );
                    if ( ( aVar->type == ASN_INTEGER ) )
                    {
                        scanRes [index].v3 = ( *(aVar->val.integer) != 6 );
                    } /* else nothing to do */
                break;
                case 5: /* swSensorInfo */
                    index = ScanResIndex ( aVar->name [15] );
                    scanRes [index].v5 = ( aVar->type == ASN_OCTET_STR );
                    scanRes [index].info = VarListStrDup ( aVar );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v2 && scanRes [i].v3 && scanRes [i].v5 )
        {
            retVal = STATE_OK;
            printf ( "OK index: %i ; info: %s\n",
                    scanRes [i].index, scanRes [i].info );
            free ( scanRes [i].info );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_type  [16];
    oid o_stat  [16];
    size_t oidLen;

    OidCpy ( o_type,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_stat,  &oidLen, baseOID, &baseOIDLen );
    o_type  [14] = 2; o_type  [15] = idx;
    o_stat  [14] = 3; o_stat  [15] = idx;
    oidLen = 16;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.type  = ReadInteger ( sess2, o_type,  oidLen );
    v.stat  = ReadInteger ( sess2, o_stat,  oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

void CreateDataFile ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

void ParseData ( const char* aData )
{
    /* performance data is not supported */
}

void WriteData ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

int CheckData ()
{
    int retVal = STATE_OK;
    char* stat = NULL;

    ReadVals ();
    if ( v.type == 3 )
    {
        switch ( v.stat )
        {
            case  1: retVal = STATE_WARNING;  stat = "'unknown'";       break;
            case  2: retVal = STATE_WARNING;  stat = "'faulty'";        break;
            case  3: retVal = STATE_CRITICAL; stat = "'below minimum'"; break;
            case  4: retVal = STATE_OK;       stat = "'nominal'";       break;
            case  5: retVal = STATE_CRITICAL; stat = "'above maximum'"; break;
            case  6: retVal = STATE_WARNING;  stat = "'absent'";        break;
            default: retVal = STATE_UNKNOWN;  stat = "invalid";         break;
        }
        PrintExit ( retVal );
        printf ( " - power supply unit is %s\n", stat );
    }
    else
    {
        retVal = STATE_UNKNOWN;
        PrintExit ( retVal );
        printf ( " - not a power supply unit\n" );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
