/*
 * File:      check_cisco_cpuusage.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int index; /* cpu index */
    int v4;    /* found cpmCPUTotal1min */
    int v5;    /* found cpmCPUTotal5min */
    int v7;    /* found cpmCPUTotal1minRev */
    int v8;    /* found cpmCPUTotal5minRev */
} tScanRes;

/* data type for storing data */
typedef struct
{
    uint usg1; /* 1 minute average cpu usage */
    uint usg5; /* 5 minute average cpu usage */
} tDataSet;


/* global input variables */
uint warn1;
uint warn5;
uint crit1;
uint crit5;
int oldOIDs = FALSE;
int isWarnSet  = FALSE;
int isCritSet  = FALSE;

/*
 * SNMP base OID
 * CISCO-PROCESS-MIB::cpmCPUTotal1min.IDX
 * = 1.3.6.1.4.1.9.9.109.1.1.1.1.4.IDX
 * CISCO-PROCESS-MIB::cpmCPUTotal5min.IDX
 * = 1.3.6.1.4.1.9.9.109.1.1.1.1.5.IDX
 * CISCO-PROCESS-MIB::cpmCPUTotal1minRev.IDX
 * = 1.3.6.1.4.1.9.9.109.1.1.1.1.7.IDX
 * CISCO-PROCESS-MIB::cpmCPUTotal5minRev.IDX
 * = 1.3.6.1.4.1.9.9.109.1.1.1.1.8.IDX
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 9, 9, 109, 1, 1, 1, 1 };
size_t baseOIDLen = 13;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_cisco_cpuusage";
    progVersion = "0.5";
    helpIndexP = "-i INDEX -o";
    helpThresP = "-w WARN1,WARN5 -c CRIT1,CRIT5";
    helpPurpose = "Checks the average cpu usage of a Cisco(R) device.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 1 * 60;
    rrdHBeat = 3 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = TRUE;
    hasAddInfParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-w, --warning INTEGER,INTEGER\n"
    "    Average cpu usage at which a warning is generated.\n"
    "\n"
    "-c, --critical INTEGER,INTEGER\n"
    "    Average cpu usage at which a critical is generated.\n"
    "\n"
    "-i, --index INTEGER\n"
    "    Index of the cpu which should be checked.\n"
    "\n"
    "-o, --old-oids\n"
    "    Usually this plugin uses cpmCPUTotal1minRev and cpmCPUTotal5minRev.\n"
    "    These OIDs however are not supported on every device. This option\n"
    "    switches the plugin to use the old (deprecated) OIDs\n"
    "    cpmCPUTotal1min and cpmCPUTotal5min which do exactly the same.\n"
    "    Don't ask, I don't know.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "To generate a warning or critical only one of the cpu usage averages\n"
    "needs to be higher than the respective level.\n"
    "\n"
    "All values are expected as well as returned in per cent.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<addInf>;<usage1>;<usage5>\n"
    "where\n"
    "    addInf ... additional information.\n"
    "    usage1 ... percentage the cpu was busy during the last minute.\n"
    "    usage5 ... percentage the cpu was busy during the last 5 minutes.\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'1 minute cpu usage'=<usage1>%%;<warn1>;<crit1> "
    "'5 minute cpu usage'=<usage5>%%;<warn5>;<crit5>\n"
    "\n"
    );
}

void InvalidWarnParam ()
{
    Usage ( "Invalid 'warn' parameter" );
}

void InvalidCritParam ()
{
    Usage ( "Invalid 'crit' parameter" );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (            !strcmp ( argv [(*i)], "-w" )
                ||  !strcmp ( argv [(*i)], "--warning" ) )
    {
        isWarnSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];

            errno = 0;
            warn1 = strtoul ( argv [(*i)], &tail, 0 );
            if ( errno ) InvalidWarnParam ();
            if ( (*tail) == ',' )
            {
                char* w5str = &(tail [1]);

                tail = w5str;
                warn5 = strtoul ( w5str, &tail, 0 );
                if ( errno || *tail ) InvalidWarnParam ();
                (*i)++;
            }
            else InvalidWarnParam();
        }
        else Usage ( "Missing 'warn' parameter" );
    } else if ( !strcmp ( argv [(*i)], "-c" )
            ||  !strcmp ( argv [(*i)], "--critical" ) )
    {
        isCritSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];

            errno = 0;
            crit1 = strtoul ( argv [(*i)], &tail, 0 );
            if ( errno ) InvalidCritParam ();
            if ( *tail == ',' )
            {
                char* c5str = &(tail [1]);

                tail = c5str;
                crit5 = strtoul ( c5str, &tail, 0 );
                if ( errno || *tail ) InvalidCritParam ();
                (*i)++;
            }
            else InvalidCritParam ();
        }
        else Usage ( "Missing 'crit' parameter" );
    } else if ( !strcmp ( argv [(*i)], "-o" )
            ||  !strcmp ( argv [(*i)], "--old-oids" ) )
    {
        oldOIDs = TRUE;
        (*i)++;
    } else ret = FALSE;
    return ( ret );
}

int ScanResIndex ( int aIndex )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 15 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [13] )
            {
                case 2: /* cpmCPUTotalPhysicalIndex */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].index = aVar->name [14];
                    scanRes [index].v4 = FALSE;
                    scanRes [index].v5 = FALSE;
                    scanRes [index].v7 = FALSE;
                    scanRes [index].v8 = FALSE;
                break;
                case 4: /* cpmCPUTotal1min */
                    index = ScanResIndex ( aVar->name [14] );
                    scanRes [index].v4 = ( aVar->type == ASN_GAUGE );
                break;
                case 5: /* cpmCPUTotal5min */
                    index = ScanResIndex ( aVar->name [14] );
                    scanRes [index].v5 = ( aVar->type == ASN_GAUGE );
                break;
                case 7: /* cpmCPUTotal1minRev */
                    index = ScanResIndex ( aVar->name [14] );
                    scanRes [index].v7 = ( aVar->type == ASN_GAUGE );
                break;
                case 8: /* cpmCPUTotal5minRev */
                    index = ScanResIndex ( aVar->name [14] );
                    scanRes [index].v8 = ( aVar->type == ASN_GAUGE );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if      ( scanRes [i].v7 && scanRes [i].v8 )
        {
            retVal = STATE_OK;
            printf ( "OK index: %i newOIDs\n", scanRes [i].index );
        }
        else if ( scanRes [i].v4 && scanRes [i].v5 )
        {
            retVal = STATE_OK;
            printf ( "OK index: %i oldOIDs\n", scanRes [i].index );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_l1  [15];
    oid o_l5  [15];
    size_t oidLen;

    OidCpy ( o_l1, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_l5, &oidLen, baseOID, &baseOIDLen );
    if ( oldOIDs )
    {
        o_l1 [13] = 4; o_l1 [14] = idx;
        o_l5 [13] = 5; o_l5 [14] = idx;
    }
    else
    {
        o_l1 [13] = 7; o_l1 [14] = idx;
        o_l5 [13] = 8; o_l5 [14] = idx;
    }
    oidLen = 15;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.usg1  = ReadGauge32 ( sess2, o_l1,  oidLen );
    v.usg5  = ReadGauge32 ( sess2, o_l5,  oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "usg1",  GAUGE, rrdHBeat, "0", "100" );
    params [i++] = CreateDSStr ( "usg5",  GAUGE, rrdHBeat, "0", "100" );
    params [i++] = RRA_MAX_7H_1M;
    params [i++] = RRA_AVG_7H_1M;
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    char buf[1024];

    chk = sscanf ( aData, "%1023s;%u;%u", buf, &(v.usg1), &(v.usg5) );
    if ( chk != 3 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
    addInf = StrDup ( buf );
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* usg1   = UIntToStr   ( v.usg1 );
    char* usg5   = UIntToStr   ( v.usg5 );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "usg1:usg5" );
    params [i++] = StrConcat ( 5, ts, ":", usg1, ":", usg5 );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( usg5 );
    free ( usg1 );
    free ( ts );
}

void PlotDataUsages ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[%]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, NULL, aTiming );
    PPAddPlot ( pp, "1 minute average",  "%", FALSE );
    PPAddPlot ( pp, "5 minutes average", "%", FALSE );
    title = StrConcat ( 4, hostName, " cpu ", addInf, " usages" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 100, FALSE );
    params [i++] = CreateDEF ( "usg1M", aFN, "usg1", MAX );
    params [i++] = CreateDEF ( "usg1A", aFN, "usg1", AVG );
    params [i++] = CreateDEF ( "usg5M", aFN, "usg5", MAX );
    params [i++] = CreateDEF ( "usg5A", aFN, "usg5", AVG );
    params [i++] = CreatePlot ( PLOT_LINE3, "usg1M", C_LOAD1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE3, "usg5M", C_LOAD2_MAX, pp, 1, MAX, TRUE );
    params [i++] = CreatePlot ( PLOT_LINE2, "usg1A", C_LOAD1_AVG, pp, 0, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "usg5A", C_LOAD2_AVG, pp, 1, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "usg1A", "usg1A", "usg1M", "usg1M", "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "usg5A", "usg5A", "usg5M", "usg5M", "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataUsages ( fn, PLOT_6H_1M,   now );
    PlotDataUsages ( fn, PLOT_24H_5M,  now );
    PlotDataUsages ( fn, PLOT_7D_15M,  now );
    PlotDataUsages ( fn, PLOT_31D_60M, now );
    PlotDataUsages ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    int retVal = STATE_OK;

    ReadVals ();
    if ( ( v.usg1 > warn1 ) && ( retVal < STATE_WARNING ) )
            retVal = STATE_WARNING;
    if ( ( v.usg5 > warn5 ) && ( retVal < STATE_WARNING ) )
            retVal = STATE_WARNING;
    if ( ( v.usg1 > crit1 ) && ( retVal < STATE_CRITICAL ) )
            retVal = STATE_CRITICAL;
    if ( ( v.usg5 > crit5 ) && ( retVal < STATE_CRITICAL ) )
            retVal = STATE_CRITICAL;
    PrintExit ( retVal );
    printf ( " - cpu %i usage: %u%%, %u%%",
            idx, v.usg1, v.usg5 );
    if ( npPerfData )
    {
        printf ( "|'1 minute cpu usage'=%u%%;%u;%u", v.usg1, warn1, crit1 );
        printf ( " '5 minute cpu usage'=%u%%;%u;%u\n", v.usg5, warn5, crit5 );
    }
    else
    {
        printf ( "|%s;%u;%u\n", addInf, v.usg1, v.usg5 );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( checkData )
    {
        if ( !isWarnSet ) Usage ( "Missing 'warn' parameters" );
        if ( !isCritSet ) Usage ( "Missing 'crit' parameters" );
    }
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ( NULL );
        if ( plotPerfData ) PlotPerfData ( NULL );
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
