/*
 * File:      check_cisco_fan.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include "globals.h"
#include "nagiosif.h"
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int index;   /* fan index */
    int v3;      /* found ciscoEnvMonFanState and is not "not present" */
    char* descr; /* description of the fan */
} tScanRes;

/* data type for storing data */
typedef struct
{
    char* descr; /* name of the fan */
    int   state; /* state of the fan */
} tDataSet;

/*
 * SNMP base OID
 * CISCO-ENVMON-MIB::ciscoEnvMonFanStatusDescr.INDEX
 * = 1.3.6.1.4.1.9.9.13.1.4.1.2.INDEX
 * CISCO-ENVMON-MIB::ciscoEnvMonFanState.INDEX
 * = 1.3.6.1.4.1.9.9.13.1.4.1.3.INDEX
 */
oid baseOID [] = { 1, 3, 6, 1, 4, 1, 9, 9, 13, 1, 4, 1 };
size_t baseOIDLen = 12;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_cisco_fan";
    progVersion = "0.4";
    helpIndexP = "-i INDEX";
    helpThresP = NULL;
    helpPurpose = "Checks the status of a fan of a Cisco(R) device.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
    /* rrdStep is unused */
    /* rrdHBeat is unused */
    supportsPerfLog = FALSE;
    hasIndexParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-i, --index INTEGER\n"
    "    Index of the fan which should be checked.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin returns the overall state of the fan.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "There is no performance data provided.\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    /* no specific parameters except --index which is handled elsewhere */
    return FALSE;
}

int ScanResIndex ( int aIndex )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 14 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [12] )
            {
                case 2: /* ciscoEnvMonFanStatusDescr */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].index = aVar->name [13];
                    scanRes [index].v3 = FALSE;
                    scanRes [index].descr = VarListStrDup ( aVar );
                break;
                case 3: /* ciscoEnvMonFanState */
                {
                    int typeOK = ( aVar->type == ASN_INTEGER );
                    int valOK = ( *(aVar->val.integer) != 5 );

                    index = ScanResIndex ( aVar->name [13] );
                    scanRes [index].v3 = ( typeOK && valOK );
                }
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if ( scanRes [i].v3 )
        {
            retVal = STATE_OK;
            printf ( "OK index: %i ; descr: %s\n",
                    scanRes [i].index, scanRes [i].descr );
            free ( scanRes [i].descr );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_descr [14];
    oid o_state [14];
    size_t oidLen;

    OidCpy ( o_descr, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_state, &oidLen, baseOID, &baseOIDLen );
    o_descr[12] = 2; o_descr[13] = idx;
    o_state[12] = 3; o_state[13] = idx;
    oidLen = 14;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.descr = ReadString  ( sess2, o_descr, oidLen );
    v.state = ReadInteger ( sess2, o_state, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

void CreateDataFile ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

void ParseData ( const char* aData )
{
    /* performance data is not supported */
}

void WriteData ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

int CheckData ()
{
    int retVal = STATE_UNKNOWN;

    ReadVals ();
    switch ( v.state )
    {
        case 1:
            retVal = STATE_OK;
            printf ( "OK - %s fan has state 'normal'\n", v.descr );
        break;
        case 2:
            retVal = STATE_WARNING;
            printf ( "WARNING - %s fan has state 'warning'\n", v.descr );
        break;
        case 3:
            retVal = STATE_CRITICAL;
            printf ( "CRITICAL - %s fan has state 'critical'\n", v.descr );
        break;
        case 4:
            retVal = STATE_WARNING;
            printf ( "WARNING - %s fan has state 'shutdown'\n", v.descr );
        break;
        case 5:
            retVal = STATE_WARNING;
            printf ( "WARNING - %s fan has state 'not present'\n", v.descr );
        break;
        case 6:
            retVal = STATE_CRITICAL;
            printf ( "CRITICAL - %s fan has state 'not functioning'\n",
                    v.descr );
        break;
        default:
            retVal = STATE_UNKNOWN;
            printf ( "UNKNOWN - %s fan has an invalid state\n", v.descr );
        break;
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
        if ( checkData ) free ( v.descr );
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
