/*
 * File:      check_cisco_if_load.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

/*
 * This is not a good example. It uses just one column and is therefor
 * somewhat different (and simplier)! There is but one feature you might
 * want to check: How to get a value from an entirely different table which
 * uses the same indices.
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include "errno.h"
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"

/* data type for storing scan results */
typedef struct
{
    int index;   /* interface index */
    char* descr; /* interface description */
} tScanRes;

/* data type for storing data */
typedef struct
{
    int  load;
} tDataSet;

/* global input variables */
int warn;
int crit;
int isWarnSet  = FALSE;
int isCritSet  = FALSE;

/*
 * SNMP base OID
 * OLD-CISCO-INTERFACES-MIB::locIfLoad.IDX  = .1.3.6.1.4.1.9.2.2.1.1.24.IDX
 *
 * This OID is needed to get the interface names:
 * IF-MIB::ifDescr.IDX                      = .1.3.6.1.2.1.2.2.1.2.IDX
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 9, 2, 2, 1, 1, 24 };
size_t baseOIDLen = 12;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_cisco_if_load";
    progVersion = "0.1";
    helpIndexP = "-i INDEX -w WARN -c CRIT";
    helpThresP = NULL;
    helpPurpose = "Checks the load of a network interface.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 1 * 60;
    rrdHBeat = 3 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = TRUE;
    hasAddInfParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-i, --index INTEGER\n"
    "    Index of the network interface which should be checked.\n"
    "\n"
    "-w, --warning INTEGER\n"
    "    Interface load at which a warning is generated (255 = 100%%)\n"
    "\n"
    "-c, --critical INTEGER\n"
    "    Interface load at which a critical is generated (255 = 100%%)\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "Generates a warning/critical if the interface load is above the\n"
    " respective level.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<addInf>;<load>\n"
    "where\n"
    "    addInf ... additional information.\n"
    "    load   ... interface load (0 to 255)\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'load'=<load>;<warn>;<crit>;0;255\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-w" )
            ||  !strcmp ( argv [(*i)], "--warning" ) )
    {
        isWarnSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];

            errno = 0;
            warn = strtol ( argv [(*i)++], &tail, 0 );
            if ( errno || (*tail) ) Usage ( "Invalid 'warn' parameter" );
        }
        else Usage ( "Missing 'warn' parameter(s)" );
    } else if ( !strcmp ( argv [(*i)], "-c" )
            ||  !strcmp ( argv [(*i)], "--critical" ) )
    {
        isCritSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];

            errno = 0;
            crit = strtol ( argv [(*i)++], &tail, 0 );
            if ( errno || (*tail) ) Usage ( "Invalid 'crit' parameter" );
        }
        else Usage ( "Missing 'crit' parameter(s)" );
    } else ret = FALSE;
    return ( ret );
}

char* ScanResGetName ( int aIdx )
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_descr [] = { 1, 3, 6, 1, 2, 1, 2, 2, 1, 2, aIdx };
    size_t oidLen = 11;
    char* ret;

    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    ret = ReadString ( sess2, o_descr, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    return ( ret );
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 13 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            index = scanResLen;
            scanResLen++;
            scanRes = realloc ( scanRes, sizeof ( tScanRes ) * scanResLen );
            if ( scanRes == NULL ) ErrorMalloc ();
            scanRes [index].index = aVar->name [12];
            scanRes [index].descr = ScanResGetName ( aVar->name [12] );
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        retVal = STATE_OK;
        printf ( "OK index: %i ; descr: %s\n",
                    scanRes [i].index, scanRes [i].descr );
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_load [13];
    size_t oidLen;

    OidCpy ( o_load, &oidLen, baseOID, &baseOIDLen );
    o_load [12] = idx;
    oidLen = 13;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.load  = ReadInteger   ( sess2, o_load, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "load",  GAUGE, rrdHBeat, "0", "U" );
    params [i++] = RRA_MAX_7H_1M;
    params [i++] = RRA_AVG_7H_1M;
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    char buf[1024];

    chk = sscanf ( aData, "%1023s;%u", buf, &(v.load) );
    if ( chk != 2 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
    addInf = StrDup ( buf );
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* load   = UIntToStr   ( v.load );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "load" );
    params [i++] = StrConcat ( 3, ts, ":", load );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( load );
    free ( ts );
}

void PlotDataLoad ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[%]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, NULL, aTiming );
    PPAddPlot ( pp, "interface load",  "%", FALSE );
    title = StrConcat ( 4, hostName, " interface ", addInf, " load" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 100, FALSE );
    params [i++] = CreateDEF ( "loadMR", aFN, "load", MAX );
    params [i++] = CreateDEF ( "loadAR", aFN, "load", AVG );
    params [i++] = StrDup ( "CDEF:loadM=loadMR,255,/,100,*" );
    params [i++] = StrDup ( "CDEF:loadA=loadAR,255,/,100,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "loadM", C_THRU_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "loadA", C_THRU_AVG, pp, 0, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "loadA", "loadA", "loadM", "loadM", "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataLoad ( fn, PLOT_6H_1M,   now );
    PlotDataLoad ( fn, PLOT_24H_5M,  now );
    PlotDataLoad ( fn, PLOT_7D_15M,  now );
    PlotDataLoad ( fn, PLOT_31D_60M, now );
    PlotDataLoad ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    int retVal = STATE_OK;

    ReadVals ();
    if ( warn < v.load ) retVal = MAXIMUM ( retVal, STATE_WARNING );
    if ( crit < v.load ) retVal = MAXIMUM ( retVal, STATE_CRITICAL );
    PrintExit ( retVal );
    printf ( " - interface load: %i", v.load );
    if ( npPerfData )
    {
        printf ( "|'interface load'=%i;%i;%i;0;255\n", v.load, warn, crit );
    }
    else
    {
        printf ( "|%s;%i\n", addInf, v.load );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( checkData )
    {
        if ( !isWarnSet ) Usage ( "Missing 'warn' parameters" );
        if ( !isCritSet ) Usage ( "Missing 'crit' parameters" );
    }
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ();
        if ( plotPerfData ) PlotPerfData ();
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
