/*
 * File:      check_cisco_mem.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int index;  /* memory pool index */
    int v3;     /* found ciscoMemoryPoolAlternate */
    int v5;     /* found ciscoMemoryPoolUsed */
    int v6;     /* found ciscoMemoryPoolFree */
    int v7;     /* found ciscoMemoryPoolLargestFree */
    char* name; /* memory pool name */
    int alter;  /* alternate pool */
} tScanRes;

/* data type for storing data */
typedef struct
{
    char* name; /* name of the memory pool */
    uint  used; /* used bytes */
    uint  free; /* unused bytes */
    uint  lFree; /* number of bytes in the largest unused contiguous block */
} tDataSet;

/* global input variables */
uint  warnS = 75;
char  warnU = '%';
uint  critS = 85;
char  critU = '%';

/*
 * SNMP base OID
 * CISCO-MEMORY-POOL-MIB::ciscoMemoryPoolName.IDX
 * = 1.3.6.1.4.1.9.9.48.1.1.1.2.IDX
 * CISCO-MEMORY-POOL-MIB::ciscoMemoryPoolAlternate.IDX
 * = 1.3.6.1.4.1.9.9.48.1.1.1.3.IDX
 * CISCO-MEMORY-POOL-MIB::ciscoMemoryPoolUsed.IDX
 * = 1.3.6.1.4.1.9.9.48.1.1.1.5.IDX
 * CISCO-MEMORY-POOL-MIB::ciscoMemoryPoolFree.IDX
 * = 1.3.6.1.4.1.9.9.48.1.1.1.6.IDX
 * CISCO-MEMORY-POOL-MIB::ciscoMemoryPoolLargestFree.IDX
 * = 1.3.6.1.4.1.9.9.48.1.1.1.7.IDX
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 9, 9, 48, 1, 1, 1 };
size_t baseOIDLen = 12;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_cisco_mem";
    progVersion = "0.4";
    helpIndexP = "-i INDEX";
    helpThresP = "[-w WARN] [-c CRIT]";
    helpPurpose = "Checks the usage of a memory pool of a Cisco(R) device.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 5 * 60;
    rrdHBeat = 11 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = TRUE;
    hasAddInfParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-w, --warning INTEGER[%%kM]\n"
    "    Memory usage at which a warning is generated. Default = %i%c.\n",
    warnS, warnU
    );
    printf (
    "\n"
    "-c, --critical INTEGER[%%kM]\n"
    "    Memory usage at which a critical is generated. Default = %i%c.\n",
    critS, critU
    );
    printf (
    "\n"
    "-i, --index INTEGER\n"
    "    Index of the memory pool which should be checked\n"
    "\n"
    );
}

void HelpExplain ()
{
    /* Haven't got anything to say ... */
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<addInf>;<used>;<size>;<lFree>\n"
    "where\n"
    "    addInf ... additional information.\n"
    "    used   ... used bytes in the pool.\n"
    "    free   ... unused bytes in the pool.\n"
    "    lFree  ... size of the largest contiguous block in the pool in\n"
    "               bytes.\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'memory'=<used>B;<warn>;<crit>;0;<size> "
    "'largest free block'=<lFree>B\n"
    "\n"
    );
}

char CheckSpaceUnit ( char* aUnit )
{
    char ret = '1';
    if      ( !strcmp ( aUnit, ""  ) ) ret = '1';
    else if ( !strcmp ( aUnit, "k" ) ) ret = 'k';
    else if ( !strcmp ( aUnit, "M" ) ) ret = 'M';
    else if ( !strcmp ( aUnit, "%" ) ) ret = '%';
    else Usage ( "invalid space unit" );
    return ( ret );
}

uint SpaceToB ( char* aWhat, uint aLevel, char aUnit, uint aSize )
{
    uint retVal = '0';
    double level = (double) aLevel;
    double size = (double) aSize;

    switch ( aUnit )
    {
        case '1': retVal = aLevel;               break;
        case 'k': retVal = aLevel * 1024;        break;
        case 'M': retVal = aLevel * 1048576;     break;
        case '%': retVal = level * size / 100.0; break;
        default:
        {
            Error ( LOG_ERR, "An invalid memory level slipped through!" );
        }
        break;
    }
    return ( retVal );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-w" )
            ||  !strcmp ( argv [(*i)], "--warning" ) )
    {
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            warnS = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno ) Usage ( "Invalid 'warn' parameter" );
            warnU = CheckSpaceUnit ( tail );
        } else Usage ( "Missing 'warn' parameter(s)" );
    } else if ( !strcmp ( argv [(*i)], "-c" )
            ||  !strcmp ( argv [(*i)], "--critical" ) )
    {
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            critS = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno ) Usage ( "Invalid 'crit' parameter" );
            critU = CheckSpaceUnit ( tail );
        } else Usage ( "Missing 'crit' parameter(s)" );
    } else ret = FALSE;
    return ( ret );
}

int ScanResIndex ( int aIndex )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 14 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [12] )
            {
                case 2: /* ciscoMemoryPoolName */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].index = aVar->name [13];
                    scanRes [index].v5 = FALSE;
                    scanRes [index].v6 = FALSE;
                    scanRes [index].v7 = FALSE;
                    scanRes [index].name = VarListStrDup ( aVar );
                break;
                case 3: /* ciscoMemoryPoolAlternate */
                    index = ScanResIndex ( aVar->name [13] );
                    scanRes [index].v3 = ( aVar->type == ASN_INTEGER );
                    scanRes [index].alter = *(aVar->val.integer);
                break;
                case 5: /* ciscoMemoryPoolUsed */
                    index = ScanResIndex ( aVar->name [13] );
                    scanRes [index].v5 = ( aVar->type == ASN_GAUGE );
                break;
                case 6: /* ciscoMemoryPoolFree */
                    index = ScanResIndex ( aVar->name [13] );
                    scanRes [index].v6 = ( aVar->type == ASN_GAUGE );
                break;
                case 7: /* ciscoMemoryPoolLargestFree */
                    index = ScanResIndex ( aVar->name [13] );
                    scanRes [index].v7 = ( aVar->type == ASN_GAUGE );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if ( scanRes [i].v5 && scanRes [i].v6 && scanRes [i].v7 )
        {
            retVal = STATE_OK;
            printf ( "OK index: %i ; alternate: %i ; name: %s\n",
                    scanRes [i].index, scanRes[i].alter, scanRes [i].name );
            free ( scanRes [i].name );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    /* TODO: Handle alternate pools */
    oid o_name  [14];
    oid o_used  [14];
    oid o_free  [14];
    oid o_lFree [14];
    size_t oidLen;

    OidCpy ( o_name,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_used,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_free,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_lFree, &oidLen, baseOID, &baseOIDLen );
    o_name  [12] = 2; o_name  [13] = idx;
    o_used  [12] = 5; o_used  [13] = idx;
    o_free  [12] = 6; o_free  [13] = idx;
    o_lFree [12] = 7; o_lFree [13] = idx;
    oidLen = 14;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.name  = ReadString  ( sess2, o_name,  oidLen );
    v.used  = ReadGauge32 ( sess2, o_used,  oidLen );
    v.free  = ReadGauge32 ( sess2, o_free,  oidLen );
    v.lFree = ReadGauge32 ( sess2, o_lFree, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "used",  GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "size",  GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "lFree", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    char buf[1024];

    chk = sscanf ( aData, "%1023s;%u;%u;%u",
            buf, &(v.used), &(v.free), &(v.lFree) );
    if ( chk != 4 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
    addInf = StrDup ( buf );
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* used   = UIntToStr   ( v.used );
    char* size   = UIntToStr   ( v.used + v.free );
    char* lFree  = UIntToStr   ( v.lFree );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "used:size:lFree" );
    params [i++] = StrConcat ( 7, ts, ":", used, ":", size, ":", lFree );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( lFree );
    free ( size );
    free ( used );
}

void PlotDataUsages ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[%]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, NULL, aTiming );
    PPAddPlot ( pp, "used memory",        "%", FALSE );
    PPAddPlot ( pp, "largest free block", "%", FALSE );
    title = StrConcat ( 4, hostName, " memory pool ", addInf, " usage" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 100, FALSE );
    params [i++] = CreateDEF ( "sizeMR",  aFN, "size",  MAX );
    params [i++] = CreateDEF ( "sizeAR",  aFN, "size",  AVG );
    params [i++] = CreateDEF ( "usedMR",  aFN, "used",  MAX );
    params [i++] = CreateDEF ( "lFreeMR", aFN, "lFree", MAX );
    params [i++] = CreateDEF ( "usedAR",  aFN, "used",  AVG );
    params [i++] = CreateDEF ( "lFreeAR", aFN, "lFree", AVG );
    params [i++] = StrDup (   "CDEF:usedM=usedMR,sizeMR,/,100,*" );
    params [i++] = StrDup ( "CDEF:lFreeM=lFreeMR,sizeMR,/,100,*" );
    params [i++] = StrDup (   "CDEF:usedA=usedAR,sizeAR,/,100,*" );
    params [i++] = StrDup ( "CDEF:lFreeA=lFreeAR,sizeAR,/,100,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "usedM",  C_USAGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "lFreeM", C_FREE_MAX,   pp, 1, MAX, TRUE );
    params [i++] = CreatePlot ( PLOT_LINE2, "usedA",  C_USAGE1_AVG, pp, 0, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "lFreeA", C_FREE_AVG,   pp, 1, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "usedA",  "usedA",  "usedM",  "usedM",  "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "lFreeA", "lFreeA", "lFreeM", "lFreeM", "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataUsages ( fn, PLOT_24H_5M,  now );
    PlotDataUsages ( fn, PLOT_7D_15M,  now );
    PlotDataUsages ( fn, PLOT_31D_60M, now );
    PlotDataUsages ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    int retVal = STATE_OK;
    uint size;
    float usedPc;
    uint warn;
    uint crit;

    ReadVals ();
    /* calculate everything in Bytes */
    size = v.used + v.free;
    usedPc = ( ( (float) v.used ) * 100.0 ) / ( (float) size );
    warn = SpaceToB ( "warning memory",  warnS, warnU, size );
    crit = SpaceToB ( "critical memory", critS, critU, size );
    if ( v.used > warn ) retVal = MAXIMUM ( STATE_WARNING, retVal );
    if ( v.used > crit ) retVal = MAXIMUM ( STATE_CRITICAL, retVal );
    PrintExit ( retVal );
    printf ( " - %s memory pool usage: %i/%i B (%.1f%%)",
            addInf, v.used, size, usedPc );
    if ( npPerfData )
    {
        printf ( "|'memory'=%uB;%u;%u;0;%u", v.used, warn, crit, size );
        printf ( " 'largest free block'=%uB\n", v.lFree );
    }
    else
    {
        printf ( "|%s;%i;%i;%i\n", addInf, v.used, size, v.lFree );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ();
        if ( plotPerfData ) PlotPerfData ();
#endif /* WITH_RRD */
        if ( checkData
#if(WITH_RRD)
                || logPerfData
#endif /* WITH_RRD */
                ) free ( v.name );
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
