/*
 * File:      check_cisco_temp.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */


#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int index;   /* temperature sensor index */
    int v3;      /* found ciscoEnvMonTemperatureStatusValue */
    int v4;      /* found ciscoEnvMonTemperatureThreshold */
    int v6;      /* found ciscoEnvMonTemperatureState */
    char* descr; /* temperature sensor description */
} tScanRes;

/* data type for storing data */
typedef struct
{
    char* descr; /* name of the testpoint */
    uint  tempC; /* current temperature */
    int   tempE; /* emergency shutdown temperature */
    int   state; /* state of the testpoint */
} tDataSet;

/* global input variables */
uint warn;
uint crit;
int isWarnSet  = FALSE;
int isCritSet  = FALSE;

/*
 * SNMP base OID
 * CISCO-ENVMON-MIB::ciscoEnvMonTemperatureStatusDescr.INDEX
 * = 1.3.6.1.4.1.9.9.13.1.3.1.2.INDEX
 * CISCO-ENVMON-MIB::ciscoEnvMonTemperatureStatusValue.INDEX
 * = 1.3.6.1.4.1.9.9.13.1.3.1.3.INDEX
 * CISCO-ENVMON-MIB::ciscoEnvMonTemperatureThreshold.INDEX
 * = 1.3.6.1.4.1.9.9.13.1.3.1.4.INDEX
 * CISCO-ENVMON-MIB::ciscoEnvMonTemperatureState.INDEX
 * = 1.3.6.1.4.1.9.9.13.1.3.1.6.INDEX
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 9, 9, 13, 1, 3, 1 };
size_t baseOIDLen = 12;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* snmp values */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_cisco_temp";
    progVersion = "0.4";
    helpIndexP = "-i INDEX";
    helpThresP = "-w WARN -c CRIT";
    helpPurpose =
            "Checks the status of a temperature sensor of a Cisco(R) device.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 5 * 60;
    rrdHBeat = 11 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = TRUE;
    hasAddInfParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-w, --warning INTEGER\n"
    "    Temperature in degree Celsius at which a warning is generated\n"
    "\n"
    "-c, --critical INTEGER\n"
    "    Temperature in degree Celsius at which a critical is generated\n"
    "\n"
    "-i, --index INTEGER\n"
    "    Index of the temperature sensor which should be checked\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin returns the worst state of the overall state and the\n"
    "current temperature compared to the warning and critical settings.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<addInf>;<tempC>;<tempE>\n"
    "where\n"
    "    addInf ... additional information.\n"
    "    tempC  ... current temperature in degree Celsius.\n"
    "    tempE  ... emergency shutdown temperature.\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'current temperature'=<tempC>;<warn>;<crit> "
    "'emergency shutdown temperature'=<tempE>\n"
    "\n"
    "Note: Nagios Plugin performance data format does not allow Degree\n"
    "Celsius as unit of measurement.\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-w" )
            ||  !strcmp ( argv [(*i)], "--warning" ) )
    {
        isWarnSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            warn = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || *tail ) Usage ( "Invalid 'warn' parameter" );
        }
        else Usage ( "Missing 'warn' parameter(s)" );
    } else if ( !strcmp ( argv [(*i)], "-c" )
            ||  !strcmp ( argv [(*i)], "--critical" ) )
    {
        isCritSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            crit = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || *tail ) Usage ( "Invalid 'crit' parameter" );
        }
        else Usage ( "Missing 'crit' parameter(s)" );
    } else ret = FALSE;
    return ( ret );
}

int ScanResIndex ( int aIndex )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 14 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [12] )
            {
                case 2: /* ciscoMemoryPoolName */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].index = aVar->name [13];
                    scanRes [index].v3 = FALSE;
                    scanRes [index].v4 = FALSE;
                    scanRes [index].v6 = FALSE;
                    scanRes [index].descr = VarListStrDup ( aVar );
                break;
                case 3: /* ciscoEnvMonTemperatureStatusValue */
                    index = ScanResIndex ( aVar->name [13] );
                    scanRes [index].v3 = ( aVar->type == ASN_GAUGE );
                break;
                case 4: /* ciscoEnvMonTemperatureThreshold */
                    index = ScanResIndex ( aVar->name [13] );
                    scanRes [index].v4 = ( aVar->type == ASN_INTEGER );
                break;
                case 6: /* ciscoEnvMonTemperatureState */
                {
                    int typeOK = ( aVar->type == ASN_INTEGER );
                    int valOK = ( *(aVar->val.integer) != 5 );

                    index = ScanResIndex ( aVar->name [13] );
                    scanRes [index].v6 = ( typeOK && valOK );
                }
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if ( scanRes [i].v3 && scanRes [i].v4 && scanRes [i].v6 )
        {
            retVal = STATE_OK;
            printf ( "OK index: %i ; descr: %s\n",
                    scanRes [i].index, scanRes [i].descr );
            free ( scanRes [i].descr );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_descr [14];
    oid o_tempC [14];
    oid o_tempE [14];
    oid o_state [14];
    size_t oidLen;

    OidCpy ( o_descr, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_tempC, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_tempE, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_state, &oidLen, baseOID, &baseOIDLen );
    o_descr [12] = 2; o_descr [13] = idx;
    o_tempC [12] = 3; o_tempC [13] = idx;
    o_tempE [12] = 4; o_tempE [13] = idx;
    o_state [12] = 6; o_state [13] = idx;
    oidLen = 14;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.descr = ReadString  ( sess2, o_descr, oidLen );
    v.tempC = ReadGauge32 ( sess2, o_tempC, oidLen ); /* strange */
    v.tempE = ReadInteger ( sess2, o_tempE, oidLen ); /* isn't it? */
    v.state = ReadInteger ( sess2, o_state, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "tempC",  GAUGE, rrdHBeat, "-20", "150" );
    params [i++] = CreateDSStr ( "tempE",  GAUGE, rrdHBeat, "-20", "150" );
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    char buf[1024];

    chk = sscanf ( aData, "%1023s;%u;%i", buf, &(v.tempC), &(v.tempE) );
    if ( chk != 3 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
    addInf = StrDup ( buf );
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* tempC  = UIntToStr   ( v.tempC );
    char* tempE  = IntToStr    ( v.tempE );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "tempC:tempE" );
    params [i++] = StrConcat ( 5, ts, ":", tempC, ":", tempE );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( tempE );
    free ( tempC );
}

void PlotDataTemp ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[deg C]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, NULL, aTiming );
    PPAddPlot ( pp, "temperature", "deg C", FALSE );
    PPAddPlot ( pp, "threshold",   "deg C", FALSE );
    title = StrConcat ( 3, hostName, " temperature ", addInf );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, FALSE );
    params [i++] = CreateDEF ( "tempCM", aFN, "tempC", MAX );
    params [i++] = CreateDEF ( "tempCA", aFN, "tempC", AVG );
    params [i++] = CreateDEF ( "tempEM", aFN, "tempE", MAX );
    params [i++] = CreatePlot ( PLOT_AREA,  "tempCM", C_GAUGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "tempCA", C_GAUGE1_AVG, pp, 0, AVG, TRUE );
    params [i++] = CreatePlot ( PLOT_LINE2, "tempEM", C_THRES_MAX,  pp, 1, MAX, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "tempCA", "tempCA", "tempCM", "tempCM", "%4.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "tempEM", "tempEM", "tempEM", "tempEM", "%4.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataTemp ( fn, PLOT_24H_5M,  now );
    PlotDataTemp ( fn, PLOT_7D_15M,  now );
    PlotDataTemp ( fn, PLOT_31D_60M, now );
    PlotDataTemp ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    int retVal = STATE_UNKNOWN;
    char* strState = NULL;

    ReadVals ();
    switch ( v.state )
    {
        case 1:
            retVal = STATE_OK;       strState = "state 'normal'";
        break;
        case 2:
            retVal = STATE_WARNING;  strState = "state 'warning'";
        break;
        case 3:
            retVal = STATE_CRITICAL; strState = "state 'critical'";
        break;
        case 4:
            retVal = STATE_WARNING;  strState = "state 'shutdown'";
        break;
        case 5:
            retVal = STATE_WARNING;  strState = "state 'not present'";
        break;
        case 6:
            retVal = STATE_CRITICAL; strState = "state 'not functioning'";
        break;
        default:
            retVal = STATE_UNKNOWN;  strState = "an invalid state";
        break;
    }
    if ( ( v.tempC > warn ) && ( retVal < STATE_WARNING ) )
            retVal = STATE_WARNING;
    if ( ( v.tempC > crit ) && ( retVal < STATE_CRITICAL ) )
            retVal = STATE_CRITICAL;
    PrintExit ( retVal );
    printf ( " - %s sensor has %s", v.descr, strState );
    printf ( ", %u/%i deg C", v.tempC, v.tempE );
    if ( npPerfData )
    {
        printf ( "|'current temperature'=%u;%u;%u", v.tempC, warn, crit );
        printf ( " 'emergency shutdown temperature'=%i\n", v.tempE );
    }
    else
    {
        printf ( "|%s;%u;%i\n", addInf, v.tempC, v.tempE );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( checkData )
    {
        if ( !isWarnSet  ) Usage ( "Missing 'warn' parameters" );
        if ( !isCritSet  ) Usage ( "Missing 'crit' parameters" );
    }
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ();
        if ( plotPerfData ) PlotPerfData ();
#endif /* WITH_RRD */
        if ( checkData
#if(WITH_RRD)
                || logPerfData
#endif /* WITH_RRD */
                ) free ( v.descr );
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
