/*
 * File:      check_cisco_voltage.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */


#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int index;   /* voltage sensor index */
    int v3;      /* found ciscoEnvMonVoltageStatusValue */
    int v4;      /* found ciscoEnvMonVoltageThresholdLow */
    int v5;      /* found ciscoEnvMonVoltageThresholdHigh */
    int v7;      /* found ciscoEnvMonVoltageState */
    char* descr; /* voltage sensor description */
} tScanRes;

/* data type for storing data */
typedef struct
{
    char* descr; /* name of the testpoint */
    int   voltC; /* current voltage */
    int   voltL; /* low threshold voltage */
    int   voltH; /* high threshold voltage */
    int   state; /* state of the testpoint */
} tDataSet;


/*
 * SNMP base OID
 * CISCO-ENVMON-MIB::ciscoEnvMonVoltageStatusDescr.INDEX
 * = .1.3.6.1.4.1.9.9.13.1.2.1.2.INDEX
 * CISCO-ENVMON-MIB::ciscoEnvMonVoltageStatusValue.INDEX
 * = .1.3.6.1.4.1.9.9.13.1.2.1.3.INDEX
 * CISCO-ENVMON-MIB::ciscoEnvMonVoltageThresholdLow.INDEX
 * = .1.3.6.1.4.1.9.9.13.1.2.1.4.INDEX
 * CISCO-ENVMON-MIB::ciscoEnvMonVoltageThresholdHigh.INDEX
 * = .1.3.6.1.4.1.9.9.13.1.2.1.5.INDEX
 * CISCO-ENVMON-MIB::ciscoEnvMonVoltageState.INDEX
 * = .1.3.6.1.4.1.9.9.13.1.2.1.7.INDEX
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 9, 9, 13, 1, 2, 1 };
size_t baseOIDLen = 12;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* snmp values */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_cisco_voltage";
    progVersion = "0.2";
    helpIndexP = "-i INDEX";
    helpThresP = NULL;
    helpPurpose =
            "Checks the status of a voltage sensor of a Cisco(R) device.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 5 * 60;
    rrdHBeat = 11 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = TRUE;
    hasAddInfParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-i, --index INTEGER\n"
    "    Index of the voltage sensor which should be checked.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin returns the state of the given voltage sensor.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<addInf>;<voltC>;<voltL>;<voltH>\n"
    "where\n"
    "    addInf ... additional information.\n"
    "    voltC  ... current voltage in millivolts.\n"
    "    voltL  ... low threshold voltage in millivolts.\n"
    "    voltH  ... high threshold voltage in millivolts.\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'current voltage'=<voltC> "
    "'low threshold voltage'=<voltL> "
    "'high threshold voltage'=<voltH>\n"
    "\n"
    "Note: Nagios Plugin performance data format does not allow Millivolt\n"
    "as units of measurement.\n"
    "Note: The threshold voltages are described as those thresholds where\n"
    "an emergency shutdown is initiated. How they relate to\n"
    "ciscoEnvMonVoltageState which is used to report the state of a sensor\n"
    "is currently unknown.\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    return FALSE;
}

int ScanResIndex ( int aIndex )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 14 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [12] )
            {
                case 2: /* ciscoEnvMonVoltageStatusDescr */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].index = aVar->name [13];
                    scanRes [index].v3 = FALSE;
                    scanRes [index].v4 = FALSE;
                    scanRes [index].v5 = FALSE;
                    scanRes [index].v7 = FALSE;
                    scanRes [index].descr = VarListStrDup ( aVar );
                break;
                case 3: /* ciscoEnvMonVoltageStatusValue */
                    index = ScanResIndex ( aVar->name [13] );
                    scanRes [index].v3 = ( aVar->type == ASN_INTEGER );
                break;
                case 4: /* ciscoEnvMonVoltageThresholdLow */
                    index = ScanResIndex ( aVar->name [13] );
                    scanRes [index].v4 = ( aVar->type == ASN_INTEGER );
                break;
                case 5: /* ciscoEnvMonVoltageThresholdHigh */
                    index = ScanResIndex ( aVar->name [13] );
                    scanRes [index].v5 = ( aVar->type == ASN_INTEGER );
                break;
                case 7: /* ciscoEnvMonVoltageState */
                {
                    int typeOK = ( aVar->type == ASN_INTEGER );
                    int valOK = ( *(aVar->val.integer) != 5 );

                    index = ScanResIndex ( aVar->name [13] );
                    scanRes [index].v7 = ( typeOK && valOK );
                }
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v3 && scanRes [i].v4 && scanRes [i].v5 &&
                scanRes [i].v7 )
        {
            retVal = STATE_OK;
            printf ( "OK index: %i ; descr: %s\n",
                    scanRes [i].index, scanRes [i].descr );
            free ( scanRes [i].descr );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_descr [14];
    oid o_voltC [14];
    oid o_voltL [14];
    oid o_voltH [14];
    oid o_state [14];
    size_t oidLen;

    OidCpy ( o_descr, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_voltC, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_voltL, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_voltH, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_state, &oidLen, baseOID, &baseOIDLen );
    o_descr [12] = 2; o_descr [13] = idx;
    o_voltC [12] = 3; o_voltC [13] = idx;
    o_voltL [12] = 4; o_voltL [13] = idx;
    o_voltH [12] = 5; o_voltH [13] = idx;
    o_state [12] = 7; o_state [13] = idx;
    oidLen = 14;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.descr = ReadString  ( sess2, o_descr, oidLen );
    v.voltC = ReadInteger ( sess2, o_voltC, oidLen );
    v.voltL = ReadInteger ( sess2, o_voltL, oidLen );
    v.voltH = ReadInteger ( sess2, o_voltH, oidLen );
    v.state = ReadInteger ( sess2, o_state, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "voltC",  GAUGE, rrdHBeat, "-60.0", "60.0" );
    params [i++] = CreateDSStr ( "voltL",  GAUGE, rrdHBeat, "-60.0", "60.0" );
    params [i++] = CreateDSStr ( "voltH",  GAUGE, rrdHBeat, "-60.0", "60.0" );
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MIN_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MIN_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MIN_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    params [i++] = RRA_MIN_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    char buf[1024];

    chk = sscanf ( aData, "%1023s;%i;%i;%i", buf,
            &(v.voltC), &(v.voltL), &(v.voltH) );
    if ( chk != 4 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
    addInf = StrDup ( buf );
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* voltC  = DoubleToStr ( (double) v.voltC / 1000.0 );
    char* voltL  = DoubleToStr ( (double) v.voltL / 1000.0 );
    char* voltH  = DoubleToStr ( (double) v.voltH / 1000.0 );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "voltC:voltL:voltH" );
    params [i++] = StrConcat ( 7, ts, ":", voltC, ":", voltL, ":", voltH );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( voltH );
    free ( voltL );
    free ( voltC );
    free ( ts );
}

void PlotDataVoltage ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[V]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, NULL, aTiming );
    PPAddPlot ( pp, "voltage",         "V", FALSE );
    PPAddPlot ( pp, "lower threshold", "V", FALSE );
    PPAddPlot ( pp, "upper threshold", "V", FALSE );
    title = StrConcat ( 4, hostName, " voltage ", addInf, "" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_AUTO, 0, 0, FALSE );
    params [i++] = CreateDEF ( "voltCM", aFN, "voltC", MAX );
    params [i++] = CreateDEF ( "voltCA", aFN, "voltC", AVG );
    params [i++] = CreateDEF ( "voltLM", aFN, "voltL", MAX );
    params [i++] = CreateDEF ( "voltHM", aFN, "voltH", MIN );
    params [i++] = CreatePlot ( PLOT_AREA,  "voltCM", C_GAUGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "voltCA", C_GAUGE1_AVG, pp, 0, AVG, TRUE );
    params [i++] = CreatePlot ( PLOT_LINE2, "voltLM", C_THRES_MAX,  pp, 1, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "voltHM", C_THRES_MAX,  pp, 2, MIN, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "voltCA", "voltCA", "voltCM", "voltCM", "%6.3lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataVoltage ( fn, PLOT_24H_5M,  now );
    PlotDataVoltage ( fn, PLOT_7D_15M,  now );
    PlotDataVoltage ( fn, PLOT_31D_60M, now );
    PlotDataVoltage ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    int retVal = STATE_UNKNOWN;
    char* strState = NULL;

    ReadVals ();
    switch ( v.state )
    {
        case 1:
            retVal = STATE_OK;       strState = "state 'normal'";
        break;
        case 2:
            retVal = STATE_WARNING;  strState = "state 'warning'";
        break;
        case 3:
            retVal = STATE_CRITICAL; strState = "state 'critical'";
        break;
        case 4:
            retVal = STATE_WARNING;  strState = "state 'shutdown'";
        break;
        case 5:
            retVal = STATE_WARNING;  strState = "state 'not present'";
        break;
        case 6:
            retVal = STATE_CRITICAL; strState = "state 'not functioning'";
        break;
        default:
            retVal = STATE_UNKNOWN;  strState = "an invalid state";
        break;
    }
    PrintExit ( retVal );
    printf ( " - %s voltage has %s", v.descr, strState );
    printf ( ", %.2f V",   (double) v.voltL / 1000.0 );
    printf ( " <= %.2f V", (double) v.voltC / 1000.0 );
    printf ( " <= %.2f V", (double) v.voltH / 1000.0 );
    if ( npPerfData )
    {
        printf ( "|'current voltage'=%i", v.voltC );
        printf ( " 'low threshold voltage'=%i", v.voltL );
        printf ( " 'high threshold voltage'=%i\n", v.voltH );
    }
    else
    {
        printf ( "| %s;%i;%i;%i\n", addInf, v.voltC, v.voltL, v.voltH );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ();
        if ( plotPerfData ) PlotPerfData ();
#endif /* WITH_RRD */
        if ( checkData
#if(WITH_RRD)
                || logPerfData
#endif /* WITH_RRD */
                ) free ( v.descr );
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
