/*
 * File:      check_cisco_vpn_conns.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int v1;  /* alActiveSessionCount */
    int v4;  /* alActiveSessionMaxUsers */
    int v6;  /* alMaxSessionCount */
    int v7;  /* alActiveLanToLanSessionCount */
    int v8;  /* alActiveManagementSessionCount */
    int v9;  /* alActiveRemoteAccessSessionCount */
} tScanRes;

/* data type for storing data */
typedef struct
{
    uint actSC; /* active session count (including management sessions) */
    uint maxUS; /* maximum user sessions (excluding management sessions) */
    uint maxSC; /* maximum concurrent session count since reboot */
    uint l2lSC; /* lan to lan session count */
    uint manSC; /* management session count */
    uint remSC; /* remote access session count */
} tDataSet;

/* global input variables */
uint   warnC = 90;
char   warnU = '%';
uint   critC = 95;
char   critU = '%';

/*
 * SNMP base OID
 * ALTIGA-SESSION-STATS-MIB::alActiveSessionCount.0
 * = .1.3.6.1.4.1.3076.2.1.2.17.1.1.0
 * ALTIGA-SESSION-STATS-MIB::alActiveSessionMaxUsers.0
 * = .1.3.6.1.4.1.3076.2.1.2.17.1.4.0
 * ALTIGA-SESSION-STATS-MIB::alMaxSessionCount.0
 * = .1.3.6.1.4.1.3076.2.1.2.17.1.6.0
 * ALTIGA-SESSION-STATS-MIB::alActiveLanToLanSessionCount.0
 * = .1.3.6.1.4.1.3076.2.1.2.17.1.7.0
 * ALTIGA-SESSION-STATS-MIB::alActiveManagementSessionCount.0
 * = .1.3.6.1.4.1.3076.2.1.2.17.1.8.0
 * ALTIGA-SESSION-STATS-MIB::alActiveRemoteAccessSessionCount.0
 * = .1.3.6.1.4.1.3076.2.1.2.17.1.9.0
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 3076, 2, 1, 2, 17, 1 };
size_t baseOIDLen = 12;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_cisco_vpn_conns";
    progVersion = "0.1";
    helpIndexP = NULL;
    helpThresP = "[-w WARN] [-c CRIT]";
    helpPurpose = "Checks the number of connections (sessions) on a Cisco (R)"
            " VPN Concentrator.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 5 * 60;
    rrdHBeat = 11 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = FALSE;
}

void HelpSpecificParams ()
{
    printf (
    "-w, --warning UINT[%%]\n"
    "    Number (if no '%%' is given) or percentage of user sessions at\n"
    "    which a warning is generated.\n"
    "    Default is %u%c.\n\n", warnC, warnU
    );
    printf (
    "-c, --critical UINT[%%]\n"
    "    Number (if no '%%' is given) or percentage of user sessions at\n"
    "    which a critical is generated.\n"
    "    Default is %u%c.\n\n", critC, critU
    );
}

void HelpExplain ()
{
    printf (
    "The plugin checks the number of currently active user sessions.\n"
    "It generates a warning or critical if the respective threshold is less\n"
    "than the number of currently active user sessions.\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<actSC>;<actUS>;<maxUS>;<maxSC>;<l2lSC>;<manSC>;<remSC>\n"
    "where\n"
    "    actSC ... number of currently active session.\n"
    "    actUS ... number of currently active user sessions.\n"
    "              This is not read from the device but l2lSC + remSC.\n"
    "    maxUS ... maximum user sessions.\n"
    "    maxSC ... maximum concurrent sessions since reboot.\n"
    "    l2lSC ... current number of LAN-to-LAN sessions.\n"
    "    manSC ... current number management sessions.\n"
    "    remSC ... current number of remote access sessions.\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'active sessions'=<actSC> "
    "'active user sessions'=<actUS>;<warn>;<crit>;;<maxUS> "
    "'maximum sessions'=<maxSC> "
    "'LAN-to-LAN sessions'=<l2lSC> "
    "'management sessions'=<manSC> "
    "'remote access sessions'=<remSC>\n"
    "\n"
    );
}

void InvalidWarnParam ()
{
    Usage ( "Invalid 'warn' parameter" );
}

void InvalidCritParam ()
{
    Usage ( "Invalid 'crit' parameter" );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (            !strcmp ( argv [(*i)], "-w" )
                ||  !strcmp ( argv [(*i)], "--warning" ) )
    {
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];

            errno = 0;
            warnC = strtoul ( argv [(*i)], &tail, 0 );
            if ( errno ) InvalidWarnParam ();
            if      ( (*tail) == '\0' ) warnU = '1';
            else if ( (*tail) == '%'  ) warnU = '%';
            else Usage ( "Invalid unit" );
            if ( warnU != '1' ) tail = &(tail [1]);
            if ( (*tail) != '\0' ) InvalidWarnParam ();
            (*i)++;
        }
        else Usage ( "Missing 'warn' parameter" );
    } else if ( !strcmp ( argv [(*i)], "-c" )
            ||  !strcmp ( argv [(*i)], "--critical" ) )
    {
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];

            errno = 0;
            critC = strtoul ( argv [(*i)], &tail, 0 );
            if ( errno ) InvalidCritParam ();
            if      ( (*tail) == '\0' ) critU = '1';
            else if ( (*tail) == '%'  ) critU = '%';
            else Usage ( "Invalid unit" );
            if ( critU != '1' ) tail = &(tail [1]);
            if ( (*tail) != '\0' ) InvalidCritParam ();
            (*i)++;
        }
        else Usage ( "Missing 'crit' parameter" );
    } else ret = FALSE;
    return ( ret );
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 14 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            /*
             * Note: aVar->name [13] is expected to be 0 always but as of now
             * this is not checked.
             */
            switch ( aVar->name [12] )
            {
                case 1: /* alActiveSessionCount */
                    scanResLen = 1;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [0].v1 = ( aVar->type == ASN_GAUGE );
                    scanRes [0].v4 = FALSE;
                    scanRes [0].v6 = FALSE;
                    scanRes [0].v7 = FALSE;
                    scanRes [0].v8 = FALSE;
                    scanRes [0].v9 = FALSE;
                break;
                case 4: /* alActiveSessionMaxUsers */
                    scanRes [0].v4 = ( aVar->type == ASN_GAUGE );
                break;
                case 6: /* alMaxSessionCount */
                    scanRes [0].v6 = ( aVar->type == ASN_COUNTER );
                break;
                case 7: /* alActiveLanToLanSessionCount */
                    scanRes [0].v7 = ( aVar->type == ASN_GAUGE );
                break;
                case 8: /* alActiveManagementSessionCount */
                    scanRes [0].v8 = ( aVar->type == ASN_GAUGE );
                break;
                case 9: /* alActiveRemoteAccessSessionCount */
                    scanRes [0].v9 = ( aVar->type == ASN_GAUGE );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v1 && scanRes [i].v4 && scanRes [i].v6 &&
                scanRes [i].v7 && scanRes [i].v8 && scanRes [i].v9 )
        {
            retVal = STATE_OK;
            printf ( "OK\n" );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_actSC [14];
    oid o_maxUS [14];
    oid o_maxSC [14];
    oid o_l2lSC [14];
    oid o_manSC [14];
    oid o_remSC [14];
    size_t oidLen;

    OidCpy ( o_actSC, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_maxUS, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_maxSC, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_l2lSC, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_manSC, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_remSC, &oidLen, baseOID, &baseOIDLen );
    o_actSC [12] = 1; o_actSC [13] = 0;
    o_maxUS [12] = 4; o_maxUS [13] = 0;
    o_maxSC [12] = 6; o_maxSC [13] = 0;
    o_l2lSC [12] = 7; o_l2lSC [13] = 0;
    o_manSC [12] = 8; o_manSC [13] = 0;
    o_remSC [12] = 9; o_remSC [13] = 0;
    oidLen = 14;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.actSC = ReadGauge32   ( sess2, o_actSC, oidLen );
    v.maxUS = ReadGauge32   ( sess2, o_maxUS, oidLen );
    v.maxSC = ReadCounter32 ( sess2, o_maxSC, oidLen );
    v.l2lSC = ReadGauge32   ( sess2, o_l2lSC, oidLen );
    v.manSC = ReadGauge32   ( sess2, o_manSC, oidLen );
    v.remSC = ReadGauge32   ( sess2, o_remSC, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "actSC", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "maxUS", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "maxSC", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "l2lSC", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "manSC", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "remSC", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    chk = sscanf ( aData, "%u;%u;%u;%u;%u;%u",
            &(v.actSC), &(v.maxUS), &(v.maxSC),
            &(v.l2lSC), &(v.manSC), &(v.remSC) );
    if ( chk != 6 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* actSC  = UIntToStr   ( v.actSC );
    char* maxUS  = UIntToStr   ( v.maxUS );
    char* maxSC  = UIntToStr   ( v.maxSC );
    char* l2lSC  = UIntToStr   ( v.l2lSC );
    char* manSC  = UIntToStr   ( v.manSC );
    char* remSC  = UIntToStr   ( v.remSC );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "actSC:maxUS:maxSC:l2lSC:manSC:remSC" );
    params [i++] = StrConcat ( 13, ts, ":", actSC, ":", maxUS, ":", maxSC,
            ":", l2lSC, ":", manSC, ":", remSC );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( remSC );
    free ( manSC );
    free ( l2lSC );
    free ( maxSC  );
    free ( maxUS );
    free ( actSC );
    free ( ts );
}

void PlotDataOvervw ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[sessions]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "overvw", aTiming );
    PPAddPlot ( pp, "active sessions",       "sessions", FALSE );
    PPAddPlot ( pp, "active user sessions",  "sessions", FALSE );
    PPAddPlot ( pp, "maximum user sessions", "sessions", FALSE );
    title = StrConcat ( 2, hostName, " sessions overview" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, FALSE );
    params [i++] = CreateDEF ( "actSCM", aFN, "actSC", MAX );
    params [i++] = CreateDEF ( "actSCA", aFN, "actSC", AVG );
    params [i++] = CreateDEF ( "maxUSM", aFN, "maxUS", MAX );
    params [i++] = CreateDEF ( "l2lSCM", aFN, "l2lSC", MAX );
    params [i++] = CreateDEF ( "l2lSCA", aFN, "l2lSC", AVG );
    params [i++] = CreateDEF ( "remSCM", aFN, "remSC", MAX );
    params [i++] = CreateDEF ( "remSCA", aFN, "remSC", AVG );
    params [i++] = StrDup ( "CDEF:actUSM=l2lSCM,remSCM,+" );
    params [i++] = StrDup ( "CDEF:actUSA=l2lSCA,remSCA,+" );
    params [i++] = CreatePlot ( PLOT_LINE3, "maxUSM", C_LOAD1_MAX, pp, 2, MAX, TRUE );
    params [i++] = CreatePlot ( PLOT_AREA,  "actSCM", C_LOAD2_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "actSCA", C_LOAD2_AVG, pp, 0, AVG, TRUE );
    params [i++] = CreatePlot ( PLOT_LINE3, "actUSM", C_LOAD3_MAX, pp, 1, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "actUSA", C_LOAD3_AVG, pp, 1, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "actSCA", "actSCA", "actSCM", "actSCM", "%6.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "actUSA", "actUSA", "actUSM", "actUSM", "%6.1lf" );
    i = PlotPrintParams ( i, params, pp, 2, "maxUSM", "maxUSM", "maxUSM", "maxUSM", "%6.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataDetail ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[sessions]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "detail", aTiming );
    PPAddPlot ( pp, "LAN-to-LAN sessions",    "sessions", FALSE );
    PPAddPlot ( pp, "management sessions",    "sessions", TRUE );
    PPAddPlot ( pp, "remote access sessions", "sessions", TRUE );
    title = StrConcat ( 2, hostName, " sessions details" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, FALSE );
    params [i++] = CreateDEF ( "l2lSCM", aFN, "l2lSC", MAX );
    params [i++] = CreateDEF ( "l2lSCA", aFN, "l2lSC", AVG );
    params [i++] = CreateDEF ( "manSCM", aFN, "manSC", MAX );
    params [i++] = CreateDEF ( "manSCA", aFN, "manSC", AVG );
    params [i++] = CreateDEF ( "remSCM", aFN, "remSC", MAX );
    params [i++] = CreateDEF ( "remSCA", aFN, "remSC", AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "l2lSCM", C_LOAD3_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "manSCM", C_LOAD2_MAX, pp, 1, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "remSCM", C_LOAD1_MAX, pp, 2, MAX, TRUE );
    params [i++] = CreatePlot ( PLOT_LINE2, "l2lSCA", C_LOAD3_AVG, pp, 0, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "manSCA", C_LOAD2_AVG, pp, 1, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "remSCA", C_LOAD1_AVG, pp, 2, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "l2lSCA", "l2lSCA", "l2lSCM", "l2lSCM", "%6.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "manSCA", "manSCA", "manSCM", "manSCM", "%6.1lf" );
    i = PlotPrintParams ( i, params, pp, 2, "remSCA", "remSCA", "remSCM", "remSCM", "%6.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ( const char* aSrvId )
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataOvervw ( fn, PLOT_6H_1M,   now );
    PlotDataDetail ( fn, PLOT_6H_1M,   now );
    PlotDataOvervw ( fn, PLOT_24H_5M,  now );
    PlotDataDetail ( fn, PLOT_24H_5M,  now );
    PlotDataOvervw ( fn, PLOT_7D_15M,  now );
    PlotDataDetail ( fn, PLOT_7D_15M,  now );
    PlotDataOvervw ( fn, PLOT_31D_60M, now );
    PlotDataDetail ( fn, PLOT_31D_60M, now );
    PlotDataOvervw ( fn, PLOT_366D_4H, now );
    PlotDataDetail ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    int retVal = STATE_OK;
    uint actUS;
    uint warn;
    uint crit;

    ReadVals ();
    actUS = v.l2lSC + v.remSC;
    warn = ( '%' == warnU ) ? v.maxUS * warnC / 100 : warnC;
    crit = ( '%' == critU ) ? v.maxUS * critC / 100 : critC;
    if ( warn < actUS ) retVal = MAXIMUM ( retVal, STATE_WARNING );
    if ( crit < actUS ) retVal = MAXIMUM ( retVal, STATE_CRITICAL );
    PrintExit ( retVal );
    printf ( " - user sessions: %u/%u", actUS, v.maxSC );
    if ( npPerfData )
    {
        printf ( "|'active sessions'=%u", v.actSC );
        printf ( " 'active user sessions'=%u;%u;%u;;%u",
                actUS, warn, crit, v.maxUS );
        printf ( " 'maximum sessions'=%u", v.maxSC );
        printf ( " 'LAN-to-LAN sessions'=%u", v.l2lSC );
        printf ( " 'management sessions'=%u", v.manSC );
        printf ( " 'remote access sessions'=%u\n", v.remSC );
    }
    else
    {
        printf ( "|%u;%u;%u%u;%u;%u;%u\n",
                v.actSC, actUS, v.maxUS, v.maxSC, v.l2lSC, v.manSC, v.remSC );
    }
    return ( retVal );
}


int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ( NULL );
        if ( plotPerfData ) PlotPerfData ( NULL );
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
