/*
 * File:      check_cisco_vpn_fan.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

/*
 * This does not use proper tables but (many) scalars for environmental
 * monitoring. So this code is rather unusual. If you are looking for
 * a generic example, better look somewhere else.
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"

/* data type for storing scan results */
typedef struct
{
    int v37;     /* found alHardwareFan1Rpm and value >= 1  */
    int v38;     /* found alHardwareFan1RpmAlarm */
    int v39;     /* found alHardwareFan1RpmCount */
    int v41;     /* found alHardwareFan2Rpm and value >= 1  */
    int v42;     /* found alHardwareFan2RpmAlarm */
    int v43;     /* found alHardwareFan2RpmCount */
    int v45;     /* found alHardwareFan3Rpm and value >= 1 */
    int v46;     /* found alHardwareFan3RpmAlarm */
    int v47;     /* found alHardwareFan3RpmCount */
} tScanRes;

/* data type for storing data */
typedef struct
{
    uint  speed; /* current speed in rounds per minute */
    int   alarm; /* alarm status */
    uint  count; /* number of alarms since boot */
} tDataSet;


/*
 * SNMP base OID
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareFan1Rpm.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.37.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareFan1RpmAlarm.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.38.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareFan1RpmCount.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.39.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareFan2Rpm.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.41.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareFan2RpmAlarm.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.42.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareFan2RpmCount.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.43.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareFan3Rpm.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.45.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareFan3RpmAlarm.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.46.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareFan3RpmCount.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.47.0
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 3076, 2, 1, 2, 22, 1 };
size_t baseOIDLen = 12;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* snmp values */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_cisco_vpn_fan";
    progVersion = "0.1";
    helpIndexP = "-i INDEX";
    helpThresP = NULL;
    helpPurpose =
            "Checks the status of a fan of a Cisco(R) VPN Concentrator.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 5 * 60;
    rrdHBeat = 11 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = TRUE;
    hasAddInfParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-i, --index INTEGER\n"
    "    Index of the fan which should be checked.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin checks whether a fan of a Cisco VPN Concentrator has an\n"
    "alarm. If the fan does have an alarm a critical alert will be\n"
    "generated. The plugin never returns a warning.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<addInf>;<speed>\n"
    "where\n"
    "    addInf ... additional information.\n"
    "    speed  ... current speed in rounds per minute.\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'current speed'=<speed>\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    return FALSE;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 14 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            /*
             * Note: aVar->name [13] is expected to be 0 always but as of now
             * this is not checked.
             */
            switch ( aVar->name [12] )
            {
                case 37: /* alHardwareFan1Rpm */
                {
                    int typeOK = ( aVar->type == ASN_GAUGE );
                    int valOK = ( *(aVar->val.integer) >= 1 );

                    scanResLen = 1;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();

                    scanRes [0].v37 = ( typeOK && valOK );
                    scanRes [0].v38 = FALSE;
                    scanRes [0].v39 = FALSE;
                    scanRes [0].v41 = FALSE;
                    scanRes [0].v42 = FALSE;
                    scanRes [0].v43 = FALSE;
                    scanRes [0].v45 = FALSE;
                    scanRes [0].v46 = FALSE;
                    scanRes [0].v47 = FALSE;
                }
                break;
                case 38: /* alHardwareFan1RpmAlarm */
                    scanRes [0].v38 = ( aVar->type == ASN_INTEGER );
                break;
                case 39: /* alHardwareFan1RpmCount */
                    scanRes [0].v39 = ( aVar->type == ASN_COUNTER );
                break;
                case 41: /* alHardwareFan2Rpm */
                {
                    int typeOK = ( aVar->type == ASN_GAUGE );
                    int valOK = ( *(aVar->val.integer) >= 1 );

                    scanRes [0].v41 = ( typeOK && valOK );
                }
                break;
                case 42: /* alHardwareFan2RpmAlarm */
                    scanRes [0].v42 = ( aVar->type == ASN_INTEGER );
                break;
                case 43: /* alHardwareFan2RpmCount */
                    scanRes [0].v43 = ( aVar->type == ASN_COUNTER );
                break;
                case 45: /* alHardwareFan3Rpm */
                {
                    int typeOK = ( aVar->type == ASN_GAUGE );
                    int valOK = ( *(aVar->val.integer) >= 1 );

                    scanRes [0].v45 = ( typeOK && valOK );
                }
                break;
                case 46: /* alHardwareFan3RpmAlarm */
                    scanRes [0].v46 = ( aVar->type == ASN_INTEGER );
                break;
                case 47: /* alHardwareFan3RpmCount */
                    scanRes [0].v47 = ( aVar->type == ASN_COUNTER );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v37 && scanRes [i].v38 && scanRes [i].v39 )
        {
            retVal = STATE_OK;
            printf ( "OK index: 1 ; descr: Fan 1\n" );
        }
        if (    scanRes [i].v41 && scanRes [i].v42 && scanRes [i].v43 )
        {
            retVal = STATE_OK;
            printf ( "OK index: 2 ; descr: Fan 2\n" );
        }
        if (    scanRes [i].v45 && scanRes [i].v46 && scanRes [i].v47 )
        {
            retVal = STATE_OK;
            printf ( "OK index: 3 ; descr: Fan 3\n" );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_speed [14];
    oid o_alarm [14];
    oid o_count [14];
    size_t oidLen;
    int offset = ( idx + 8 ) * 4;

    OidCpy ( o_speed, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_alarm, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_count, &oidLen, baseOID, &baseOIDLen );
    o_speed [12] = offset + 1; o_speed [13] = 0;
    o_alarm [12] = offset + 2; o_alarm [13] = 0;
    o_count [12] = offset + 3; o_count [13] = 0;
    oidLen = 14;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.speed = ReadGauge32   ( sess2, o_speed, oidLen );
    v.alarm = ReadInteger   ( sess2, o_alarm, oidLen );
    v.count = ReadCounter32 ( sess2, o_count, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "speed",  GAUGE, rrdHBeat, "0", "20000" );
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MIN_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MIN_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MIN_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    params [i++] = RRA_MIN_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    char buf[1024];

    chk = sscanf ( aData, "%1023s;%u", buf, &(v.speed) );
    if ( chk != 2 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
    addInf = StrDup ( buf );
    v.alarm = 2;
    v.count = 0;
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* speed  = UIntToStr   ( v.speed );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "speed" );
    params [i++] = StrConcat ( 3, ts, ":", speed );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( speed );
    free ( ts );
}

void PlotDataFan ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[rpm]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, NULL, aTiming );
    PPAddPlot ( pp, "speed", "rpm", FALSE );
    title = StrConcat ( 4, hostName, " fan ", addInf, " speed" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_AUTO, 0, 0, FALSE );
    params [i++] = CreateDEF ( "speedM", aFN, "speed", MAX );
    params [i++] = CreateDEF ( "speedA", aFN, "speed", AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "speedM", C_GAUGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "speedA", C_GAUGE1_AVG, pp, 0, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "speedA", "speedA", "speedM", "speedM", "%.0lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataFan ( fn, PLOT_24H_5M,  now );
    PlotDataFan ( fn, PLOT_7D_15M,  now );
    PlotDataFan ( fn, PLOT_31D_60M, now );
    PlotDataFan ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    int retVal = STATE_UNKNOWN;
    char* strState = NULL;

    ReadVals ();
    if ( 1 == v.alarm )
    {
        retVal = STATE_CRITICAL;
        strState = "an alarm";
    }
    else if ( 2 == v.alarm )
    {
        retVal = STATE_OK;
        strState = "no alarm";
    }
    else
    {
        retVal = STATE_UNKNOWN;
        strState = "an invalid alarm state";
    }
    PrintExit ( retVal );
    printf ( " - fan has %s", strState );
    printf ( ", %u rpm", v.speed );
    printf ( ", %u alarms since boot", v.count );
    if ( npPerfData )
    {
        printf ( "|'current speed'=%u\n", v.speed );
    }
    else
    {
        printf ( "| %s;%u\n", addInf, v.speed );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ();
        if ( plotPerfData ) PlotPerfData ();
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
