/*
 * File:      check_cisco_vpn_temp.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

/*
 * This does not use proper tables but (many) scalars for environmental
 * monitoring. So this code is rather unusual. If you are looking for
 * a generic example, better look somewhere else.
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"

/* data type for storing scan results */
typedef struct
{
    int v29;     /* found alHardwareCpuTemp */
    int v30;     /* found alHardwareCpuTempAlarm */
    int v31;     /* found alHardwareCpuTempCount */
    int v33;     /* found alHardwareCageTemp */
    int v34;     /* found alHardwareCageTempAlarm */
    int v35;     /* found alHardwareCageTempCount */
} tScanRes;

/* data type for storing data */
typedef struct
{
    int   tempC; /* current temperature in degree Celsius */
    int   alarm; /* alarm status */
    uint  count; /* number of alarms since boot */
} tDataSet;


/*
 * SNMP base OID
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareCpuTemp.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.29.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareCpuTempAlarm.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.30.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareCpuTempCount.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.31.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareCageTemp.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.33.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareCageTempAlarm.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.34.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareCageTempCount.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.35.0
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 3076, 2, 1, 2, 22, 1 };
size_t baseOIDLen = 12;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* snmp values */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_cisco_vpn_temp";
    progVersion = "0.1";
    helpIndexP = "-i INDEX";
    helpThresP = NULL;
    helpPurpose = "Checks the status of a temperature sensor of a Cisco(R)"
            " VPN Concentrator.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 5 * 60;
    rrdHBeat = 11 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = TRUE;
    hasAddInfParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-i, --index INTEGER\n"
    "    Index of the voltage sensor which should be checked.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin checks whether a temperature sensor of a Cisco VPN Concentrator\n"
    "has an alarm. If the temperature sensor does have an alarm a critical\n"
    "alert will be generated. The plugin never returns a warning.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<addInf>;<tempC>\n"
    "where\n"
    "    addInf ... additional information.\n"
    "    tempC  ... current temperature in degree Centrigrade.\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'current temperature'=<tempC>\n"
    "\n"
    "Note: Nagios Plugin performance data format does not allow Degree\n"
    "Celsius as units of measurement.\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    return FALSE;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 14 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            /*
             * Note: aVar->name [13] is expected to be 0 always but as of now
             * this is not checked.
             */
            switch ( aVar->name [12] )
            {
                case 29: /* alHardwareCpuTemp */
                    scanResLen = 1;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [0].v29 = ( aVar->type == ASN_INTEGER );
                    scanRes [0].v30 = FALSE;
                    scanRes [0].v31 = FALSE;
                    scanRes [0].v33 = FALSE;
                    scanRes [0].v34 = FALSE;
                    scanRes [0].v35 = FALSE;
                break;
                case 30: /* alHardwareCpuTempAlarm */
                    scanRes [0].v30 = ( aVar->type == ASN_INTEGER );
                break;
                case 31: /* alHardwareCpuTempCount */
                    scanRes [0].v31 = ( aVar->type == ASN_COUNTER );
                break;
                case 33: /* alHardwareCageTemp */
                    scanRes [0].v33 = ( aVar->type == ASN_INTEGER );
                break;
                case 34: /* alHardwareCageTempAlarm */
                    scanRes [0].v34 = ( aVar->type == ASN_INTEGER );
                break;
                case 35: /* alHardwareCageTempCount */
                    scanRes [0].v35 = ( aVar->type == ASN_COUNTER );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v29 && scanRes [i].v30 && scanRes [i].v31 )
        {
            retVal = STATE_OK;
            printf ( "OK index: 1 ; descr: CPU\n" );
        }
        if (    scanRes [i].v33 && scanRes [i].v34 && scanRes [i].v35 )
        {
            retVal = STATE_OK;
            printf ( "OK index: 2 ; descr: Cage\n" );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_tempC [14];
    oid o_alarm [14];
    oid o_count [14];
    size_t oidLen;
    int offset = ( idx + 6 ) * 4;

    OidCpy ( o_tempC, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_alarm, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_count, &oidLen, baseOID, &baseOIDLen );
    o_tempC [12] = offset + 1; o_tempC [13] = 0;
    o_alarm [12] = offset + 2; o_alarm [13] = 0;
    o_count [12] = offset + 3; o_count [13] = 0;
    oidLen = 14;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.tempC = ReadInteger   ( sess2, o_tempC, oidLen );
    v.alarm = ReadInteger   ( sess2, o_alarm, oidLen );
    v.count = ReadCounter32 ( sess2, o_count, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "tempC",  GAUGE, rrdHBeat, "-20", "150" );
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    char buf[1024];

    chk = sscanf ( aData, "%1023s;%i", buf, &(v.tempC) );
    if ( chk != 2 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
    addInf = StrDup ( buf );
    v.alarm = 2;
    v.count = 0;
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* tempC  = IntToStr    ( v.tempC );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "tempC" );
    params [i++] = StrConcat ( 3, ts, ":", tempC );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( tempC );
    free ( ts );
}

void PlotDataTemp ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[deg C]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, NULL, aTiming );
    PPAddPlot ( pp, "temperature", "deg C", FALSE );
    title = StrConcat ( 4, hostName, " ", addInf, " temperature" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, FALSE );
    params [i++] = CreateDEF ( "tempCM", aFN, "tempC", MAX );
    params [i++] = CreateDEF ( "tempCA", aFN, "tempC", AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "tempCM", C_GAUGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "tempCA", C_GAUGE1_AVG, pp, 0, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "tempCA", "tempCA", "tempCM", "tempCM", "%4.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataTemp ( fn, PLOT_24H_5M,  now );
    PlotDataTemp ( fn, PLOT_7D_15M,  now );
    PlotDataTemp ( fn, PLOT_31D_60M, now );
    PlotDataTemp ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    int retVal = STATE_UNKNOWN;
    char* strState = NULL;

    ReadVals ();
    if ( 1 == v.alarm )
    {
        retVal = STATE_CRITICAL;
        strState = "an alarm";
    }
    else if ( 2 == v.alarm )
    {
        retVal = STATE_OK;
        strState = "no alarm";
    }
    else
    {
        retVal = STATE_UNKNOWN;
        strState = "an invalid alarm state";
    }
    PrintExit ( retVal );
    printf ( " - temperature has %s", strState );
    printf ( ", %i deg C", v.tempC );
    printf ( ", %u alarms since boot", v.count );
    if ( npPerfData )
    {
        printf ( "|'current temperature'=%i\n", v.tempC );
    }
    else
    {
        printf ( "| %s;%i\n", addInf, v.tempC );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ();
        if ( plotPerfData ) PlotPerfData ();
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
