/*
 * File:      check_cisco_vpn_thru.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int v0;    /* found alGeneralGaugeThroughput */
} tScanRes;

/* data type for storing data */
typedef struct
{
    uint thru; /* throughput utilized percentage */
} tDataSet;


/* global input variables */
uint warn;
uint crit;
int isWarnSet  = FALSE;
int isCritSet  = FALSE;

/*
 * SNMP base OID
 * ALTIGA-GENERAL-STATS-MIB::alGeneralGaugeThroughput.0
 * = .1.3.6.1.4.1.3076.2.1.2.25.1.4.0
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 3076, 2, 1, 2, 25, 1, 4 };
size_t baseOIDLen = 13;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_cisco_vpn_thru";
    progVersion = "0.1";
    helpIndexP = NULL;
    helpThresP = "-w WARN -c CRIT";
    helpPurpose = "Checks the throughput of a Cisco(R) VPN Concentrator.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 1 * 60;
    rrdHBeat = 3 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = FALSE;
}

void HelpSpecificParams ()
{
    printf (
    "-w, --warning INTEGER\n"
    "    Throughput utilization at which a warning is generated.\n"
    "\n"
    "-c, --critical INTEGER\n"
    "    Throughput utilization at which a critical is generated.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "Generates an alert if the throughput utilization is greater than the\n"
    "respective threshold level.\n"
    "\n"
    "All values are expected as well as returned in percent.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<utilization>\n"
    "where\n"
    "    utilization ... percentage the throughput utilized.\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'throughput utilization'=<utilization>%%;<warn>;<crit>\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (            !strcmp ( argv [(*i)], "-w" )
                ||  !strcmp ( argv [(*i)], "--warning" ) )
    {
        isWarnSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];

            errno = 0;
            warn = strtoul ( argv [(*i)], &tail, 0 );
            if ( errno || *tail ) Usage ( "Invalid 'warn' parameter" );
            (*i)++;
        }
        else Usage ( "Missing 'warn' parameter" );
    } else if ( !strcmp ( argv [(*i)], "-c" )
            ||  !strcmp ( argv [(*i)], "--critical" ) )
    {
        isCritSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];

            errno = 0;
            crit = strtoul ( argv [(*i)], &tail, 0 );
            if ( errno || *tail ) Usage ( "Invalid 'crit' parameter" );
            (*i)++;
        }
        else Usage ( "Missing 'crit' parameter" );
    } else ret = FALSE;
    return ( ret );
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 14 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            switch ( aVar->name [13] )
            {
                case 0: /* alGeneralGaugeCpuUtil.0 */
                    scanResLen = 1;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [0].v0 = ( aVar->type == ASN_GAUGE );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if ( scanRes [i].v0 )
        {
            retVal = STATE_OK;
            printf ( "OK\n" );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_thru [14];
    size_t oidLen;

    OidCpy ( o_thru, &oidLen, baseOID, &baseOIDLen );
    o_thru [13] = 0;
    oidLen = 14;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.thru  = ReadGauge32 ( sess2, o_thru, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "thru",  GAUGE, rrdHBeat, "-1", "101" );
    params [i++] = RRA_MAX_7H_1M;
    params [i++] = RRA_AVG_7H_1M;
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;

    chk = sscanf ( aData, "%u", &(v.thru) );
    if ( chk != 1 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* thru   = UIntToStr   ( v.thru );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "thru" );
    params [i++] = StrConcat ( 3, ts, ":", thru );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( thru );
    free ( ts );
}

void PlotDataUtilization ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[%]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, NULL, aTiming );
    PPAddPlot ( pp, "utilization",  "%", FALSE );
    title = StrConcat ( 2, hostName, " throughput utilization" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 100, FALSE );
    params [i++] = CreateDEF ( "thruM", aFN, "thru", MAX );
    params [i++] = CreateDEF ( "thruA", aFN, "thru", AVG );
    params [i++] = CreatePlot ( PLOT_LINE3, "thruM", C_THRU_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "thruA", C_THRU_AVG, pp, 0, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "thruA", "thruA", "thruM", "thruM", "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataUtilization ( fn, PLOT_6H_1M,   now );
    PlotDataUtilization ( fn, PLOT_24H_5M,  now );
    PlotDataUtilization ( fn, PLOT_7D_15M,  now );
    PlotDataUtilization ( fn, PLOT_31D_60M, now );
    PlotDataUtilization ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    int retVal = STATE_OK;

    ReadVals ();
    if ( warn < v.thru ) retVal = MAXIMUM ( retVal, STATE_WARNING );
    if ( crit < v.thru ) retVal = MAXIMUM ( retVal, STATE_WARNING );
    PrintExit ( retVal );
    printf ( " - throughput utilization: %u%%", v.thru );
    if ( npPerfData )
    {
        printf ( "|'throughput utilization'=%u%%;%u;%u\n", v.thru, warn, crit );
    }
    else
    {
        printf ( "|%u\n", v.thru );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( checkData )
    {
        if ( !isWarnSet ) Usage ( "Missing 'warn' parameters" );
        if ( !isCritSet ) Usage ( "Missing 'crit' parameters" );
    }
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ( NULL );
        if ( plotPerfData ) PlotPerfData ( NULL );
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
