/*
 * File:      check_cisco_vpn_voltage.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

/*
 * This does not use proper tables but (many) scalars for environmental
 * monitoring. So this code is rather unusual. If you are looking for
 * a generic example, better look somewhere else.
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"

/* data type for storing scan results */
typedef struct
{
    int v1;      /* found alHardwareCpuVoltage */
    int v2;      /* found alHardwareCpuVoltageAlarm */
    int v3;      /* found alHardwareCpuVoltageCount */
    int v5;      /* found alHardwarePs1Voltage3v */
    int v6;      /* found alHardwarePs1Voltage3vAlarm */
    int v7;      /* found alHardwarePs1Voltage3vCount */
    int v9;      /* found alHardwarePs1Voltage5v */
    int v10;     /* found alHardwarePs1Voltage5vAlarm */
    int v11;     /* found alHardwarePs1Voltage5vCount */
    int v13;     /* found alHardwarePs2Voltage3v */
    int v14;     /* found alHardwarePs2Voltage3vAlarm */
    int v15;     /* found alHardwarePs2Voltage3vCount */
    int v17;     /* found alHardwarePs2Voltage5v */
    int v18;     /* found alHardwarePs2Voltage5vAlarm */
    int v19;     /* found alHardwarePs2Voltage5vCount */
    int v21;     /* found alHardwareBoardVoltage3v */
    int v22;     /* found alHardwareBoardVoltage3vAlarm */
    int v23;     /* found alHardwareBoardVoltage3vCount */
    int v25;     /* found alHardwareBoardVoltage5v */
    int v26;     /* found alHardwareBoardVoltage5vAlarm */
    int v27;     /* found alHardwareBoardVoltage5vCount */
    int v49;     /* found alHardwarePs1Type and value == ac(2) */
    int v50;     /* found alHardwarePs2Type and value == ac(2) */
} tScanRes;

/* data type for storing data */
typedef struct
{
    uint  voltC; /* current voltage in centivolts */
    int   alarm; /* alarm status */
    uint  count; /* number of alarms since boot */
} tDataSet;


/*
 * SNMP base OID
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareCpuVoltage.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.1.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareCpuVoltageAlarm.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.2.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareCpuVoltageCount.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.3.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwarePs1Voltage3v.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.5.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwarePs1Voltage3vAlarm.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.6.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwarePs1Voltage3vCount.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.7.0 = Counter32: 0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwarePs1Voltage5v.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.9.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwarePs1Voltage5vAlarm.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.10.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwarePs1Voltage5vCount.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.11.0 = Counter32: 0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwarePs2Voltage3v.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.13.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwarePs2Voltage3vAlarm.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.14.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwarePs2Voltage3vCount.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.15.0 = Counter32: 0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwarePs2Voltage5v.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.17.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwarePs2Voltage5vAlarm.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.18.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwarePs2Voltage5vCount.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.19.0 = Counter32: 0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareBoardVoltage3v.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.21.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareBoardVoltage3vAlarm.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.22.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareBoardVoltage3vCount.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.23.0 = Counter32: 0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareBoardVoltage5v.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.25.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareBoardVoltage5vAlarm.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.26.0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwareBoardVoltage5vCount.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.27.0 = Counter32: 0
 * ALTIGA-HARDWARE-STATS-MIB::alHardwarePs1Type.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.49.0 = INTEGER: ac(2)
 * ALTIGA-HARDWARE-STATS-MIB::alHardwarePs2Type.0
 * = .1.3.6.1.4.1.3076.2.1.2.22.1.50.0 = INTEGER: none(1)
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 3076, 2, 1, 2, 22, 1 };
size_t baseOIDLen = 12;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* snmp values */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_cisco_vpn_voltage";
    progVersion = "0.1";
    helpIndexP = "-i INDEX";
    helpThresP = NULL;
    helpPurpose = "Checks the status of a voltage sensor of a Cisco(R)"
            " VPN Concentrator.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 5 * 60;
    rrdHBeat = 11 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = TRUE;
    hasAddInfParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-i, --index INTEGER\n"
    "    Index of the voltage sensor which should be checked.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin returns the state of the given voltage sensor.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<addInf>;<voltC>\n"
    "where\n"
    "    addInf ... additional information.\n"
    "    voltC  ... current voltage in millivolts.\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'current voltage'=<voltC>\n"
    "\n"
    "Note: Nagios Plugin performance data format does not allow Millivolt\n"
    "as units of measurement.\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    return FALSE;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 14 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            /*
             * Note: aVar->name [13] is expected to be 0 always but as of now
             * this is not checked.
             */
            switch ( aVar->name [12] )
            {
                case 1: /* alHardwareCpuVoltage */
                    scanResLen = 1;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [0].v1 = ( aVar->type == ASN_GAUGE );
                    scanRes [0].v2 = FALSE;
                    scanRes [0].v3 = FALSE;
                    scanRes [0].v5 = FALSE;
                    scanRes [0].v6 = FALSE;
                    scanRes [0].v7 = FALSE;
                    scanRes [0].v9 = FALSE;
                    scanRes [0].v10 = FALSE;
                    scanRes [0].v11 = FALSE;
                    scanRes [0].v13 = FALSE;
                    scanRes [0].v14 = FALSE;
                    scanRes [0].v15 = FALSE;
                    scanRes [0].v17 = FALSE;
                    scanRes [0].v18 = FALSE;
                    scanRes [0].v19 = FALSE;
                    scanRes [0].v21 = FALSE;
                    scanRes [0].v22 = FALSE;
                    scanRes [0].v23 = FALSE;
                    scanRes [0].v25 = FALSE;
                    scanRes [0].v26 = FALSE;
                    scanRes [0].v27 = FALSE;
                    scanRes [0].v49 = FALSE;
                    scanRes [0].v50 = FALSE;
                break;
                case 2: /* alHardwareCpuVoltageAlarm */
                    scanRes [0].v2 = ( aVar->type == ASN_INTEGER );
                break;
                case 3: /* alHardwareCpuVoltageCount */
                    scanRes [0].v3 = ( aVar->type == ASN_COUNTER );
                break;
                case 5: /* alHardwarePs1Voltage3v */
                    scanRes [0].v5 = ( aVar->type == ASN_GAUGE );
                break;
                case 6: /* alHardwarePs1Voltage3vAlarm */
                    scanRes [0].v6 = ( aVar->type == ASN_INTEGER );
                break;
                case 7: /* alHardwarePs1Voltage3vCount */
                    scanRes [0].v7 = ( aVar->type == ASN_COUNTER );
                break;
                case 9: /* alHardwarePs1Voltage5v */
                    scanRes [0].v9 = ( aVar->type == ASN_GAUGE );
                break;
                case 10: /* alHardwarePs1Voltage5vAlarm */
                    scanRes [0].v10 = ( aVar->type == ASN_INTEGER );
                break;
                case 11: /* alHardwarePs1Voltage5vCount */
                    scanRes [0].v11 = ( aVar->type == ASN_COUNTER );
                break;
                case 13: /* alHardwarePs2Voltage3v */
                    scanRes [0].v13 = ( aVar->type == ASN_GAUGE );
                break;
                case 14: /* alHardwarePs2Voltage3vAlarm */
                    scanRes [0].v14 = ( aVar->type == ASN_INTEGER );
                break;
                case 15: /* alHardwarePs2Voltage3vCount */
                    scanRes [0].v15 = ( aVar->type == ASN_COUNTER );
                break;
                case 17: /* alHardwarePs2Voltage5v */
                    scanRes [0].v17 = ( aVar->type == ASN_GAUGE );
                break;
                case 18: /* alHardwarePs2Voltage5vAlarm */
                    scanRes [0].v18 = ( aVar->type == ASN_INTEGER );
                break;
                case 19: /* alHardwarePs2Voltage5vCount */
                    scanRes [0].v19 = ( aVar->type == ASN_COUNTER );
                break;
                case 21: /* alHardwareBoardVoltage3v */
                    scanRes [0].v21 = ( aVar->type == ASN_GAUGE );
                break;
                case 22: /* alHardwareBoardVoltage3vAlarm */
                    scanRes [0].v22 = ( aVar->type == ASN_INTEGER );
                break;
                case 23: /* alHardwareBoardVoltage3vCount */
                    scanRes [0].v23 = ( aVar->type == ASN_COUNTER );
                break;
                case 25: /* alHardwareBoardVoltage5v */
                    scanRes [0].v25 = ( aVar->type == ASN_GAUGE );
                break;
                case 26: /* alHardwareBoardVoltage5vAlarm */
                    scanRes [0].v26 = ( aVar->type == ASN_INTEGER );
                break;
                case 27: /* alHardwareBoardVoltage5vCount */
                    scanRes [0].v27 = ( aVar->type == ASN_COUNTER );
                break;
                case 49: /* alHardwarePs1Type */
                {
                    int typeOK = ( aVar->type == ASN_INTEGER );
                    int valOK = ( *(aVar->val.integer) == 2 );

                    scanRes [0].v49 = ( typeOK && valOK );
                }
                case 50: /* alHardwarePs2Type */
                {
                    int typeOK = ( aVar->type == ASN_INTEGER );
                    int valOK = ( *(aVar->val.integer) == 2 );

                    scanRes [0].v50 = ( typeOK && valOK );
                }
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v1  && scanRes [i].v2  && scanRes [i].v3  )
        {
            retVal = STATE_OK;
            printf ( "OK index: 1 ; descr: CPU\n" );
        }
        if (    scanRes [i].v5  && scanRes [i].v6  && scanRes [i].v7  &&
                scanRes [i].v49 )
        {
            retVal = STATE_OK;
            printf ( "OK index: 2 ; descr: PS 1/3V\n" );
        }
        if (    scanRes [i].v9  && scanRes [i].v10 && scanRes [i].v11 &&
                scanRes [i].v49 )
        {
            retVal = STATE_OK;
            printf ( "OK index: 3 ; descr: PS 1/5V\n" );
        }
        if (    scanRes [i].v13 && scanRes [i].v14 && scanRes [i].v15 &&
                scanRes [i].v50 )
        {
            retVal = STATE_OK;
            printf ( "OK index: 4 ; descr: PS 2/3V\n" );
        }
        if (    scanRes [i].v17 && scanRes [i].v18 && scanRes [i].v19 &&
                scanRes [i].v50 )
        {
            retVal = STATE_OK;
            printf ( "OK index: 5 ; descr: PS 2/5V\n" );
        }
        if (    scanRes [i].v21 && scanRes [i].v22 && scanRes [i].v23 )
        {
            retVal = STATE_OK;
            printf ( "OK index: 6 ; descr: Board/3V\n" );
        }
        if (    scanRes [i].v25 && scanRes [i].v26 && scanRes [i].v27 )
        {
            retVal = STATE_OK;
            printf ( "OK index: 7 ; descr: Board/5V\n" );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_voltC [14];
    oid o_alarm [14];
    oid o_count [14];
    size_t oidLen;
    int offset = ( idx - 1 ) * 4;

    OidCpy ( o_voltC, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_alarm, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_count, &oidLen, baseOID, &baseOIDLen );
    o_voltC [12] = offset + 1; o_voltC [13] = 0;
    o_alarm [12] = offset + 2; o_alarm [13] = 0;
    o_count [12] = offset + 3; o_count [13] = 0;
    oidLen = 14;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.voltC = ReadGauge32   ( sess2, o_voltC, oidLen );
    v.alarm = ReadInteger   ( sess2, o_alarm, oidLen );
    v.count = ReadCounter32 ( sess2, o_count, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "voltC",  GAUGE, rrdHBeat, "-60.0", "60.0" );
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MIN_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MIN_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MIN_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    params [i++] = RRA_MIN_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    char buf[1024];

    chk = sscanf ( aData, "%1023s;%u", buf, &(v.voltC) );
    if ( chk != 2 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
    addInf = StrDup ( buf );
    v.voltC /= 100;
    v.alarm = 2;
    v.count = 0;
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* voltC  = DoubleToStr ( (double) v.voltC / 100.0 );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "voltC" );
    params [i++] = StrConcat ( 3, ts, ":", voltC );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( voltC );
    free ( ts );
}

void PlotDataVoltage ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[V]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, NULL, aTiming );
    PPAddPlot ( pp, "voltage",         "V", FALSE );
    title = StrConcat ( 4, hostName, " voltage ", addInf, "" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_AUTO, 0, 0, FALSE );
    params [i++] = CreateDEF ( "voltCM", aFN, "voltC", MAX );
    params [i++] = CreateDEF ( "voltCA", aFN, "voltC", AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "voltCM", C_GAUGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "voltCA", C_GAUGE1_AVG, pp, 0, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "voltCA", "voltCA", "voltCM", "voltCM", "%6.3lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataVoltage ( fn, PLOT_24H_5M,  now );
    PlotDataVoltage ( fn, PLOT_7D_15M,  now );
    PlotDataVoltage ( fn, PLOT_31D_60M, now );
    PlotDataVoltage ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    int retVal = STATE_UNKNOWN;
    char* strState = NULL;

    ReadVals ();
    if ( 1 == v.alarm )
    {
        retVal = STATE_CRITICAL;
        strState = "an alarm";
    }
    else if ( 2 == v.alarm )
    {
        retVal = STATE_OK;
        strState = "no alarm";
    }
    else
    {
        retVal = STATE_UNKNOWN;
        strState = "an invalid alarm state";
    }
    PrintExit ( retVal );
    printf ( " - voltage has %s", strState );
    printf ( ", %.2f V", (double) v.voltC / 100.0 );
    printf ( ", %u alarms since boot", v.count );
    if ( npPerfData )
    {
        printf ( "|'current voltage'=%u\n", v.voltC * 10 );
    }
    else
    {
        printf ( "| %s;%u\n", addInf, v.voltC * 10 );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ();
        if ( plotPerfData ) PlotPerfData ();
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
