/*
 * File:      check_cpq_fan.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "decoders.h"
#include "globals.h"
#include "nagiosif.h"
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int chass;   /* chassis index */
    int index;   /* fan index */
    int v3;      /* found cpqHeFltTolFanLocale */
    int v4;      /* found cpqHeFltTolFanPresent and is OK */
    int v5;      /* found cpqHeFltTolFanType */
    int v6;      /* found cpqHeFltTolFanSpeed */
    int v9;      /* found cpqHeFltTolFanCondition */
    int v10;     /* found cpqHeFltTolFanHotPlug */
    char* local; /* fan locale */
} tScanRes;

/* data type for storing data */
typedef struct
{
    int   local; /* locale */
    int   prsnt; /* presence */
    int   type;
    int   speed;
    int   cond;  /* condition */
    int   hotP;  /* hot pluggable */
} tDataSet;


/* global input variables */
int chass;
int isChassSet  = FALSE;

/*
 * SNMP base OID
 * CPQHLTH-MIB::cpqHeFltTolFanLocale.CHA.IDX
 * = .1.3.6.1.4.1.232.6.2.6.7.1.3.CHA.IDX
 * CPQHLTH-MIB::cpqHeFltTolFanPresent.CHA.IDX
 * = .1.3.6.1.4.1.232.6.2.6.7.1.4.CHA.IDX
 * CPQHLTH-MIB::cpqHeFltTolFanType.CHA.IDX
 * = .1.3.6.1.4.1.232.6.2.6.7.1.5.CHA.IDX
 * CPQHLTH-MIB::cpqHeFltTolFanSpeed.CHA.IDX
 * = .1.3.6.1.4.1.232.6.2.6.7.1.6.CHA.IDX
 * CPQHLTH-MIB::cpqHeFltTolFanCondition.CHA.IDX
 * = .1.3.6.1.4.1.232.6.2.6.7.1.9.CHA.IDX
 * CPQHLTH-MIB::cpqHeFltTolFanHotPlug.CHA.IDX
 * = .1.3.6.1.4.1.232.6.2.6.7.1.10.CHA.IDX
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 232, 6, 2, 6, 7, 1 };
size_t baseOIDLen = 12;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_cpq_fan";
    progVersion = "0.8";
    helpIndexP = "-s SYSTEM -i INDEX";
    helpThresP = NULL;
    helpPurpose = "Checks the status of a fan of Compaq/HP server.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
    /* rrdStep is unused */
    /* rrdHBeat is unused */
    supportsPerfLog = FALSE;
    hasIndexParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-s, --system-chassis INTEGER\n"
    "    Index of the system chassis.\n"
    "\n"
    "-i, --index INTEGER\n"
    "    Index of the fan which should be checked.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin returns the overall state of the fan.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "There is no performance data provided.\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-s" )
            ||  !strcmp ( argv [(*i)], "--system-chassis" ) )
    {
        isChassSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            chass = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || *tail )
                    Usage ( "Invalid 'system chassis' parameter" );
        }
        else Usage ( "Missing 'system chassis' parameter(s)" );
    } else ret = FALSE;
    return ( ret );
}

int ScanResIndex ( int aChass, int aIndex )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].chass == aChass )
                && ( scanRes [retVal].index == aIndex );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 15 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [12] )
            {
                case 2: /* cpqHeFltTolFanIndex */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].chass = aVar->name [13];
                    scanRes [index].index = aVar->name [14];
                    scanRes [index].v3 = FALSE;
                    scanRes [index].v4 = FALSE;
                    scanRes [index].v5 = FALSE;
                    scanRes [index].v6 = FALSE;
                    scanRes [index].v9 = FALSE;
                    scanRes [index].v10 = FALSE;
                    scanRes [index].local = NULL;
                break;
                case 3: /* cpqHeFltTolFanLocale */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    if ( ( aVar->type == ASN_INTEGER ) )
                    {
                        scanRes [index].v3 = TRUE;
                        scanRes [index].local =
                                DecodeCPQLocal ( *(aVar->val.integer) );
                    } /* else nothing to do */
                break;
                case 4: /* cpqHeFltTolFanPresent */
                {
                    int typeOK = ( aVar->type == ASN_INTEGER );
                    int valOK = ( *(aVar->val.integer) == 3 );

                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v4 = ( typeOK && valOK );
                }
                break;
                case 5: /* cpqHeFltTolFanType */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v5 = ( aVar->type == ASN_INTEGER );
                break;
                case 6: /* cpqHeFltTolFanSpeed */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v6 = ( aVar->type == ASN_INTEGER );
                break;
                case 9: /* cpqHeFltTolFanCondition */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v9 = ( aVar->type == ASN_INTEGER );
                break;
                case 10: /* cpqHeFltTolFanHotPlug */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v10 = ( aVar->type == ASN_INTEGER );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v3 && scanRes [i].v4 && scanRes [i].v5 &&
                scanRes [i].v6 && scanRes [i].v9 && scanRes [i].v10 )
        {
            retVal = STATE_OK;
            printf ( "OK chassis: %i ; index: %i ; local: %s\n",
                    scanRes [i].chass, scanRes [i].index, scanRes [i].local );
            free ( scanRes [i].local );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_local [15];
    oid o_prsnt [15];
    oid o_type  [15];
    oid o_speed [15];
    oid o_cond  [15];
    oid o_hotP  [15];
    size_t oidLen;

    OidCpy ( o_local, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_prsnt, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_type,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_speed, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_cond,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_hotP,  &oidLen, baseOID, &baseOIDLen );
    o_local [12] =  3; o_local [13] = chass; o_local [14] = idx;
    o_prsnt [12] =  4; o_prsnt [13] = chass; o_prsnt [14] = idx;
    o_type  [12] =  5; o_type  [13] = chass; o_type  [14] = idx;
    o_speed [12] =  6; o_speed [13] = chass; o_speed [14] = idx;
    o_cond  [12] =  9; o_cond  [13] = chass; o_cond  [14] = idx;
    o_hotP  [12] = 10; o_hotP  [13] = chass; o_hotP  [14] = idx;
    oidLen = 15;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.local = ReadInteger ( sess2, o_local, oidLen );
    v.prsnt = ReadInteger ( sess2, o_prsnt, oidLen );
    v.type  = ReadInteger ( sess2, o_type,  oidLen );
    v.speed = ReadInteger ( sess2, o_speed, oidLen );
    v.cond  = ReadInteger ( sess2, o_cond,  oidLen );
    v.hotP  = ReadInteger ( sess2, o_hotP,  oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

void CreateDataFile ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

void ParseData ( const char* aData )
{
    /* performance data is not supported */
}

void WriteData ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

char* DecodeCPQFanType ( int aType )
{
    char* retVal = NULL;

    switch ( aType )
    {
        case  1: retVal = StrDup ( "unknown" );            break;
        case  2: retVal = StrDup ( "temperature driven" ); break;
        case  3: retVal = StrDup ( "spin detecting" );     break;
        default: retVal = StrDup ( "invalid" );            break;
    }
    return retVal;
}

char* DecodeCPQFanSpeed ( int aSpeed )
{
    char* retVal = NULL;

    switch ( aSpeed )
    {
        case  1: retVal = StrDup ( "unknown" ); break;
        case  2: retVal = StrDup ( "normal" );  break;
        case  3: retVal = StrDup ( "high" );    break;
        default: retVal = StrDup ( "invalid" ); break;
    }
    return retVal;
}

char* DecodeCPQHotplug ( int aHotP )
{
    char* retVal = NULL;

    switch ( aHotP )
    {
        case  1: retVal = StrDup ( "unknown" ); break;
        case  2: retVal = StrDup ( "without" ); break;
        case  3: retVal = StrDup ( "with" );    break;
        default: retVal = StrDup ( "invalid" ); break;
    }
    return retVal;
}

char* Evaluate ( int* aRetVal )
{
    char* retMsg = NULL;

    switch ( v.prsnt )
    {
        case  1:
            retMsg = StrDup ( "presence of the fan could not be determined" );
            *aRetVal = STATE_WARNING;
        break;
        case  2:
            retMsg = StrDup ( "fan is not present" );
            *aRetVal = STATE_WARNING;
        break;
        case  3:
        {
            char* local = DecodeCPQLocal    ( v.local );
            char* type  = DecodeCPQFanType  ( v.type );
            char* speed = DecodeCPQFanSpeed ( v.speed );
            char* hotP  = DecodeCPQHotplug  ( v.hotP );

            switch ( v.cond )
            {
                case 1:  *aRetVal = STATE_WARNING;  break;
                case 2:  *aRetVal = STATE_OK;       break;
                case 3:  *aRetVal = STATE_WARNING;  break;
                case 4:  *aRetVal = STATE_CRITICAL; break;
                default: *aRetVal = STATE_UNKNOWN;  break;
            }
            retMsg = StrConcat ( 8, local, " fan, ", type, " type, ", speed,
                    " speed, ", hotP, " hot plug capability" );
            free ( hotP );
            free ( speed );
            free ( type );
            free ( local );
        }
        break;
        default:
            retMsg = StrDup ( "presence of this fan is invalid" );
            *aRetVal = STATE_WARNING;
        break;
    }
    return retMsg;
}

int CheckData ()
{
    int retVal = STATE_UNKNOWN;
    char* retStr = NULL;

    ReadVals ();
    retStr = Evaluate ( &retVal );
    PrintExit ( retVal );
    printf ( " - %s\n", retStr );
    free ( retStr );
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( checkData )
    {
        if ( !isChassSet ) Usage ( "Missing 'system chassis' parameters" );
    }
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
