/*
 * File:      check_cpq_fcaaccel.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int index;   /* box index */
    int slot;    /* IO slot index */
    int v3;      /* found cpqFcaAccelStatus */
    int v4;      /* found cpqFcaAccelBadData */
    int v5;      /* found cpqFcaAccelErrCode */
    int v6;      /* found cpqFcaAccelBatteryStatus */
    int v9;      /* found cpqFcaAccelCondition */
    char* model; /* model name */
} tScanRes;

/* data type for storing data */
typedef struct
{
    int   stat;  /* status */
    int   badD;  /* bad data */
    int   errC;  /* error code */
    int   batS;  /* battery status */
    int   cond;  /* condition */
} tDataSet;


/* global input variables */
int slot;
int isSlotSet  = FALSE;

/*
 * SNMP base OID
 * CPQFCA-MIB::cpqFcaAccelBoxIndex.BOX.SLOT =
 * .1.3.6.1.4.1.232.16.2.2.2.1.1.BOX.SLOT =
 * CPQFCA-MIB::cpqFcaAccelBoxIoSlot.BOX.SLOT =
 * .1.3.6.1.4.1.232.16.2.2.2.1.2.BOX.SLOT =
 * CPQFCA-MIB::cpqFcaAccelStatus.BOX.SLOT =
 * .1.3.6.1.4.1.232.16.2.2.2.1.3.BOX.SLOT =
 * CPQFCA-MIB::cpqFcaAccelBadData.BOX.SLOT =
 * .1.3.6.1.4.1.232.16.2.2.2.1.4.BOX.SLOT =
 * CPQFCA-MIB::cpqFcaAccelErrCode.BOX.SLOT =
 * .1.3.6.1.4.1.232.16.2.2.2.1.5.BOX.SLOT =
 * CPQFCA-MIB::cpqFcaAccelBatteryStatus.BOX.SLOT =
 * .1.3.6.1.4.1.232.16.2.2.2.1.6.BOX.SLOT =
 * CPQFCA-MIB::cpqFcaAccelCondition.BOX.SLOT =
 * .1.3.6.1.4.1.232.16.2.2.2.1.9.BOX.SLOT =
 */
oid baseOID [] = { 1, 3, 6, 1, 4, 1, 232, 16, 2, 2, 2, 1 };
size_t baseOIDLen = 12;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_cpq_fcaaccel";
    progVersion = "0.2";
    helpIndexP = "-i INDEX -s SLOT";
    helpThresP = NULL;
    helpPurpose =
            "Checks the status of a Compaq/HP External Array Accelerator.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
    /* rrdStep is unused */
    /* rrdHBeat is unused */
    supportsPerfLog = FALSE;
    hasIndexParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-i, --index INTEGER\n"
    "    Index of the storage box the accelerator resides in.\n"
    "\n"
    "-s, --slot INTEGER\n"
    "    Index of the physical IO slot the accelerator resides in.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin returns the overall state of the array accelerator.\n"
    "Additional information like battery status and error code are written\n"
    "to stdout but they are not considered when determining the exit code.\n"
    "The accel board is expected to consider those values when calculating\n"
    "its condition.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "There is no performance data provided.\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-s" )
            ||  !strcmp ( argv [(*i)], "--slot" ) )
    {
        isSlotSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            slot = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || *tail )
                    Usage ( "Invalid 'slot' parameter" );
        }
        else Usage ( "Missing 'slot' parameter(s)" );
    } else ret = FALSE;
    return ( ret );
}

int ScanResIndex ( int aIndex, int aSlot )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex )
                && ( scanRes [retVal].slot == aSlot );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 15 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [12] )
            {
                case 1: /* cpqFcaAccelBoxIndex */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].index = aVar->name [13];
                    scanRes [index].slot = aVar->name [14];
                    scanRes [index].v3 = FALSE;
                    scanRes [index].v4 = FALSE;
                    scanRes [index].v5 = FALSE;
                    scanRes [index].v6 = FALSE;
                    scanRes [index].v9 = FALSE;
                break;
                case 3: /* cpqFcaAccelStatus */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v3 = ( aVar->type == ASN_INTEGER );
                break;
                case 4: /* cpqFcaAccelBadData */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v4 = ( aVar->type == ASN_INTEGER );
                break;
                case 5: /* cpqFcaAccelErrCode */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v5 = ( aVar->type == ASN_INTEGER );
                break;
                case 6: /* cpqFcaAccelBatteryStatus */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v6 = ( aVar->type == ASN_INTEGER );
                break;
                case 9: /* cpqFcaAccelCondition */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v9 = ( aVar->type == ASN_INTEGER );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v3 && scanRes [i].v4 && scanRes [i].v5 &&
                scanRes [i].v6 && scanRes [i].v9 )
        {
            retVal = STATE_OK;
            printf ( "OK index: %i ; slot: %i\n",
                    scanRes [i].index, scanRes [i].slot );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_stat [15];
    oid o_badD [15];
    oid o_errC [15];
    oid o_batS [15];
    oid o_cond [15];
    size_t oidLen;

    OidCpy ( o_stat, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_badD, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_errC, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_batS, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_cond, &oidLen, baseOID, &baseOIDLen );
    o_stat [12] = 3; o_stat [13] = idx; o_stat [14] = slot;
    o_badD [12] = 4; o_badD [13] = idx; o_badD [14] = slot;
    o_errC [12] = 5; o_errC [13] = idx; o_errC [14] = slot;
    o_batS [12] = 6; o_batS [13] = idx; o_batS [14] = slot;
    o_cond [12] = 9; o_cond [13] = idx; o_cond [14] = slot;
    oidLen = 15;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.stat = ReadInteger ( sess2, o_stat, oidLen );
    v.badD = ReadInteger ( sess2, o_badD, oidLen );
    v.errC = ReadInteger ( sess2, o_errC, oidLen );
    v.batS = ReadInteger ( sess2, o_batS, oidLen );
    v.cond = ReadInteger ( sess2, o_cond, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

void CreateDataFile ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

void ParseData ( const char* aData )
{
    /* performance data is not supported */
}

void WriteData ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

int CheckData ()
{
    int retVal = STATE_OK;
    char* stat = NULL;
    char* badD = NULL;
    char* errC = NULL;
    char* batS = NULL;
    char* cond = NULL;

    ReadVals ();
    switch ( v.stat )
    {
        case 1:  stat = "'other'";        break;
        case 2:  stat = "'invalid'";      break;
        case 3:  stat = "'enabled'";      break;
        case 4:  stat = "'tmpDisabled'";  break;
        case 5:  stat = "'permDisabled'"; break;
        default: stat = "invalid";        break;
    }
    switch ( v.badD )
    {
        case 1:  badD = "'other'";    break;
        case 2:  badD = "'none'";     break;
        case 3:  badD = "'possible'"; break;
        default: badD = "invalid";    break;
    }
    switch ( v.errC )
    {
        case  1: errC = "'other'";                 break;
        case  2: errC = "'invalid' (that's good)"; break;
        case  3: errC = "'badConfig'";             break;
        case  4: errC = "'lowBattery'";            break;
        case  5: errC = "'disableCmd'";            break;
        case  6: errC = "'noResources'";           break;
        case  7: errC = "'notConnected'";          break;
        case  8: errC = "'badMirrorData'";         break;
        case  9: errC = "'readErr'";               break;
        case 10: errC = "'writeErr'";              break;
        case 11: errC = "'configCmd'";             break;
        case 12: errC = "'expandInProgress'";      break;
        case 13: errC = "'snapshotInProgress'";    break;
        case 14: errC = "'redundantLowBattery'";   break;
        case 15: errC = "'redundantSizeMismatch'"; break;
        case 16: errC = "'redundantCacheFailure'"; break;
        case 17: errC = "'excessiveEccErrors'";    break;
        case 19: errC = "'postEccErrors'";         break;
        default: errC = "invalid";                 break;
    }
    switch ( v.batS )
    {
        case 1:  batS = "'other'";      break;
        case 2:  batS = "'ok'";         break;
        case 3:  batS = "'recharging'"; break;
        case 4:  batS = "'failed'";     break;
        case 5:  batS = "'degraded'";   break;
        case 6:  batS = "'notPresent'"; break;
        default: batS = "invalid";      break;
    }
    switch ( v.cond )
    {
        case 1:  retVal = STATE_WARNING;  cond = "'other'";    break;
        case 2:  retVal = STATE_OK;       cond = "'ok'";       break;
        case 3:  retVal = STATE_WARNING;  cond = "'degraded'"; break;
        case 4:  retVal = STATE_CRITICAL; cond = "'failed'";   break;
        default: retVal = STATE_UNKNOWN;  cond = "invalid";    break;
    }

    PrintExit ( retVal );
    printf ( " - Condition: %s, Status: %s, BadData: %s", cond, stat, badD );
    printf ( ", ErrCode: %s, BatteryStatus: %s\n", errC, batS );
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( checkData )
    {
        if ( !isSlotSet ) Usage ( "Missing 'slot' parameters" );
    }
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
