/*
 * File:      check_cpq_fcaeac.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int index;   /* box index */
    int slot;    /* IO slot index */
    int v3;      /* found cpqFcaCntlrModel */
    int v5;      /* found cpqFcaCntlrStatus */
    int v6;      /* found cpqFcaCntlrCondition */
    char* model; /* model name */
} tScanRes;

/* data type for storing data */
typedef struct
{
    int   stat;  /* status */
    int   cond;  /* condition */
} tDataSet;


/* global input variables */
int slot;
int isSlotSet  = FALSE;

/*
 * SNMP base OID
 * CPQFCA-MIB::cpqFcaCntlrBoxIndex.BOX.SLOT
 * = 1.3.6.1.4.1.232.16.2.2.1.1.1.BOX.SLOT
 * CPQFCA-MIB::cpqFcaCntlrBoxIoSlot.BOX.SLOT
 * = 1.3.6.1.4.1.232.16.2.2.1.1.2.BOX.SLOT
 * CPQFCA-MIB::cpqFcaCntlrModel.BOX.SLOT
 * = 1.3.6.1.4.1.232.16.2.2.1.1.3.BOX.SLOT
 * CPQFCA-MIB::cpqFcaCntlrStatus.BOX.SLOT
 * = 1.3.6.1.4.1.232.16.2.2.1.1.5.BOX.SLOT
 * CPQFCA-MIB::cpqFcaCntlrCondition.BOX.SLOT
 * = 1.3.6.1.4.1.232.16.2.2.1.1.6.BOX.SLOT
 */
oid baseOID [] = { 1, 3, 6, 1, 4, 1, 232, 16, 2, 2, 1, 1 };
size_t baseOIDLen = 12;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_cpq_eac";
    progVersion = "0.2";
    helpIndexP = "-i INDEX -s SLOT";
    helpThresP = NULL;
    helpPurpose =
            "Checks the status of a Compaq/HP External Array Controller.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
    /* rrdStep is unused */
    /* rrdHBeat is unused */
    supportsPerfLog = FALSE;
    hasIndexParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-i, --index INTEGER\n"
    "    Index of the storage box the controller resides in.\n"
    "\n"
    "-s, --slot INTEGER\n"
    "    Index of the physical IO slot the controller resides in.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin returns the overall state of a controller and array\n"
    "accelerators.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "There is no performance data provided.\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-s" )
            ||  !strcmp ( argv [(*i)], "--slot" ) )
    {
        isSlotSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            slot = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || *tail )
                    Usage ( "Invalid 'slot' parameter" );
        }
        else Usage ( "Missing 'slot' parameter(s)" );
    } else ret = FALSE;
    return ( ret );
}

char* DecodeCPQEacModel ( int aType )
{
    char* retVal = NULL;

    switch ( aType )
    {
        case  1:
            retVal = StrDup ( "other" );
        break;
        case  2:
            retVal = StrDup ( "RAID Array 4000" );
        break;
        case  3:
            retVal = StrDup ( "Modular Smart Array 1000" );
        break;
        case  4:
            retVal = StrDup ( "Modular Smart Array 500 Controller" );
        break;
        case  5:
            retVal = StrDup ( "HSG80 RAID Array Controller" );
        break;
        case  6:
            retVal = StrDup ( "HSV110 Virtual Array Controller" );
        break;
        case  7:
            retVal = StrDup ( "Modular Smart Array 500 G2 Controller" );
        break;
        case  8:
            retVal = StrDup ( "Modular Smart Array 20 Controller" );
        break;
        default:
            retVal = StrDup ( "invalid" );
        break;
    }
    return retVal;
}

int ScanResIndex ( int aIndex, int aSlot )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex )
                && ( scanRes [retVal].slot == aSlot );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 15 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [12] )
            {
                case 1: /* cpqFcaCntlrBoxIndex */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].index = aVar->name [13];
                    scanRes [index].slot = aVar->name [14];
                    scanRes [index].v3 = FALSE;
                    scanRes [index].v5 = FALSE;
                    scanRes [index].v6 = FALSE;
                    scanRes [index].model = NULL;
                break;
                case 3: /* cpqFcaCntlrModel */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    if ( ( aVar->type == ASN_INTEGER ) )
                    {
                        scanRes [index].v3 = TRUE;
                        scanRes [index].model =
                                DecodeCPQEacModel ( *(aVar->val.integer) );
                    } /* else nothing to do */
                break;
                case 5: /* cpqFcaCntlrStatus */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v5 = ( aVar->type == ASN_INTEGER );
                break;
                case 6: /* cpqFcaCntlrCondition */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v6 = ( aVar->type == ASN_INTEGER );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if ( scanRes [i].v3 && scanRes [i].v5 && scanRes [i].v6 )
        {
            retVal = STATE_OK;
            printf ( "OK index: %i ; slot: %i ; model: %s\n",
                    scanRes [i].index, scanRes [i].slot, scanRes [i].model );
            free ( scanRes [i].model );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_stat [15];
    oid o_cond [15];
    size_t oidLen;

    OidCpy ( o_stat, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_cond, &oidLen, baseOID, &baseOIDLen );
    o_stat [12] = 5; o_stat [13] = idx; o_stat [14] = slot;
    o_cond [12] = 6; o_cond [13] = idx; o_cond [14] = slot;
    oidLen = 15;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.stat = ReadInteger ( sess2, o_stat, oidLen );
    v.cond = ReadInteger ( sess2, o_cond, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

void CreateDataFile ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

void ParseData ( const char* aData )
{
    /* performance data is not supported */
}

void WriteData ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

int CheckData ()
{
    int retVal = STATE_OK;
    char* cond = NULL;
    char* stat = NULL;

    ReadVals ();
    switch ( v.cond )
    {
        case 1:  retVal = STATE_WARNING;  cond = "'other'";    break;
        case 2:  retVal = STATE_OK;       cond = "'ok'";       break;
        case 3:  retVal = STATE_WARNING;  cond = "'degraded'"; break;
        case 4:  retVal = STATE_CRITICAL; cond = "'failed'";   break;
        default: retVal = STATE_UNKNOWN;  cond = "invalid";    break;
    }
    switch ( v.stat )
    {
        case 1:
            retVal = MAXIMUM ( retVal, STATE_WARNING);
            stat = "'other'";
        break;
        case 2:
            stat = "'ok'";
        break;
        case 3:
            retVal = MAXIMUM ( retVal, STATE_CRITICAL);
            stat = "'failed'";
        break;
        case 4:
            retVal = MAXIMUM ( retVal, STATE_CRITICAL);
            stat = "'offline'";
        break;
        case 5:
            retVal = MAXIMUM ( retVal, STATE_WARNING);
            stat = "'redundantPathOffline'";
        break;
        default:
            retVal = MAXIMUM ( retVal, STATE_UNKNOWN);
            stat = "invalid";
        break;
    }
    PrintExit ( retVal );
    printf ( " - Condition is %s, state is %s\n", cond, stat );
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( checkData )
    {
        if ( !isSlotSet ) Usage ( "Missing 'slot' parameters" );
    }
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
