/*
 * File:      check_cpq_fcalogdrv.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int index;   /* box index */
    int lDrv;    /* logical drive index */
    int v4;      /* found cpqFcaLogDrvStatus */
    int v6;      /* found cpqFcaLogDrvPercentRebuild */
    int v11;     /* found cpqFcaLogDrvCondition */
    char* fTol;  /* Fault tolerance */
} tScanRes;

/* data type for storing data */
typedef struct
{
    int   stat;  /* status */
    uint  perc;  /* percentage rebuilding/expanding is done */
    int   cond;  /* condition */
} tDataSet;


/* global input variables */
int lDrv;
int isLDrvSet  = FALSE;

/*
 * SNMP base OID
 * CPQFCA-MIB::cpqFcaLogDrvBoxIndex.BOX.LDRV
 * = .1.3.6.1.4.1.232.16.2.3.1.1.1.BOX.LDRV
 * CPQFCA-MIB::cpqFcaLogDrvIndex.BOX.LDRV
 * = .1.3.6.1.4.1.232.16.2.3.1.1.2.BOX.LDRV
 * CPQFCA-MIB::cpqFcaLogDrvFaultTol.BOX.LDRV
 * = .1.3.6.1.4.1.232.16.2.3.1.1.3.BOX.LDRV
 * CPQFCA-MIB::cpqFcaLogDrvStatus.BOX.LDRV
 * = .1.3.6.1.4.1.232.16.2.3.1.1.4.BOX.LDRV
 * CPQFCA-MIB::cpqFcaLogDrvPercentRebuild.BOX.LDRV
 * = .1.3.6.1.4.1.232.16.2.3.1.1.6.BOX.LDRV
 * CPQFCA-MIB::cpqFcaLogDrvCondition.BOX.LDRV
 * = .1.3.6.1.4.1.232.16.2.3.1.1.11.BOX.LDRV
 */
oid baseOID [] = { 1, 3, 6, 1, 4, 1, 232, 16, 2, 3, 1, 1 };
size_t baseOIDLen = 12;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_cpq_fcalogdrv";
    progVersion = "0.2";
    helpIndexP = "-i INDEX -d DRIVE";
    helpThresP = NULL;
    helpPurpose =
            "Checks the status of a Compaq/HP External Array logical drive.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
    /* rrdStep is unused */
    /* rrdHBeat is unused */
    supportsPerfLog = FALSE;
    hasIndexParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-i, --index INTEGER\n"
    "    Index of the storage box the logical drive resides in.\n"
    "\n"
    "-d, --drive INTEGER\n"
    "    Index of the logical drive.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin returns the overall condition of a logical drive and any\n"
    "associated physical drives.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "There is no performance data provided.\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-d" )
            ||  !strcmp ( argv [(*i)], "--drive" ) )
    {
        isLDrvSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            lDrv = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || *tail )
                    Usage ( "Invalid 'drive' parameter" );
        }
        else Usage ( "Missing 'drive' parameter(s)" );
    } else ret = FALSE;
    return ( ret );
}

char* DecodeCPQFcaFTol ( int aFTol )
{
    char* retVal = NULL;

    switch ( aFTol )
    {
        case  1:
            retVal = StrDup ( "other" );
        break;
        case  2:
            retVal = StrDup ( "none" );
        break;
        case  3:
            retVal = StrDup ( "mirroring (RAID 1)" );
        break;
        case  4:
            retVal = StrDup ( "Data Guard (RAID 4)" );
        break;
        case  5:
            retVal = StrDup ( "Distributed Data Guard (RAID 5)" );
        break;
        case  7:
            retVal = StrDup ( "Advanced Data Guard (RAID ADG)" );
        break;
        default:
            retVal = StrDup ( "invalid" );
        break;
    }
    return retVal;
}

int ScanResIndex ( int aIndex, int aLDrv )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex )
                && ( scanRes [retVal].lDrv == aLDrv );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 15 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [12] )
            {
                case 1: /* cpqFcaLogDrvBoxIndex */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].index = aVar->name [13];
                    scanRes [index].lDrv = aVar->name [14];
                    scanRes [index].v4 = FALSE;
                    scanRes [index].v6 = FALSE;
                    scanRes [index].v11 = FALSE;
                    scanRes [index].fTol = NULL;
                break;
                case 3: /* cpqFcaLogDrvFaultTol */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    if ( ( aVar->type == ASN_INTEGER ) )
                    {
                        scanRes [index].fTol =
                                DecodeCPQFcaFTol ( *(aVar->val.integer) );
                    } /* else nothing to do */
                break;
                case 4: /* cpqFcaLogDrvStatus */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v4 = ( aVar->type == ASN_INTEGER );
                break;
                case 6: /* cpqFcaLogDrvPercentRebuild */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v6 = ( aVar->type == ASN_GAUGE );
                break;
                case 11: /* cpqFcaLogDrvCondition */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v11 = ( aVar->type == ASN_INTEGER );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if ( scanRes [i].v4  && scanRes [i].v6  && scanRes [i].v11 )
        {
            retVal = STATE_OK;
            printf ( "OK index: %i ; drive: %i ; faultTolerance: %s\n",
                    scanRes [i].index, scanRes [i].lDrv, scanRes [i].fTol );
            free ( scanRes [i].fTol );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_stat [15];
    oid o_perc [15];
    oid o_cond [15];
    size_t oidLen;

    OidCpy ( o_stat, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_perc, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_cond, &oidLen, baseOID, &baseOIDLen );
    o_stat [12] =  4; o_stat [13] = idx; o_stat [14] = lDrv;
    o_perc [12] =  6; o_perc [13] = idx; o_perc [14] = lDrv;
    o_cond [12] = 11; o_cond [13] = idx; o_cond [14] = lDrv;
    oidLen = 15;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.stat = ReadInteger ( sess2, o_stat, oidLen );
    v.perc = ReadGauge32 ( sess2, o_perc, oidLen );
    v.cond = ReadInteger ( sess2, o_cond, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

void CreateDataFile ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

void ParseData ( const char* aData )
{
    /* performance data is not supported */
}

void WriteData ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

int CheckData ()
{
    int retVal = STATE_OK;
    char* cond = NULL;
    char* stat = NULL;

    ReadVals ();
    switch ( v.cond )
    {
        case 1:  retVal = STATE_WARNING;  cond = "'other'";    break;
        case 2:  retVal = STATE_OK;       cond = "'ok'";       break;
        case 3:  retVal = STATE_WARNING;  cond = "'degraded'"; break;
        case 4:  retVal = STATE_CRITICAL; cond = "'failed'";   break;
        default: retVal = STATE_UNKNOWN;  cond = "invalid";    break;
    }
    switch ( v.stat )
    {
        case  1: stat = StrDup ( "'other'" );              break;
        case  2: stat = StrDup ( "'ok'" );                 break;
        case  3: stat = StrDup ( "'failed'" );             break;
        case  4: stat = StrDup ( "'unconfigured'" );       break;
        case  5: stat = StrDup ( "'recovering'" );         break;
        case  6: stat = StrDup ( "'readyForRebuild'" );    break;
        case  7:
        {
            char* perc  = IntToStr    ( v.perc );

            stat = StrConcat ( 3, "rebuilding (%", perc, "%)" );
            free ( perc );
        }
        break;
        case  8: stat = StrDup ( "'wrongDrive'" );         break;
        case  9: stat = StrDup ( "'badConnect'" );         break;
        case 10: stat = StrDup ( "'overheating'" );        break;
        case 11: stat = StrDup ( "'shutdown'" );           break;
        case 12:
        {
            char* perc  = IntToStr    ( v.perc );

            stat = StrConcat ( 3, "expanding (", perc, "%)" );
            free ( perc );
        }
        break;
        case 13: stat = StrDup ( "'notAvailable'" );       break;
        case 14: stat = StrDup ( "'queuedForExpansion'" ); break;
        case 15: stat = StrDup ( "'hardError'" );          break;
        default: stat = StrDup ( "invalid" );              break;
    }
    PrintExit ( retVal );
    printf ( " - Condition is %s, state is %s\n", cond, stat );
    free ( stat );
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( checkData )
    {
        if ( !isLDrvSet ) Usage ( "Missing 'drive' parameters" );
    }
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
