/*
 * File:      check_cpq_fcaspare.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int index;   /* box index */
    int sDrv;    /* spare drive index */
    int v3;      /* found cpqFcaSpareStatus */
    int v4;      /* found cpqFcaSpareReplacedDrvBay */
    int v5;      /* found cpqFcaSparePercentRebuild */
    int v6;      /* found cpqFcaSpareCondition */
    int v8;      /* found cpqFcaSpareBay */
    int bay;     /* drive bay */
} tScanRes;

/* data type for storing data */
typedef struct
{
    int   stat;  /* status */
    int   rBay;  /* replaced drive in bay */
    uint  perc;  /* percentage rebuilding is done */
    int   cond;  /* condition */
    int   bay;   /* drive bay */
} tDataSet;


/* global input variables */
int sDrv;
int isSDrvSet  = FALSE;

/*
 * SNMP base OID
 * CPQFCA-MIB::cpqFcaSpareBoxIndex.BOX.SDRV
 * = .1.3.6.1.4.1.232.16.2.4.1.1.1.BOX.SDRV
 * CPQFCA-MIB::cpqFcaSparePhyDrvIndex.BOX.SDRV
 * = .1.3.6.1.4.1.232.16.2.4.1.1.2.BOX.SDRV
 * CPQFCA-MIB::cpqFcaSpareStatus.BOX.SDRV
 * = .1.3.6.1.4.1.232.16.2.4.1.1.3.BOX.SDRV
 * CPQFCA-MIB::cpqFcaSpareReplacedDrvBay.BOX.SDRV
 * = .1.3.6.1.4.1.232.16.2.4.1.1.4.BOX.SDRV
 * CPQFCA-MIB::cpqFcaSparePercentRebuild.BOX.SDRV
 * = .1.3.6.1.4.1.232.16.2.4.1.1.5.BOX.SDRV
 * CPQFCA-MIB::cpqFcaSpareCondition.BOX.SDRV
 * = .1.3.6.1.4.1.232.16.2.4.1.1.6.BOX.SDRV
 * CPQFCA-MIB::cpqFcaSpareBay.BOX.SDRV
 * = .1.3.6.1.4.1.232.16.2.4.1.1.8.BOX.SDRV
 */
oid baseOID [] = { 1, 3, 6, 1, 4, 1, 232, 16, 2, 4, 1, 1 };
size_t baseOIDLen = 12;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_cpq_fcaspare";
    progVersion = "0.2";
    helpIndexP = "-i INDEX -d DRIVE";
    helpThresP = NULL;
    helpPurpose =
            "Checks the status of a Compaq/HP External Array spare drive.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
    /* rrdStep is unused */
    /* rrdHBeat is unused */
    supportsPerfLog = FALSE;
    hasIndexParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-i, --index INTEGER\n"
    "    Index of the storage box the drive resides in.\n"
    "\n"
    "-d, --drive INTEGER\n"
    "    Index of the spare drive.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin returns the overall condition of a spare drive and any\n"
    "associated physical drives.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "There is no performance data provided.\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-d" )
            ||  !strcmp ( argv [(*i)], "--drive" ) )
    {
        isSDrvSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            sDrv = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || *tail )
                    Usage ( "Invalid 'drive' parameter" );
        }
        else Usage ( "Missing 'drive' parameter(s)" );
    } else ret = FALSE;
    return ( ret );
}

int ScanResIndex ( int aIndex, int aSDrv )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex )
                && ( scanRes [retVal].sDrv == aSDrv );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 15 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [12] )
            {
                case 1: /* cpqFcaSpareBoxIndex */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].index = aVar->name [13];
                    scanRes [index].sDrv = aVar->name [14];
                    scanRes [index].v3 = FALSE;
                    scanRes [index].v4 = FALSE;
                    scanRes [index].v5 = FALSE;
                    scanRes [index].v6 = FALSE;
                    scanRes [index].v8 = FALSE;
                    scanRes [index].bay = -1;
                break;
                case 3: /* cpqFcaSpareStatus */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v3 = ( aVar->type == ASN_INTEGER );
                break;
                case 4: /* cpqFcaSpareReplacedDrvBay */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v4 = ( aVar->type == ASN_INTEGER );
                break;
                case 5: /* cpqFcaSparePercentRebuild */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v5 = ( aVar->type == ASN_GAUGE );
                break;
                case 6: /* cpqFcaSpareCondition */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v6 = ( aVar->type == ASN_INTEGER );
                break;
                case 8: /* cpqFcaSpareBay */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v8 = ( aVar->type == ASN_INTEGER );
                    scanRes [index].bay = *(aVar->val.integer);
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v3 && scanRes [i].v4 && scanRes [i].v5 &&
                scanRes [i].v6 && scanRes [i].v8 )
        {
            retVal = STATE_OK;
            printf ( "OK index: %i ; drive: %i ; bay: %i\n",
                    scanRes [i].index, scanRes [i].sDrv, scanRes [i].bay );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_stat [15];
    oid o_rBay [15];
    oid o_perc [15];
    oid o_cond [15];
    oid o_bay  [15];
    size_t oidLen;

    OidCpy ( o_stat, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_rBay, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_perc, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_cond, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_bay,  &oidLen, baseOID, &baseOIDLen );
    o_stat [12] = 3; o_stat [13] = idx; o_stat [14] = sDrv;
    o_rBay [12] = 4; o_rBay [13] = idx; o_rBay [14] = sDrv;
    o_perc [12] = 5; o_perc [13] = idx; o_perc [14] = sDrv;
    o_cond [12] = 6; o_cond [13] = idx; o_cond [14] = sDrv;
    o_bay  [12] = 8; o_bay  [13] = idx; o_bay  [14] = sDrv;
    oidLen = 15;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.stat = ReadInteger ( sess2, o_stat, oidLen );
    v.rBay = ReadInteger ( sess2, o_rBay, oidLen );
    v.perc = ReadGauge32 ( sess2, o_perc, oidLen );
    v.cond = ReadInteger ( sess2, o_cond, oidLen );
    v.bay  = ReadInteger ( sess2, o_bay,  oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

void CreateDataFile ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

void ParseData ( const char* aData )
{
    /* performance data is not supported */
}

void WriteData ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

int CheckData ()
{
    int retVal = STATE_OK;
    char* cond = NULL;
    char* stat = NULL;

    ReadVals ();
    switch ( v.cond )
    {
        case 1:  retVal = STATE_WARNING;  cond = "'other'";    break;
        case 2:  retVal = STATE_OK;       cond = "'ok'";       break;
        case 3:  retVal = STATE_WARNING;  cond = "'degraded'"; break;
        case 4:  retVal = STATE_CRITICAL; cond = "'failed'";   break;
        default: retVal = STATE_UNKNOWN;  cond = "invalid";    break;
    }
    switch ( v.stat )
    {
        case 1:  stat = StrDup ( "'other'" );    break;
        case 2:  stat = StrDup ( "'inactive'" ); break;
        case 3:  stat = StrDup ( "'failed'" );   break;
        case 4:
        {
            char* perc  = IntToStr    ( v.perc );
            char* rBay  = IntToStr    ( v.rBay );

            stat = StrConcat ( 4, "building (", perc,
                    "%), replacing drive in bay ", rBay );
            free ( rBay );
            free ( perc );
        }
        break;
        case 5:
        {
            char* rBay  = IntToStr    ( v.rBay );

            stat = StrConcat ( 2,
                    "active, replacing drive in bay ", rBay );
            free ( rBay );
        }
        break;
        default: stat = StrDup ( "invalid" );    break;
    }
    PrintExit ( retVal );
    printf ( " - bay %i: cond: %s, stat: %s\n", v.bay, cond, stat );
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( checkData )
    {
        if ( !isSDrvSet ) Usage ( "Missing 'drive' parameters" );
    }
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
