/*
 * File:      check_cpq_ida.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include "decoders.h"
#include "globals.h"
#include "nagiosif.h"
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int index;   /* controller index */
    int v2;      /* found cpqDaCntlrModel */
    int v6;      /* found cpqDaCntlrCondition */
    char* model; /* model name */
} tScanRes;

/* data type for storing data */
typedef struct
{
    int   cond;  /* condition */
} tDataSet;


/*
 * SNMP base OID
 * CPQIDA-MIB::cpqDaCntlrIndex.IDX
 * = 1.3.6.1.4.1.232.3.2.2.1.1.1.IDX
 * CPQIDA-MIB::cpqDaCntlrModel.IDX
 * = 1.3.6.1.4.1.232.3.2.2.1.1.2.IDX
 * CPQIDA-MIB::cpqDaCntlrCondition.IDX
 * = 1.3.6.1.4.1.232.3.2.2.1.1.6.IDX
 */
oid baseOID [] = { 1, 3, 6, 1, 4, 1, 232, 3, 2, 2, 1, 1 };
size_t baseOIDLen = 12;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_cpq_ida";
    progVersion = "0.8";
    helpIndexP = "-i INDEX";
    helpThresP = NULL;
    helpPurpose = "Checks the status of a Compaq/HP IDA Controller.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
    /* rrdStep is unused */
    /* rrdHBeat is unused */
    supportsPerfLog = FALSE;
    hasIndexParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-i, --index INTEGER\n"
    "    Index of the controller which should be checked\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin returns the overall state of this controller, any\n"
    "associated logical drives, physical drives and array accelerators.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "There is no performance data provided.\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    /* no specific parameters except --index which is handled elsewhere */
    return FALSE;
}

char* DecodeCPQIdaModel ( int aType )
{
    char* retVal = NULL;

    switch ( aType )
    {
        case  1: retVal = StrDup ( "other" );                  break;
        case  2: retVal = StrDup ( "IDA" );                    break;
        case  3: retVal = StrDup ( "IDA Expansion" );          break;
        case  4: retVal = StrDup ( "IDA - 2" );                break;
        case  5: retVal = StrDup ( "SMART" );                  break;
        case  6: retVal = StrDup ( "SMART - 2/E" );            break;
        case  7: retVal = StrDup ( "SMART - 2/P" );            break;
        case  8: retVal = StrDup ( "SMART - 2SL" );            break;
        case  9: retVal = StrDup ( "Smart - 3100ES" );         break;
        case 10: retVal = StrDup ( "Smart - 3200" );           break;
        case 11: retVal = StrDup ( "SMART - 2DH" );            break;
        case 12: retVal = StrDup ( "Smart - 221" );            break;
        case 13: retVal = StrDup ( "Smart Array 4250ES" );     break;
        case 14: retVal = StrDup ( "Smart Array 4200" );       break;
        case 15: retVal = StrDup ( "Integrated Smart Array" ); break;
        case 16: retVal = StrDup ( "Smart Array 431" );        break;
        case 17: retVal = StrDup ( "Smart Array 5300" );       break;
        case 18: retVal = StrDup ( "RAID LC2 Controller" );    break;
        case 19: retVal = StrDup ( "Smart Array 5i" );         break;
        case 20: retVal = StrDup ( "Smart Array 532" );        break;
        case 21: retVal = StrDup ( "Smart Array 5312" );       break;
        case 22: retVal = StrDup ( "Smart Array 641" );        break;
        case 23: retVal = StrDup ( "Smart Array 642" );        break;
        case 24: retVal = StrDup ( "Smart Array 6400" );       break;
        case 25: retVal = StrDup ( "Smart Array 6400 EM" );    break;
        case 26: retVal = StrDup ( "Smart Array 6i" );         break;
        default: retVal = StrDup ( "invalid" );                break;
    }
    return retVal;
}

int ScanResIndex ( int aIndex )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 14 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [12] )
            {
                case 1: /* cpqDaCntlrIndex */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].index = aVar->name [13];
                    scanRes [index].v2 = FALSE;
                    scanRes [index].v6 = FALSE;
                    scanRes [index].model = NULL;
                break;
                case 2: /* cpqHeFltTolFanLocale */
                    index = ScanResIndex ( aVar->name [13] );
                    if ( ( aVar->type == ASN_INTEGER ) )
                    {
                        scanRes [index].v2 = TRUE;
                        scanRes [index].model =
                                DecodeCPQIdaModel ( *(aVar->val.integer) );
                    } /* else nothing to do */
                case 6: /* cpqDaCntlrCondition */
                    index = ScanResIndex ( aVar->name [13] );
                    scanRes [index].v6 = ( aVar->type == ASN_INTEGER );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if ( scanRes [i].v2 && scanRes [i].v6 )
        {
            retVal = STATE_OK;
            printf ( "OK index: %i ; model: %s\n",
                    scanRes [i].index, scanRes [i].model );
            free ( scanRes [i].model );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_cond [14];
    size_t oidLen;

    OidCpy ( o_cond, &oidLen, baseOID, &baseOIDLen );
    o_cond [12] = 6; o_cond [13] = idx;
    oidLen = 14;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.cond = ReadInteger ( sess2, o_cond, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

void CreateDataFile ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

void ParseData ( const char* aData )
{
    /* performance data is not supported */
}

void WriteData ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

int CheckData ()
{
    int retVal = STATE_UNKNOWN;

    ReadVals ();
    switch ( v.cond )
    {
        case 1:
            retVal = STATE_WARNING;
            printf ( "WARNING - IDA state is 'other'\n" );
        break;
        case 2:
            retVal = STATE_OK;
            printf ( "OK - IDA state is 'ok'\n" );
        break;
        case 3:
            retVal = STATE_WARNING;
            printf ( "WARNING - IDA state is 'degraded'\n" );
        break;
        case 4:
            retVal = STATE_CRITICAL;
            printf ( "CRITICAL - IDA state is 'failed'\n" );
        break;
        default:
            retVal = STATE_UNKNOWN;
            printf ( "UNKNOWN - IDA state is invalid\n" );
        break;
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
