/*
 * File:      check_cpq_phydrv.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int index;   /* controller index */
    int pDrv;    /* physical drive index */
    int v5;      /* found cpqDaPhyDrvBay */
    int v6;      /* found cpqDaPhyDrvStatus */
    int v9;      /* found cpqDaPhyDrvRefHours */
    int v10;     /* found cpqDaPhyDrvHReads */
    int v11;     /* found cpqDaPhyDrvReads */
    int v12;     /* found cpqDaPhyDrvHWrites */
    int v13;     /* found cpqDaPhyDrvWrites */
    int v16;     /* found cpqDaPhyDrvHardReadErrs */
    int v17;     /* found cpqDaPhyDrvRecvReadErrs */
    int v18;     /* found cpqDaPhyDrvHardWriteErrs */
    int v19;     /* found cpqDaPhyDrvRecvWriteErrs */
    int v35;     /* found cpqDaPhyDrvThreshPassed */
    int v36;     /* found cpqDaPhyDrvHasMonInfo and is OK */
    int v37;     /* found cpqDaPhyDrvCondition */
    int v49;     /* found cpqDaPhyDrvPlacement */
    int v51;     /* found cpqDaPhyDrvSerialNum */
    int v57;     /* found cpqDaPhyDrvSmartStatus */
    int place;   /* placement */
    char* ser;   /* serial number */
    int bay;     /* bay */
} tScanRes;

/* data type for storing data */
typedef struct
{
    int      bay;   /* bay the drive is in */
    int      stat;  /* status of the drive */
    int      cond;  /* condition of the drive */
    int      smart; /* smart status of the drive */
    int      thres; /* thresholds status of the drive */
    uint     hours; /* operating hours of the drive */
    uint     read;  /* reads from the drive */
    uint     writ;  /* writes to the drive */
    uint     haRd;  /* hard read errors */
    uint     reRd;  /* recoverable read errors */
    uint     haWr;  /* hard write errors */
    uint     reWr;  /* recoverable write errors */
} tDataSet;

/* global input variables */
int pDrv;
int isPDrvSet  = FALSE;

/*
 * SNMP base OID
 * CPQIDA-MIB::cpqDaPhyDrvBay.CTRL.PDRV
 * = 1.3.6.1.4.1.232.3.2.5.1.1.5.CTRL.PDRV
 * CPQIDA-MIB::cpqDaPhyDrvStatus.CTRL.PDRV
 * = 1.3.6.1.4.1.232.3.2.5.1.1.6.CTRL.PDRV
 * CPQIDA-MIB::cpqDaPhyDrvCondition.CTRL.PDRV
 * = 1.3.6.1.4.1.232.3.2.5.1.1.37.CTRL.PDRV
 * CPQIDA-MIB::cpqDaPhyDrvSmartStatus.CTRL.PDRV
 * = 1.3.6.1.4.1.232.3.2.5.1.1.57.CTRL.PDRV
 * CPQIDA-MIB::cpqDaPhyDrvThreshPassed.CTRL.PDRV
 * = 1.3.6.1.4.1.232.3.2.5.1.1.35.CTRL.PDRV
 * CPQIDA-MIB::cpqDaPhyDrvRefHours.CTRL.PDRV
 * = 1.3.6.1.4.1.232.3.2.5.1.1.9.CTRL.PDRV
 * CPQIDA-MIB::cpqDaPhyDrvHReads.CTRL.PDRV
 * = 1.3.6.1.4.1.232.3.2.5.1.1.10.CTRL.PDRV
 * CPQIDA-MIB::cpqDaPhyDrvReads.CTRL.PDRV
 * = 1.3.6.1.4.1.232.3.2.5.1.1.11.CTRL.PDRV
 * CPQIDA-MIB::cpqDaPhyDrvHWrites.CTRL.PDRV
 * = 1.3.6.1.4.1.232.3.2.5.1.1.12.CTRL.PDRV
 * CPQIDA-MIB::cpqDaPhyDrvWrites.CTRL.PDRV
 * = 1.3.6.1.4.1.232.3.2.5.1.1.13.CTRL.PDRV
 * CPQIDA-MIB::cpqDaPhyDrvHardReadErrs.CTRL.PDRV
 * = 1.3.6.1.4.1.232.3.2.5.1.1.16.CTRL.PDRV
 * CPQIDA-MIB::cpqDaPhyDrvRecvReadErrs.CTRL.PDRV
 * = 1.3.6.1.4.1.232.3.2.5.1.1.17.CTRL.PDRV
 * CPQIDA-MIB::cpqDaPhyDrvHardWriteErrs.CTRL.PDRV
 * = 1.3.6.1.4.1.232.3.2.5.1.1.18.CTRL.PDRV
 * CPQIDA-MIB::cpqDaPhyDrvRecvWriteErrs.CTRL.PDRV
 * = 1.3.6.1.4.1.232.3.2.5.1.1.19.CTRL.PDRV
 * CPQIDA-MIB:cpqDaPhyDrvPlacement:.CTRL.PDRV
 * = 1.3.6.1.4.1.232.3.2.5.1.1.49.CTRL.PDRV
 * CPQIDA-MIB:cpqDaPhyDrvSerialNum:.CTRL.PDRV
 * = 1.3.6.1.4.1.232.3.2.5.1.1.51.CTRL.PDRV
 */
oid baseOID [] = { 1, 3, 6, 1, 4, 1, 232, 3, 2, 5, 1, 1 };
size_t baseOIDLen = 12;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_cpq_phydrv";
    progVersion = "0.8";
    helpIndexP = "-i INDEX -d DRIVE";
    helpThresP = NULL;
    helpPurpose =
            "Checks the state of a physical drive connected to a Compaq\n"
            "Intelligent Drive Array Controller.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 5 * 60;
    rrdHBeat = 11 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = TRUE;
    hasAddInfParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-i, --index INTEGER\n"
    "    Index of the controller to which the drive which should be checked\n"
    "    is connected.\n"
    "\n"
    "-d, --drive INTEGER\n"
    "    Index of the physical drive.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin returns the overall state of the physical drive."
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<addInf>;<read>;<writ>;<haRd>;<reRd>;<haWr>;<reRw>\n"
    "where\n"
    "    addInf ... additional information.\n"
    "    read   ... Sectors read\n"
    "    writ   ... Sectors written\n"
    "    haRd   ... Hard read errors\n"
    "    reRd   ... Recovered read errors\n"
    "    haWr   ... Hard write errors\n"
    "    reWr   ... Recovered write errors\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'sectors read'=<read>c "
    "'sectors written'=<writ>c "
    "'hard read errors'=<haRd>c "
    "'recovered read errors'=<reRd>c "
    "'hard write errors'=<haWr>c "
    "'recovered write errors'=<reRw>c\n"
    "\n"
    );
    printf (
    "For more information about those counters see CPQIDA-MIB.txt\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-d" )
            ||  !strcmp ( argv [(*i)], "--drive" ) )
    {
        isPDrvSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            pDrv = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || *tail ) Usage ( "Invalid 'drive' parameter" );
        }
        else Usage ( "Missing 'drive' parameter(s)" );
    } else ret = FALSE;
    return ( ret );
}

int ScanResIndex ( int aIndex, int aPDrv )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex )
                && ( scanRes [retVal].pDrv == aPDrv );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 15 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [12] )
            {
                case 2: /* cpqDaPhyDrvIndex */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].index = aVar->name [13];
                    scanRes [index].pDrv = aVar->name [14];
                    scanRes [index].v5 = FALSE;
                    scanRes [index].v6 = FALSE;
                    scanRes [index].v9 = FALSE;
                    scanRes [index].v11 = FALSE;
                    scanRes [index].v13 = FALSE;
                    scanRes [index].v16 = FALSE;
                    scanRes [index].v17 = FALSE;
                    scanRes [index].v18 = FALSE;
                    scanRes [index].v19 = FALSE;
                    scanRes [index].v35 = FALSE;
                    scanRes [index].v36 = FALSE;
                    scanRes [index].v37 = FALSE;
                    scanRes [index].v57 = FALSE;
                break;
                case 5: /* cpqDaPhyDrvBay */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v5 = ( aVar->type == ASN_INTEGER );
                    scanRes [index].bay = *(aVar->val.integer);
                break;
                case 6: /* cpqDaPhyDrvStatus */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v6 = ( aVar->type == ASN_INTEGER );
                break;
                case 9: /* cpqDaPhyDrvRefHours */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v9 = ( aVar->type == ASN_COUNTER );
                break;
                case 11: /* cpqDaPhyDrvReads */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v11 = ( aVar->type == ASN_COUNTER );
                break;
                case 13: /* cpqDaPhyDrvWrites */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v13 = ( aVar->type == ASN_COUNTER );
                break;
                case 16: /* cpqDaPhyDrvHardReadErrs */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v16 = ( aVar->type == ASN_COUNTER );
                break;
                case 17: /* cpqDaPhyDrvRecvReadErrs */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v17 = ( aVar->type == ASN_COUNTER );
                break;
                case 18: /* cpqDaPhyDrvHardWriteErrs */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v18 = ( aVar->type == ASN_COUNTER );
                break;
                case 19: /* cpqDaPhyDrvRecvWriteErrs */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v19 = ( aVar->type == ASN_COUNTER );
                break;
                case 35: /* cpqDaPhyDrvThreshPassed */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v35 = ( aVar->type == ASN_INTEGER );
                break;
                case 36: /* cpqDaPhyDrvHasMonInfo */
                {
                    int typeOK = ( aVar->type == ASN_INTEGER );
                    int valOK = ( *(aVar->val.integer) == 2 );

                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v36 = ( typeOK && valOK );
                }
                break;
                case 37: /* cpqDaPhyDrvCondition */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v37 = ( aVar->type == ASN_INTEGER );
                break;
                case 49: /* cpqDaPhyDrvPlacement */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v49 = ( aVar->type == ASN_INTEGER );
                    scanRes [index].place = *(aVar->val.integer);
                break;
                case 51: /* cpqDaPhyDrvSerialNum */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v51 = ( aVar->type == ASN_OCTET_STR );
                    scanRes [index].ser = VarListStrDup ( aVar );
                break;
                case 57: /* cpqDaPhyDrvSmartStatus */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v57 = ( aVar->type == ASN_INTEGER );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v5  && scanRes [i].v6  && scanRes [i].v9  &&
                scanRes [i].v11 && scanRes [i].v13 && scanRes [i].v16 &&
                scanRes [i].v17 && scanRes [i].v18 && scanRes [i].v19 &&
                scanRes [i].v35 && scanRes [i].v36 && scanRes [i].v37 &&
                scanRes [i].v49 && scanRes [i].v51 && scanRes [i].v57 )
        {
            char* place = NULL;

            switch ( scanRes[i].place )
            {
                case  1: place = "other";    break;
                case  2: place = "int";      break;
                case  3: place = "ext";      break;
                default: place = "invalid";  break;
            }
            retVal = STATE_OK;
            printf ( "OK index: %i ; drive: %i ; place: %s ; bay: %i",
                    scanRes [i].index, scanRes [i].pDrv, place,
                    scanRes [i].bay );
            printf ( " ; ser: %s\n",
                    scanRes [i].ser );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_bay   [15];
    oid o_stat  [15];
    oid o_cond  [15];
    oid o_smart [15];
    oid o_thres [15];
    oid o_hours [15];
    oid o_read  [15];
    oid o_writ  [15];
    oid o_haRd  [15];
    oid o_reRd  [15];
    oid o_haWr  [15];
    oid o_reWr  [15];
    size_t oidLen;

    OidCpy ( o_bay,   &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_stat,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_cond,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_smart, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_thres, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_hours, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_read,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_writ,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_haRd,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_reRd,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_haWr,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_reWr,  &oidLen, baseOID, &baseOIDLen );
    o_bay   [12] =  5; o_bay   [13] = idx; o_bay   [14] = pDrv;
    o_stat  [12] =  6; o_stat  [13] = idx; o_stat  [14] = pDrv;
    o_cond  [12] = 37; o_cond  [13] = idx; o_cond  [14] = pDrv;
    o_smart [12] = 57; o_smart [13] = idx; o_smart [14] = pDrv;
    o_thres [12] = 35; o_thres [13] = idx; o_thres [14] = pDrv;
    o_hours [12] =  9; o_hours [13] = idx; o_hours [14] = pDrv;
    o_read  [12] = 11; o_read  [13] = idx; o_read  [14] = pDrv;
    o_writ  [12] = 13; o_writ  [13] = idx; o_writ  [14] = pDrv;
    o_haRd  [12] = 16; o_haRd  [13] = idx; o_haRd  [14] = pDrv;
    o_reRd  [12] = 17; o_reRd  [13] = idx; o_reRd  [14] = pDrv;
    o_haWr  [12] = 18; o_haWr  [13] = idx; o_haWr  [14] = pDrv;
    o_reWr  [12] = 19; o_reWr  [13] = idx; o_reWr  [14] = pDrv;
    oidLen = 15;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.bay   = ReadInteger   ( sess2, o_bay,   oidLen );
    v.stat  = ReadInteger   ( sess2, o_stat,  oidLen );
    v.cond  = ReadInteger   ( sess2, o_cond,  oidLen );
    v.smart = ReadInteger   ( sess2, o_smart, oidLen );
    v.thres = ReadInteger   ( sess2, o_thres, oidLen );
    v.hours = ReadCounter32 ( sess2, o_hours, oidLen );
    v.read  = ReadCounter32 ( sess2, o_read,  oidLen );
    v.writ  = ReadCounter32 ( sess2, o_writ,  oidLen );
    v.haRd  = ReadCounter32 ( sess2, o_haRd,  oidLen );
    v.reRd  = ReadCounter32 ( sess2, o_reRd,  oidLen );
    v.haWr  = ReadCounter32 ( sess2, o_haWr,  oidLen );
    v.reWr  = ReadCounter32 ( sess2, o_reWr,  oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "read", COUNTER, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "writ", COUNTER, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "haRd", COUNTER, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "reRd", COUNTER, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "haWr", COUNTER, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "reWr", COUNTER, rrdHBeat, "0", "U" );
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    char buf[1024];

    chk = sscanf ( aData, "%1023s;%u;%u;%u;%u;%u;%u", buf,
            &(v.read), &(v.writ), &(v.haRd), &(v.reRd), &(v.haWr), &(v.reWr) );
    if ( chk != 7 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
    addInf = StrDup ( buf );
    v.bay = -1;
    v.stat = -1;
    v.cond = -1;
    v.smart = -1;
    v.thres = -1;
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* read   = UIntToStr   ( v.read );
    char* writ   = UIntToStr   ( v.writ );
    char* haRd   = UIntToStr   ( v.haRd );
    char* reRd   = UIntToStr   ( v.reRd );
    char* haWr   = UIntToStr   ( v.haWr );
    char* reWr   = UIntToStr   ( v.reWr );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "read:writ:haRd:reRd:haWr:reWr" );
    params [i++] = StrConcat ( 13, ts, ":", read, ":", writ, ":", haRd,
            ":", reRd, ":", haWr, ":", reWr );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( reWr );
    free ( haWr );
    free ( reRd );
    free ( haRd );
    free ( writ );
    free ( read );
    free ( ts );
}

void PlotDataRdWr ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[sectors/s]  <-- wr | rd -->";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "rdwr", aTiming );
    PPAddPlot ( pp, "read",    "sectors/s", FALSE );
    PPAddPlot ( pp, "written", "sectors/s", FALSE );
    title = StrConcat ( 4, hostName, " drive ", addInf, " I/O operations" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1024, PLOT_SI_AUTO, 0, 0, TRUE );
    params [i++] = CreateDEF ( "readMR", aFN, "read", MAX );
    params [i++] = CreateDEF ( "readAR", aFN, "read", AVG );
    params [i++] = CreateDEF ( "writMR", aFN, "writ", MAX );
    params [i++] = CreateDEF ( "writAR", aFN, "writ", AVG );
    params [i++] = StrDup ( "CDEF:readM=readMR" );
    params [i++] = StrDup ( "CDEF:readA=readAR" );
    params [i++] = StrDup ( "CDEF:writM=writMR,-1,*" );
    params [i++] = StrDup ( "CDEF:writA=writAR,-1,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "readM", C_THRU_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "readA", C_THRU_AVG, pp, 0, AVG, TRUE );
    params [i++] = CreatePlot ( PLOT_AREA,  "writM", C_THRU_MAX, pp, 1, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "writA", C_THRU_AVG, pp, 1, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "readAR", "readAR", "readMR", "readMR", "%4.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "writAR", "writAR", "writMR", "writMR", "%4.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataECnt ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[errors/s]  <-- wr | rd -->";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "ecnt", aTiming );
    PPAddPlot ( pp, "recovered read errors",  "errors/s", FALSE );
    PPAddPlot ( pp, "hard read errors",       "errors/s", FALSE );
    PPAddPlot ( pp, "recovered write errors", "errors/s", FALSE );
    PPAddPlot ( pp, "hard write errors",      "errors/s", FALSE );
    title = StrConcat ( 4, hostName, " drive ", addInf, " errors" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, TRUE );
    params [i++] = CreateDEF ( "reRdMR", aFN, "reRd", MAX );
    params [i++] = CreateDEF ( "reRdAR", aFN, "reRd", AVG );
    params [i++] = CreateDEF ( "haRdMR", aFN, "haRd", MAX );
    params [i++] = CreateDEF ( "haRdAR", aFN, "haRd", AVG );
    params [i++] = CreateDEF ( "reWrMR", aFN, "reWr", MAX );
    params [i++] = CreateDEF ( "reWrAR", aFN, "reWr", AVG );
    params [i++] = CreateDEF ( "haWrMR", aFN, "haWr", MAX );
    params [i++] = CreateDEF ( "haWrAR", aFN, "haWr", AVG );
    params [i++] = StrDup ( "CDEF:reRdM=reRdMR" );
    params [i++] = StrDup ( "CDEF:reRdA=reRdAR" );
    params [i++] = StrDup ( "CDEF:haRdM=haRdMR" );
    params [i++] = StrDup ( "CDEF:haRdA=haRdAR" );
    params [i++] = StrDup ( "CDEF:reWrM=reWrMR,-1,*" );
    params [i++] = StrDup ( "CDEF:reWrA=reWrAR,-1,*" );
    params [i++] = StrDup ( "CDEF:haWrM=haWrMR,-1,*" );
    params [i++] = StrDup ( "CDEF:haWrA=haWrAR,-1,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "reRdM", C_ERR2_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "reRdA", C_ERR2_AVG, pp, 0, AVG, TRUE );
    params [i++] = CreatePlot ( PLOT_LINE2, "haRdM", C_ERR1_MAX, pp, 1, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "haRdA", C_ERR1_AVG, pp, 1, AVG, TRUE );
    params [i++] = CreatePlot ( PLOT_AREA,  "reWrM", C_ERR2_MAX, pp, 2, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "reWrA", C_ERR2_AVG, pp, 2, AVG, TRUE );
    params [i++] = CreatePlot ( PLOT_LINE2, "haWrM", C_ERR1_MAX, pp, 3, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "haWrA", C_ERR1_AVG, pp, 3, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "reRdAR", "reRdAR", "reRdMR", "reRdMR", "%4.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "haRdAR", "haRdAR", "haRdMR", "haRdMR", "%4.1lf" );
    i = PlotPrintParams ( i, params, pp, 2, "reWrAR", "reWrAR", "reWrMR", "reWrMR", "%4.1lf" );
    i = PlotPrintParams ( i, params, pp, 3, "haWrAR", "haWrAR", "haWrMR", "haWrMR", "%4.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataRdWr ( fn, PLOT_24H_5M,  now );
    PlotDataECnt ( fn, PLOT_24H_5M,  now );
    PlotDataRdWr ( fn, PLOT_7D_15M,  now );
    PlotDataECnt ( fn, PLOT_7D_15M,  now );
    PlotDataRdWr ( fn, PLOT_31D_60M, now );
    PlotDataECnt ( fn, PLOT_31D_60M, now );
    PlotDataRdWr ( fn, PLOT_366D_4H, now );
    PlotDataECnt ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

char* CheckStatus ( int* aRetVal )
{
    char* aRetStr = NULL;

    switch ( v.stat )
    {
        case 1:
            aRetStr = StrDup ( "other" );
            *aRetVal = MAXIMUM ( *aRetVal, STATE_WARNING );
        break;
        case 2:
            aRetStr = StrDup ( "ok" );
        break;
        case 3:
            aRetStr = StrDup ( "critical" );
            *aRetVal = MAXIMUM ( *aRetVal, STATE_CRITICAL );
        break;
        case 4:
            aRetStr = StrDup ( "predictive failure" );
            *aRetVal = MAXIMUM ( *aRetVal, STATE_CRITICAL );
        break;
        default:
            aRetStr = StrDup ( "invalid" );
            *aRetVal = MAXIMUM ( *aRetVal, STATE_WARNING );
        break;
    }
    return ( aRetStr );
}

char* CheckCond ( int* aRetVal )
{
    char* aRetStr = NULL;

    switch ( v.cond )
    {
        case 1:
            aRetStr = StrDup ( "other" );
            *aRetVal = MAXIMUM ( *aRetVal, STATE_WARNING );
        break;
        case 2:
            aRetStr = StrDup ( "ok" );
        break;
        case 3:
            aRetStr = StrDup ( "degraded" );
            *aRetVal = MAXIMUM ( *aRetVal, STATE_WARNING );
        break;
        case 4:
            aRetStr = StrDup ( "failed" );
            *aRetVal = MAXIMUM ( *aRetVal, STATE_CRITICAL );
        break;
        default:
            aRetStr = StrDup ( "invalid" );
            *aRetVal = MAXIMUM ( *aRetVal, STATE_WARNING );
        break;
    }
    return ( aRetStr );
}

char* CheckSmart ( int* aRetVal )
{
    char* aRetStr = NULL;

    switch ( v.smart )
    {
        case 1:
            aRetStr = StrDup ( "other" );
            *aRetVal = MAXIMUM ( *aRetVal, STATE_WARNING );
        break;
        case 2:
            aRetStr = StrDup ( "ok" );
        break;
        case 3:
            aRetStr = StrDup ( "replace drive" );
            *aRetVal = MAXIMUM ( *aRetVal, STATE_CRITICAL );
        break;
        default:
            aRetStr = StrDup ( "invalid" );
            *aRetVal = MAXIMUM ( *aRetVal, STATE_WARNING );
        break;
    }
    return ( aRetStr );
}

char* CheckThreshold ( int* aRetVal )
{
    char* aRetStr = NULL;

    switch ( v.thres )
    {
        case 1:
            aRetStr = StrDup ( "ok" );
        break;
        case 2:
            aRetStr = StrDup ( "exceeded" );
            *aRetVal = MAXIMUM ( *aRetVal, STATE_WARNING );
        break;
        default:
            aRetStr = StrDup ( "invalid" );
            *aRetVal = MAXIMUM ( *aRetVal, STATE_WARNING );
        break;
    }
    return ( aRetStr );
}

int CheckData ()
{
    int retVal = STATE_OK;
    char* stat = NULL;
    char* cond = NULL;
    char* smart = NULL;
    char* thres = NULL;

    ReadVals ();
    stat  = CheckStatus    ( &retVal );
    cond  = CheckCond      ( &retVal );
    smart = CheckSmart     ( &retVal );
    thres = CheckThreshold ( &retVal );
    PrintExit ( retVal );
    printf ( " - bay %i: %ud %uh, stat: %s, cond: %s, SMART: %s, thres: %s",
            v.bay, v.hours / 24, v.hours % 24, stat, cond, smart, thres );
    if ( npPerfData )
    {
        printf ( "|'sectors read'=%uc", v.read );
        printf ( " 'sectors written'=%uc", v.writ );
        printf ( " 'hard read errors'=%uc", v.haRd );
        printf ( " 'recovered read errors'=%uc", v.reRd );
        printf ( " 'hard write errors'=%uc", v.haWr );
        printf ( " 'recovered write errors'=%uc\n", v.reWr );
    }
    else
    {
        printf ( "|%s;%u;%u;%u;%u;%u;%u\n",
                addInf, v.read, v.writ, v.haRd, v.reRd, v.haWr, v.reWr );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );

    if ( checkData )
    {
        if ( !isPDrvSet ) Usage ( "Missing 'drive' parameters" );
    }
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ();
        if ( plotPerfData ) PlotPerfData ();
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return retVal;
}
