/*
 * File:      check_cpq_temp.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "decoders.h"
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int chass;   /* chassis index */
    int index;   /* temperature sensor index */
    int v3;      /* found cpqHeTemperatureLocale */
    int v4;      /* found cpqHeTemperatureCelsius */
    int v5;      /* found cpqHeTemperatureThreshold */
    int v7;      /* found cpqHeTemperatureThresholdType */
    char* local; /* temperature sensor locale */
    int tempT;   /* threshold temperature */
} tScanRes;

/* data type for storing data */
typedef struct
{
    int   local; /* name of the testpoint */
    int   tempC; /* current temperature */
    int   tempT; /* threshold temperature */
    int   type;  /* threshold type */
} tDataSet;

/* global input variables */
uint warn;
uint crit;
int chass;
int isWarnSet  = FALSE;
int isCritSet  = FALSE;
int isChassSet  = FALSE;

/*
 * SNMP base OID
 * CPQHLTH-MIB::cpqHeTemperatureLocale.CHA.IDX
 * = .1.3.6.1.4.1.232.6.2.6.8.1.3.CHA.IDX
 * CPQHLTH-MIB::cpqHeTemperatureCelsius.CHA.IDX
 * = .1.3.6.1.4.1.232.6.2.6.8.1.4.CHA.IDX
 * CPQHLTH-MIB::cpqHeTemperatureThreshold.CHA.IDX
 * = .1.3.6.1.4.1.232.6.2.6.8.1.5.CHA.IDX
 * CPQHLTH-MIB::cpqHeTemperatureThresholdType.CHA.IDX
 * = .1.3.6.1.4.1.232.6.2.6.8.1.7.CHA.IDX
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 232, 6, 2, 6, 8, 1 };
size_t baseOIDLen = 12;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_cpq_temp";
    progVersion = "0.8";
    helpIndexP = " -s SYSTEM -i INDEX";
    helpThresP = "-w WARN -c CRIT";
    helpPurpose =
            "Checks the status of a temperature sensor of a Compaq server.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 5 * 60;
    rrdHBeat = 11 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = TRUE;
    hasAddInfParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-w, --warning INTEGER\n"
    "    Temperature in degree Celsius at which a warning is generated\n"
    "\n"
    "-c, --critical INTEGER\n"
    "    Temperature in degree Celsius at which a critical is generated\n"
    "\n"
    "-s, --system-chassis INTEGER\n"
    "    Index of the system chassis\n"
    "\n"
    "-i, --index INTEGER\n"
    "    Index of the temperature sensor which should be checked\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin returns how the current temperature compares to the warning\n"
    "and critical settings.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<addInf>;<tempC>;<tempE>\n"
    "where\n"
    "    addInf ... additional information.\n"
    "    tempC  ... current temperature in degree Celsius.\n"
    "    tempT  ... threshold temperature.\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'current temperature'=<tempC>;<warn>;<crit> "
    "'threshold temperature'=<tempT>\n"
    "\n"
    "Note: Nagios Plugin performance data format does not allow Degree\n"
    "Centigrade as unit of measurement.\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-w" )
            ||  !strcmp ( argv [(*i)], "--warning" ) )
    {
        isWarnSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            warn = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || *tail ) Usage ( "Invalid 'warn' parameter" );
        }
        else Usage ( "Missing 'warn' parameter(s)" );
    } else if ( !strcmp ( argv [(*i)], "-c" )
            ||  !strcmp ( argv [(*i)], "--critical" ) )
    {
        isCritSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            crit = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || *tail ) Usage ( "Invalid 'crit' parameter" );
        }
        else Usage ( "Missing 'crit' parameter(s)" );
    } else if ( !strcmp ( argv [(*i)], "-s" )
            ||  !strcmp ( argv [(*i)], "--system-chassis" ) )
    {
        isChassSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            chass = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || *tail )
                    Usage ( "Invalid 'system chassis' parameter" );
        }
        else Usage ( "Missing 'system chassis' parameter(s)" );
    } else ret = FALSE;
    return ( ret );
}

int ScanResIndex ( int aChass, int aIndex )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].chass == aChass )
                && ( scanRes [retVal].index == aIndex );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 15 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [12] )
            {
                case 2: /* cpqHeTemperatureIndex */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].chass = aVar->name [13];
                    scanRes [index].index = aVar->name [14];
                    scanRes [index].v3 = FALSE;
                    scanRes [index].v4 = FALSE;
                    scanRes [index].v5 = FALSE;
                    scanRes [index].v7 = FALSE;
                    scanRes [index].local = NULL;
                    scanRes [index].tempT = -1;
                break;
                case 3: /* cpqHeTemperatureLocale */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    if ( ( aVar->type == ASN_INTEGER ) )
                    {
                        scanRes [index].v3 = TRUE;
                        scanRes [index].local =
                                DecodeCPQLocal ( *(aVar->val.integer) );
                    } /* else nothing to do */
                break;
                case 4: /* cpqHeTemperatureCelsius */
                {
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v4 = ( aVar->type == ASN_INTEGER );
                }
                break;
                case 5: /* cpqHeTemperatureThreshold */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v5 = ( aVar->type == ASN_INTEGER );
                    scanRes [index].tempT = *(aVar->val.integer);
                break;
                case 7: /* cpqHeTemperatureThresholdType */
                    index = ScanResIndex ( aVar->name [13], aVar->name [14] );
                    scanRes [index].v7 = ( aVar->type == ASN_INTEGER );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v3 && scanRes [i].v4 && scanRes [i].v5 &&
                scanRes [i].v7  )
        {
            retVal = STATE_OK;
            printf ( "OK chassis: %i ; index: %i ; tempT: %i ; local: %s\n",
                    scanRes [i].chass, scanRes [i].index, scanRes [i].tempT,
                    scanRes [i].local );
            free ( scanRes [i].local );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_local [15];
    oid o_tempC [15];
    oid o_tempT [15];
    oid o_type  [15];
    size_t oidLen;

    OidCpy ( o_local, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_tempC, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_tempT, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_type,  &oidLen, baseOID, &baseOIDLen );
    o_local [12] = 3; o_local [13] = chass; o_local [14] = idx;
    o_tempC [12] = 4; o_tempC [13] = chass; o_tempC [14] = idx;
    o_tempT [12] = 5; o_tempT [13] = chass; o_tempT [14] = idx;
    o_type  [12] = 7; o_type  [13] = chass; o_type  [14] = idx;
    oidLen = 15;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.local = ReadInteger ( sess2, o_local, oidLen );
    v.tempC = ReadInteger ( sess2, o_tempC, oidLen );
    v.tempT = ReadInteger ( sess2, o_tempT, oidLen );
    v.type  = ReadInteger ( sess2, o_type,  oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "tempC",  GAUGE, rrdHBeat, "-20", "150" );
    params [i++] = CreateDSStr ( "tempT",  GAUGE, rrdHBeat, "-20", "150" );
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    char buf[1024];

    chk = sscanf ( aData, "%1023s;%i;%i", buf, &(v.tempC), &(v.tempT) );
    if ( chk != 3 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
    addInf = StrDup ( buf );
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* tempC  = IntToStr    ( v.tempC );
    char* tempT  = IntToStr    ( v.tempT );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "tempC:tempT" );
    params [i++] = StrConcat ( 5, ts, ":", tempC, ":", tempT );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( tempT );
    free ( tempC );
    free ( ts );
}

void PlotDataTemp ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[deg C]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, NULL, aTiming );
    PPAddPlot ( pp, "temperature", "deg C", FALSE );
    PPAddPlot ( pp, "threshold",   "deg C", FALSE );
    title = StrConcat ( 4, hostName, " temperature ", addInf, "" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, FALSE );
    params [i++] = CreateDEF ( "tempCM", aFN, "tempC", MAX );
    params [i++] = CreateDEF ( "tempCA", aFN, "tempC", AVG );
    params [i++] = CreateDEF ( "tempTM", aFN, "tempT", MAX );
    params [i++] = CreatePlot ( PLOT_AREA,  "tempCM", C_GAUGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "tempCA", C_GAUGE1_AVG, pp, 0, AVG, TRUE );
    params [i++] = CreatePlot ( PLOT_LINE2, "tempTM", C_ERR2_MAX,   pp, 1, MAX, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "tempCA", "tempCA", "tempCM", "tempCM", "%4.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "tempTM", "tempTM", "tempTM", "tempTM", "%4.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataTemp ( fn, PLOT_24H_5M,  now );
    PlotDataTemp ( fn, PLOT_7D_15M,  now );
    PlotDataTemp ( fn, PLOT_31D_60M, now );
    PlotDataTemp ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

char* DecodeCPQTempType ( int aType )
{
    char* ret = NULL;

    switch ( aType )
    {
        case  1: ret = StrDup ( "unknown" );    break;
        case  5: ret = StrDup ( "'blowout'" );  break;
        case  9: ret = StrDup ( "'caution'" );  break;
        case 15: ret = StrDup ( "'critical'" ); break;
        default: ret = StrDup ( "invalid" );    break;
    }
    return ret;
}

int CheckData ()
{
    int retVal = STATE_OK;
    char* local = NULL;
    char* type = NULL;

    ReadVals ();
    if ( ( v.tempC > warn ) && ( retVal < STATE_WARNING ) )
            retVal = STATE_WARNING;
    if ( ( v.tempC > crit ) && ( retVal < STATE_CRITICAL ) )
            retVal = STATE_CRITICAL;
    local = DecodeCPQLocal ( v.local );
    type = DecodeCPQTempType ( v.type );
    PrintExit ( retVal );
    printf ( " - %s sensor, %s type, %i/%i deg C", local, type,
            v.tempC, v.tempT );
    if ( npPerfData )
    {
        printf ( "|'current temperature'=%i;%u;%u", v.tempC, warn, crit );
        printf ( "'threshold temperature'=%i\n", v.tempT );
    }
    else
    {
        printf ( "|%s;%i;%i\n", addInf, v.tempC, v.tempT );
    }
    free ( type );
    free ( local );
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( checkData )
    {
        if ( !isWarnSet  ) Usage ( "Missing 'warn' parameters" );
        if ( !isCritSet  ) Usage ( "Missing 'crit' parameters" );
        if ( !isChassSet ) Usage ( "Missing 'system chassis' parameters" );
    }
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ();
        if ( plotPerfData ) PlotPerfData ();
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
