/*
 * File:      check_cpq_thermal.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include "decoders.h"
#include "globals.h"
#include "nagiosif.h"
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int v0;      /* found cpqHeThermalCondition */
} tScanRes;

/* data type for storing data */
typedef struct
{
    int   cond;  /* condition */
} tDataSet;


/*
 * CPQHLTH-MIB::cpqHeThermalCondition.0 = .1.3.6.1.4.1.232.6.2.6.1.0
 */
oid baseOID [] = { 1, 3, 6, 1, 4, 1, 232, 6, 2, 6, 1 };
size_t baseOIDLen = 11;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_cpq_ida";
    progVersion = "0.8";
    helpIndexP = NULL;
    helpThresP = NULL;
    helpPurpose = "Checks the status of a Compaq/HP IDA Controller.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
    /* rrdStep is unused */
    /* rrdHBeat is unused */
    supportsPerfLog = FALSE;
    hasIndexParam = FALSE;
}

void HelpSpecificParams ()
{
    /* No specific parameters */
}

void HelpExplain ()
{
    printf (
    "The plugin returns the overall thermal state of a Compaq server.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "There is no performance data provided.\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    return FALSE;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 12 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [11] )
            {
                case 0: /* cpqHeThermalCondition.0 */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].v0 = TRUE;
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if ( scanRes [i].v0 )
        {
            retVal = STATE_OK;
            printf ( "OK\n" );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_cond [12];
    size_t oidLen;

    OidCpy ( o_cond, &oidLen, baseOID, &baseOIDLen );
    o_cond [11] = 0;
    oidLen = 12;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.cond = ReadInteger ( sess2, o_cond, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

void CreateDataFile ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

void ParseData ( const char* aData )
{
    /* performance data is not supported */
}

void WriteData ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

int CheckData ()
{
    int retVal = STATE_UNKNOWN;

    ReadVals ();
    switch ( v.cond )
    {
        case 1:
            retVal = STATE_WARNING;
            printf ( "WARNING - Thermal state is 'other'\n" );
        break;
        case 2:
            retVal = STATE_OK;
            printf ( "OK - Thermal state is 'ok'\n" );
        break;
        case 3:
            retVal = STATE_WARNING;
            printf ( "OK - Thermal state is 'degraded'\n" );
        break;
        case 4:
            retVal = STATE_CRITICAL;
            printf ( "OK - Thermal state is 'failed'\n" );
        break;
        default:
            retVal = STATE_UNKNOWN;
            printf ( "OK - Thermal state is invalid\n" );
        break;
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
