/*
 * File:      check_if_by_snmp.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

/* ### Comments like this describe what to put where if you want to    ### */
/* ### write your own plugin. Usually they will be above the line(s)   ### */
/* ### which has to be changed. There is one thing which is to change  ### */
/* ### above: The "File:"-line.                                        ### */

#if HAVE_CONFIG_H
#include <config.h>
#endif

/* ### You'll probably have to '#include <errno.h>'                    ### */
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    /* ### Put an int here for every level of indices used in the      ### */
    /* ### oids.                                                       ### */
    int index;   /* interface index */
    /* ### Put an int (used as boolean) here for every variable the    ### */
    /* ### plugin accesses when testing the service.                   ### */
    int v2;      /* found ifDescr */
    int v3;      /* found ifType */
    int v5;      /* found ifSpeed */
    int v7;      /* found ifAdminStatus */
    int v8;      /* found ifOperStatus */
    int v10;     /* found ifInOctets */
    int v11;     /* found ifInUcastPkts */
    int v14;     /* found ifInErrors */
    int v16;     /* found ifOutOctets */
    int v17;     /* found ifOutUcastPkts */
    int v20;     /* found ifOutErrors */
    /* ### Put a field of whatever type is appropriate here for each   ### */
    /* ### variable which is written to stdout by scanning.            ### */
    char* descr; /* interface description */
    int type;    /* interface type */
    uint speed;  /* current or nominal speed */
    int aStat;   /* administrative status */
} tScanRes;

/* data type for storing data */
typedef struct
{
    /* ### Put a field of whatever type is appropriate here for each   ### */
    /* ### variable which is needed for testing. For the "appropriate  ### */
    /* ### type" see snmpif.h, Read*() functions.                      ### */
    int  aStat;
    int  oStat;
    uint speed;
    uint iOct;
    uint iPkt;
    uint iErr;
    uint oOct;
    uint oPkt;
    uint oErr;
} tDataSet;


/* ### This is just a comment but I'd strongly suggest you put you     ### */
/* ### OIDs here.                                                      ### */
/*
 * SNMP base OID
 * IF-MIB::ifAdminStatus.IDX  = .1.3.6.1.2.1.2.2.1.7.IDX
 * IF-MIB::ifOperStatus.IDX   = .1.3.6.1.2.1.2.2.1.8.IDX
 * IF-MIB::ifSpeed.IDX        = .1.3.6.1.2.1.2.2.1.5.IDX
 * IF-MIB::ifInOctets.IDX     = .1.3.6.1.2.1.2.2.1.10.IDX
 * IF-MIB::ifInUcastPkts.IDX  = .1.3.6.1.2.1.2.2.1.11.IDX
 * IF-MIB::ifInErrors.IDX     = .1.3.6.1.2.1.2.2.1.14.IDX
 * IF-MIB::ifOutOctets.IDX    = .1.3.6.1.2.1.2.2.1.16.IDX
 * IF-MIB::ifOutUcastPkts.IDX = .1.3.6.1.2.1.2.2.1.17.IDX
 * IF-MIB::ifOutErrors.IDX    = .1.3.6.1.2.1.2.2.1.20.IDX
 */
/* ### The "baseOID" is the common part of all OIDs used for testing.  ### */
oid baseOID[] = { 1, 3, 6, 1, 2, 1, 2, 2, 1 };
/* ### The "baseOIDLen" is the length of the array above.              ### */
size_t baseOIDLen = 9;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    /* ### The service name.                                           ### */
    srvName = "check_if_by_snmp";
    /* ### Version of this plugin.                                     ### */
    progVersion = "0.9";
    /* ### Synopsis of any index used for testing.                     ### */
    helpIndexP = "-i INDEX";
    /* ### Synopsis of any other parameter used for testing.           ### */
    helpThresP = NULL;
    /* ### (Very) short description of the programs purpuse.           ### */
    helpPurpose = "Checks the status of a network interface.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    /* ### RRD Step in seconds (usually 5*60, 1*60 with CPU usages)    ### */
    rrdStep = 5 * 60;
    /* ### RRD Heartbeat in seconds ( = ( rrdStep * 2 + 1 ) * 60 ).    ### */
    rrdHBeat = 11 * 60;
#endif /* WITH_RRD */
    /* ### TRUE if the plugin supports logging and plotting.           ### */
    supportsPerfLog = TRUE;
    /* ### TRUE if the plugin has the "-i index" parameter.            ### */
    hasIndexParam = TRUE;
    /* ### TRUE if the plugin has the "-I addInf" parameter.           ### */
    hasAddInfParam = TRUE;
}

void HelpSpecificParams ()
{
    /* ### Help text for the program specific parameters.              ### */
    printf (
    "-i, --index INTEGER\n"
    "    Index of the network interface which should be checked.\n"
    "\n"
    );
}

void HelpExplain ()
{
    /* ### Help text describing the plugin.                            ### */
    printf (
    "If the interface is up, then the plugin generates an OK. If it is\n"
    "administratively down, then the plugin generates a warning. If it is\n"
    "down and administratively enabled, then the plugin returns critical.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    /* ### Description of the performance data provided. ### */
    printf (
    "Native performance data is provided in the format:\n"
    "<addInf>;<speed>;<oStat>;<iOct>;<iPkt>;<iErr>;<oOct>;<oPkt>;<oErr>\n"
    "where\n"
    "    addInf ... additional information.\n"
    "    speed  ... estimated maximum speed of the interface\n"
    "    oStat  ... operational state, c.f. RFC1573, pp. 28-29\n"
    "    iOct   ... Input octet counter\n"
    "    iPkt   ... Input packet counter\n"
    "    iErr   ... Input error counter\n"
    "    oOct   ... Output octet counter\n"
    "    oPkt   ... Output packet counter\n"
    "    oErr   ... Output error counter\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'estimated speed'=<speed> "
    "'input octets'=<iOct>c "
    "'input packets'=<iPkt>c "
    "'input errors'=<iErr>c "
    "'output octets'=<oOct>c "
    "'output packets'=<oPkt>c "
    "'output errors'=<oErr>c\n"
    "\n"
    "Note: Nagios Plugin performance data format does not allow bits per\n"
    "second as unit of measurement.\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    /* ### Checks for program specific parameters. See e.g.            ### */
    /* ### check_cpq_phydrv.c check_ucd_snmp_load.c and                ### */
    /* ### check_netapp_vol.c for examplex                             ### */
    return FALSE;
}

/* ### For a "one index" plugin this will always look like this.       ### */
/* ### Several indices are used in e.g. check_cpq_phydrv.c. "Named"    ### */
/* ### OIDs are used in check_winf_cpuusage.c.                         ### */
int ScanResIndex ( int aIndex )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

/* ### Okay, now this is a bit difficult. This function gets called ### */
/* ### for each variable encountered during the walk.               ### */
void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    /* ### Checks the length of the oid returned by the walk. This  ### */
    /* ### is usually a redundant check. The length ("11") has to   ### */
    /* ### be set appropriately                                     ### */
    if ( aVar->name_length == 11 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            /* ### Switch by the "type" part of the object ID. This is ### */
            /* ### usually the number at [baseOIDLen]                  ### */
            switch ( aVar->name [9] )
            {
                /* ### Do whatever is needed -- sorry, you'll have     ### */
                /* ### to understand the code.                         ### */
                case 1: /* ifIndex */
                    /* set the index of result */
                    index = scanResLen;
                    /* allocate a new result record */
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    /* initialize new resource record */
                    scanRes [index].index = aVar->name [10];
                    scanRes [index].v2 = FALSE;
                    scanRes [index].v3 = FALSE;
                    scanRes [index].v5 = FALSE;
                    scanRes [index].v7 = FALSE;
                    scanRes [index].v8 = FALSE;
                    scanRes [index].v10 = FALSE;
                    scanRes [index].v11 = FALSE;
                    scanRes [index].v14 = FALSE;
                    scanRes [index].v16 = FALSE;
                    scanRes [index].v17 = FALSE;
                    scanRes [index].v20 = FALSE;
                    scanRes [index].descr = NULL;
                    scanRes [index].type = -1;
                    scanRes [index].speed = 0;
                    scanRes [index].aStat = -1;
                break;
                /* ### Just checking a variable is simplier.           ### */
                case 2: /* ifDescr */
                    /* find the index of the result record */
                    index = ScanResIndex ( aVar->name [10] );
                    /* check the type of variable */
                    /* ### Change field name and expected type.        ### */
                    scanRes [index].v2 = ( aVar->type == ASN_OCTET_STR );
                    /* If the value is used for printing, then save it. */
                    /* ### Change field name and right side term.      ### */
                    scanRes [index].descr = VarListStrDup ( aVar );
                break;
                case 3: /* ifType */
                    index = ScanResIndex ( aVar->name [10] );
                    scanRes [index].v3 = ( aVar->type == ASN_INTEGER );
                    scanRes [index].type = *(aVar->val.integer);
                break;
                case 5: /* ifSpeed */
                    index = ScanResIndex ( aVar->name [10] );
                    scanRes [index].v5 = ( aVar->type == ASN_GAUGE );
                    scanRes [index].speed = (uint) *(aVar->val.integer);
                break;
                case 7: /* ifAdminStatus */
                    index = ScanResIndex ( aVar->name [10] );
                    scanRes [index].v7 = ( aVar->type == ASN_INTEGER );
                    scanRes [index].aStat = *(aVar->val.integer);
                break;
                case 8: /* ifOperStatus */
                    index = ScanResIndex ( aVar->name [10] );
                    scanRes [index].v8 = ( aVar->type == ASN_INTEGER );
                break;
                case 10: /*ifInOctets  */
                    index = ScanResIndex ( aVar->name [10] );
                    scanRes [index].v10 = ( aVar->type == ASN_COUNTER );
                break;
                case 11: /* ifInUcastPkts */
                    index = ScanResIndex ( aVar->name [10] );
                    scanRes [index].v11 = ( aVar->type == ASN_COUNTER );
                break;
                case 14: /* ifInErrors */
                    index = ScanResIndex ( aVar->name [10] );
                    scanRes [index].v14 = ( aVar->type == ASN_COUNTER );
                break;
                case 16: /* ifOutOctets */
                    index = ScanResIndex ( aVar->name [10] );
                    scanRes [index].v16 = ( aVar->type == ASN_COUNTER );
                break;
                case 17: /* ifOutUcastPkts */
                    index = ScanResIndex ( aVar->name [10] );
                    scanRes [index].v17 = ( aVar->type == ASN_COUNTER );
                break;
                case 20: /* ifOutErrors */
                    index = ScanResIndex ( aVar->name [10] );
                    scanRes [index].v20 = ( aVar->type == ASN_COUNTER );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        /* ### Add the field for each "found" check.                   ### */
        if (    scanRes [i].v2  && scanRes [i].v3  && scanRes [i].v5  &&
                scanRes [i].v7  && scanRes [i].v8  && scanRes [i].v10 &&
                scanRes [i].v11 && scanRes [i].v14 && scanRes [i].v16 &&
                scanRes [i].v17 && scanRes [i].v20 )
        {
            retVal = STATE_OK;
            /* ### Print the scanner result line.                      ### */
            printf ( "OK index: %i ; type: %i ; aStat: %i ; descr: %s\n",
                    scanRes [i].index, scanRes [i].type, scanRes [i].aStat,
                    scanRes [i].descr );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    /* ### Declare an oid for each object which should be queried.     ### */
    oid o_aStat [11];
    oid o_oStat [11];
    oid o_speed [11];
    oid o_iOct  [11];
    oid o_iPkt  [11];
    oid o_iErr  [11];
    oid o_oOct  [11];
    oid o_oPkt  [11];
    oid o_oErr  [11];
    size_t oidLen;

    /* ### Copy the contents of the base object ID to each of the      ### */
    /* ### oids declared above.                                        ### */
    OidCpy ( o_aStat, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_oStat, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_speed, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_iOct, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_iPkt, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_iErr, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_oOct, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_oPkt, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_oErr, &oidLen, baseOID, &baseOIDLen );
    /* ### Set the "type" and "index" part of the oids.                ### */
    o_aStat [9] = 7;  o_aStat [10] = idx;
    o_oStat [9] = 8;  o_oStat [10] = idx;
    o_speed [9] = 5;  o_speed [10] = idx;
    o_iOct  [9] = 10; o_iOct  [10] = idx;
    o_iPkt  [9] = 11; o_iPkt  [10] = idx;
    o_iErr  [9] = 14; o_iErr  [10] = idx;
    o_oOct  [9] = 16; o_oOct  [10] = idx;
    o_oPkt  [9] = 17; o_oPkt  [10] = idx;
    o_oErr  [9] = 20; o_oErr  [10] = idx;
    /* ### Set the length of the object IDs.                           ### */
    oidLen = 11;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    /* ### Read each value needed.                                     ### */
    v.aStat = ReadInteger   ( sess2, o_aStat, oidLen );
    v.oStat = ReadInteger   ( sess2, o_oStat, oidLen );
    v.speed = ReadGauge32   ( sess2, o_speed, oidLen );
    v.iOct  = ReadCounter32 ( sess2, o_iOct,  oidLen );
    v.iPkt  = ReadCounter32 ( sess2, o_iPkt,  oidLen );
    v.iErr  = ReadCounter32 ( sess2, o_iErr,  oidLen );
    v.oOct  = ReadCounter32 ( sess2, o_oOct,  oidLen );
    v.oPkt  = ReadCounter32 ( sess2, o_oPkt,  oidLen );
    v.oErr  = ReadCounter32 ( sess2, o_oErr,  oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
/* ### You probably want to check "man 1 rrdcreate"                    ### */
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    /* ### Create an Data Source for each value you want to log.       ### */
    /* ### Change the name of the Data Source ("iOct" etc.).           ### */
    /* ### Change the type the Data Source ("COUNTER", c.f. rrdif.h).  ### */
    /* ### Change the lower and upper bounds. "U" means undefined.     ### */
    params [i++] = CreateDSStr ( "iOct",  COUNTER, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "iPkt",  COUNTER, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "iErr",  COUNTER, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "oOct",  COUNTER, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "oPkt",  COUNTER, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "oErr",  COUNTER, rrdHBeat, "0", "U" );
    /* ### Create an RRA statement. Usually those mentioned here are   ### */
    /* ### sufficient, only with cpu usages there are additional       ### */
    /* ### timings. See rrdif.h for further information.               ### */
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

/* Parse performance data */
void ParseData ( const char* aData )
{
    int chk;
    char buf[1024];

    /* ### Set up the format string to match the performance data      ### */
    /* ### format.                                                     ### */
    chk = sscanf ( aData, "%1023s;%u;%i;%u;%u;%u;%u;%u;%u",
            buf, &(v.speed), &(v.oStat), &(v.iOct), &(v.iPkt), &(v.iErr),
            &(v.oOct), &(v.oPkt), &(v.oErr) );
    /* ### Set up the number of assignments ("9").                     ### */
    if ( chk != 9 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
    /* ### Assign the string buffer to the appropriate internal        ### */
    /* ### variable.                                                   ### */
    addInf = StrDup ( buf );
}

/* Write data to the RRD, c.f. man 1 rrdupdate. */
void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* iOct   = UIntToStr   ( v.iOct );
    char* iPkt   = UIntToStr   ( v.iPkt );
    char* iErr   = UIntToStr   ( v.iErr );
    char* oOct   = UIntToStr   ( v.oOct );
    char* oPkt   = UIntToStr   ( v.oPkt );
    char* oErr   = UIntToStr   ( v.oErr );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    /* ### Set up the template.                                        ### */
    params [i++] = StrDup ( "iOct:iPkt:iErr:oOct:oPkt:oErr" );
    /* ### Set up the data parameter, don't forget the timestamp.      ### */
    params [i++] = StrConcat ( 13, ts, ":", iOct, ":", iPkt, ":", iErr,
            ":", oOct, ":", oPkt, ":", oErr );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( oErr );
    free ( oPkt );
    free ( oOct );
    free ( iErr );
    free ( iPkt );
    free ( iOct );
    free ( ts );
}

/* Create one plot. See man 1 rrdgraph for further information. There are */
/* probably more of these functions.                                      */
void PlotDataOct ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    /* ### Set up the label of the y-Axis                              ### */
    char* vLabel = "[O/s]  <-- out | in -->";
    char* params[1024]; /* Now that should be enough! */

    /* ### Change the plot designation ("oct"). Use "NULL" if only one ### */
    /* ### plot is made.                                               ### */
    FNSetPNGData ( aFN, "oct", aTiming );
    /* ### Set up a bit of information, one line for each plot.        ### */
    /* ### Set up the name of the plot ("input"), the unit ("O/s") and ### */
    /* ### if the plot will be stacked.                                ### */
    PPAddPlot ( pp, "input",  "O/s", FALSE );
    PPAddPlot ( pp, "output", "O/s", FALSE );
    /* ### Set up the title of the plot.                               ### */
    title = StrConcat ( 4, hostName, " interface ", addInf, " transfer (octets)" );
    /* ### Add basic parameters for the plot. Replace 1000 with 1024   ### */
    /* ### if binary prefixes should be used. Replace PLOT_SI_AUTO     ### */
    /* ### with the appropriate PLOT_SI_-constant for fixed prefixes   ### */
    /* ### (c.f. rrdif.h). "0, 0" are fixed lower and upper bounds     ### */
    /* ### which are only used if lower < upper. Use TRUE if the plot  ### */
    /* ### used negative values.                                       ### */
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_AUTO, 0, 0, TRUE );
    /* ### Add a "DEF" (a virtual data source). "iOctMR" is the        ### */
    /* ### virtual name of the data source. "iOct" is the name of the  ### */
    /* ### data source. "MAX" is the consolidation function used (c.f. ### */
    /* ### rrdif.h).                                                   ### */
    params [i++] = CreateDEF ( "iOctMR", aFN, "iOct", MAX );
    params [i++] = CreateDEF ( "iOctAR", aFN, "iOct", AVG );
    params [i++] = CreateDEF ( "oOctMR", aFN, "oOct", MAX );
    params [i++] = CreateDEF ( "oOctAR", aFN, "oOct", AVG );
    /* ### Add whatever math is necessary.                             ### */
    params [i++] = StrDup ( "CDEF:iOctM=iOctMR" );
    params [i++] = StrDup ( "CDEF:iOctA=iOctAR" );
    params [i++] = StrDup ( "CDEF:oOctM=oOctMR,-1,*" );
    params [i++] = StrDup ( "CDEF:oOctA=oOctAR,-1,*" );
    /* ### Add plot statements. "PLOT_AREA" is the type of the plot    ### */
    /* ### (c.f. rrdif.h -- you guessed that, didn't you?). "iOctM" is ### */
    /* ### the virtual data source which has to be plotted.            ### */
    /* ### "C_THRU_MAX" is just a color specification (c.f. rrdif.h).  ### */
    /* ### "0" is the index into pp[]. "MAX" is the consilidation      ### */
    /* ### function. "FALSE" is used if there is no line feed to be    ### */
    /* ### added to the legend.                                        ### */
    params [i++] = CreatePlot ( PLOT_AREA,  "iOctM", C_THRU_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "iOctA", C_THRU_AVG, pp, 0, AVG, TRUE );
    params [i++] = CreatePlot ( PLOT_AREA,  "oOctM", C_THRU_MAX, pp, 1, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "oOctA", C_THRU_AVG, pp, 1, AVG, TRUE );
    /* ### Print the minimum, average, maximum and last values of a    ### */
    /* ### plot. Replace the data sources as appropriate. "%5.2lf" are ### */
    /* ### printf-like formats for the values printed.                 ### */
    i = PlotPrintParams ( i, params, pp, 0, "iOctAR", "iOctAR", "iOctMR", "iOctMR", "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "oOctAR", "oOctAR", "oOctMR", "oOctMR", "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    /* ### Free the title (if needed).                                 ### */
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataPkt ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[packets/s]  <-- out | in -->";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "pkt", aTiming );
    PPAddPlot ( pp, "input",  "packets/s", FALSE );
    PPAddPlot ( pp, "output", "packets/s", FALSE );
    title = StrConcat ( 4, hostName, " interface ", addInf, " transfer (packets)" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_AUTO, 0, 0, TRUE );
    params [i++] = CreateDEF ( "iPktMR", aFN, "iPkt", MAX );
    params [i++] = CreateDEF ( "iPktAR", aFN, "iPkt", AVG );
    params [i++] = CreateDEF ( "oPktMR", aFN, "oPkt", MAX );
    params [i++] = CreateDEF ( "oPktAR", aFN, "oPkt", AVG );
    params [i++] = StrDup ( "CDEF:iPktM=iPktMR" );
    params [i++] = StrDup ( "CDEF:iPktA=iPktAR" );
    params [i++] = StrDup ( "CDEF:oPktM=oPktMR,-1,*" );
    params [i++] = StrDup ( "CDEF:oPktA=oPktAR,-1,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "iPktM", C_THRU_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "iPktA", C_THRU_AVG, pp, 0, AVG, TRUE );
    params [i++] = CreatePlot ( PLOT_AREA,  "oPktM", C_THRU_MAX, pp, 1, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "oPktA", C_THRU_AVG, pp, 1, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "iPktAR", "iPktAR", "iPktMR", "iPktMR", "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "oPktAR", "oPktAR", "oPktMR", "oPktMR", "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataErr ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[errors/s]  <-- out | in -->";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "err", aTiming );
    PPAddPlot ( pp, "input",  "errors/s", FALSE );
    PPAddPlot ( pp, "output", "errors/s", FALSE );
    title = StrConcat ( 4, hostName, " interface ", addInf, " errors" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_AUTO, 0, 0, TRUE );
    params [i++] = CreateDEF ( "iErrMR", aFN, "iErr", MAX );
    params [i++] = CreateDEF ( "iErrAR", aFN, "iErr", AVG );
    params [i++] = CreateDEF ( "oErrMR", aFN, "oErr", MAX );
    params [i++] = CreateDEF ( "oErrAR", aFN, "oErr", AVG );
    params [i++] = StrDup ( "CDEF:iErrM=iErrMR" );
    params [i++] = StrDup ( "CDEF:iErrA=iErrAR" );
    params [i++] = StrDup ( "CDEF:oErrM=oErrMR,-1,*" );
    params [i++] = StrDup ( "CDEF:oErrA=oErrAR,-1,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "iErrM", C_ERR2_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "iErrA", C_ERR2_AVG, pp, 0, AVG, TRUE );
    params [i++] = CreatePlot ( PLOT_AREA,  "oErrM", C_ERR2_MAX, pp, 1, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "oErrA", C_ERR2_AVG, pp, 1, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "iErrAR", "iErrAR", "iErrMR", "iErrMR", "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "oErrAR", "oErrAR", "oErrMR", "oErrMR", "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    /* ### Call all plot functions.                                    ### */
    PlotDataOct ( fn, PLOT_24H_5M,  now );
    PlotDataPkt ( fn, PLOT_24H_5M,  now );
    PlotDataErr ( fn, PLOT_24H_5M,  now );
    PlotDataOct ( fn, PLOT_7D_15M,  now );
    PlotDataPkt ( fn, PLOT_7D_15M,  now );
    PlotDataErr ( fn, PLOT_7D_15M,  now );
    PlotDataOct ( fn, PLOT_31D_60M, now );
    PlotDataPkt ( fn, PLOT_31D_60M, now );
    PlotDataErr ( fn, PLOT_31D_60M, now );
    PlotDataOct ( fn, PLOT_366D_4H, now );
    PlotDataPkt ( fn, PLOT_366D_4H, now );
    PlotDataErr ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

/* ### Sorry, this one has to be written anew most of the times.       ### */
/* ### There is actually a bunch of things done here:                  ### */
/* ###  - read the values (i.e. call ReadVals()).                      ### */
/* ###  - figure out whether to raise an alert.                        ### */
/* ###  - write output.                                                ### */
int CheckData ()
{
    int retVal = STATE_OK;

    /* Read the values. */
    ReadVals ();
    /* Consolidate status and print output line. */
    if ( v.aStat == 1 ) /* that's "up" */
    {
        if ( v.oStat == 1 ) /* that's "up" */
        {
            printf ( "OK - %s: I: %u kO, %u Pkt; O: %u kO, %u Pkt",
                    addInf, v.iOct / 1024, v.iPkt, v.oOct / 1024, v.oPkt );
        }
        else if ( v.oStat == 5 ) /* that's "dormant" */
        {
            printf ( "OK - %s is dormant", addInf );
        }
        else
        {
            char* stat = NULL;

            switch ( v.oStat )
            {
                /* case 1: already done */
                case 2: stat = "is down"; break;
                case 3: stat = "is testing"; break;
                case 4: stat = "status is unknown"; break;
                /* case 5: already done */
                case 6: stat = "misses some component"; break;
                case 7: stat = "is down due to some lower layer"; break;
                default: stat = "status is invalid"; break;
            }
            printf ( "CRITICAL - %s %s", addInf, stat );
            retVal = STATE_CRITICAL;
        }
    }
    else
    {
        printf ( "WARNING - %s is administratively down", addInf );
        retVal = STATE_WARNING;
    }
    /* ### There are always two ways of writing performance data.      ### */
    /* ###  - the "Nagios Plugins" compatible way (if (npPerfData))    ### */
    /* ###  - the native way (if (!npPerfData))                        ### */
    /* ### Make sure the output matches the descriptions given in      ### */
    /* ### HelpPerformanceData()!                                      ### */
    if ( npPerfData )
    {
        printf ( "|'estimated speed'=%u", v.speed );
        printf ( " 'input octets'=%uc", v.iOct );
        printf ( " 'input packets'=%uc", v.iPkt );
        printf ( " 'input errors'=%uc", v.iErr );
        printf ( " 'output octets'=%uc", v.oOct );
        printf ( " 'output packets'=%uc", v.oPkt );
        printf ( " 'output errors'=%uc\n", v.oErr );
    }
    else
    {
        printf ( "|%s;%u;%i;%u;%u;%u;%u;%u;%u\n",
                addInf, v.speed, v.oStat,
                v.iOct, v.iPkt, v.iErr, v.oOct, v.oPkt, v.oErr );
    }
    /* Return the status */
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    /* ### Add checks if parameters are set                            ### */
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ();
        if ( plotPerfData ) PlotPerfData ();
#endif /* WITH_RRD */
    }
    /* ### Free whatever is left.                                      ### */
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
