/*
 * File:      check_mail_server.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


typedef struct
{
    int v1;
    int v101;
    int v102;
    int v103;
    int v104;
    int v105;
    int v106;
    int v107;
    int v108;
    int v109;
    int v110;
    int v111;
    int v2;
    int v201;
    int v202;
    int v203;
    int v204;
    int v205;
    int v206;
    int v207;
    int v208;
    int v209;
    int v210;
    int v211;
} tScanRes;

/* data type for storing data */
typedef struct
{
    uint miQ;  /* messages in queue */
    uint soQ;  /* size of queue */
    uint mRec; /* messages received */
    uint succ; /* successful deliveries */
    uint fail; /* failed deliveries */
    uint def;  /* deferred deliveries */
    uint spam; /* identified spams */
    uint ham;  /* identified hams */
    uint mTL;  /* message too large */
    uint rDD;  /* relaying denied (because of invalid) domain */
    uint rDU;  /* relaying denied (because of invalid) user */
} tDataSet;

/* global input variables */
uint warn;
uint crit;
int isWarnSet  = FALSE;
int isCritSet  = FALSE;

/*
 * The plugins have to be configured using
 * pass .1.3.6.1.4.1.2021.255 /usr/local/sbin/qmail-snmp
 * or
 * pass .1.3.6.1.4.1.2021.255 /usr/local/sbin/postfix-snmp
 * in /etc/snmpd.conf which makes the agent use a part of the UCD enterprise
 * MIB for them.
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 2021, 255 };
size_t baseOIDLen = 8;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_mail_server";
    progVersion = "0.8";
    helpIndexP = NULL;
    helpThresP = "-w WARN -c CRIT";
    helpPurpose = "Checks the state of a mail relay/spam filter.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 5 * 60;
    rrdHBeat = 11 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = FALSE;
}

void HelpSpecificParams ()
{
    printf (
    "-w, --warning INTEGER\n"
    "    Number of messages in the queue at which a warning is generated.\n"
    "\n"
    "-c, --critical INTEGER\n"
    "    Number of messages in the queue at which a critical is generated.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "This plugin uses an extension to the net-snmmp or ucd-snmp on the\n"
    "remote host which needs to be installed and configured.\n"
    "\n"
    "The extension a/o mail server may not support all performance values.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<miQ>;<soQ>;<mRec>;<succ>;<fail>;<def>;<spam>;<ham>;<mTL>;<rDD>;<rDU>\n"
    "where\n"
    "    miQ  ... Messages in queue\n"
    "    soQ  ... Size of queued messages (KB)\n"
    "    mRec ... Messages received\n"
    "    succ ... Successful deliveries\n"
    );
    printf (
    "    fail ... Failed deliveries\n"
    "    def  ... Deferred deliveries\n"
    "    spam ... Messages marked as Spam\n"
    "    ham  ... Messages marked as Ham\n"
    "    mTL  ... Messag too large\n"
    "    rDD  ... Relaying denied (because of invalid) domain\n"
    "    rDU  ... Recipient unknown (because of invalid) user\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'messages in queue'=<miQ>;<warn>;<crit> "
    "'size of queued messages'=<soQ>KB "
    "'messages received'=<mRec>c "
    "'successful deliveries'=<succ>c "
    "'failed deliveries'=<fail>c "
    "'deferred deliveries'=<def>c "
    "'messages marked as Spam'=<spam>c "
    "'messages marked as Ham'=<ham>c "
    "'messag too large'=<mTL>c "
    "'relaying denied'=<rDD>c "
    "'recipient unknown'=<rDU>c\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;
    if (        !strcmp ( argv [(*i)], "-w" )
            ||  !strcmp ( argv [(*i)], "--warning" ) )
    {
        isWarnSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            warn = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || *tail ) Usage ( "Invalid 'warn' parameter" );
        }
        else Usage ( "Missing 'warn' parameter(s)" );
    } else if ( !strcmp ( argv [(*i)], "-c" )
            ||  !strcmp ( argv [(*i)], "--critical" ) )
    {
        isCritSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            crit = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || *tail ) Usage ( "Invalid 'crit' parameter" );
        }
        else Usage ( "Missing 'crit' parameter(s)" );
    } else ret = FALSE;
    return ( ret );
}

void SnmpWalkHandleVariable1 ( netsnmp_variable_list* aVar )
{
    char* str = NULL;

    if ( aVar->type == ASN_OCTET_STR )
    {
        str = VarListStrDup ( aVar );
        switch ( aVar->name [9] )
        {
            case  1:
                scanRes [0].v101 = !strcmp ( str, "messages in queue" );
            break;
            case  2:
                scanRes [0].v102 = !strcmp ( str, "size of queued messages" );
            break;
            case  3:
                scanRes [0].v103 = !strcmp ( str, "messages received" );
            break;
            case  4:
                scanRes [0].v104 = !strcmp ( str, "messages delivered" );
            break;
            case  5:
                scanRes [0].v105 = !strcmp ( str, "deliveries failed" );
            break;
            case  6:
                scanRes [0].v106 = !strcmp ( str, "deliveries deferred" );
            break;
            case  7:
                scanRes [0].v107 = !strcmp ( str, "messages marked as spam" );
            break;
            case  8:
                scanRes [0].v108 = !strcmp ( str, "messages marked as ham" );
            break;
            case  9:
                scanRes [0].v109 = !strcmp ( str, "messages too large" );
            break;
            case 10:
                scanRes [0].v110 = !strcmp ( str, "relaying denied" );
            break;
            case 11:
                scanRes [0].v111 = !strcmp ( str, "recipient unknown" );
            break;
            default:
                Log ( LOG_CRIT, "Unexpected OID while scanning services" );
            break;
        }
    } /* else do nothing */
    free ( str );
}

void SnmpWalkHandleVariable2 ( netsnmp_variable_list* aVar )
{
    switch ( aVar->name [9] )
    {
        case  1:
            scanRes [0].v201 = ( aVar->type == ASN_GAUGE );
        break;
        case  2:
            scanRes [0].v202 = ( aVar->type == ASN_GAUGE );
        break;
        case  3:
            scanRes [0].v203 = ( aVar->type == ASN_COUNTER );
        break;
        case  4:
            scanRes [0].v204 = ( aVar->type == ASN_COUNTER );
        break;
        case  5:
            scanRes [0].v205 = ( aVar->type == ASN_COUNTER );
        break;
        case  6:
            scanRes [0].v206 = ( aVar->type == ASN_COUNTER );
        break;
        case  7:
            scanRes [0].v207 = ( aVar->type == ASN_COUNTER );
        break;
        case  8:
            scanRes [0].v208 = ( aVar->type == ASN_COUNTER );
        break;
        case  9:
            scanRes [0].v209 = ( aVar->type == ASN_COUNTER );
        break;
        case 10:
            scanRes [0].v210 = ( aVar->type == ASN_COUNTER );
        break;
        case 11:
            scanRes [0].v211 = ( aVar->type == ASN_COUNTER );
        break;
        default:
            Log ( LOG_CRIT, "Unexpected OID while scanning services" );
        break;
    }
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    char* str = NULL;

    if ( aVar->type == ASN_OCTET_STR )
    {
        str = malloc ( aVar->val_len + 1 );
        if ( str == NULL ) ErrorMalloc ();
        strncpy ( str, (char*) (aVar->val.string), aVar->val_len );
        str [aVar->val_len] = '\0';
    } /* else do nothing */
    if ( ( aVar->name_length == 9 ) || ( aVar->name_length == 10 ) )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            switch ( aVar->name [8] )
            {
                case 1:
                    if ( aVar->name_length == 9 )
                    {
                        scanResLen = 1;
                        scanRes = realloc ( scanRes,
                                sizeof ( tScanRes ) * scanResLen );
                        if ( scanRes == NULL ) ErrorMalloc ();
                        scanRes [0].v1 = !strcmp ( str, "Descriptions" );
                        scanRes [0].v101 = FALSE;
                        scanRes [0].v102 = FALSE;
                        scanRes [0].v103 = FALSE;
                        scanRes [0].v104 = FALSE;
                        scanRes [0].v105 = FALSE;
                        scanRes [0].v106 = FALSE;
                        scanRes [0].v107 = FALSE;
                        scanRes [0].v108 = FALSE;
                        scanRes [0].v109 = FALSE;
                        scanRes [0].v110 = FALSE;
                        scanRes [0].v111 = FALSE;
                        scanRes [0].v2 = FALSE;
                        scanRes [0].v201 = FALSE;
                        scanRes [0].v202 = FALSE;
                        scanRes [0].v203 = FALSE;
                        scanRes [0].v204 = FALSE;
                        scanRes [0].v205 = FALSE;
                        scanRes [0].v206 = FALSE;
                        scanRes [0].v207 = FALSE;
                        scanRes [0].v208 = FALSE;
                        scanRes [0].v209 = FALSE;
                        scanRes [0].v210 = FALSE;
                        scanRes [0].v211 = FALSE;
                    }
                    else SnmpWalkHandleVariable1 ( aVar );
                break;
                case 2:
                    if ( aVar->name_length == 9 )
                    {
                        scanRes [0].v2 = !strcmp ( str, "Counters" );
                    }
                    else SnmpWalkHandleVariable2 ( aVar );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
    free ( str );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v1   && scanRes [i].v101 && scanRes [i].v102 &&
                scanRes [i].v103 && scanRes [i].v104 && scanRes [i].v105 &&
                scanRes [i].v106 && scanRes [i].v107 && scanRes [i].v108 &&
                scanRes [i].v109 && scanRes [i].v110 && scanRes [i].v111 &&
                scanRes [i].v2   && scanRes [i].v201 && scanRes [i].v202 &&
                scanRes [i].v203 && scanRes [i].v204 && scanRes [i].v205 &&
                scanRes [i].v206 && scanRes [i].v207 && scanRes [i].v208 &&
                scanRes [i].v209 && scanRes [i].v210 && scanRes [i].v211 )
        {
            retVal = STATE_OK;
            printf ( "OK\n" );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_miQ  [10];
    oid o_soQ  [10];
    oid o_mRec [10];
    oid o_succ [10];
    oid o_fail [10];
    oid o_def  [10];
    oid o_spam [10];
    oid o_ham  [10];
    oid o_mTL  [10];
    oid o_rDD  [10];
    oid o_rDU  [10];
    size_t oidLen;

    OidCpy ( o_miQ,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_soQ,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_mRec, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_succ, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_fail, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_def,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_spam, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_ham,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_mTL,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_rDD,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_rDU,  &oidLen, baseOID, &baseOIDLen );
    o_miQ  [8] = 2; o_miQ  [9] =  1;
    o_soQ  [8] = 2; o_soQ  [9] =  2;
    o_mRec [8] = 2; o_mRec [9] =  3;
    o_succ [8] = 2; o_succ [9] =  4;
    o_fail [8] = 2; o_fail [9] =  5;
    o_def  [8] = 2; o_def  [9] =  6;
    o_spam [8] = 2; o_spam [9] =  7;
    o_ham  [8] = 2; o_ham  [9] =  8;
    o_mTL  [8] = 2; o_mTL  [9] =  9;
    o_rDD  [8] = 2; o_rDD  [9] = 10;
    o_rDU  [8] = 2; o_rDU  [9] = 11;
    oidLen = 10;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.miQ  = ReadGauge32 ( sess2, o_miQ,  oidLen );
    v.soQ  = ReadGauge32 ( sess2, o_soQ,  oidLen );
    v.mRec = ReadCounter32 ( sess2, o_mRec,  oidLen );
    v.succ = ReadCounter32 ( sess2, o_succ, oidLen );
    v.fail = ReadCounter32 ( sess2, o_fail, oidLen );
    v.def  = ReadCounter32 ( sess2, o_def,  oidLen );
    v.spam = ReadCounter32 ( sess2, o_spam, oidLen );
    v.ham  = ReadCounter32 ( sess2, o_ham,  oidLen );
    v.mTL  = ReadCounter32 ( sess2, o_mTL,  oidLen );
    v.rDD  = ReadCounter32 ( sess2, o_rDD,  oidLen );
    v.rDU  = ReadCounter32 ( sess2, o_rDU,  oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "MiQ",  GAUGE,   rrdHBeat, "0", "10000" );
    params [i++] = CreateDSStr ( "SoQ",  GAUGE,   rrdHBeat, "0", "10000000" );
    params [i++] = CreateDSStr ( "MRec", COUNTER, rrdHBeat, "0", "10000" );
    params [i++] = CreateDSStr ( "Succ", COUNTER, rrdHBeat, "0", "10000" );
    params [i++] = CreateDSStr ( "Fail", COUNTER, rrdHBeat, "0", "10000" );
    params [i++] = CreateDSStr ( "Def",  COUNTER, rrdHBeat, "0", "10000" );
    params [i++] = CreateDSStr ( "Spam", COUNTER, rrdHBeat, "0", "10000" );
    params [i++] = CreateDSStr ( "Ham",  COUNTER, rrdHBeat, "0", "10000" );
    params [i++] = CreateDSStr ( "MTL",  COUNTER, rrdHBeat, "0", "10000" );
    params [i++] = CreateDSStr ( "RDn",  COUNTER, rrdHBeat, "0", "10000" );
    params [i++] = CreateDSStr ( "RcU",  COUNTER, rrdHBeat, "0", "10000" );
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    chk = sscanf ( aData,
            "%u;%u;%u;%u;%u;%u;%u;%u;%u;%u;%u",
            &(v.miQ), &(v.soQ), &(v.mRec), &(v.succ), &(v.fail), &(v.def),
            &(v.spam), &(v.ham), &(v.mTL), &(v.rDD), &(v.rDU) );
    if ( chk != 11 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* miQ    = UIntToStr   ( v.miQ );
    char* soQ    = UIntToStr   ( v.soQ );
    char* mRec   = UIntToStr   ( v.mRec );
    char* succ   = UIntToStr   ( v.succ );
    char* fail   = UIntToStr   ( v.fail );
    char* def    = UIntToStr   ( v.def );
    char* spam   = UIntToStr   ( v.spam );
    char* ham    = UIntToStr   ( v.ham );
    char* mTL    = UIntToStr   ( v.mTL );
    char* rDD    = UIntToStr   ( v.rDD );
    char* rDU    = UIntToStr   ( v.rDU );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup (
            "MiQ:SoQ:MRec:Succ:Fail:Def:Spam:Ham:MTL:RDn:RcU" );

    params [i++] = StrConcat ( 23, ts, ":", miQ, ":", soQ, ":", mRec,
            ":", succ, ":", fail, ":", def, ":", spam, ":", ham, ":", mTL,
            ":", rDD, ":", rDU );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( rDU );
    free ( rDD );
    free ( mTL );
    free ( ham );
    free ( spam );
    free ( def );
    free ( fail );
    free ( succ );
    free ( mRec );
    free ( soQ );
    free ( miQ );
    free ( ts );
}

void PlotDataQueue ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "<-- [kByte] | [messages] -->";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "queue", aTiming );
    PPAddPlot ( pp, "Messages in queue",         "messages", FALSE );
    PPAddPlot ( pp, "Size of messages in queue", "kByte",    FALSE );
    title = StrConcat ( 2, hostName, " mail queue" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, TRUE );
    params [i++] = CreateDEF ( "MiQM",  aFN, "MiQ", MAX );
    params [i++] = CreateDEF ( "MiQA",  aFN, "MiQ", AVG );
    params [i++] = CreateDEF ( "SoQMR", aFN, "SoQ", MAX );
    params [i++] = CreateDEF ( "SoQAR", aFN, "SoQ", AVG );
    params [i++] = StrDup ( "CDEF:SoQM=SoQMR,1024,/" );
    params [i++] = StrDup ( "CDEF:SoQMP=SoQM,-1,*" );
    params [i++] = StrDup ( "CDEF:SoQA=SoQAR,1024,/" );
    params [i++] = StrDup ( "CDEF:SoQAP=SoQA,-1,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "MiQM",  C_USAGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "MiQA",  C_USAGE1_AVG, pp, 0, AVG, TRUE );
    params [i++] = CreatePlot ( PLOT_AREA,  "SoQMP", C_USAGE2_MAX, pp, 1, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "SoQAP", C_USAGE2_AVG, pp, 1, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "MiQA",  "MiQA",  "MiQM",  "MiQM",  "%7.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "SoQAR", "SoQAR", "SoQMR", "SoQMR", "%7.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataTrans ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[messages/h]  <-- tx | rx -->";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "trans", aTiming );
    PPAddPlot ( pp, "Received", "messages/h", FALSE );
    PPAddPlot ( pp, "Failed",   "messages/h", FALSE );
    PPAddPlot ( pp, "Deferred", "messages/h", TRUE );
    PPAddPlot ( pp, "Sent",     "messages/h", TRUE );
    title = StrConcat ( 2, hostName, " mail transfer" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, TRUE );
    params [i++] = CreateDEF ( "MRecMR", aFN, "MRec", MAX );
    params [i++] = CreateDEF ( "MRecAR", aFN, "MRec", AVG );
    params [i++] = CreateDEF ( "FailMR", aFN, "Fail", MAX );
    params [i++] = CreateDEF ( "DefMR",  aFN, "Def",  MAX );
    params [i++] = CreateDEF ( "SuccMR", aFN, "Succ", MAX );
    params [i++] = CreateDEF ( "FailAR", aFN, "Fail", AVG );
    params [i++] = CreateDEF ( "DefAR",  aFN, "Def",  AVG );
    params [i++] = CreateDEF ( "SuccAR", aFN, "Succ", AVG );
    params [i++] = StrDup ( "CDEF:MRecM=MRecMR,3600,*" );
    params [i++] = StrDup ( "CDEF:MRecA=MRecAR,3600,*" );
    params [i++] = StrDup ( "CDEF:FailM=FailMR,3600,*" );
    params [i++] = StrDup ( "CDEF:DefM=DefMR,3600,*" );
    params [i++] = StrDup ( "CDEF:SuccM=SuccMR,3600,*" );
    params [i++] = StrDup ( "CDEF:FailA=FailAR,3600,*" );
    params [i++] = StrDup ( "CDEF:DefA=DefAR,3600,*" );
    params [i++] = StrDup ( "CDEF:SuccA=SuccAR,3600,*" );
    params [i++] = StrDup ( "CDEF:FailMP=FailM,-1,*" );
    params [i++] = StrDup ( "CDEF:DefMP=DefM,-1,*" );
    params [i++] = StrDup ( "CDEF:SuccMP=SuccM,-1,*" );
    params [i++] = StrDup ( "CDEF:FailAP=FailA,-1,*" );
    params [i++] = StrDup ( "CDEF:DefAP=DefA,-1,*" );
    params [i++] = StrDup ( "CDEF:SuccAP=SuccA,-1,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "MRecM",  C_THRU_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "MRecA",  C_THRU_AVG, pp, 0, AVG, TRUE );
    params [i++] = CreatePlot ( PLOT_AREA,  "FailMP", C_ERR1_MAX, pp, 1, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "DefMP",  C_ERR2_MAX, pp, 2, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "SuccMP", C_THRU_MAX, pp, 3, MAX, TRUE );
    params [i++] = CreatePlot ( PLOT_LINE2, "FailAP", C_ERR1_AVG, pp, 1, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "DefAP",  C_ERR2_AVG, pp, 2, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "SuccAP", C_THRU_AVG, pp, 3, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "MRecA", "MRecA", "MRecM", "MRecM", "%7.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "FailA", "FailA", "FailM", "FailM", "%7.1lf" );
    i = PlotPrintParams ( i, params, pp, 2, "DefA",  "DefA",  "DefM",  "DefM",  "%7.1lf" );
    i = PlotPrintParams ( i, params, pp, 3, "SuccA", "SuccA", "SuccM", "SuccM", "%7.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataSpHam ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[messages/h]  <-- Spam | Ham -->";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "spham", aTiming );
    PPAddPlot ( pp, "ham",  "messages/h", FALSE );
    PPAddPlot ( pp, "spam", "messages/h", FALSE );
    title = StrConcat ( 2, hostName, " spam and ham" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, TRUE );
    params [i++] = CreateDEF ( "HamMR",  aFN, "Ham",  MAX );
    params [i++] = CreateDEF ( "HamAR",  aFN, "Ham",  AVG );
    params [i++] = CreateDEF ( "SpamMR", aFN, "Spam", MAX );
    params [i++] = CreateDEF ( "SpamAR", aFN, "Spam", AVG );
    params [i++] = StrDup ( "CDEF:HamM=HamMR,3600,*" );
    params [i++] = StrDup ( "CDEF:HamA=HamAR,3600,*" );
    params [i++] = StrDup ( "CDEF:SpamM=SpamMR,3600,*" );
    params [i++] = StrDup ( "CDEF:SpamA=SpamAR,3600,*" );
    params [i++] = StrDup ( "CDEF:SpamMP=SpamM,-1,*" );
    params [i++] = StrDup ( "CDEF:SpamAP=SpamA,-1,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "HamM",   C_THRU_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "HamA",   C_THRU_AVG, pp, 0, AVG, TRUE );
    params [i++] = CreatePlot ( PLOT_AREA,  "SpamMP", C_ERR2_MAX, pp, 1, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "SpamAP", C_ERR2_AVG, pp, 1, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "HamA",  "HamA",  "HamM",  "HamM",  "%7.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "SpamA", "SpamA", "SpamM", "SpamM", "%7.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataMTL ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[messages/h]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "mtl", aTiming );
    PPAddPlot ( pp, "rejected", "messages/h", FALSE );
    title = StrConcat ( 2, hostName, " rejected mails because of size" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, FALSE );
    params [i++] = CreateDEF ( "MTLMR", aFN, "MTL",  MAX );
    params [i++] = CreateDEF ( "MTLAR", aFN, "MTL",  AVG );
    params [i++] = StrDup ( "CDEF:MTLM=MTLMR,3600,*" );
    params [i++] = StrDup ( "CDEF:MTLA=MTLAR,3600,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "MTLM", C_ERR2_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "MTLA", C_ERR2_AVG, pp, 0, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "MTLA",  "MTLA",  "MTLM",  "MTLM",  "%7.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataRDn ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[messages/h]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "rdn", aTiming );
    PPAddPlot ( pp, "rejected", "messages/h", FALSE );
    title = StrConcat ( 2, hostName, " rejected mails because of invalid destination domain" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, FALSE );
    params [i++] = CreateDEF ( "RDnMR",  aFN, "RDn",  MAX );
    params [i++] = CreateDEF ( "RDnAR",  aFN, "RDn",  AVG );
    params [i++] = StrDup ( "CDEF:RDnM=RDnMR,3600,*" );
    params [i++] = StrDup ( "CDEF:RDnA=RDnAR,3600,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "RDnM", C_ERR2_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "RDnA", C_ERR2_AVG, pp, 0, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "RDnA",  "RDnA",  "RDnM",  "RDnM",  "%7.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataRcU ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[messages/h]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "rcu", aTiming );
    PPAddPlot ( pp, "rejected", "messages/h", FALSE );
    title = StrConcat ( 2, hostName, " rejected mails because of unknown user" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, FALSE );
    params [i++] = CreateDEF ( "RcUMR",  aFN, "RcU",  MAX );
    params [i++] = CreateDEF ( "RcUAR",  aFN, "RcU",  AVG );
    params [i++] = StrDup ( "CDEF:RcUM=RcUMR,3600,*" );
    params [i++] = StrDup ( "CDEF:RcUA=RcUAR,3600,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "RcUM", C_ERR2_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "RcUA", C_ERR2_AVG, pp, 0, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "RcUA",  "RcUA",  "RcUM",  "RcUM",  "%7.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataQueue ( fn, PLOT_24H_5M,  now );
    PlotDataTrans ( fn, PLOT_24H_5M,  now );
    PlotDataSpHam ( fn, PLOT_24H_5M,  now );
    PlotDataMTL   ( fn, PLOT_24H_5M,  now );
    PlotDataRDn   ( fn, PLOT_24H_5M,  now );
    PlotDataRcU   ( fn, PLOT_24H_5M,  now );
    PlotDataQueue ( fn, PLOT_7D_15M,  now );
    PlotDataTrans ( fn, PLOT_7D_15M,  now );
    PlotDataSpHam ( fn, PLOT_7D_15M,  now );
    PlotDataMTL   ( fn, PLOT_7D_15M,  now );
    PlotDataRDn   ( fn, PLOT_7D_15M,  now );
    PlotDataRcU   ( fn, PLOT_7D_15M,  now );
    PlotDataQueue ( fn, PLOT_31D_60M, now );
    PlotDataTrans ( fn, PLOT_31D_60M, now );
    PlotDataSpHam ( fn, PLOT_31D_60M, now );
    PlotDataMTL   ( fn, PLOT_31D_60M, now );
    PlotDataRDn   ( fn, PLOT_31D_60M, now );
    PlotDataRcU   ( fn, PLOT_31D_60M, now );
    PlotDataQueue ( fn, PLOT_366D_4H, now );
    PlotDataTrans ( fn, PLOT_366D_4H, now );
    PlotDataSpHam ( fn, PLOT_366D_4H, now );
    PlotDataMTL   ( fn, PLOT_366D_4H, now );
    PlotDataRDn   ( fn, PLOT_366D_4H, now );
    PlotDataRcU   ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    int retVal = STATE_OK;

    ReadVals ();
    if ( ( v.miQ > warn ) && ( retVal < STATE_WARNING ) )
            retVal = STATE_WARNING;
    if ( ( v.miQ > crit ) && ( retVal < STATE_CRITICAL ) )
            retVal = STATE_CRITICAL;
    PrintExit ( retVal );
    printf ( " - %i messages with %i kBytes in queue", v.miQ, v.soQ );
    if ( npPerfData )
    {
        printf ( "|'messages in queue'=%u;%u;%u", v.miQ, warn, crit );
        printf ( " 'size of queued messages'=%uKB", v.soQ );
        printf ( " 'messages received'=%uc", v.mRec );
        printf ( " 'successful deliveries'=%uc", v.succ );
        printf ( " 'failed deliveries'=%uc", v.fail );
        printf ( " 'deferred deliveries'=%uc", v.def );
        printf ( " 'messages marked as Spam'=%uc", v.spam );
        printf ( " 'messages marked as Ham'=%uc", v.ham );
        printf ( " 'messag too large'=%uc", v.mTL );
        printf ( " 'relaying denied'=%uc", v.rDD );
        printf ( " 'recipient unknown'=%uc\n", v.rDU );
    }
    else
    {
        printf ( "|%u;%u;%u;%u;%u;%u;%u;%u;%u;%u;%u\n",
                v.miQ, v.soQ, v.mRec, v.succ, v.fail, v.def,
                v.spam, v.ham, v.mTL, v.rDD, v.rDU );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( checkData )
    {
        if ( !isWarnSet ) Usage ( "Missing 'warn' parameters" );
        if ( !isCritSet ) Usage ( "Missing 'crit' parameters" );
    }
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ( NULL );
        if ( plotPerfData ) PlotPerfData ( NULL );
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return retVal;
}
