/*
 * File:      check_netapp_du.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int index;  /* file system index */
    int v10;    /* found dfMountedOn */
    int v11;    /* found dfMaxFilesAvail */
    int v12;    /* found dfMaxFilesUsed */
    int v14;    /* found dfHighTotalKBytes */
    int v15;    /* found dfLowTotalKBytes */
    int v16;    /* found dfHighUsedKBytes */
    int v17;    /* found dfLowUsedKBytes */
    char* name; /* file system name */
} tScanRes;

/* data type for storing data */
typedef struct
{
    double usedS; /* used storage */
    double sizeS; /* size storage */
    double usedF; /* used files */
    double sizeF; /* size files */
} tDataSet;

/* global input variables */
double warnSS = 75;
char   warnSU = '%';
double critSS = 85;
char   critSU = '%';
double warnFS = 75;
char   warnFU = '%';
double critFS = 85;
char   critFU = '%';

/*
 * SNMP base OID
 * NETWORK-APPLIANCE-MIB::dfMaxFilesAvail.IDX
 * = 1.3.6.1.4.1.789.1.5.4.1.11.IDX
 * NETWORK-APPLIANCE-MIB::dfMaxFilesUsed.IDX
 * = 1.3.6.1.4.1.789.1.5.4.1.12.IDX
 * NETWORK-APPLIANCE-MIB::dfHighTotalKBytes.IDX
 * = 1.3.6.1.4.1.789.1.5.4.1.14.IDX
 * NETWORK-APPLIANCE-MIB::dfLowTotalKBytes.IDX
 * = 1.3.6.1.4.1.789.1.5.4.1.15.IDX
 * NETWORK-APPLIANCE-MIB::dfHighUsedKBytes.IDX
 * = 1.3.6.1.4.1.789.1.5.4.1.16.IDX
 * NETWORK-APPLIANCE-MIB::dfLowUsedKBytes.IDX
 * = 1.3.6.1.4.1.789.1.5.4.1.17.IDX
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 789, 1, 5, 4, 1 };
size_t baseOIDLen = 11;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_netapp_du";
    progVersion = "0.8";
    helpIndexP = "-i INDEX";
    helpThresP = "[-w WARNM,WARNF] [-c CRITM,CRITF]";
    helpPurpose = "Checks the usage of a NetApp(R) Filer volume.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 5 * 60;
    rrdHBeat = 11 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = TRUE;
    hasAddInfParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-w, --warning INTEGER[%%kMG],INTEGER[%%]\n"
    "    Storage and files usage at which a warning is generated.\n"
    "    Default is %.0f%c,%.0f%c.\n", warnSS, warnSU, warnFS, warnFU
    );
    printf (
    "\n"
    "-c, --critical INTEGER[%%kMG],INTEGER[%%]\n"
    "    Storage and files usage at which a critical is generated.\n"
    "    Default is %.0f%c,%.0f%c.\n", critSS, critSU, critFS, critFU
    );
    printf (
    "\n"
    "-i, --index INTEGER\n"
    "    Index of the volume which should be checked.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The first value of warning and critical parameters is the used storage\n"
    "value. If it is given without unit, then the unit 'Bytes' is assumed.\n"
    "'%%', 'k', 'M' and 'G' are interpreted as percent of the whole volume,\n"
    "Kilobytes (1024 Bytes), Megabytes (1024*1024 Bytes) and Gigabytes\n"
    "(1024*1024*1024 Bytes) respectively. The second parameter is the number\n"
    "of files used, given either as the absolute number (when no unit is\n"
    "given) or as percent of the maximum number of files.\n"
    );
    printf (
    "\n"
    "The program uses double precision floats for calculations. Nevertheless\n"
    "parameters are supposed to be integers (plus optional units).\n"
    "\n"
    "To generate a warning or critical only one of the storage or files\n"
    "usages needs to be higher than the respective level.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<addInf>;<usedS>;<sizeS>;<usedF>;<sizeF>\n"
    "where\n"
    "    addInf ... additional information.\n"
    "    usedS  ... used storage in Megabytes.\n"
    "    sizeS  ... size of the storage in Megabytes.\n"
    "    usedF  ... used files.\n"
    "    sizeF  ... number of files available.\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'used space'=<usedS>MB;<warnS>;<critS>;0;<sizeS> "
    "'used files'=<usedF>;<warnF>;<critF>;0;<sizeF>\n"
    "\n"
    );
}

void InvalidWarnParam ()
{
    Usage ( "Invalid 'warn' parameter" );
}

void InvalidCritParam ()
{
    Usage ( "Invalid 'crit' parameter" );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (            !strcmp ( argv [(*i)], "-w" )
                ||  !strcmp ( argv [(*i)], "--warning" ) )
    {
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];

            errno = 0;
            warnSS = strtod ( argv [(*i)], &tail );
            if ( errno || ( (*tail) == '\0' ) ) InvalidWarnParam ();
            if      ( (*tail) == ',' ) warnSU = '1';
            else if ( (*tail) == 'k' ) warnSU = 'k';
            else if ( (*tail) == 'M' ) warnSU = 'M';
            else if ( (*tail) == 'G' ) warnSU = 'G';
            else if ( (*tail) == '%' ) warnSU = '%';
            else Usage ( "Invalid space unit" );
            if ( warnSU != '1' ) tail = &(tail [1]);
            if ( (*tail) == ',' )
            {
                char* wfstr = &(tail [1]);

                tail = wfstr;
                warnFS = strtod ( wfstr, &tail );
                if ( errno ) InvalidWarnParam ();
                if      ( (*tail) == '\0' ) warnFU = '1';
                else if ( (*tail) == '%'  ) warnFU = '%';
                else Usage ( "Invalid files unit" );
                if ( warnFU != '1' ) tail = &(tail [1]);
                if ( (*tail) != '\0' ) InvalidWarnParam ();
                (*i)++;
            }
            else InvalidWarnParam();
        }
        else Usage ( "Missing 'warn' parameter" );
    } else if ( !strcmp ( argv [(*i)], "-c" )
            ||  !strcmp ( argv [(*i)], "--critical" ) )
    {
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];

            errno = 0;
            critSS = strtod ( argv [(*i)], &tail );
            if ( errno || ( (*tail) == '\0' ) ) InvalidCritParam ();
            if      ( (*tail) == ',' ) critSU = '1';
            else if ( (*tail) == 'k' ) critSU = 'k';
            else if ( (*tail) == 'M' ) critSU = 'M';
            else if ( (*tail) == 'G' ) critSU = 'G';
            else if ( (*tail) == '%' ) critSU = '%';
            else Usage ( "Invalid space unit" );
            if ( critSU != '1' ) tail = &(tail [1]);
            if ( (*tail) == ',' )
            {
                char* wfstr = &(tail [1]);

                tail = wfstr;
                critFS = strtod ( wfstr, &tail );
                if ( errno ) InvalidCritParam ();
                if      ( (*tail) == '\0' ) critFU = '1';
                else if ( (*tail) == '%'  ) critFU = '%';
                else Usage ( "Invalid files unit" );
                if ( critFU != '1' ) tail = &(tail [1]);
                if ( (*tail) != '\0' ) InvalidCritParam ();
                (*i)++;
            }
            else InvalidCritParam();
        }
        else Usage ( "Missing 'crit' parameter" );
    } else ret = FALSE;
    return ( ret );
}

int ScanResIndex ( int aIndex )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 13 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [11] )
            {
                case 1: /* dfIndex */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].index = aVar->name [12];
                    scanRes [index].v10 = FALSE;
                    scanRes [index].v11 = FALSE;
                    scanRes [index].v12 = FALSE;
                    scanRes [index].v14 = FALSE;
                    scanRes [index].v15 = FALSE;
                    scanRes [index].v16 = FALSE;
                    scanRes [index].v17 = FALSE;
                    scanRes [index].name = NULL;
                break;
                case 10: /* dfMountedOn */
                    index = ScanResIndex ( aVar->name [12] );
                    scanRes [index].v10 = ( aVar->type == ASN_OCTET_STR );
                    scanRes [index].name = VarListStrDup ( aVar );
                break;
                case 11: /* dfMaxFilesAvail */
                    index = ScanResIndex ( aVar->name [12] );
                    scanRes [index].v11 = ( aVar->type == ASN_INTEGER );
                break;
                case 12: /* dfMaxFilesUsed */
                    index = ScanResIndex ( aVar->name [12] );
                    scanRes [index].v12 = ( aVar->type == ASN_INTEGER );
                break;
                case 14: /* dfHighTotalKBytes */
                    index = ScanResIndex ( aVar->name [12] );
                    scanRes [index].v14 = ( aVar->type == ASN_INTEGER );
                break;
                case 15: /* dfLowTotalKBytes */
                    index = ScanResIndex ( aVar->name [12] );
                    scanRes [index].v15 = ( aVar->type == ASN_INTEGER );
                break;
                case 16: /* dfHighUsedKBytes */
                    index = ScanResIndex ( aVar->name [12] );
                    scanRes [index].v16 = ( aVar->type == ASN_INTEGER );
                break;
                case 17: /* dfLowUsedKBytes */
                    index = ScanResIndex ( aVar->name [12] );
                    scanRes [index].v17 = ( aVar->type == ASN_INTEGER );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v10 && scanRes [i].v11 && scanRes [i].v12 &&
                scanRes [i].v14 && scanRes [i].v15 && scanRes [i].v16 &&
                scanRes [i].v17 )
        {
            retVal = STATE_OK;
            printf ( "OK index: %i ; name: %s\n",
                    scanRes [i].index, scanRes [i].name );
            free ( scanRes [i].name );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_fSize  [13];
    oid o_fUsed  [13];
    oid o_sSizeH [13];
    oid o_sSizeL [13];
    oid o_sUsedH [13];
    oid o_sUsedL [13];
    size_t oidLen;
    double sizeH;
    double sizeL;
    double usedH;
    double usedL;
    double mult = 256.0 * 256.0 * 256.0 * 256.0;

    OidCpy ( o_fSize,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_fUsed,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_sSizeH, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_sSizeL, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_sUsedH, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_sUsedL, &oidLen, baseOID, &baseOIDLen );
    o_fSize  [11] = 11; o_fSize  [12] = idx;
    o_fUsed  [11] = 12; o_fUsed  [12] = idx;
    o_sSizeH [11] = 14; o_sSizeH [12] = idx;
    o_sSizeL [11] = 15; o_sSizeL [12] = idx;
    o_sUsedH [11] = 16; o_sUsedH [12] = idx;
    o_sUsedL [11] = 17; o_sUsedL [12] = idx;
    oidLen = 13;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    /*
     * Netapp screwed it. Now that is type casting ...
     */
    v.sizeF = (double) ( (uint) ReadInteger ( sess2, o_fSize,  oidLen ) );
    v.usedF = (double) ( (uint) ReadInteger ( sess2, o_fUsed,  oidLen ) );
    sizeH   = (double) ( (uint) ReadInteger ( sess2, o_sSizeH, oidLen ) );
    sizeL   = (double) ( (uint) ReadInteger ( sess2, o_sSizeL, oidLen ) );
    usedH   = (double) ( (uint) ReadInteger ( sess2, o_sUsedH, oidLen ) );
    usedL   = (double) ( (uint) ReadInteger ( sess2, o_sUsedL, oidLen ) );
    v.sizeS = sizeH * mult + sizeL;
    v.usedS = usedH * mult + usedL;
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "usedS", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "sizeS", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "usedF", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "sizeF", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    char buf[1024];

    chk = sscanf ( aData, "%1023s;%lf;%lf;%lf;%lf",
            buf, &(v.usedS), &(v.sizeS), &(v.usedF), &(v.sizeF) );
    if ( chk != 5 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
    addInf  = StrDup ( buf );
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* usedS  = DoubleToStr ( v.usedS );
    char* sizeS  = DoubleToStr ( v.sizeS );
    char* usedF  = DoubleToStr ( v.usedF );
    char* sizeF  = DoubleToStr ( v.sizeF );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "usedS:sizeS:usedF:sizeF" );
    params [i++] = StrConcat ( 9, ts, ":", usedS, ":", sizeS, ":", usedF,
            ":", sizeF );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( sizeF );
    free ( usedF );
    free ( sizeS );
    free ( usedS );
    free ( ts );
}

void PlotDataStorage ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[%]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "stor", aTiming );
    PPAddPlot ( pp, "used storage", "%", FALSE );
    title = StrConcat ( 4, hostName, " volume ", addInf, " storage usage" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 100, FALSE );
    params [i++] = CreateDEF ( "sizeMR", aFN, "sizeS", MAX );
    params [i++] = CreateDEF ( "sizeAR", aFN, "sizeS", AVG );
    params [i++] = CreateDEF ( "usedMR", aFN, "usedS", MAX );
    params [i++] = CreateDEF ( "usedAR", aFN, "usedS", AVG );
    params [i++] = StrDup ( "CDEF:usedM=usedMR,sizeMR,/,100,*" );
    params [i++] = StrDup ( "CDEF:usedA=usedAR,sizeAR,/,100,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "usedM", C_USAGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "usedA", C_USAGE1_AVG, pp, 0, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "usedA",  "usedA",  "usedM",  "usedM",  "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataFiles ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[%]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "file", aTiming );
    PPAddPlot ( pp, "used files", "%", FALSE );
    title = StrConcat ( 4, hostName, " volume ", addInf, " files usage" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 100, FALSE );
    params [i++] = CreateDEF ( "sizeMR", aFN, "sizeF", MAX );
    params [i++] = CreateDEF ( "sizeAR", aFN, "sizeF", AVG );
    params [i++] = CreateDEF ( "usedMR", aFN, "usedF", MAX );
    params [i++] = CreateDEF ( "usedAR", aFN, "usedF", AVG );
    params [i++] = StrDup ( "CDEF:usedM=usedMR,sizeMR,/,100,*" );
    params [i++] = StrDup ( "CDEF:usedA=usedAR,sizeAR,/,100,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "usedM", C_USAGE2_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "usedA", C_USAGE2_AVG, pp, 0, AVG, TRUE );
    i = PlotPrintParams ( i, params,  pp, 0, "usedA",  "usedA",  "usedM",  "usedM",  "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataStorage ( fn, PLOT_24H_5M,  now );
    PlotDataFiles   ( fn, PLOT_24H_5M,  now );
    PlotDataStorage ( fn, PLOT_7D_15M,  now );
    PlotDataFiles   ( fn, PLOT_7D_15M,  now );
    PlotDataStorage ( fn, PLOT_31D_60M, now );
    PlotDataFiles   ( fn, PLOT_31D_60M, now );
    PlotDataStorage ( fn, PLOT_366D_4H, now );
    PlotDataFiles   ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

double SpaceToMB ( char* aWhat, double aLevel, char aUnit, double aSize )
{
    double retVal = '0';

    switch ( aUnit )
    {
        case '1': retVal = aLevel / 1048576.0;         break;
        case 'k': retVal = aLevel / 1024.0;            break;
        case 'M': retVal = aLevel;                     break;
        case 'G': retVal = aLevel * 1024.0;            break;
        case '%': retVal = ( aLevel * aSize ) / 100.0; break;
        default:
        {
            Error ( LOG_ERR, "An invalid level slipped through!" );
        }
        break;
    }
    return ( retVal );
}

double FilesToFiles ( char* aWhat, double aFiles, char aUnit, double aSize )
{
    double retVal = '0';

    switch ( aUnit )
    {
        case '1': retVal = aFiles;                 break;
        case '%': retVal = aFiles * aSize / 100.0; break;
        default:
        {
            Error ( LOG_ERR, "An invalid level slipped through!" );
        }
        break;
    }
    return ( retVal );
}

int CheckData ()
{
    int retVal = STATE_OK;
    double sizeS;
    double usedS;
    double usedFPc;
    double usedSPc;
    double warnS;
    double critS;
    double warnF;
    double critF;

    ReadVals ();
    /* calculate everything in Megabytes */
    sizeS = v.sizeS / 1024.0;
    usedS = v.usedS / 1024.0;
    usedFPc = ( v.usedF * 100.0 ) / v.sizeF;
    usedSPc = ( v.usedS * 100.0 ) / v.sizeS;
    warnS = SpaceToMB    ( "warning space",  warnSS, warnSU, sizeS );
    critS = SpaceToMB    ( "critical space", critSS, critSU, sizeS );
    warnF = FilesToFiles ( "warning files",  warnFS, warnFU, v.sizeF );
    critF = FilesToFiles ( "critical files", critFS, critFU, v.sizeF );
    if ( ( v.usedF > warnF ) && ( retVal < STATE_WARNING ) )
            retVal = STATE_WARNING;
    if ( ( v.usedF > critF ) && ( retVal < STATE_CRITICAL ) )
            retVal = STATE_CRITICAL;
    if ( ( usedS > warnS ) && ( retVal < STATE_WARNING ) )
            retVal = STATE_WARNING;
    if ( ( usedS > critS ) && ( retVal < STATE_CRITICAL ) )
            retVal = STATE_CRITICAL;
    PrintExit ( retVal );
    printf ( " - %s usage:", addInf );
    printf ( " %.0f/%.0f MB (%.1f%%)",    usedS,   sizeS,   usedSPc );
    printf ( " %.0f/%.0f files (%.1f%%)", v.usedF, v.sizeF, usedFPc );
    if ( npPerfData )
    {
        printf ( "|'used space'=%.0fMB;%.0f;%.0f;0;%.0f",
                usedS, warnS, critS, sizeS );
        printf ( " 'used files'=%.0f;%.0f;%.0f;0;%.0f\n",
                v.usedF, warnF, critF, v.sizeF );
    }
    else
    {
        printf ( " | %s;%.0f;%.0f;%.0f;%.0f\n",
                addInf, usedS, sizeS, v.usedF, v.sizeF );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ();
        if ( plotPerfData ) PlotPerfData ();
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
