/*
 * File:      check_netapp_ops.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int v6;  /* miscLowNfsOps */
    int v8;  /* miscLowCifsOps */
    int v10; /* miscLowHttpOps */
} tScanRes;

/* data type for storing data */
typedef struct
{
    uint nfs;  /* nfs ops */
    uint cifs; /* cifs ops */
    uint http; /* http ops */
} tDataSet;

/*
 * SNMP base OID
 * NETWORK-APPLIANCE-MIB::miscLowNfsOps.0  = .1.3.6.1.4.1.789.1.2.2.6.0
 * NETWORK-APPLIANCE-MIB::miscLowCifsOps.0 = .1.3.6.1.4.1.789.1.2.2.8.0
 * NETWORK-APPLIANCE-MIB::miscLowHttpOps.0 = .1.3.6.1.4.1.789.1.2.2.10.0
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 789, 1, 2, 2 };
size_t baseOIDLen = 10;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_netapp_ops";
    progVersion = "0.4";
    helpIndexP = NULL;
    helpThresP = NULL;
    helpPurpose = "Get the number of operations from a NetApp(R) Filer.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 5 * 60;
    rrdHBeat = 11 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = FALSE;
}

void HelpSpecificParams ()
{
    /* No specific parameters */
}

void HelpExplain ()
{
    printf (
    "The plugin returns the number of operations from a NetApp(R) Filer.\n"
    "It is designed to gather performance data and does not return warnings\n"
    "or critical alerts.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<nfs>;<cifs>;<http>\n"
    "where\n"
    "    nfs  ... number of NFS ops,\n"
    "             see NETWORK-APPLIANCE-MIB::miscLowNfsOps.\n"
    "    cifs ... number of CIFS ops,\n"
    "             see NETWORK-APPLIANCE-MIB::miscLowCifsOps.\n"
    "    http ... number of HTTP ops,\n"
    "             see NETWORK-APPLIANCE-MIB::miscLowHttpOps.\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'nfs operations'=<nfs>c "
    "'cifs operations'=<cifs>c "
    "'http operations'=<http>c\n"
    "\n"
    );
    printf (
    "For more information about those counters see netapp.mib\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    return FALSE;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 12 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            switch ( aVar->name [10] )
            {
                case 6: /* miscLowNfsOps */
                    scanResLen = 1;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [0].v6 = ( aVar->type == ASN_COUNTER );
                    scanRes [0].v8 = FALSE;
                    scanRes [0].v10 = FALSE;
                break;
                case 8: /* miscLowCifsOps */
                    scanRes [0].v8 = ( aVar->type == ASN_COUNTER );
                break;
                case 10: /* miscLowHttpOps */
                    scanRes [0].v10 = ( aVar->type == ASN_COUNTER );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if ( scanRes [i].v6 && scanRes [i].v8 && scanRes [i].v10 )
        {
            retVal = STATE_OK;
            printf ( "OK\n" );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_nfs  [12];
    oid o_cifs [12];
    oid o_http [12];
    size_t oidLen;

    OidCpy ( o_nfs,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_cifs, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_http, &oidLen, baseOID, &baseOIDLen );
    o_nfs  [10] =  6; o_nfs  [11] = 0;
    o_cifs [10] =  8; o_cifs [11] = 0;
    o_http [10] = 10; o_http [11] = 0;
    oidLen = 12;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.nfs  = ReadCounter32 ( sess2, o_nfs,  oidLen );
    v.cifs = ReadCounter32 ( sess2, o_cifs, oidLen );
    v.http = ReadCounter32 ( sess2, o_http, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "nfs",  COUNTER, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "cifs", COUNTER, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "http", COUNTER, rrdHBeat, "0", "U" );
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    chk = sscanf ( aData, "%u;%u;%u", &(v.nfs), &(v.cifs), &(v.http) );
    if ( chk != 3 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* nfs    = UIntToStr   ( v.nfs );
    char* cifs   = UIntToStr   ( v.cifs );
    char* http   = UIntToStr   ( v.http );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "nfs:cifs:http" );
    params [i++] = StrConcat ( 7, ts, ":", nfs, ":", cifs, ":", http );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( http );
    free ( cifs );
    free ( nfs );
    free ( ts );
}

void PlotDataOps ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[ops]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, NULL, aTiming );
    PPAddPlot ( pp, "NFS ops",  "ops", FALSE );
    PPAddPlot ( pp, "CIFS ops", "ops", FALSE );
    PPAddPlot ( pp, "HTTP ops", "ops", FALSE );
    title = StrConcat ( 2, hostName, " operations" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, FALSE );
    params [i++] = CreateDEF ( "nfsM",  aFN, "nfs",  MAX );
    params [i++] = CreateDEF ( "nfsA",  aFN, "nfs",  AVG );
    params [i++] = CreateDEF ( "cifsM", aFN, "cifs", MAX );
    params [i++] = CreateDEF ( "cifsA", aFN, "cifs", AVG );
    params [i++] = CreateDEF ( "httpM", aFN, "http", MAX );
    params [i++] = CreateDEF ( "httpA", aFN, "http", AVG );
    params [i++] = CreatePlot ( PLOT_LINE3, "nfsM",  C_LOAD1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE3, "cifsM", C_LOAD2_MAX, pp, 1, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE3, "httpM", C_LOAD3_MAX, pp, 2, MAX, TRUE );
    params [i++] = CreatePlot ( PLOT_LINE2, "nfsA",  C_LOAD1_AVG, pp, 0, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "cifsA", C_LOAD2_AVG, pp, 1, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "httpA", C_LOAD3_AVG, pp, 2, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "nfsA",  "nfsA",  "nfsM",  "nfsM",  "%6.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "cifsA", "cifsA", "cifsM", "cifsM", "%6.1lf" );
    i = PlotPrintParams ( i, params, pp, 2, "httpA", "httpA", "httpM", "httpM", "%6.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ( const char* aSrvId )
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataOps ( fn, PLOT_6H_1M,   now );
    PlotDataOps ( fn, PLOT_24H_5M,  now );
    PlotDataOps ( fn, PLOT_7D_15M,  now );
    PlotDataOps ( fn, PLOT_31D_60M, now );
    PlotDataOps ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    int retVal = STATE_OK;

    ReadVals ();
    printf ( "OK - operations: %u, %u, %u", v.nfs, v.cifs, v.http );
    if ( npPerfData )
    {
        printf ( "|'nfs operations'=%uc ", v.nfs );
        printf ( " 'cifs operations'=%uc ", v.cifs );
        printf ( " 'http operations'=%uc\n", v.http );
    }
    else
    {
        printf ( "|%u;%u;%u\n", v.nfs, v.cifs, v.http );
    }
    return ( retVal );
}


int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ( NULL );
        if ( plotPerfData ) PlotPerfData ( NULL );
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
