/*
 * File:      check_netapp_raiddrv.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int index;   /* volume index */
    int group;   /* group index */
    int pDrv;    /* physical drive index */
    int v1;      /* found raidVIndex */
    int v3;      /* found raidVStatus */
    int v11;     /* found raidVCompletionPerCent */
    int v19;     /* found raidVShelf */
    int v20;     /* found raidVBay*/
    int shelf;   /* shelf */
    int bay;     /* bay */
} tScanRes;

/* data type for storing data */
typedef struct
{
    int  stat;  /* status of the drive */
    int  perc;  /* reconstruction, verification or scrubbing completeness */
    int  shlf;  /* condition of the drive */
    int  bay;   /* bay the drive is in */
} tDataSet;

/* global input variables */
int grp;
int isGrpSet = FALSE;
int pDrv;
int isPDrvSet  = FALSE;

/*
 * SNMP base OID
 * raidVStatus.IDX.GRP.PDRV
 * = 1.3.6.1.4.1.789.1.6.2.1.3.IDX.GRP.PDRV
 * raidVCompletionPerCent.IDX.GRP.PDRV
 * = 1.3.6.1.4.1.789.1.6.2.1.11.IDX.GRP.PDRV
 * raidVShelf.IDX.GRP.PDRV
 * = 1.3.6.1.4.1.789.1.6.2.1.19.IDX.GRP.PDRV
 * raidVBay.IDX.GRP.PDRV
 * = 1.3.6.1.4.1.789.1.6.2.1.20.IDX.GRP.PDRV
 */
oid baseOID [] = { 1, 3, 6, 1, 4, 1, 789, 1, 6, 2, 1 };
size_t baseOIDLen = 11;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_netapp_raiddrv";
    progVersion = "0.3";
    helpIndexP = "-i INDEX -g GROUP -d DRIVE";
    helpThresP = NULL;
    helpPurpose =
            "Checks the state of a physical drive of a NetApp(R) Filer.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
    /* rrdStep is unused */
    /* rrdHBeat is unused */
    supportsPerfLog = FALSE;
    hasIndexParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-i, --index INTEGER\n"
    "    Index of the volume of the drive.\n"
    "\n"
    "-g, --group INTEGER\n"
    "    Index of the group of the drive.\n"
    "\n"
    "-d, --drive INTEGER\n"
    "    Index of the physical drive.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin checks the overall state of a RAID drive of a NetApp(R)\n"
    "Filer."
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "There is no performance data provided.\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;
    if (        !strcmp ( argv [(*i)], "-g" )
            ||  !strcmp ( argv [(*i)], "--group" ) )
    {
        isGrpSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            grp = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || *tail ) Usage ( "Invalid 'group' parameter" );
        }
        else Usage ( "Missing 'group' parameter(s)" );
    } else if ( !strcmp ( argv [(*i)], "-d" )
            ||  !strcmp ( argv [(*i)], "--drive" ) )
    {
        isPDrvSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            pDrv = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || *tail ) Usage ( "Invalid 'drive' parameter" );
        }
        else Usage ( "Missing 'drive' parameter(s)" );
    } else ret = FALSE;
    return ( ret );
}

int ScanResIndex ( int aIndex, int aGroup, int aPDrv )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex )
                && ( scanRes [retVal].group == aGroup )
                && ( scanRes [retVal].pDrv == aPDrv );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 15 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [11] )
            {
                case 1: /* raidVIndex */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].index = aVar->name [12];
                    scanRes [index].group = aVar->name [13];
                    scanRes [index].pDrv = aVar->name [14];
                    scanRes [index].v1 = ( aVar->type == ASN_INTEGER );
                    scanRes [index].v3 = FALSE;
                    scanRes [index].v11 = FALSE;
                    scanRes [index].v19 = FALSE;
                    scanRes [index].v20 = FALSE;
                break;
                case 3: /* raidVStatus */
                    index = ScanResIndex ( aVar->name [12], aVar->name [13],
                            aVar->name [14] );
                    scanRes [index].v3 = ( aVar->type == ASN_INTEGER );
                break;
                case 11: /* raidVCompletionPerCent */
                    index = ScanResIndex ( aVar->name [12], aVar->name [13],
                            aVar->name [14] );
                    scanRes [index].v11 = ( aVar->type == ASN_INTEGER );
                break;
                case 19: /* raidVShelf */
                    index = ScanResIndex ( aVar->name [12], aVar->name [13],
                            aVar->name [14] );
                    scanRes [index].v19 = ( aVar->type == ASN_INTEGER );
                    scanRes [index].shelf = *(aVar->val.integer);
                break;
                case 20: /* raidVBay */
                    index = ScanResIndex ( aVar->name [12], aVar->name [13],
                            aVar->name [14] );
                    scanRes [index].v20 = ( aVar->type == ASN_INTEGER );
                    scanRes [index].bay = *(aVar->val.integer);
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v1  && scanRes [i].v3  && scanRes [i].v11 &&
                scanRes [i].v19 && scanRes [i].v20 )
        {
            retVal = STATE_OK;
            printf ( "OK volume: %i ; group: %i ; drive: %i",
                    scanRes [i].index, scanRes [i].group, scanRes [i].pDrv );
            printf ( " ; shelf: %i ; bay: %i\n",
                    scanRes [i].shelf, scanRes [i].bay );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_stat [15];
    oid o_perc [15];
    oid o_shlf [15];
    oid o_bay  [15];
    size_t oidLen;

    OidCpy ( o_stat, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_perc, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_shlf, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_bay,  &oidLen, baseOID, &baseOIDLen );
    o_stat [11] =  3; o_stat [12] = idx; o_stat [13] = grp; o_stat [14] = pDrv;
    o_perc [11] = 11; o_perc [12] = idx; o_perc [13] = grp; o_perc [14] = pDrv;
    o_shlf [11] = 19; o_shlf [12] = idx; o_shlf [13] = grp; o_shlf [14] = pDrv;
    o_bay  [11] = 20; o_bay  [12] = idx; o_bay  [13] = grp; o_bay  [14] = pDrv;
    oidLen = 15;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.stat = ReadInteger   ( sess2, o_stat, oidLen );
    v.perc = ReadInteger   ( sess2, o_perc, oidLen );
    v.shlf = ReadInteger   ( sess2, o_shlf, oidLen );
    v.bay  = ReadInteger   ( sess2, o_bay,  oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

void CreateDataFile ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

void ParseData ( const char* aData )
{
    /* performance data is not supported */
}

void WriteData ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

int CheckData ()
{
    int retVal = STATE_OK;
    char* mStat = NULL;
    char* perc = IntToStr ( v.perc );

    ReadVals ();
    switch ( v.stat )
    {
        case 1:
            mStat = StrDup ( "active" );
        break;
        case 2:
            retVal = ( retVal < STATE_WARNING ) ? STATE_WARNING : retVal;
            mStat = StrConcat ( 3, "reconstruction in progress, ",
                    v.perc, "% done" );
        break;
        case 3:
            retVal = ( retVal < STATE_WARNING ) ? STATE_WARNING : retVal;
            mStat = StrConcat ( 3, "parity reconstruction in progress, ",
                    v.perc, "% done" );
        break;
        case 4:
            retVal = ( retVal < STATE_WARNING ) ? STATE_WARNING : retVal;
            mStat = StrConcat ( 3, "parity verification in progress, ",
                    v.perc, "% done" );
        break;
        case 5:
            mStat = StrConcat ( 3, "scrubbing in progress, ",
                    v.perc, "% done" );
        break;
        case 6:
            retVal = ( retVal < STATE_CRITICAL ) ? STATE_CRITICAL : retVal;
            mStat = StrDup ( "failed" );
        break;
        default:
            retVal = ( retVal < STATE_WARNING ) ? STATE_WARNING : retVal;
            mStat = StrDup ( "invalid" );
        break;
    }
    PrintExit ( retVal );
    printf ( " - shelf %i, bay %i: %s\n", v.shlf, v.bay, mStat );
    free ( perc );
    free ( mStat );
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( checkData )
    {
        if ( !isGrpSet )  Usage ( "Missing 'group' parameters" );
        if ( !isPDrvSet ) Usage ( "Missing 'drive' parameters" );
    }
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
