/*
 * File:      check_netapp_spare.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int pDrv ;   /* spare drive index */
    int v1;      /* found spareIndex */
    int v3;      /* found spareStatus */
    int v12;     /* found spareShelf */
    int v13;     /* found spareBay */
    int shelf;   /* shelf */
    int bay;     /* bay */
} tScanRes;

/* data type for storing data */
typedef struct
{
    int  stat;  /* status of the drive */
    int  shlf;  /* shelf the drive is in */
    int  bay;   /* bay the drive is in */
} tDataSet;

/* global input variables */
int pDrv;
int isPDrvSet  = FALSE;

/*
 * NETWORK-APPLIANCE-MIB::spareStatus.PDRV = 1.3.6.1.4.1.789.1.6.3.1.3.PDRV
 * NETWORK-APPLIANCE-MIB::spareShelf.PDRV  = 1.3.6.1.4.1.789.1.6.3.1.12.PDRV
 * NETWORK-APPLIANCE-MIB::spareBay.PDRV    = 1.3.6.1.4.1.789.1.6.3.1.13.PDRV
 */
oid baseOID [] = { 1, 3, 6, 1, 4, 1, 789, 1, 6, 3, 1 };
size_t baseOIDLen = 11;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_netapp_spare";
    progVersion = "0.3";
    helpIndexP = "-d DRIVE";
    helpThresP = NULL;
    helpPurpose =
            "Checks the state of a spare drive of a NetApp(R) Filer.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
    /* rrdStep is unused */
    /* rrdHBeat is unused */
    supportsPerfLog = FALSE;
    hasIndexParam = FALSE;
}

void HelpSpecificParams ()
{
    printf (
    "-d, --drive INTEGER\n"
    "    Index of the spare drive.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin checks the overall state of a spare drive of a NetApp(R)\n"
    "Filer."
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "There is no performance data provided.\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;
    if (        !strcmp ( argv [(*i)], "-d" )
            ||  !strcmp ( argv [(*i)], "--drive" ) )
    {
        isPDrvSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            pDrv = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || *tail ) Usage ( "Invalid 'drive' parameter" );
        }
        else Usage ( "Missing 'drive' parameter(s)" );
    } else ret = FALSE;
    return ( ret );
}

int ScanResIndex ( int aIndex )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].pDrv == aIndex );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 13 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [11] )
            {
                case 1: /* spareIndex */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].pDrv = aVar->name [12];
                    scanRes [index].v1 = ( aVar->type == ASN_INTEGER );
                    scanRes [index].v3 = FALSE;
                    scanRes [index].v12 = FALSE;
                    scanRes [index].v13 = FALSE;
                break;
                case 3: /* spareStatus */
                    index = ScanResIndex ( aVar->name [12] );
                    scanRes [index].v3 = ( aVar->type == ASN_INTEGER );
                break;
                case 12: /* spareShelf */
                    index = ScanResIndex ( aVar->name [12] );
                    scanRes [index].v12 = ( aVar->type == ASN_INTEGER );
                    scanRes [index].shelf = *(aVar->val.integer);
                break;
                case 13: /* spareBay */
                    index = ScanResIndex ( aVar->name [12] );
                    scanRes [index].v13 = ( aVar->type == ASN_INTEGER );
                    scanRes [index].bay = *(aVar->val.integer);
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v1  && scanRes [i].v3  && scanRes [i].v12 &&
                scanRes [i].v13 )
        {
            retVal = STATE_OK;
            printf ( "OK drive: %i ; shelf: %i ; bay: %i\n",
                    scanRes [i].pDrv, scanRes [i].shelf, scanRes [i].bay );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_stat [13];
    oid o_shlf [13];
    oid o_bay  [13];
    size_t oidLen;

    OidCpy ( o_stat, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_shlf, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_bay,  &oidLen, baseOID, &baseOIDLen );
    o_stat [11] =  3; o_stat [12] = pDrv;
    o_shlf [11] = 12; o_shlf [12] = pDrv;
    o_bay  [11] = 13; o_bay  [12] = pDrv;
    oidLen = 13;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.stat = ReadInteger   ( sess2, o_stat, oidLen );
    v.shlf = ReadInteger   ( sess2, o_shlf, oidLen );
    v.bay  = ReadInteger   ( sess2, o_bay,  oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

void CreateDataFile ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

void ParseData ( const char* aData )
{
    /* performance data is not supported */
}

void WriteData ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

int CheckData ()
{
    int retVal = STATE_OK;
    char* mStat = NULL;

    ReadVals ();
    switch ( v.stat )
    {
        case 1:
            mStat = StrDup ( "spare" );
        break;
        case 2:
            retVal = ( retVal < STATE_WARNING ) ? STATE_WARNING : retVal;
            mStat = StrDup ( "adding spare" );
        break;
        case 3:
            retVal = ( retVal < STATE_WARNING ) ? STATE_WARNING : retVal;
            mStat = StrDup ( "bypassed" );
        break;
        case 4:
            retVal = ( retVal < STATE_WARNING ) ? STATE_WARNING : retVal;
            mStat = StrDup ( "unknown" );
        break;
        default:
            retVal = ( retVal < STATE_WARNING ) ? STATE_WARNING : retVal;
            mStat = StrDup ( "invalid" );
        break;
    }
    PrintExit ( retVal );
    printf ( " - shelf %i, bay %i: %s\n", v.shlf, v.bay, mStat );
    free ( mStat );
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( checkData )
    {
        if ( !isPDrvSet ) Usage ( "Missing 'drive' parameters" );
    }
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
