/*
 * File:      check_netapp_vol.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include "globals.h"
#include "nagiosif.h"
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int index;  /* volume index */
    int v2;     /* found qvStateName */
    int v3;     /* found qvStateStat */
    int v4;     /* found qvStateInitPercent */
    char* name; /* volume name */
} tScanRes;

/* data type for storing data */
typedef struct
{
    char* name; /* name of the volume */
    int   stat; /* quota status */
    int   perc; /* percentage quota computation is done */
} tDataSet;

/* global input variables */
char* name;
int isNameSet = FALSE;


/*
 * SNMP base OID
 * NETWORK-APPLIANCE-MIB::qvStateName.IDX
 * = 1.3.6.1.4.1.789.1.4.4.1.2.IDX
 * NETWORK-APPLIANCE-MIB::qvStateStat.IDX
 * = 1.3.6.1.4.1.789.1.4.4.1.3.IDX
 * NETWORK-APPLIANCE-MIB::qvStateInitPercent.IDX
 * = 1.3.6.1.4.1.789.1.4.4.1.4.IDX
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 789, 1, 4, 4, 1 };
size_t baseOIDLen = 11;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_netapp_vol";
    progVersion = "0.3";
    helpIndexP = "-i INDEX";
    helpThresP = "-n NAME";
    helpPurpose = "Checks the status of a volume of a NetApp(R) Filer.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
    /* rrdStep is unused */
    /* rrdHBeat is unused */
    /* rrdStep is unused */
    /* rrdHBeat is unused */
    supportsPerfLog = FALSE;
    hasIndexParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-i, --index INTEGER\n"
    "    Index of the volume of the drive.\n"
    "\n"
    "-n, --name STRING\n"
    "    Expected name of the volume.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin returns the quota status of a NetApp(R) Filer volume. It\n"
    "returns a warning when the quotas are initializing or off.\n"
    "\n"
    "It also returns a warning if the name of the volume is not the\n"
    "expected name of the volume. This may happen if volumes are added or\n"
    "deleted.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "There is no performance data provided.\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-n" )
            ||  !strcmp ( argv [(*i)], "--name" ) )
    {
        isNameSet = TRUE;
        (*i)++;
        if ( (*i) < argc ) name = StrDup ( argv [(*i)++] );
        else Usage ( "Missing 'name' parameter" );
    } else ret = FALSE;
    return ( ret );
}

int ScanResIndex ( int aIndex )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 13 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [11] )
            {
                case 1: /* qvStateVolume */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].index = aVar->name [12];
                    scanRes [index].v2 = FALSE;
                    scanRes [index].v3 = FALSE;
                    scanRes [index].v4 = FALSE;
                    scanRes [index].name = NULL;
                break;
                case 2: /* qvStateName */
                    index = ScanResIndex ( aVar->name [12] );
                    scanRes [index].v2 = ( aVar->type == ASN_OCTET_STR );
                    scanRes [index].name = VarListStrDup ( aVar );
                break;
                case 3: /* qvStateStat */
                    index = ScanResIndex ( aVar->name [12] );
                    scanRes [index].v3 = ( aVar->type == ASN_INTEGER );
                break;
                case 4: /* qvStateInitPercent */
                    index = ScanResIndex ( aVar->name [12] );
                    scanRes [index].v4 = ( aVar->type == ASN_INTEGER );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if ( scanRes [i].v2 && scanRes [i].v3 && scanRes [i].v4 )
        {
            retVal = STATE_OK;
            printf ( "OK index: %i ; name: %s\n",
                    scanRes [i].index, scanRes [i].name );
            free ( scanRes [i].name );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_name [13];
    oid o_stat [13];
    oid o_perc [13];
    size_t oidLen;

    OidCpy ( o_name, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_stat, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_perc, &oidLen, baseOID, &baseOIDLen );
    o_name [11] = 2; o_name [12] = idx;
    o_stat [11] = 3; o_stat [12] = idx;
    o_perc [11] = 4; o_perc [12] = idx;
    oidLen = 13;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.name = ReadString  ( sess2, o_name, oidLen );
    v.stat = ReadInteger ( sess2, o_stat, oidLen );
    v.perc = ReadInteger ( sess2, o_perc, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

void CreateDataFile ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

void ParseData ( const char* aData )
{
    /* performance data is not supported */
}

void WriteData ( const char* aRRDFPath )
{
    /* performance data is not supported */
}

int CheckData ()
{
    int retVal = STATE_OK;

    ReadVals ();
    if ( strcmp ( v.name, name ) )
    {
        retVal = STATE_WARNING;
        printf ( "WARNING - name does not match expected name\n" );
    } else
    {
        switch ( v.stat )
        {
            case 1:
                retVal = STATE_WARNING;
                printf ( "WARNING - Quotas are off\n" );
            break;
            case 2:
                printf ( "OK - Quotas are on\n" );
            break;
            case 3:
                retVal = STATE_WARNING;
                printf ( "WARNING - Quotas are initializing, %i%% done\n",
                        v.perc );
            break;
            default:
                retVal = STATE_UNKNOWN;
                printf ( "UNKNOWN - Quotas state is invalid\n" );
            break;
        }
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( checkData )
    {
        if ( !isNameSet )  Usage ( "Missing 'name' parameters" );
    }
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
        free ( v.name );
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
