/*
 * File:      check_storage_by_snmp.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int index;   /* storage index */
    int v2;      /* found hrStorageType */
    int v3;      /* found hrStorageDescr */
    int v4;      /* found hrStorageAllocationUnits */
    int v5;      /* found hrStorageSize */
    int v6;      /* found hrStorageUsed */
    char* type;  /* storage type OID */
    char* descr; /* storage description */
} tScanRes;

/* data type for storing data */
typedef struct
{
    uint   alloU; /* allocation unit */
    double used;  /* used storage in allocation units */
    double size;  /* size storage in allocation units */
} tDataSet;


/* global input variables */
ulong  warnS = 75;
char   warnU = '%';
ulong  critS = 85;
char   critU = '%';

/*
 * SNMP base OID
 * HOST-RESOURCES-MIB::hrStorageAllocationUnits.IDX
 * = 1.3.6.1.2.1.25.2.3.1.4.IDX
 * HOST-RESOURCES-MIB::hrStorageSize.IDX
 * = 1.3.6.1.2.1.25.2.3.1.5.IDX
 * HOST-RESOURCES-MIB::hrStorageUsed.IDX
 * = 1.3.6.1.2.1.25.2.3.1.6.IDX
 */
oid baseOID[] = { 1, 3, 6, 1, 2, 1, 25, 2, 3, 1 };
size_t baseOIDLen = 10;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_storage_by_snmp";
    progVersion = "0.8";
    helpIndexP = "-i INDEX";
    helpThresP = "[-w WARN] [-c CRIT]";
    helpPurpose = "Checks the usage of some storage device (disk, ram, ...).";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 5 * 60;
    rrdHBeat = 11 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = TRUE;
    hasAddInfParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-w, --warning INTEGER[%%kMG]\n"
    "    Storage usage at which a warning is generated.\n"
    "    Default is %lu%c.\n", warnS, warnU
    );
    printf (
    "\n"
    "-c, --critical INTEGER[%%kMG]\n"
    "    Storage usage at which a critical is generated.\n"
    "    Default is %lu%c.\n", critS, critU
    );
    printf (
    "\n"
    "-i, --index INTEGER\n"
    "    Index of the storage device which should be checked.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "If the warning and critical parameters are given without unit, then the\n"
    "unit Bytes is assumed. '%%', 'k', 'M' and 'G' are interpreted as\n"
    "percent of the whole volume, Kilobytes (1024 Bytes), Megabytes\n"
    "(1024*1024 Bytes) and Gigabytes (1024*1024*1024 Bytes) respectively.\n"
    "\n"
    "The program uses double precision floats for calculations. Nevertheless\n"
    "parameters are supposed to be integers (plus optional units).\n"
    "\n"
    "For further information about the objects used for this plugin, see\n"
    "RFC 1514 about hrStorageEntry.\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<addInf>;<used>;<size>\n"
    "where\n"
    "    addInf ... service ident.\n"
    "    used   ... used storage in Megabytes.\n"
    "    size   ... size of the storage in Megabytes.\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'used storage'=<used>MB;<warn>;<crit>;0;<size>\n"
    "\n"
    );
}

char CheckSpaceUnit ( char* aUnit )
{
    char ret = '1';
    if      ( !strcmp ( aUnit, ""  ) ) ret = '1';
    else if ( !strcmp ( aUnit, "k" ) ) ret = 'k';
    else if ( !strcmp ( aUnit, "M" ) ) ret = 'M';
    else if ( !strcmp ( aUnit, "%" ) ) ret = '%';
    else Usage ( "invalid space unit" );
    return ( ret );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-w" )
            ||  !strcmp ( argv [(*i)], "--warning" ) )
    {
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            warnS = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno ) Usage ( "Invalid 'warn' parameter" );
            warnU = CheckSpaceUnit ( tail );
        } else Usage ( "Missing 'warn' parameter(s)" );
    } else if ( !strcmp ( argv [(*i)], "-c" )
            ||  !strcmp ( argv [(*i)], "--critical" ) )
    {
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            critS = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno ) Usage ( "Invalid 'crit' parameter" );
            critU = CheckSpaceUnit ( tail );
        } else Usage ( "Missing 'crit' parameter(s)" );
    } else ret = FALSE;
    return ( ret );
}

int ScanResIndex ( int aIndex )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 12 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [10] )
            {
                case 1: /* hrStorageIndex */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].index = aVar->name [11];
                    scanRes [index].v2 = FALSE;
                    scanRes [index].v3 = FALSE;
                    scanRes [index].v4 = FALSE;
                    scanRes [index].v5 = FALSE;
                    scanRes [index].v6 = FALSE;
                    scanRes [index].type = NULL;
                    scanRes [index].descr = NULL;
                break;
                case 2: /* hrStorageType */
                {
                    char buffer [1024];

                    index = ScanResIndex ( aVar->name [11] );
                    scanRes [index].v2 = ( aVar->type == ASN_OBJECT_ID );
                    OIDToString ( buffer, 1024,
                            aVar->val.objid, aVar->val_len / sizeof (oid) );
                    scanRes [index].type = StrDup ( buffer );
                }
                break;
                case 3: /* hrStorageDescr */
                    index = ScanResIndex ( aVar->name [11] );
                    scanRes [index].v3 = ( aVar->type == ASN_OCTET_STR );
                    scanRes [index].descr = VarListStrDup ( aVar );
                break;
                case 4: /* hrStorageAllocationUnits */
                    index = ScanResIndex ( aVar->name [11] );
                    scanRes [index].v4 = ( aVar->type == ASN_INTEGER );
                break;
                case 5: /* hrStorageSize */
                    index = ScanResIndex ( aVar->name [11] );
                    scanRes [index].v5 = ( aVar->type == ASN_INTEGER );
                break;
                case 6: /* hrStorageUsed */
                    index = ScanResIndex ( aVar->name [11] );
                    scanRes [index].v6 = ( aVar->type == ASN_INTEGER );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v2 && scanRes [i].v3 && scanRes [i].v4 &&
                scanRes [i].v5 && scanRes [i].v6 )
        {
            retVal = STATE_OK;
            printf ( "OK index: %i ; type: %s ; descr: %s\n",
                    scanRes [i].index, scanRes [i].type, scanRes [i].descr );
            free ( scanRes [i].descr );
            free ( scanRes [i].type );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_alloU [12];
    oid o_size  [12];
    oid o_used  [12];
    size_t oidLen;

    OidCpy ( o_alloU, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_size,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_used,  &oidLen, baseOID, &baseOIDLen );
    o_alloU [10] = 4; o_alloU [11] = idx;
    o_size  [10] = 5; o_size  [11] = idx;
    o_used  [10] = 6; o_used  [11] = idx;
    oidLen = 12;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.alloU = ReadInteger ( sess2, o_alloU, oidLen );
    v.size  = ReadInteger ( sess2, o_size, oidLen );
    v.used  = ReadInteger ( sess2, o_used, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}


#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "used",  GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "size",  GAUGE, rrdHBeat, "0", "U" );
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    char buf[1024];

    chk = sscanf ( aData, "%1023s;%lf;%lf", buf, &(v.used), &(v.size) );
    if ( chk != 4 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
    addInf = StrDup ( buf );
    v.alloU = 1024.0 * 1024.0;
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* used   = DoubleToStr ( (double) v.alloU * (double) v.used / ( 1024.0 * 1024.0 ) );
    char* size   = DoubleToStr ( (double) v.alloU * (double) v.size / ( 1024.0 * 1024.0 ) );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "used:size" );
    params [i++] = StrConcat ( 5, ts, ":", used, ":", size );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( size );
    free ( used );
    free ( ts );
}

void PlotDataUsages ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[%]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, NULL, aTiming );
    PPAddPlot ( pp, "used storage", "%", FALSE );
    title = StrConcat ( 4, hostName, " storage ", addInf, " usage" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 100, FALSE );
    params [i++] = CreateDEF ( "sizeMR", aFN, "size",  MAX );
    params [i++] = CreateDEF ( "sizeAR", aFN, "size",  AVG );
    params [i++] = CreateDEF ( "usedMR", aFN, "used",  MAX );
    params [i++] = CreateDEF ( "usedAR", aFN, "used",  AVG );
    params [i++] = StrDup (   "CDEF:usedM=usedMR,sizeMR,/,100,*" );
    params [i++] = StrDup (   "CDEF:usedA=usedAR,sizeAR,/,100,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "usedM", C_USAGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "usedA", C_USAGE1_AVG, pp, 0, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "usedA", "usedA",  "usedM",  "usedM",  "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataUsages ( fn, PLOT_24H_5M,  now );
    PlotDataUsages ( fn, PLOT_7D_15M,  now );
    PlotDataUsages ( fn, PLOT_31D_60M, now );
    PlotDataUsages ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

double LevelToMB ( char* aWhat, int aLevel, char aUnit, double aSize )
{
    double retVal = 0.0;
    double level = (double) aLevel;

    switch ( aUnit )
    {
        case '1': retVal = level / ( 1024.0 * 1024.0 ); break;
        case 'k': retVal = level / 1024.0;              break;
        case 'M': retVal = level;                       break;
        case 'G': retVal = level * 1024.0;              break;
        case '%': retVal = level * aSize / 100.0;       break;
        default:
        {
            Error ( LOG_ERR, "An invalid level slipped through!" );
        }
        break;
    }
    return ( retVal );
}

int CheckData ()
{
    int retVal = STATE_OK;
    double size;
    double used;
    double usedPc;
    double warn;
    double crit;

    ReadVals ();
    /* calculate everything in Megabytes */
    if ( v.alloU && v.size )
    {
        size = (double) v.size * (double) v.alloU / ( 1024.0 * 1024.0 );
        used = (double) v.used * (double) v.alloU / ( 1024.0 * 1024.0 );
        usedPc = used / size * 100.0;
    }
    else
    {
        size = 0.0;
        used = 0.0;
        usedPc = 0.0;
    }
    warn = LevelToMB ( "warning level",  warnS, warnU, size );
    crit = LevelToMB ( "critical level", critS, critU, size );
    if ( ( used > warn ) && ( retVal < STATE_WARNING ) )
            retVal = STATE_WARNING;
    if ( ( used > crit ) && ( retVal < STATE_CRITICAL ) )
            retVal = STATE_CRITICAL;
    PrintExit ( retVal );
    printf ( " - %s usage: ", addInf );
    printf ( "%.0f/%.0f MB (%3.1f%%)", used, size, usedPc );
    if ( npPerfData )
    {
        printf ( "|'used storage'=%.0fMB;%.0f;%.0f;0;%.0f\n",
                used, warn, crit, size );
    }
    else
    {
        printf ( "| %s;%.0f;%.0f\n", addInf, used, size );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ();
        if ( plotPerfData ) PlotPerfData ();
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
