/*
 * File:      check_ucd_lms_fan.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int index;  /* sensor index */
    int v2;     /* found lmFanSensorsDevice */
    int v3;     /* found lmFanSensorsValue */
    char* name; /* name of the sensor */
} tScanRes;

/* data type for storing data */
typedef struct
{
    uint  speed; /* current speed in rounds per minute */
} tDataSet;


/* global input variables */
uint warn;
uint crit;
int isWarnSet  = FALSE;
int isCritSet  = FALSE;


/*
 * SNMP base OID
 * LM-SENSORS-MIB::lmFanSensorsIndex.INDEX
 * = .1.3.6.1.4.1.2021.13.16.3.1.1.INDEX
 * LM-SENSORS-MIB::lmFanSensorsDevice.INDEX
 * = .1.3.6.1.4.1.2021.13.16.3.1.2.INDEX
 * LM-SENSORS-MIB::lmFanSensorsValue.INDEX
 * = .1.3.6.1.4.1.2021.13.16.3.1.3.INDEX
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 2021, 13, 16, 3, 1 };
size_t baseOIDLen = 11;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_ucd_lms_fan";
    progVersion = "0.1";
    helpIndexP = "-i INDEX";
    helpThresP = "-w WARN -c CRIT";
    helpPurpose =
            "Checks the value of a lm_sensors monitored fan speed sensor.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 5 * 60;
    rrdHBeat = 11 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = TRUE;
    hasAddInfParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-i, --index INTEGER\n"
    "    Index of the fan which should be checked\n"
    "\n"
    "-w, --warning UINT\n"
    "    Fan speed in RPM below which a warning is generated.\n"
    "\n"
    "-c, --critical UINT\n"
    "    Fan speed in RPM below which a critical is generated.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin checks how the speed of the given sensor compares to the\n"
    "threshold speeds (should be above).\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<addInf>;<speed>\n"
    "where\n"
    "    addInf ... additional information.\n"
    "    speed  ... current speed in rounds per minute.\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'current speed'=<speed>;<warn>;<crit>\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-w" )
            ||  !strcmp ( argv [(*i)], "--warning" ) )
    {
        isWarnSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];

            errno = 0;
            warn = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || (*tail) )
            Usage ( "Invalid 'warn' parameter" );
        }
        else Usage ( "Missing 'warn' parameter(s)" );
    } else if ( !strcmp ( argv [(*i)], "-c" )
            ||  !strcmp ( argv [(*i)], "--critical" ) )
    {
        isCritSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];

            errno = 0;
            crit = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || (*tail) ) Usage ( "Invalid 'crit' parameter" );
        }
        else Usage ( "Missing 'crit' parameter(s)" );
    } else ret = FALSE;
    return ( ret );
}

int ScanResIndex ( int aIndex )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 13 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [11] )
            {
                case 1: /* lmFanSensorsIndex */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].index = aVar->name [12];
                    scanRes [index].v2 = FALSE;
                    scanRes [index].v3 = FALSE;
                    scanRes [index].name = NULL;
                break;
                case 2: /* lmFanSensorsDevice */
                    index = ScanResIndex ( aVar->name [12] );
                    scanRes [index].v2 = ( aVar->type == ASN_OCTET_STR );
                    scanRes [index].name = VarListStrDup ( aVar );
                break;
                case 3: /* lmFanSensorsValue */
                    index = ScanResIndex ( aVar->name [12] );
                    scanRes [index].v3 = ( aVar->type == ASN_GAUGE );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v2 && scanRes [i].v3 )
        {
            retVal = STATE_OK;
            printf ( "OK index: %i ; name: %s\n",
                    scanRes [i].index, scanRes [i].name );
            free ( scanRes [i].name );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_speed [13];
    size_t oidLen;

    OidCpy ( o_speed, &oidLen, baseOID, &baseOIDLen );
    o_speed [11] = 3; o_speed [12] = idx;
    oidLen = 13;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.speed = ReadGauge32 ( sess2, o_speed, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "speed",  GAUGE, rrdHBeat, "0", "30000" );
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MIN_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MIN_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MIN_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    params [i++] = RRA_MIN_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    char buf[1024];

    chk = sscanf ( aData, "%1023s;%i", buf, &(v.speed) );
    if ( chk != 2 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
    addInf = StrDup ( buf );
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* speed  = UIntToStr   ( v.speed );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "speed" );
    params [i++] = StrConcat ( 3, ts, ":", speed );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( speed );
    free ( ts );
}

void PlotDataFan ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[rpm]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, NULL, aTiming );
    PPAddPlot ( pp, "speed", "rpm", FALSE );
    title = StrConcat ( 4, hostName, " fan ", addInf, " speed" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 7000, FALSE );
    params [i++] = CreateDEF ( "speedM", aFN, "speed", MAX );
    params [i++] = CreateDEF ( "speedA", aFN, "speed", AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "speedM", C_GAUGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "speedA", C_GAUGE1_AVG, pp, 0, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "speedA", "speedA", "speedM", "speedM", "%4.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataFan ( fn, PLOT_24H_5M,  now );
    PlotDataFan ( fn, PLOT_7D_15M,  now );
    PlotDataFan ( fn, PLOT_31D_60M, now );
    PlotDataFan ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    int retVal = STATE_OK;

    ReadVals ();
    if ( v.speed < warn ) retVal = MAXIMUM ( retVal, STATE_WARNING );
    if ( v.speed < crit ) retVal = MAXIMUM ( retVal, STATE_CRITICAL );
    PrintExit ( retVal );
    printf ( " - fan speed %i: %u RPM", idx, v.speed );
    if ( npPerfData )
    {
        printf ( "|'current speed'=%u;%u;%u\n", v.speed, warn, crit );
    }
    else
    {
        printf ( "|%s;%u\n", addInf, v.speed );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( checkData )
    {
        if ( !isWarnSet ) Usage ( "Missing 'warn' parameters" );
        if ( !isCritSet ) Usage ( "Missing 'crit' parameters" );
    }
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ();
        if ( plotPerfData ) PlotPerfData ();
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
