/*
 * File:      check_ucd_lms_voltage.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */


#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int index;   /* voltage sensor index */
    int v2;      /* found lmVoltSensorsDevice */
    int v3;      /* found lmVoltSensorsValue */
    char* name; /* name of the sensor */
} tScanRes;

/* data type for storing data */
typedef struct
{
    int  volt; /* current voltage in mV */
} tDataSet;


/* global input variables */
int warnL;
int warnH;
int critL;
int critH;
int isWarnSet  = FALSE;
int isCritSet  = FALSE;

/*
 * SNMP base OID
 * LM-SENSORS-MIB::lmVoltSensorsIndex.INDEX
 * = .1.3.6.1.4.1.2021.13.16.4.1.1.INDEX
 * LM-SENSORS-MIB::lmVoltSensorsDevice.INDEX
 * = .1.3.6.1.4.1.2021.13.16.4.1.2.INDEX
 * LM-SENSORS-MIB::lmVoltSensorsValue.INDEX
 * = .1.3.6.1.4.1.2021.13.16.4.1.3.INDEX
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 2021, 13, 16, 4, 1 };
size_t baseOIDLen = 11;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_ucd_lms_voltage";
    progVersion = "0.1";
    helpIndexP = "-i INDEX";
    helpThresP = "-w WARNL,WARNH -c CRITL,CRITH";
    helpPurpose =
            "Checks the value of a lm_sensors monitored voltage sensor.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 5 * 60;
    rrdHBeat = 11 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = TRUE;
    hasAddInfParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-i, --index INTEGER\n"
    "    Index of the voltage sensor which should be checked.\n"
    "\n"
    "-w, --warning INTEGER,INTEGER\n"
    "    Voltages in mV below which (WARNL) and above which (WARNH) a\n"
    "    warning is generated .\n"
    "\n"
    "-c, --critical INTEGER,INTEGER\n"
    "    Voltages in mV below which (CRITL) and above which (CRITH) a\n"
    "    critical is generated.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin checks how the voltage of the given sensor compares to the\n"
    "threshold voltages (should be above the 'low' thresholds and below the\n"
    "'high' thresholds).\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<addInf>;<volt>\n"
    "where\n"
    "    addInf ... additional information.\n"
    "    voltC  ... current voltage in millivolts.\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'current voltage'=<voltC>;<warnL>:<warnH>;<critL>:<critH>"
    "\n"
    "Note: Nagios Plugin performance data format does not allow millivolt\n"
    "as units of measurement.\n"
    );
}

void InvalidWarnParam ()
{
    Usage ( "Invalid 'warn' parameter" );
}

void InvalidCritParam ()
{
    Usage ( "Invalid 'crit' parameter" );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;
    if (        !strcmp ( argv [(*i)], "-w" )
            ||  !strcmp ( argv [(*i)], "--warning" ) )
    {
        isWarnSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            warnL = strtol ( argv [(*i)++], &tail, 0 );
            if ( errno || ( (*tail) == '\0' ) ) InvalidWarnParam ();
            if ( (*tail) == ',' )
            {
                char* upper = &(tail [1]);

                tail = upper;
                warnH = strtol ( upper, &tail, 0 );
                if ( errno || ( (*tail) != '\0' ) ) InvalidWarnParam ();
            }
            else InvalidWarnParam();
        }
        else Usage ( "Missing 'warn' parameter(s)" );
    } else if ( !strcmp ( argv [(*i)], "-c" )
            ||  !strcmp ( argv [(*i)], "--critical" ) )
    {
        isCritSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            critL = strtol ( argv [(*i)++], &tail, 0 );
            if ( errno || ( (*tail) == '\0' ) ) InvalidCritParam ();
            if ( (*tail) == ',' )
            {
                char* upper = &(tail [1]);

                tail = upper;
                critH = strtol ( upper, &tail, 0 );
                if ( errno || ( (*tail) != '\0' ) ) InvalidCritParam ();
            }
            else InvalidCritParam();
        }
        else Usage ( "Missing 'crit' parameter(s)" );
    } else ret = FALSE;
    return ( ret );
}

int ScanResIndex ( int aIndex )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 13 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [11] )
            {
                case 1: /* lmVoltSensorsIndex */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].index = aVar->name [12];
                    scanRes [index].v2 = FALSE;
                    scanRes [index].v3 = FALSE;
                    scanRes [index].name = NULL;
                break;
                case 2: /* lmVoltSensorsDevice */
                    index = ScanResIndex ( aVar->name [12] );
                    scanRes [index].v2 = ( aVar->type == ASN_OCTET_STR );
                    scanRes [index].name = VarListStrDup ( aVar );
                break;
                case 3: /* lmVoltSensorsValue */
                    index = ScanResIndex ( aVar->name [12] );
                    scanRes [index].v3 = ( aVar->type == ASN_GAUGE );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v2 && scanRes [i].v3 )
        {
            retVal = STATE_OK;
            printf ( "OK index: %i ; name: %s\n",
                    scanRes [i].index, scanRes [i].name );
            free ( scanRes [i].name );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_volt  [13];
    size_t oidLen;

    OidCpy ( o_volt, &oidLen, baseOID, &baseOIDLen );
    o_volt [11] = 3; o_volt [12] = idx;
    oidLen = 13;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.volt = (int) ReadGauge32 ( sess2, o_volt, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "volt",  GAUGE, rrdHBeat, "-60.0", "60.0" );
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MIN_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MIN_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MIN_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    params [i++] = RRA_MIN_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    char buf[1024];

    chk = sscanf ( aData, "%1023s;%i", buf, &(v.volt) );
    if ( chk != 2 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
    addInf = StrDup ( buf );
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* volt   = DoubleToStr ( (double) v.volt / 1000.0 );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "volt" );
    params [i++] = StrConcat ( 3, ts, ":", volt );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( volt );
    free ( ts );
}

void PlotDataVoltage ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[V]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, NULL, aTiming );
    PPAddPlot ( pp, "voltage",         "V", FALSE );
    title = StrConcat ( 3, hostName, " voltage ", addInf );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_AUTO, 0, 0, FALSE );
    params [i++] = CreateDEF ( "voltM", aFN, "volt", MAX );
    params [i++] = CreateDEF ( "voltA", aFN, "volt", AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "voltM", C_GAUGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "voltA", C_GAUGE1_AVG, pp, 0, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "voltA", "voltA", "voltM", "voltM", "%6.3lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataVoltage ( fn, PLOT_24H_5M,  now );
    PlotDataVoltage ( fn, PLOT_7D_15M,  now );
    PlotDataVoltage ( fn, PLOT_31D_60M, now );
    PlotDataVoltage ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    int retVal = STATE_OK;

    ReadVals ();
    if ( v.volt < warnL  ) retVal = MAXIMUM ( retVal, STATE_WARNING );
    if ( warnH  < v.volt ) retVal = MAXIMUM ( retVal, STATE_WARNING );
    if ( v.volt < critL  ) retVal = MAXIMUM ( retVal, STATE_CRITICAL );
    if ( critH  < v.volt ) retVal = MAXIMUM ( retVal, STATE_CRITICAL );
    PrintExit ( retVal );
    printf ( " - voltage %i: %4.1f V", idx, (double) v.volt / 1000.0 );
    if ( npPerfData )
    {
        printf ( "|'current voltage'=%i;%i:%i;%i:%i\n",
                v.volt, warnL, warnH, critL, critH );
    }
    else
    {
        printf ( "|%s;%i\n", addInf, v.volt );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( checkData )
    {
        if ( !isWarnSet ) Usage ( "Missing 'warn' parameters" );
        if ( !isCritSet ) Usage ( "Missing 'crit' parameters" );
    }
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ();
        if ( plotPerfData ) PlotPerfData ();
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
