/*
 * File:      check_ucd_snmp_cpu.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int v50; /* found ifDescr */
    int v51; /* found ifType */
    int v52; /* found ifSpeed */
    int v53; /* found ifAdminStatus */
} tScanRes;

/* data type for storing data */
typedef struct
{
    uint user;
    uint nice;
    uint syst;
    uint idle;
} tDataSet;


/*
 * SNMP base OID
 * UCD-SNMP-MIB::ssCpuRawUser.0   = .1.3.6.1.4.1.2021.11.50.0
 * UCD-SNMP-MIB::ssCpuRawNice.0   = .1.3.6.1.4.1.2021.11.51.0
 * UCD-SNMP-MIB::ssCpuRawSystem.0 = .1.3.6.1.4.1.2021.11.52.0
 * UCD-SNMP-MIB::ssCpuRawIdle.0   = .1.3.6.1.4.1.2021.11.53.0
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 2021, 11 };
size_t baseOIDLen = 8;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_ucd_snmp_cpu";
    progVersion = "0.2";
    helpIndexP = "";
    helpThresP = NULL;
    helpPurpose = "Reports the cpu usage.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 1 * 60;
    rrdHBeat = 3 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = FALSE;
}

void HelpSpecificParams ()
{
    /* Does not have any specific parameters */
}

void HelpExplain ()
{
    printf (
    "This plugin is used only for logging cpu usages. It does not check any\n"
    "thresholds.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<user>;<nice>;<system>;<idle>\n"
    "where\n"
    "    user   ... user CPU time.\n"
    "    nice   ... nice CPU time.\n"
    "    system ... system CPU time.\n"
    "    idle   ... idle CPU time.\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'user'=<user>c "
    "'nice'=<nice>c "
    "'system'=<system>c "
    "'idle'=<idle>c\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    return FALSE;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 10 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            switch ( aVar->name [8] )
            {
                case 50: /* ssCpuRawUser */
                    scanResLen = 1;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [0].v50 = ( aVar->type == ASN_COUNTER );
                    scanRes [0].v51 = FALSE;
                    scanRes [0].v52 = FALSE;
                    scanRes [0].v53 = FALSE;
                break;
                case 51: /* ssCpuRawNice */
                    scanRes [0].v51 = ( aVar->type == ASN_COUNTER );
                break;
                case 52: /* ssCpuRawSystem */
                    scanRes [0].v52 = ( aVar->type == ASN_COUNTER );
                break;
                case 53: /* ssCpuRawIdle */
                    scanRes [0].v53 = ( aVar->type == ASN_COUNTER );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v50 && scanRes [i].v51 && scanRes [i].v52 &&
                scanRes [i].v53 )
        {
            retVal = STATE_OK;
            printf ( "OK\n" );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_user [10];
    oid o_nice [10];
    oid o_syst [10];
    oid o_idle [10];
    size_t oidLen;

    OidCpy ( o_user, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_nice, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_syst, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_idle, &oidLen, baseOID, &baseOIDLen );
    o_user [8] = 50; o_user [9] = 0;
    o_nice [8] = 51; o_nice [9] = 0;
    o_syst [8] = 52; o_syst [9] = 0;
    o_idle [8] = 53; o_idle [9] = 0;
    oidLen = 10;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.user = ReadCounter32 ( sess2, o_user, oidLen );
    v.nice = ReadCounter32 ( sess2, o_nice, oidLen );
    v.syst = ReadCounter32 ( sess2, o_syst, oidLen );
    v.idle = ReadCounter32 ( sess2, o_idle, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "user",  COUNTER, rrdHBeat, "0", "25000" );
    params [i++] = CreateDSStr ( "nice",  COUNTER, rrdHBeat, "0", "25000" );
    params [i++] = CreateDSStr ( "syst",  COUNTER, rrdHBeat, "0", "25000" );
    params [i++] = CreateDSStr ( "idle",  COUNTER, rrdHBeat, "0", "25000" );
    params [i++] = RRA_MAX_7H_1M;
    params [i++] = RRA_AVG_7H_1M;
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;

    chk = sscanf ( aData, "%u;%u;%u;%u",
            &(v.user), &(v.nice), &(v.syst), &(v.idle) );
    if ( chk != 4 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* user   = UIntToStr   ( v.user );
    char* nice   = UIntToStr   ( v.nice );
    char* syst   = UIntToStr   ( v.syst );
    char* idle   = UIntToStr   ( v.idle );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "user:nice:syst:idle" );
    params [i++] = StrConcat ( 9, ts, ":", user, ":", nice, ":", syst,
            ":", idle );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( idle );
    free ( syst );
    free ( nice );
    free ( user );
    free ( ts );
}

void PlotDataUsage ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[%]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, NULL, aTiming );
    PPAddPlot ( pp, "user", "%", FALSE );
    PPAddPlot ( pp, "nice", "%", TRUE );
    PPAddPlot ( pp, "sys",  "%", TRUE );
    PPAddPlot ( pp, "idle", "%", TRUE );
    title = StrConcat ( 2, hostName, " cpu usage" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_AUTO, 0, 0, FALSE );
    params [i++] = CreateDEF ( "userMR", aFN, "user", MAX );
    params [i++] = CreateDEF ( "userAR", aFN, "user", AVG );
    params [i++] = CreateDEF ( "niceMR", aFN, "nice", MAX );
    params [i++] = CreateDEF ( "niceAR", aFN, "nice", AVG );
    params [i++] = CreateDEF ( "systMR", aFN, "syst", MAX );
    params [i++] = CreateDEF ( "systAR", aFN, "syst", AVG );
    params [i++] = CreateDEF ( "idleMR", aFN, "idle", MAX );
    params [i++] = CreateDEF ( "idleAR", aFN, "idle", AVG );
    params [i++] = StrDup ( "CDEF:totalM=userMR,niceMR,systMR,idleMR,+,+,+" );
    params [i++] = StrDup ( "CDEF:totalA=userAR,niceAR,systAR,idleAR,+,+,+" );
    params [i++] = StrDup ( "CDEF:totalMM=100.0,totalM,/" );
    params [i++] = StrDup ( "CDEF:totalAM=100.0,totalA,/" );
    params [i++] = StrDup ( "CDEF:userM=userMR,totalMM,*" );
    params [i++] = StrDup ( "CDEF:userA=userAR,totalAM,*" );
    params [i++] = StrDup ( "CDEF:niceM=niceMR,totalMM,*" );
    params [i++] = StrDup ( "CDEF:niceA=niceAR,totalAM,*" );
    params [i++] = StrDup ( "CDEF:systM=systMR,totalMM,*" );
    params [i++] = StrDup ( "CDEF:systA=systAR,totalAM,*" );
    params [i++] = StrDup ( "CDEF:idleM=idleMR,totalMM,*" );
    params [i++] = StrDup ( "CDEF:idleA=idleAR,totalAM,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "userM", C_LOAD3_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "niceM", C_LOAD4_MAX, pp, 1, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "systM", C_LOAD1_MAX, pp, 2, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "idleM", C_LOAD2_MAX, pp, 3, MAX, TRUE );
    params [i++] = CreatePlot ( PLOT_LINE2, "userA", C_LOAD3_AVG, pp, 0, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "niceA", C_LOAD4_AVG, pp, 1, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "systA", C_LOAD1_AVG, pp, 2, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "idleA", C_LOAD2_AVG, pp, 3, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "userA", "userA", "userM", "userM", "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "niceA", "niceA", "niceM", "niceM", "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 3, "systA", "systA", "systM", "systM", "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 4, "idleA", "idleA", "idleM", "idleM", "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataUsage ( fn, PLOT_6H_1M,   now );
    PlotDataUsage ( fn, PLOT_24H_5M,  now );
    PlotDataUsage ( fn, PLOT_7D_15M,  now );
    PlotDataUsage ( fn, PLOT_31D_60M, now );
    PlotDataUsage ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    ReadVals ();
    printf ( "OK - cpu counters: %u, %u, %u, %u",
        v.user, v.nice, v.syst, v.idle );
    if ( npPerfData )
    {
        printf ( "|'user'=%uc", v.user );
        printf ( " 'nice'=%uc", v.nice );
        printf ( " 'system'=%uc", v.syst );
        printf ( " 'idle'=%uc\n", v.idle );
    }
    else
    {
        printf ( "|%u;%u;%u;%u\n", v.user, v.nice, v.syst, v.idle );
    }
    return ( STATE_OK );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ();
        if ( plotPerfData ) PlotPerfData ();
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
