/*
 * File:      check_load_by_snmp.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int v1; /* laLoadInt.1 */
    int v2; /* laLoadInt.2 */
    int v3; /* laLoadInt.3 */
} tScanRes;

/* data type for storing data */
typedef struct
{
    float load1;  /*  1 minute average load */
    float load5;  /*  5 minute average load */
    float load15; /* 15 minute average load */
} tDataSet;

/* global input variables */
float warn1;
float warn5;
float warn15;
float crit1;
float crit5;
float crit15;
int isWarnSet  = FALSE;
int isCritSet  = FALSE;

/*
 * SNMP base OID
 * UCD-SNMP-MIB::laLoadInt.1 = 1.3.6.1.4.1.2021.10.1.5.1
 * UCD-SNMP-MIB::laLoadInt.2 = 1.3.6.1.4.1.2021.10.1.5.2
 * UCD-SNMP-MIB::laLoadInt.3 = 1.3.6.1.4.1.2021.10.1.5.3
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 2021, 10, 1, 5 };
size_t baseOIDLen = 10;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_ucd_snmp_load";
    progVersion = "0.8";
    helpIndexP = NULL;
    helpThresP = "-w WARN1,WARN5,WARN15 -c CRIT1,CRIT5,CRIT15";
    helpPurpose = "Checks the average system loads using SNMP.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 1 * 60;
    rrdHBeat = 3 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = FALSE;
}


void HelpSpecificParams ()
{
    printf (
    "-w, --warning FLOAT,FLOAT,FLOAT\n"
    "    Average system loads at which a warning is generated\n"
    "\n"
    "-c, --critical FLOAT,FLOAT,FLOAT\n"
    "    Average system loads at which a critical is generated\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "To generate a warning or critical only one of the average system loads\n"
    "needs to be higher than the respective level.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<load1>;<load5>;<load15>\n"
    "where\n"
    "    load1  ... system load during the last minute\n"
    "    load5  ... system load during the last 5 minutes\n"
    "    load15 ... system load during the last 15 minutes\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'1 minute load'=<load1>;<warn1>;<crit1> "
    "'5 minute load'=<load5>;<warn5>;<crit5> "
    "'15 minute load'=<load15>;<warn15>;<crit15>\n"
    "\n"
    );
}

void InvalidWarnParam ()
{
    Usage ( "Invalid 'warn' parameter" );
}

void InvalidCritParam ()
{
    Usage ( "Invalid 'crit' parameter" );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (            !strcmp ( argv [(*i)], "-w" )
                ||  !strcmp ( argv [(*i)], "--warning" ) )
    {
        isWarnSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];

            errno = 0;
            warn1 = strtod ( argv [(*i)], &tail );
            if ( errno ) InvalidWarnParam();
            if ( (*tail) == ',' )
            {
                char* w5str = &(tail [1]);

                tail = w5str;
                warn5 = strtod ( w5str, &tail );
                if ( errno ) InvalidWarnParam ();
                if ( (*tail) == ',' )
                {
                    char* w15str = &(tail [1]);

                    tail = w15str;
                    warn15 = strtod ( w15str, &tail );
                    if ( errno || *tail ) InvalidWarnParam ();
                    (*i)++;
                }
                else InvalidWarnParam();
            }
            else InvalidWarnParam();
        }
        else Usage ( "Missing 'warn' parameter" );
    } else if ( !strcmp ( argv [(*i)], "-c" )
            ||  !strcmp ( argv [(*i)], "--critical" ) )
    {
        isCritSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];

            errno = 0;
            crit1 = strtod ( argv [(*i)], &tail );
            if ( errno ) InvalidCritParam ();
            if ( *tail == ',' )
            {
                char* c5str = &(tail [1]);

                tail = c5str;
                crit5 = strtod ( c5str, &tail );
                if ( errno ) InvalidCritParam ();
                if ( *tail == ',' )
                {
                    char* c15str = &(tail [1]);

                    tail = c15str;
                    crit15 = strtod ( c15str, &tail );
                    if ( errno || *tail ) InvalidCritParam ();
                    (*i)++;
                }
                else InvalidCritParam ();
            }
            else InvalidCritParam ();
        }
        else Usage ( "Missing 'crit' parameter" );
    } else ret = FALSE;
    return ( ret );
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 11 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            switch ( aVar->name [10] )
            {
                case 1: /* laLoadInt.1 */
                    scanResLen = 1;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [0].v1 = ( aVar->type == ASN_INTEGER );
                    scanRes [0].v2 = FALSE;
                    scanRes [0].v3 = FALSE;
                break;
                case 2: /* laLoadInt.2 */
                    scanRes [0].v2 = ( aVar->type == ASN_INTEGER );
                break;
                case 3: /* laLoadInt.3 */
                    scanRes [0].v3 = ( aVar->type == ASN_INTEGER );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if ( scanRes [i].v1 && scanRes [i].v2 && scanRes [i].v3 )
        {
            retVal = STATE_OK;
            printf ( "OK\n" );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_load1  [11];
    oid o_load5  [11];
    oid o_load15 [11];
    size_t oidLen;

    OidCpy ( o_load1,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_load5,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_load15, &oidLen, baseOID, &baseOIDLen );
    o_load1  [10] = 1;
    o_load5  [10] = 2;
    o_load15 [10] = 3;
    oidLen = 11;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.load1  = ( (float) ReadInteger ( sess2, o_load1,  oidLen ) ) / 100.0;
    v.load5  = ( (float) ReadInteger ( sess2, o_load5,  oidLen ) ) / 100.0;
    v.load15 = ( (float) ReadInteger ( sess2, o_load15, oidLen ) ) / 100.0;
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "load1",  GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "load5",  GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "load15", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = RRA_MAX_7H_1M;
    params [i++] = RRA_AVG_7H_1M;
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    chk = sscanf ( aData, "%f;%f;%f", &(v.load1), &(v.load5), &(v.load15) );
    if ( chk != 3 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* load1  = DoubleToStr ( v.load1 );
    char* load5  = DoubleToStr ( v.load5 );
    char* load15 = DoubleToStr ( v.load15 );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "load1:load5:load15" );
    params [i++] = StrConcat ( 7, ts, ":", load1, ":", load5, ":", load15 );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( load15 );
    free ( load5 );
    free ( load1 );
    free ( ts );
}

void PlotDataLoad ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[Processes]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, NULL, aTiming );
    PPAddPlot ( pp, "1 min load",  "Processes", FALSE );
    PPAddPlot ( pp, "5 min load",  "Processes", FALSE );
    PPAddPlot ( pp, "15 min load", "Processes", FALSE );
    title = StrConcat ( 2, hostName, " system load" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, FALSE );
    params [i++] = CreateDEF ( "load1M",  aFN, "load1",  MAX );
    params [i++] = CreateDEF ( "load1A",  aFN, "load1",  AVG );
    params [i++] = CreateDEF ( "load5M",  aFN, "load5",  MAX );
    params [i++] = CreateDEF ( "load5A",  aFN, "load5",  AVG );
    params [i++] = CreateDEF ( "load15M", aFN, "load15", MAX );
    params [i++] = CreateDEF ( "load15A", aFN, "load15", AVG );
    params [i++] = CreatePlot ( PLOT_LINE3, "load1M",  C_LOAD1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE3, "load5M",  C_LOAD2_MAX, pp, 1, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE3, "load15M", C_LOAD3_MAX, pp, 2, MAX, TRUE );
    params [i++] = CreatePlot ( PLOT_LINE2, "load1A",  C_LOAD1_AVG, pp, 0, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "load5A",  C_LOAD2_AVG, pp, 1, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "load15A", C_LOAD3_AVG, pp, 2, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "load1A",  "load1A",  "load1M",  "load1M",  "%5.2lf" );
    i = PlotPrintParams ( i, params, pp, 1, "load5A",  "load5A",  "load5M",  "load5M",  "%5.2lf" );
    i = PlotPrintParams ( i, params, pp, 2, "load15A", "load15A", "load15M", "load15M", "%5.2lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ( const char* aSrvId )
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataLoad ( fn, PLOT_6H_1M,   now );
    PlotDataLoad ( fn, PLOT_24H_5M,  now );
    PlotDataLoad ( fn, PLOT_7D_15M,  now );
    PlotDataLoad ( fn, PLOT_31D_60M, now );
    PlotDataLoad ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    int retVal = STATE_OK;

    ReadVals ();
    if ( ( v.load1 > warn1 ) && ( retVal < STATE_WARNING ) )
            retVal = STATE_WARNING;
    if ( ( v.load5 > warn5 ) && ( retVal < STATE_WARNING ) )
            retVal = STATE_WARNING;
    if ( ( v.load15 > warn15 ) && ( retVal < STATE_WARNING ) )
            retVal = STATE_WARNING;
    if ( ( v.load1 > crit1 ) && ( retVal < STATE_CRITICAL ) )
            retVal = STATE_CRITICAL;
    if ( ( v.load5 > crit5 ) && ( retVal < STATE_CRITICAL ) )
            retVal = STATE_CRITICAL;
    if ( ( v.load15 > crit15 ) && ( retVal < STATE_CRITICAL ) )
            retVal = STATE_CRITICAL;
    PrintExit ( retVal );
    printf ( " - load average: %.2f, %.2f, %.2f",
            v.load1, v.load5, v.load15 );
    if ( npPerfData )
    {
        printf ( "|'1 minute load'=%.2f;%.2f;%.2f",
                v.load1, warn1, crit1 );
        printf ( " '5 minute load'=%.2f;%.2f;%.2f",
                v.load5, warn5, crit5 );
        printf ( " '15 minute load'=%.2f;%.2f;%.2f\n",
                v.load15, warn15, crit15 );
    }
    else
    {
        printf ( "|%.2f;%.2f;%.2f\n", v.load1, v.load5, v.load15 );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( checkData )
    {
        if ( !isWarnSet ) Usage ( "Missing 'warn' parameters" );
        if ( !isCritSet ) Usage ( "Missing 'crit' parameters" );
    }
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ( NULL );
        if ( plotPerfData ) PlotPerfData ( NULL );
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
