/*
 * File:      check_ucd_mem.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
/* Note: VMware ESX reports negative values for this. I've got to catch it. */
typedef struct
{
    int v3;      /* found memTotalSwap and is OK */
    int v4;      /* found memAvailSwap and is OK */
    int v5;      /* found memTotalReal and is OK */
    int v6;      /* found memAvailReal and is OK */
    int v11;     /* found memTotalFree and is OK */
    int v13;     /* found memShared and is OK */
    int v14;     /* found memBuffer and is OK */
    int v15;     /* found memCached and is OK */
} tScanRes;

/* data type for storing data */
typedef struct
{
    int    sTotal; /* total swap space */
    int    sAvail; /* available swap space */
    int    rTotal; /* total real/physical memory */
    int    rAvail; /* available real/physical memory */
    int    tAvail; /* total available (free) memory */
    int    shared; /* shared memory */
    int    buffer; /* buffered memory */
    int    cached; /* cache memory */
} tDataSet;


/* global input variables */
ulong  warnS = 75; /* used virtual (=real+swap) memory */
char   warnU = '%';
ulong  critS = 85;
char   critU = '%';

/*
 * SNMP base OID
 * UCD-SNMP-MIB::memTotalSwap.0 = 1.3.6.1.4.1.2021.4.3.0
 * UCD-SNMP-MIB::memAvailSwap.0 = 1.3.6.1.4.1.2021.4.4.0
 * UCD-SNMP-MIB::memTotalReal.0 = 1.3.6.1.4.1.2021.4.5.0
 * UCD-SNMP-MIB::memAvailReal.0 = 1.3.6.1.4.1.2021.4.6.0
 * UCD-SNMP-MIB::memTotalFree.0 = 1.3.6.1.4.1.2021.4.11.0
 * UCD-SNMP-MIB::memShared.0    = 1.3.6.1.4.1.2021.4.13.0
 * UCD-SNMP-MIB::memBuffer.0    = 1.3.6.1.4.1.2021.4.14.0
 * UCD-SNMP-MIB::memCached.0    = 1.3.6.1.4.1.2021.4.15.0
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 2021, 4 };
size_t baseOIDLen = 8;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_ucd_mem";
    progVersion = "0.1";
    helpIndexP = NULL;
    helpThresP = "[-w WARN] [-c CRIT]";
    helpPurpose = "Checks the memory usage of a u*ix computer.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 1 * 60;
    rrdHBeat = 3 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = FALSE;
    hasAddInfParam = FALSE;
}

void HelpSpecificParams ()
{
    printf (
    "-w, --warning INTEGER[%%kMG]\n"
    "    Virtual (= swap + real) memory usage at which a warning is "
    "generated.\n"
    "    Default is %lu%c.\n", warnS, warnU
    );
    printf (
    "\n"
    "-c, --critical INTEGER[%%kMG]\n"
    "    Virtual (= swap + real) memory usage at which a critical is "
    "generated.\n"
    "    Default is %lu%c.\n", critS, critU
    );
}

void HelpExplain ()
{
    printf (
    "If the warning and critical parameters are given without unit, then the\n"
    "unit Bytes is assumed. '%%', 'k', 'M' and 'G' are interpreted as\n"
    "percent of the total space, Kilobytes (1024 Bytes), Megabytes\n"
    "(1024*1024 Bytes) and Gigabytes (1024*1024*1024 Bytes) respectively.\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format (all in one line):\n"
    "<sTotal>;<sAvail>;<rTotal>;<rAvail>;<tTotal>;<tAvail>;<shared>;\n"
    "<buffer>;<cached>;\n"
    "where\n"
    "    sTotal ... total swap space.\n"
    "    sAvail ... available swap space.\n"
    "    rTotal ... total real/physical memory.\n"
    "    rAvail ... available real/physical memory.\n"
    "    tTotal ... total memory.\n"
    "    tAvail ... total available memory.\n"
    "    shared ... shared memory.\n"
    "    buffer ... buffered memory.\n"
    "    cached ... cache memory.\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'swap space'=<sAvail>KB;;;0;<sTotal> "
    "'real memory'=<rAvail>KB;;;0;<rTotal> "
    "'available memory'=<tAvail>KB;<warn>;<crit>;0;<tTotal> "
    "'shared memory'=<shared>KB "
    "'buffered memory'=<buffer>KB "
    "'cache memory'=<cached>KB\n"
    "\n"
    );
}

char CheckSpaceUnit ( char* aUnit )
{
    char ret = '1';
    if      ( !strcmp ( aUnit, ""  ) ) ret = '1';
    else if ( !strcmp ( aUnit, "k" ) ) ret = 'k';
    else if ( !strcmp ( aUnit, "M" ) ) ret = 'M';
    else if ( !strcmp ( aUnit, "%" ) ) ret = '%';
    else Usage ( "invalid space unit" );
    return ( ret );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-w" )
            ||  !strcmp ( argv [(*i)], "--warning" ) )
    {
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            warnS = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno ) Usage ( "Invalid 'warn' parameter" );
            warnU = CheckSpaceUnit ( tail );
        } else Usage ( "Missing 'warn' parameter(s)" );
    } else if ( !strcmp ( argv [(*i)], "-c" )
            ||  !strcmp ( argv [(*i)], "--critical" ) )
    {
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            critS = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno ) Usage ( "Invalid 'crit' parameter" );
            critU = CheckSpaceUnit ( tail );
        } else Usage ( "Missing 'crit' parameter(s)" );
    } else ret = FALSE;
    return ( ret );
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 10 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            if ( aVar->name [9] == 0 )
            {
                int typeOK = FALSE;
                int valOK = FALSE;

                switch ( aVar->name [8] )
                {
                    case 3: /* memTotalSwap */
                        scanResLen = 1;
                        scanRes = realloc ( scanRes,
                                sizeof ( tScanRes ) * scanResLen );
                        if ( scanRes == NULL ) ErrorMalloc ();
                        typeOK = ( aVar->type == ASN_INTEGER );
                        valOK = ( *(aVar->val.integer) > 0 );
                        scanRes [0].v3 = ( typeOK && valOK );
                        scanRes [0].v4 = FALSE;
                        scanRes [0].v5 = FALSE;
                        scanRes [0].v6 = FALSE;
                        scanRes [0].v11 = FALSE;
                        scanRes [0].v13 = FALSE;
                        scanRes [0].v14 = FALSE;
                        scanRes [0].v15 = FALSE;
                    break;
                    case 4: /* memAvailSwap */
                        typeOK = ( aVar->type == ASN_INTEGER );
                        valOK = ( *(aVar->val.integer) > 0 );
                        scanRes [0].v4 = ( typeOK && valOK );
                    break;
                    case 5: /* memTotalReal */
                        typeOK = ( aVar->type == ASN_INTEGER );
                        valOK = ( *(aVar->val.integer) > 0 );
                        scanRes [0].v5 = ( typeOK && valOK );
                    break;
                    case 6: /* memAvailReal */
                        typeOK = ( aVar->type == ASN_INTEGER );
                        valOK = ( *(aVar->val.integer) > 0 );
                        scanRes [0].v6 = ( typeOK && valOK );
                    break;
                    case 11: /* memTotalFree */
                        typeOK = ( aVar->type == ASN_INTEGER );
                        valOK = ( *(aVar->val.integer) > 0 );
                        scanRes [0].v11 = ( typeOK && valOK );
                    break;
                    case 13: /* memShared */
                        typeOK = ( aVar->type == ASN_INTEGER );
                        valOK = ( *(aVar->val.integer) > 0 );
                        scanRes [0].v13 = ( typeOK && valOK );
                    break;
                    case 14: /* memBuffer */
                        typeOK = ( aVar->type == ASN_INTEGER );
                        valOK = ( *(aVar->val.integer) > 0 );
                        scanRes [0].v14 = ( typeOK && valOK );
                    break;
                    case 15: /* memCached */
                        typeOK = ( aVar->type == ASN_INTEGER );
                        valOK = ( *(aVar->val.integer) > 0 );
                        scanRes [0].v15 = ( typeOK && valOK );
                    break;
                    default:
                        /* don't care */
                    break;
                }
            }
            else Log ( LOG_CRIT, "Unexpected index (!= 0) while scanning" );
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v3  && scanRes [i].v4  && scanRes [i].v5  &&
                scanRes [i].v6  && scanRes [i].v11 && scanRes [i].v13 &&
                scanRes [i].v14 && scanRes [i].v15 )
        {
            retVal = STATE_OK;
            printf ( "OK\n" );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_sTotal [10];
    oid o_sAvail [10];
    oid o_rTotal [10];
    oid o_rAvail [10];
    oid o_tAvail [10];
    oid o_shared [10];
    oid o_buffer [10];
    oid o_cached [10];
    size_t oidLen;

    OidCpy ( o_sTotal,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_sAvail,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_rTotal,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_rAvail,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_tAvail,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_shared,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_buffer,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_cached,  &oidLen, baseOID, &baseOIDLen );
    o_sTotal [8] =  3; o_sTotal [9] = 0;
    o_sAvail [8] =  4; o_sAvail [9] = 0;
    o_rTotal [8] =  5; o_rTotal [9] = 0;
    o_rAvail [8] =  6; o_rAvail [9] = 0;
    o_tAvail [8] = 11; o_tAvail [9] = 0;
    o_shared [8] = 13; o_shared [9] = 0;
    o_buffer [8] = 14; o_buffer [9] = 0;
    o_cached [8] = 15; o_cached [9] = 0;
    oidLen = 10;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.sTotal = ReadInteger ( sess2, o_sTotal, oidLen );
    v.sAvail = ReadInteger ( sess2, o_sAvail, oidLen );
    v.rTotal = ReadInteger ( sess2, o_rTotal, oidLen );
    v.rAvail = ReadInteger ( sess2, o_rAvail, oidLen );
    v.tAvail = ReadInteger ( sess2, o_tAvail, oidLen );
    v.shared = ReadInteger ( sess2, o_shared, oidLen );
    v.buffer = ReadInteger ( sess2, o_buffer, oidLen );
    v.cached = ReadInteger ( sess2, o_cached, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}


#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "sTotal", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "sAvail", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "rTotal", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "rAvail", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "tAvail", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "shared", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "buffer", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "cached", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    int junk;

    chk = sscanf ( aData, "%i;%i;%i;%i;%i;%i;%i;%i;%i",
            &(v.sTotal), &(v.sAvail), &(v.rTotal), &(v.rAvail),
            &(junk),
            &(v.tAvail), &(v.shared), &(v.buffer), &(v.cached) );
    if ( chk != 8 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* sTotal = IntToStr    ( v.sTotal );
    char* sAvail = IntToStr    ( v.sAvail );
    char* rTotal = IntToStr    ( v.rTotal );
    char* rAvail = IntToStr    ( v.rAvail );
    char* tAvail = IntToStr    ( v.tAvail );
    char* shared = IntToStr    ( v.shared );
    char* buffer = IntToStr    ( v.buffer );
    char* cached = IntToStr    ( v.cached );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup (
            "sTotal:sAvail:rTotal:rAvail:"
            "tAvail:shared:buffer:cached" );
    params [i++] = StrConcat ( 17, ts, ":",
            sTotal, ":", sAvail, ":", rTotal, ":", rAvail, ":",
            tAvail, ":", shared, ":", buffer, ":", cached  );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( cached );
    free ( buffer );
    free ( shared );
    free ( tAvail );
    free ( rAvail );
    free ( rTotal );
    free ( sAvail );
    free ( sTotal );
    free ( ts );
}

void PlotDataReal ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[%]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "real", aTiming );
    PPAddPlot ( pp, "used",   "%", FALSE );
    PPAddPlot ( pp, "shared", "%", TRUE );
    PPAddPlot ( pp, "buffer", "%", TRUE );
    PPAddPlot ( pp, "cached", "%", TRUE );
    PPAddPlot ( pp, "tAvail", "%", TRUE );
    title = StrConcat ( 2, hostName, " real memory" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 100, FALSE );
    params [i++] = CreateDEF ( "sizeMR", aFN, "rTotal",  MAX );
    params [i++] = CreateDEF ( "sizeAR", aFN, "rTotal",  AVG );
    params [i++] = CreateDEF ( "sharMR", aFN, "shared",  MAX );
    params [i++] = CreateDEF ( "sharAR", aFN, "shared",  AVG );
    params [i++] = CreateDEF ( "buffMR", aFN, "buffer",  MAX );
    params [i++] = CreateDEF ( "buffAR", aFN, "buffer",  AVG );
    params [i++] = CreateDEF ( "cachMR", aFN, "cached",  MAX );
    params [i++] = CreateDEF ( "cachAR", aFN, "cached",  AVG );
    params [i++] = CreateDEF ( "freeMR", aFN, "rAvail",  MAX );
    params [i++] = CreateDEF ( "freeAR", aFN, "rAvail",  AVG );
    params [i++] = StrDup ( "CDEF:usedMR=sizeMR,sharMR,-,buffMR,-,cachMR,-,freeMR,-" );
    params [i++] = StrDup ( "CDEF:usedAR=sizeAR,sharAR,-,buffAR,-,cachAR,-,freeAR,-" );
    params [i++] = StrDup ( "CDEF:usedM=usedMR,sizeMR,/,100,*" );
    params [i++] = StrDup ( "CDEF:usedA=usedAR,sizeAR,/,100,*" );
    params [i++] = StrDup ( "CDEF:sharM=sharMR,sizeMR,/,100,*" );
    params [i++] = StrDup ( "CDEF:sharA=sharAR,sizeAR,/,100,*" );
    params [i++] = StrDup ( "CDEF:buffM=buffMR,sizeMR,/,100,*" );
    params [i++] = StrDup ( "CDEF:buffA=buffAR,sizeAR,/,100,*" );
    params [i++] = StrDup ( "CDEF:cachM=cachMR,sizeMR,/,100,*" );
    params [i++] = StrDup ( "CDEF:cachA=cachAR,sizeAR,/,100,*" );
    params [i++] = StrDup ( "CDEF:freeM=freeMR,sizeMR,/,100,*" );
    params [i++] = StrDup ( "CDEF:freeA=freeAR,sizeAR,/,100,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "usedM", C_LOAD3_MAX,  pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "sharM", C_THRU_MAX,   pp, 1, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "buffM", C_USAGE2_MAX, pp, 2, MAX, TRUE  );
    params [i++] = CreatePlot ( PLOT_STACK, "cachM", C_LOAD1_MAX,  pp, 3, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "freeM", C_FREE_MAX,   pp, 4, MAX, TRUE  );
    params [i++] = CreatePlot ( PLOT_LINE3, "usedA", C_LOAD3_AVG,  pp, 0, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "sharA", C_THRU_AVG,   pp, 1, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "buffA", C_USAGE2_AVG, pp, 2, AVG, TRUE  );
    params [i++] = CreatePlot ( PLOT_STACK, "cachA", C_LOAD1_AVG,  pp, 3, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "freeA", C_FREE_AVG,   pp, 4, AVG, TRUE  );
    i = PlotPrintParams ( i, params, pp, 0, "usedA", "usedA", "usedM", "usedM", "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "sharA", "sharA", "sharM", "sharM", "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 2, "buffA", "buffA", "buffM", "buffM", "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 3, "cachA", "cachA", "cachM", "cachM", "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 4, "freeA", "freeA", "freeM", "freeM", "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataSwap ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[%]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "swap", aTiming );
    PPAddPlot ( pp, "used", "%", FALSE );
    PPAddPlot ( pp, "free", "%", TRUE );
    title = StrConcat ( 2, hostName, " swap memory" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 100, FALSE );
    params [i++] = CreateDEF ( "sizeMR", aFN, "sTotal", MAX );
    params [i++] = CreateDEF ( "sizeAR", aFN, "sTotal", AVG );
    params [i++] = CreateDEF ( "freeMR", aFN, "sAvail", MAX );
    params [i++] = CreateDEF ( "freeAR", aFN, "sAvail", AVG );
    params [i++] = StrDup ( "CDEF:usedMR=sizeMR,freeMR,-" );
    params [i++] = StrDup ( "CDEF:usedAR=sizeAR,freeAR,-" );
    params [i++] = StrDup ( "CDEF:usedM=usedMR,sizeMR,/,100,*" );
    params [i++] = StrDup ( "CDEF:usedA=usedAR,sizeAR,/,100,*" );
    params [i++] = StrDup ( "CDEF:freeM=freeMR,sizeMR,/,100,*" );
    params [i++] = StrDup ( "CDEF:freeA=freeAR,sizeAR,/,100,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "usedM", C_USAGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "freeM", C_FREE_MAX,   pp, 1, MAX, TRUE  );
    params [i++] = CreatePlot ( PLOT_LINE2, "usedA", C_USAGE1_AVG, pp, 0, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "freeA", C_FREE_AVG,   pp, 1, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "usedA", "usedA", "usedM", "usedM", "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "freeA", "freeA", "freeM", "freeM", "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataReal ( fn, PLOT_6H_1M,   now );
    PlotDataSwap ( fn, PLOT_6H_1M,   now );
    PlotDataReal ( fn, PLOT_24H_5M,  now );
    PlotDataSwap ( fn, PLOT_24H_5M,  now );
    PlotDataReal ( fn, PLOT_7D_15M,  now );
    PlotDataSwap ( fn, PLOT_7D_15M,  now );
    PlotDataReal ( fn, PLOT_31D_60M, now );
    PlotDataSwap ( fn, PLOT_31D_60M, now );
    PlotDataReal ( fn, PLOT_366D_4H, now );
    PlotDataSwap ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int LevelToKB ( char* aWhat, int aLevel, char aUnit, int aSize )
{
    double retVal = 0.0;
    double level = (double) aLevel;

    switch ( aUnit )
    {
        case '1': retVal = level / 1024 ;           break;
        case 'k': retVal = level;                   break;
        case 'M': retVal = level * 1024;            break;
        case 'G': retVal = level * 1024 * 1024;     break;
        case '%': retVal = ( level * aSize ) / 100; break;
        default:
        {
            Error ( LOG_ERR, "An invalid level slipped through!" );
        }
        break;
    }
    return ( retVal );
}

int CheckData ()
{
    int retVal = STATE_OK;
    int tTotal;
    int tUsed;
    int tWarn;
    int tCrit;
    int rUsed;
    int sUsed;
    double tUsedP;
    double rUsedP;
    double sUsedP;


    ReadVals ();
    /* calculate everything in Kibibytes */
    tTotal = v.sTotal + v.rTotal;
    tUsed = tTotal - v.tAvail;
    tWarn = LevelToKB ( "warning level",  warnS, warnU, tTotal );
    tCrit = LevelToKB ( "critical level", critS, critU, tTotal );
    rUsed = v.rTotal - v.rAvail;
    sUsed = v.sTotal - v.sAvail;
    tUsedP = ( ( (double) tUsed ) * 100.0 ) / (double) tTotal;
    sUsedP = ( ( (double) sUsed ) * 100.0 ) / (double) v.sTotal;
    rUsedP = ( ( (double) rUsed ) * 100.0 ) / (double) v.rTotal;
    if ( tWarn < tUsed ) retVal = MAXIMUM ( retVal, STATE_WARNING );
    if ( tCrit < tUsed ) retVal = MAXIMUM ( retVal, STATE_CRITICAL );
    PrintExit ( retVal );
    printf ( " - available/total real: %i/%iKB (%3.1f%%), swap: %i/%iKB (%3.1f%%)",
            v.rAvail, v.rTotal, rUsedP, v.sAvail, v.sTotal, sUsedP );
    if ( npPerfData )
    {
        printf ( "|'swap space'=%iKB;;;0;%i",      v.sAvail, v.sTotal );
        printf ( " 'real memory'=%iKB;;;0;%i",     v.rAvail, v.rTotal );
        printf ( " 'free memory'=%iKB;%i;%i;0;%i",
                v.tAvail, tWarn, tCrit, tTotal );
        printf ( " 'shared memory'=%iKB",          v.shared );
        printf ( " 'buffered memory'=%iKB",        v.buffer );
        printf ( " 'cache memory'=%iKB\n",         v.cached );
    }
    else
    {
        printf ( "|%i;%i;%i;%i;%i;%i;%i;%i;%i\n",
                v.sTotal, v.sAvail, v.rTotal, v.rAvail,
                tTotal,   v.tAvail, v.shared, v.buffer, v.cached );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ();
        if ( plotPerfData ) PlotPerfData ();
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
