/*
 * File:      check_ups_battery.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int v1; /* found upsBatteryStatus */
    int v2; /* found upsSecondsOnBattery */
    int v3; /* found upsEstimatedMinutesRemaining */
    int v4; /* found upsEstimatedChargeRemaining */
    int v5; /* found upsBatteryVoltage */
    int v6; /* found upsBatteryCurrent */
    int v7; /* found upsBatteryTemperature */
} tScanRes;

/* data type for storing data */
typedef struct
{
    int      stat;  /* status of the battery */
    int      tbat;  /* time (seconds) on battery */
    int      trem;  /* estimated time (seconds!) remaining */
    int      crem;  /* charge remaining (percent) */
    double   volt;  /* battery voltage */
    double   curr;  /* battery current */
    int      temp;  /* battery temperature */
} tDataSet;

/* global input variables */

/*
 * SNMP base OID
 * UPS-MIB::upsBatteryStatus.0             = .1.3.6.1.2.1.33.1.2.1.0
 * UPS-MIB::upsSecondsOnBattery.0          = .1.3.6.1.2.1.33.1.2.2.0
 * UPS-MIB::upsEstimatedMinutesRemaining.0 = .1.3.6.1.2.1.33.1.2.3.0
 * UPS-MIB::upsEstimatedChargeRemaining.0  = .1.3.6.1.2.1.33.1.2.4.0
 * UPS-MIB::upsBatteryVoltage.0            = .1.3.6.1.2.1.33.1.2.5.0
 * UPS-MIB::upsBatteryCurrent.0            = .1.3.6.1.2.1.33.1.2.6.0
 * UPS-MIB::upsBatteryTemperature.0        = .1.3.6.1.2.1.33.1.2.7.0
 */
oid baseOID [] = { 1, 3, 6, 1, 2, 1, 33, 1, 2 };
size_t baseOIDLen = 9;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_ups_battery";
    progVersion = "0.2";
    helpIndexP = NULL;
    helpThresP = NULL;
    helpPurpose = "Checks the battery status of an UPS.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 5 * 60;
    rrdHBeat = 11 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = FALSE;
    hasAddInfParam = FALSE;
}

void HelpSpecificParams ()
{
    /* No specific parameters */
}

void HelpExplain ()
{
    printf (
    "The plugin returns the status of the battery/batteries of an UPS."
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<tbat>;<trem>;<crem>;<volt>;<curr>;<temp>\n"
    "where\n"
    "    tbat ... time (seconds) running on battery, 0 if not applicable.\n"
    "    trem ... estimated time (seconds) remaining.\n"
    "    crem ... estimated charge (percent) remaining.\n"
    "    volt ... battery voltage (V).\n"
    "    curr ... battery current (A).\n"
    "    temp ... battery temperature (deg C).\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'time on battery'=<tbat>s "
    "'estimated time remaining'=<trem>s "
    "'estimated charge remaining'=<crem>%% "
    "'battery voltage'=<volt> "
    "'battery current'=<curr> "
    "'battery temperature'=<temp>\n"
    "\n"
    );
    printf (
    "Note: SNMP agents return trem in minutes. Thus this value will change\n"
    "in multiples of 60.\n"
    "Note: Nagios Plugin performance data format does not allow Volt, Ampere\n"
    "and Degree Centigrade as units of measurement.\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    return FALSE;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 11 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            if ( aVar->name [10] == 0 )
            {
                switch ( aVar->name [9] )
                {
                    case 1: /* upsBatteryStatus */
                        scanResLen = 1;
                        scanRes = realloc ( scanRes,
                                sizeof ( tScanRes ) * scanResLen );
                        if ( scanRes == NULL ) ErrorMalloc ();
                        scanRes [0].v1 = ( aVar->type == ASN_INTEGER );
                        scanRes [0].v2 = FALSE;
                        scanRes [0].v3 = FALSE;
                        scanRes [0].v4 = FALSE;
                        scanRes [0].v5 = FALSE;
                        scanRes [0].v6 = FALSE;
                        scanRes [0].v7 = FALSE;
                    break;
                    case 2: /* upsSecondsOnBattery */
                        scanRes [0].v2 = ( aVar->type == ASN_INTEGER );
                    break;
                    case 3: /* upsEstimatedMinutesRemaining */
                        scanRes [0].v3 = ( aVar->type == ASN_INTEGER );
                    break;
                    case 4: /* upsEstimatedChargeRemaining */
                        scanRes [0].v4 = ( aVar->type == ASN_INTEGER );
                    break;
                    case 5: /* upsBatteryVoltage */
                        scanRes [0].v5 = ( aVar->type == ASN_INTEGER );
                    break;
                    case 6: /* upsBatteryCurrent */
                        scanRes [0].v6 = ( aVar->type == ASN_INTEGER );
                    break;
                    case 7: /* upsBatteryTemperature */
                        scanRes [0].v7 = ( aVar->type == ASN_INTEGER );
                    break;
                    default:
                        /* don't care */
                    break;
                }
            } else Log ( LOG_CRIT, "Unexpected index while scanning services" );
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v1 && scanRes [i].v2 && scanRes [i].v3 &&
                scanRes [i].v4 && scanRes [i].v5 && scanRes [i].v6 &&
                scanRes [i].v7 )
        {
            retVal = STATE_OK;
            printf ( "OK\n" );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_stat [15];
    oid o_tbat [15];
    oid o_trem [15];
    oid o_crem [15];
    oid o_volt [15];
    oid o_curr [15];
    oid o_temp [15];
    size_t oidLen;

    OidCpy ( o_stat, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_tbat, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_trem, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_crem, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_volt, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_curr, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_temp, &oidLen, baseOID, &baseOIDLen );
    o_stat [9] = 1; o_stat [10] = 0;
    o_tbat [9] = 2; o_tbat [10] = 0;
    o_trem [9] = 3; o_trem [10] = 0;
    o_crem [9] = 4; o_crem [10] = 0;
    o_volt [9] = 5; o_volt [10] = 0;
    o_curr [9] = 6; o_curr [10] = 0;
    o_temp [9] = 7; o_temp [10] = 0;
    oidLen = 11;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.stat  =                ReadInteger ( sess2, o_stat, oidLen );
    v.tbat  =                ReadInteger ( sess2, o_tbat, oidLen );
    v.trem  = 60 *           ReadInteger ( sess2, o_trem, oidLen );
    v.crem  =                ReadInteger ( sess2, o_crem, oidLen );
    v.volt  = 0.1 * (double) ReadInteger ( sess2, o_volt, oidLen );
    v.curr  = 0.1 * (double) ReadInteger ( sess2, o_curr, oidLen );
    v.temp  =                ReadInteger ( sess2, o_temp, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "tbat", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "trem", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "crem", GAUGE, rrdHBeat, "0", "101" );
    params [i++] = CreateDSStr ( "volt", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "curr", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "temp", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;

    chk = sscanf ( aData, "%i;%i;%i;%lf;%lf;%i",
            &(v.tbat), &(v.trem), &(v.crem), &(v.volt), &(v.curr), &(v.temp) );
    if ( chk != 6 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
    v.stat = -1;
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* tbat   = IntToStr    ( v.tbat );
    char* trem   = IntToStr    ( v.trem );
    char* crem   = IntToStr    ( v.crem );
    char* volt   = DoubleToStr ( v.volt );
    char* curr   = DoubleToStr ( v.curr );
    char* temp   = IntToStr    ( v.temp );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "tbat:trem:crem:volt:curr:temp" );
    params [i++] = StrConcat ( 13 , ts, ":", tbat, ":", trem, ":", crem, ":",
            volt, ":", curr, ":", temp );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( temp );
    free ( curr );
    free ( volt );
    free ( crem );
    free ( trem );
    free ( tbat );
    free ( ts );
}

void PlotDataTime ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[m]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "time", aTiming );
    PPAddPlot ( pp, "time on battery",     "m", FALSE );
    PPAddPlot ( pp, "est. time remaining", "m", TRUE );
    title = StrConcat ( 2, hostName, " times" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, TRUE );
    params [i++] = CreateDEF ( "tbatMR", aFN, "tbat", MAX );
    params [i++] = CreateDEF ( "tbatAR", aFN, "tbat", AVG );
    params [i++] = CreateDEF ( "tremMR", aFN, "trem", MAX );
    params [i++] = CreateDEF ( "tremAR", aFN, "trem", AVG );
    params [i++] = StrDup ( "CDEF:tbatM=tbatMR,60,/" );
    params [i++] = StrDup ( "CDEF:tbatA=tbatAR,60,/" );
    params [i++] = StrDup ( "CDEF:tremM=tremMR,60,/" );
    params [i++] = StrDup ( "CDEF:tremA=tremAR,60,/" );
    params [i++] = CreatePlot ( PLOT_AREA,  "tbatM", C_ERR1_MAX,   pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "tremM", C_GAUGE1_MAX, pp, 1, MAX, TRUE );
    params [i++] = CreatePlot ( PLOT_LINE2, "tbatA", C_ERR1_AVG,   pp, 0, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "tremA", C_GAUGE1_AVG, pp, 1, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "tbatA", "tbatA", "tbatM", "tbatM", "%4.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "tremA", "tremA", "tremM", "tremM", "%4.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataCRem ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[%]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "crem", aTiming );
    PPAddPlot ( pp, "est. charge remaining", "%", FALSE );
    title = StrConcat ( 2, hostName, " charge" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, TRUE );
    params [i++] = CreateDEF ( "cremM", aFN, "crem", MAX );
    params [i++] = CreateDEF ( "cremA", aFN, "crem", AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "cremM", C_GAUGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "cremA", C_GAUGE1_AVG, pp, 0, AVG, FALSE );
    i = PlotPrintParams ( i, params, pp, 1, "cremA", "cremA", "cremM", "cremM", "%4.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataVolt ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[V]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "volt", aTiming );
    PPAddPlot ( pp, "battery voltage", "V", FALSE );
    title = StrConcat ( 2, hostName, " battery voltage" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, TRUE );
    params [i++] = CreateDEF ( "voltM", aFN, "volt", MAX );
    params [i++] = CreateDEF ( "voltA", aFN, "volt", AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "voltM", C_GAUGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "voltA", C_GAUGE1_AVG, pp, 0, AVG, FALSE );
    i = PlotPrintParams ( i, params, pp, 1, "voltA", "voltA", "voltM", "voltM", "%4.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataCurr ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[A]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "curr", aTiming );
    PPAddPlot ( pp, "battery current", "A", FALSE );
    title = StrConcat ( 2, hostName, " battery current" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, TRUE );
    params [i++] = CreateDEF ( "currM", aFN, "curr", MAX );
    params [i++] = CreateDEF ( "currA", aFN, "curr", AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "currM", C_GAUGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "currA", C_GAUGE1_AVG, pp, 0, AVG, FALSE );
    i = PlotPrintParams ( i, params, pp, 1, "currA", "currA", "currM", "currM", "%4.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataTemp ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[deg C]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "temp", aTiming );
    PPAddPlot ( pp, "battery temperature", "deg C", FALSE );
    title = StrConcat ( 2, hostName, " battery temperature" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, TRUE );
    params [i++] = CreateDEF ( "tempM", aFN, "temp", MAX );
    params [i++] = CreateDEF ( "tempA", aFN, "temp", AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "tempM", C_GAUGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "tempA", C_GAUGE1_AVG, pp, 0, AVG, FALSE );
    i = PlotPrintParams ( i, params, pp, 1, "tempA", "tempA", "tempM", "tempM", "%4.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataTime ( fn, PLOT_24H_5M,  now );
    PlotDataCRem ( fn, PLOT_24H_5M,  now );
    PlotDataVolt ( fn, PLOT_24H_5M,  now );
    PlotDataCurr ( fn, PLOT_24H_5M,  now );
    PlotDataTemp ( fn, PLOT_24H_5M,  now );
    PlotDataTime ( fn, PLOT_7D_15M,  now );
    PlotDataCRem ( fn, PLOT_7D_15M,  now );
    PlotDataVolt ( fn, PLOT_7D_15M,  now );
    PlotDataCurr ( fn, PLOT_7D_15M,  now );
    PlotDataTemp ( fn, PLOT_7D_15M,  now );
    PlotDataTime ( fn, PLOT_31D_60M, now );
    PlotDataCRem ( fn, PLOT_31D_60M, now );
    PlotDataVolt ( fn, PLOT_31D_60M, now );
    PlotDataCurr ( fn, PLOT_31D_60M, now );
    PlotDataTemp ( fn, PLOT_31D_60M, now );
    PlotDataTime ( fn, PLOT_366D_4H, now );
    PlotDataCRem ( fn, PLOT_366D_4H, now );
    PlotDataVolt ( fn, PLOT_366D_4H, now );
    PlotDataCurr ( fn, PLOT_366D_4H, now );
    PlotDataTemp ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    int retVal = STATE_UNKNOWN;

    ReadVals ();
    switch ( v.stat )
    {
        case 1:
            retVal = STATE_UNKNOWN;
            printf ( "UNKNOWN - 'unknown'" );
        break;
        case 2:
            retVal = STATE_OK;
            printf ( "OK - 'batteryNormal'" );
        break;
        case 3:
            retVal = STATE_WARNING;
            printf ( "WARNING - 'batteryLow'" );
        break;
        case 4:
            retVal = STATE_CRITICAL;
            printf ( "CRITICAL - 'batteryDepleted'" );
        break;
        default:
            retVal = STATE_UNKNOWN;
            printf ( "UNKNOWN - status is invalid" );
        break;
    }
    printf ( ": %im on battery, %im remaining, %i%% remaining",
            v.tbat / 60, v.trem / 60, v.crem );
    if ( npPerfData )
    {
        printf ( "|'time on battery'=%is", v.tbat );
        printf ( " 'estimated time remaining'=%is", v.trem );
        printf ( " 'estimated charge remaining'=%i%%", v.crem );
        printf ( " 'battery voltage'=%.1f", v.volt );
        printf ( " 'battery current'=%.1f", v.curr );
        printf ( " 'battery temperature'=%i\n", v.temp );
    }
    else
    {
        printf ( "|%i;%i;%i;%.1f;%.1f;%i\n",
                v.tbat, v.trem, v.crem, v.volt, v.curr, v.temp );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ();
        if ( plotPerfData ) PlotPerfData ();
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return retVal;
}
