/*
 * File:      check_ups_bypass.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int index; /* bypass line index */
    int v2;    /* found upsBypassVoltage */
    int v3;    /* found upsBypassCurrent */
    int v4;    /* found upsBypassPower */
    int volt;  /* present bypass voltage */
} tScanRes;

/* data type for storing data */
typedef struct
{
    int      volt; /* present bypass voltage */
    double   curr; /* present bypass current */
    int      powr; /* present true bypass power */
} tDataSet;

/* global input variables */

/*
 * SNMP base OID
 * UPS-MIB::upsBypassLineIndex.IDX = .1.3.6.1.2.1.33.1.5.3.1.1.IDX
 * UPS-MIB::upsBypassVoltage.IDX   = .1.3.6.1.2.1.33.1.5.3.1.2.IDX
 * UPS-MIB::upsBypassCurrent.IDX   = .1.3.6.1.2.1.33.1.5.3.1.3.IDX
 * UPS-MIB::upsBypassPower.IDX     = .1.3.6.1.2.1.33.1.5.3.1.4.IDX
 */
oid baseOID [] = { 1, 3, 6, 1, 2, 1, 33, 1, 5, 3, 1 };
size_t baseOIDLen = 11;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_ups_bypass";
    progVersion = "0.2";
    helpIndexP = "-i INDEX";
    helpThresP = NULL;
    helpPurpose = "Checks the status of an UPS bypass line.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 5 * 60;
    rrdHBeat = 11 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = TRUE;
    hasAddInfParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-i, --index INTEGER\n"
    "    Index of bypass line.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "The plugin returns the status of an UPS bypass line. Is is made only\n"
    "for logging data and no warnings or criticals will be generated.\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<addInf>;<Ubyp>;<Ibyp>;<Pbyp>\n"
    "where\n"
    "    addInf ... additional information.\n"
    "    Ubyp   ... present bypass voltage (V).\n"
    "    Ibyp   ... present bypass current (A).\n"
    "    Pbyp   ... present bypass power (W).\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'bypass voltage'=<Ubyp> "
    "'bypass current'=<Ibyp> "
    "'bypass power'=<Pbyp>\n"
    "\n"
    );
    printf (
    "Note: Some UPSes do not report all of the values. Therefor performance\n"
    "data may seem inconsistent.\n"
    "Note: Nagios Plugin performance data format does not allow Volt, Ampere\n"
    "and Watt as units of measurement.\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    return ( FALSE );
}

int ScanResIndex ( int aIndex )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        found = ( scanRes [retVal].index == aIndex );
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 13 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [11] )
            {
                case 1: /* upsBypassLineIndex */
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    scanRes [index].index = aVar->name [12];
                    scanRes [index].v2 = FALSE;
                    scanRes [index].v3 = FALSE;
                    scanRes [index].v4 = FALSE;
                    scanRes [index].volt = -1;
                break;
                case 2: /* upsBypassVoltage */
                    index = ScanResIndex ( aVar->name [12] );
                    scanRes [index].v2 = ( aVar->type == ASN_INTEGER );
                    scanRes [index].volt = *(aVar->val.integer);
                break;
                case 3: /* upsBypassCurrent */
                    index = ScanResIndex ( aVar->name [12] );
                    scanRes [index].v3 = ( aVar->type == ASN_INTEGER );
                break;
                case 4: /* upsBypassPower */
                    index = ScanResIndex ( aVar->name [12] );
                    scanRes [index].v4 = ( aVar->type == ASN_INTEGER );
                break;
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if ( scanRes [i].v2 && scanRes [i].v3 && scanRes [i].v4 )
        {
            retVal = STATE_OK;
            printf ( "OK index: %i ; voltage: %i\n",
                    scanRes [i].index, scanRes [i].volt );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_volt [13];
    oid o_curr [13];
    oid o_powr [13];
    size_t oidLen;

    OidCpy ( o_volt, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_curr, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_powr, &oidLen, baseOID, &baseOIDLen );
    o_volt [11] = 2; o_volt [12] = idx;
    o_curr [11] = 3; o_curr [12] = idx;
    o_powr [11] = 4; o_powr [12] = idx;
    oidLen = 13;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.volt =                ReadInteger ( sess2, o_volt, oidLen );
    v.curr = 0.1 * (double) ReadInteger ( sess2, o_curr, oidLen );
    v.powr =                ReadInteger ( sess2, o_powr, oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "volt", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "curr", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "powr", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    char buf[1024];

    chk = sscanf ( aData, "%1023s;%i;%lf;%i", buf,
            &(v.volt), &(v.curr), &(v.powr) );
    if ( chk != 4 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
    addInf = StrDup ( buf );
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* volt   = IntToStr    ( v.volt );
    char* curr   = DoubleToStr ( v.curr );
    char* powr   = IntToStr    ( v.powr );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup ( "volt:curr:powr" );
    params [i++] = StrConcat ( 7, ts, ":", volt, ":", curr, ":", powr );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( powr );
    free ( curr );
    free ( volt );
    free ( ts );
}

void PlotDataVolt ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[V]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "volt", aTiming );
    PPAddPlot ( pp, "bypass voltage", "V", FALSE );
    title = StrConcat ( 2, hostName, " bypass voltage" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, TRUE );
    params [i++] = CreateDEF ( "voltM", aFN, "volt", MAX );
    params [i++] = CreateDEF ( "voltA", aFN, "volt", AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "voltM", C_GAUGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "voltA", C_GAUGE1_AVG, pp, 0, AVG, FALSE );
    i = PlotPrintParams ( i, params, pp, 1, "voltA", "voltA", "voltM", "voltM", "%4.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataCurr ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[A]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "curr", aTiming );
    PPAddPlot ( pp, "bypass current", "A", FALSE );
    title = StrConcat ( 2, hostName, " bypass current" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, TRUE );
    params [i++] = CreateDEF ( "currM", aFN, "curr", MAX );
    params [i++] = CreateDEF ( "currA", aFN, "curr", AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "currM", C_GAUGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "currA", C_GAUGE1_AVG, pp, 0, AVG, FALSE );
    i = PlotPrintParams ( i, params, pp, 1, "currA", "currA", "currM", "currM", "%4.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataPowr ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[W]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "powr", aTiming );
    PPAddPlot ( pp, "bypass power", "W", FALSE );
    title = StrConcat ( 2, hostName, " bypass power" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, TRUE );
    params [i++] = CreateDEF ( "powrM", aFN, "powr", MAX );
    params [i++] = CreateDEF ( "powrA", aFN, "powr", AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "powrM", C_GAUGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "powrA", C_GAUGE1_AVG, pp, 0, AVG, FALSE );
    i = PlotPrintParams ( i, params, pp, 1, "powrA", "powrA", "powrM", "powrM", "%4.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataVolt ( fn, PLOT_24H_5M,  now );
    PlotDataCurr ( fn, PLOT_24H_5M,  now );
    PlotDataPowr ( fn, PLOT_24H_5M,  now );
    PlotDataVolt ( fn, PLOT_7D_15M,  now );
    PlotDataCurr ( fn, PLOT_7D_15M,  now );
    PlotDataPowr ( fn, PLOT_7D_15M,  now );
    PlotDataVolt ( fn, PLOT_31D_60M, now );
    PlotDataCurr ( fn, PLOT_31D_60M, now );
    PlotDataPowr ( fn, PLOT_31D_60M, now );
    PlotDataVolt ( fn, PLOT_366D_4H, now );
    PlotDataCurr ( fn, PLOT_366D_4H, now );
    PlotDataPowr ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    int retVal = STATE_OK;

    ReadVals ();
    PrintExit ( retVal );
    printf ( " - bypass line %i: %iV, %.1fA, %iW",
        idx, v.volt, v.curr, v.powr );
    if ( npPerfData )
    {
        printf ( "|'bypass voltage'=%i", v.volt );
        printf ( " 'bypass current'=%.1f", v.curr );
        printf ( " 'bypass power'=%i\n", v.powr );
    }
    else
    {
        printf ( "|%s;%i;%.1f;%i\n", addInf, v.volt, v.curr, v.powr );
    }
    return ( STATE_OK );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ();
        if ( plotPerfData ) PlotPerfData ();
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return retVal;
}
