/*
 * File:      check_winf_cpuusage.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int  len;   /* length of the cpu name */
    oid* name;  /* cpu name, no trailing zero */
    int  v2;    /* found cpuPercentDPCTime */
    int  v3;    /* found cpuPercentInterruptTime */
    int  v4;    /* found cpuPercentPrivilegedTime */
    int  v5;    /* found cpuPercentProcessorTime */
    int  v6;    /* found cpuPercentUserTime */
    int  v10;   /* found cpuDPCsQueuedPerSec */
    int  v11;   /* found cpuInterruptsPerSec */
} tScanRes;

/* data type for storing data */
typedef struct
{
    uint pDPCT;  /* percentage of time receiving and servicing DPCs */
    uint pIntT;  /* percentage of time receiving and servicing interrups */
    uint pPrivT; /* percentage of non-idle time spent in privileged mode */
    uint pProcT; /* percentage of time not idle */
    uint pUserT; /* percentage of non-idle time spent in user mode */
    uint dpcQpS; /* DPCs queued per second during the sample interval */
    uint intpS;  /* Interrupts per second */
} tDataSet;


/* global input variables */
char* name;
int isNameSet = FALSE;


/*
 * SNMP base OID
 * INFORMANT-STD::cpuInstance.NAME
 * = .1.3.6.1.4.1.9600.1.1.5.1.1.NAME
 * INFORMANT-STD::cpuPercentDPCTime.NAME
 * = .1.3.6.1.4.1.9600.1.1.5.1.2.NAME
 * INFORMANT-STD::cpuPercentInterruptTime.NAME
 * = .1.3.6.1.4.1.9600.1.1.5.1.3.NAME
 * INFORMANT-STD::cpuPercentPrivilegedTime.NAME
 * = .1.3.6.1.4.1.9600.1.1.5.1.4.NAME
 * INFORMANT-STD::cpuPercentProcessorTime.NAME
 * = .1.3.6.1.4.1.9600.1.1.5.1.5.NAME
 * INFORMANT-STD::cpuPercentUserTime.NAME
 * = .1.3.6.1.4.1.9600.1.1.5.1.6.NAME
 * INFORMANT-STD::cpuDPCsQueuedPerSec.NAME
 * = .1.3.6.1.4.1.9600.1.1.5.1.10.NAME
 * INFORMANT-STD::cpuInterruptsPerSec.NAME
 * = .1.3.6.1.4.1.9600.1.1.5.1.11.NAME
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 9600, 1, 1, 5, 1 };
size_t baseOIDLen = 11;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_winf_cpuusage";
    progVersion = "0.2";
    helpIndexP = NULL;
    helpThresP = "-n NAME";
    helpPurpose = "Reports the cpu usage of a Windows(R) computer.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 1 * 60;
    rrdHBeat = 3 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = FALSE;
    hasAddInfParam = TRUE;
}

void HelpSpecificParams ()
{
    printf (
    "-n, --name STRING\n"
    "    Name of the cpu which should be checked.\n"
    "\n"
    );
}

void HelpExplain ()
{
    printf (
    "This plugin reports the cpu usage of a Windows (R) computer using SNMP\n"
    "Informant(TM).\n"
    "\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format:\n"
    "<addInf>;<pDPCT>;<pIntT>;<pPrivT>;<pProcT>;<pUserT>;<dpcQpS>;<intpS>\n"
    "where\n"
    "    addInf ... additional information.\n"
    "    pDPCT  ... percentage of time receiving and servicing DPCs.\n"
    "    pIntT  ... percentage of time receiving and servicing interrups.\n"
    "    pPrivT ... percentage of non-idle time spent in privileged mode.\n"
    "    pProcT ... percentage of time not idle.\n"
    );
    printf (
    "    pUserT ... percentage of non-idle time spent in user mode.\n"
    "    dpcQpS ... DPCs queued per second during the sample interval.\n"
    "    intpS  ... Interrupts per second.\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'DPCs receiving and servicing'=<pDPCT>%% "
    "'interrupts receiving and servicing'=<pIntT>%% "
    "'privileged mode non-idle time'=<pPrivT>%% "
    "'not idle'=<pProcT>%% "
    "'user mode non-idle time'=<pUserT>%% "
    "'DPCs queued per second'=<dpcQpS> "
    "'Interrupts per second'=<intpS>\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-n" )
            ||  !strcmp ( argv [(*i)], "--name" ) )
    {
        isNameSet = TRUE;
        (*i)++;
        if ( (*i) < argc ) name = StrDup ( argv [(*i)++] );
        else Usage ( "Missing 'name' parameter" );
    } else ret = FALSE;
    return ( ret );
}

/*
 * This MIB does not use ordinary indices. Instead the "name" of the cpu is
 * encoded into the OID. Makes things a bit more difficult ...
 */
int ScanResCheckName ( netsnmp_variable_list* aVar )
{
    int retVal = 0;
    int found = FALSE;

    while ( ( retVal < scanResLen ) && !found )
    {
        int vLen1 = aVar->name_length - 13;
        int vLen2 = aVar->name[12];

        if ( ( vLen1 == scanRes [retVal].len ) && ( vLen1 == vLen2 ) )
        {
            int i = 0;  /* index into scanRes [retVal].name */
            int j = 13; /* index into aVar */
            found = TRUE;
            while ( found && ( i < vLen1 ) )
            {
                found = ( scanRes [retVal].name[i++] == aVar->name[j++] );
            }
        } /* else that's quite normal */
        if ( !found ) retVal++;
    }
    if ( !found ) retVal = -1;
    return retVal;
}

void ScanResBuildName ( netsnmp_variable_list* aVar, int aIndex )
{
    int vLen1 = aVar->name_length - 13;
    int vLen2 = aVar->name[12];

    if ( vLen1 == vLen2 )
    {
        int i; /* index into scanRes [index].name */
        int j = 13; /* index into aVar */
        scanRes [aIndex].len = vLen1;
        scanRes [aIndex].name = malloc ( sizeof ( oid ) * vLen1 );
        if ( scanRes [aIndex].name == NULL ) ErrorMalloc ();
        for ( i = 0; i < vLen1; i++ )
        {
            scanRes [aIndex].name[i] = aVar->name[j++];
        }
    } else Log ( LOG_DEBUG, "Length did not match" );
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length >= 14 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            int index;

            switch ( aVar->name [11] )
            {
                case 1: /* cpmCPUTotalPhysicalIndex */
                {
                    index = scanResLen;
                    scanResLen++;
                    scanRes = realloc ( scanRes,
                            sizeof ( tScanRes ) * scanResLen );
                    if ( scanRes == NULL ) ErrorMalloc ();
                    ScanResBuildName ( aVar, index );
                    scanRes [index].v2  = FALSE;
                    scanRes [index].v3  = FALSE;
                    scanRes [index].v4  = FALSE;
                    scanRes [index].v5  = FALSE;
                    scanRes [index].v6  = FALSE;
                    scanRes [index].v10 = FALSE;
                    scanRes [index].v11 = FALSE;
                }
                break;
                case 2:  /* cpuPercentDPCTime */
                    index = ScanResCheckName ( aVar );
                    scanRes [index].v2 = ( aVar->type == ASN_GAUGE );
                case 3:  /* cpuPercentInterruptTime */
                    index = ScanResCheckName ( aVar );
                    scanRes [index].v3 = ( aVar->type == ASN_GAUGE );
                case 4:  /* cpuPercentPrivilegedTime */
                    index = ScanResCheckName ( aVar );
                    scanRes [index].v4 = ( aVar->type == ASN_GAUGE );
                case 5:  /* cpuPercentProcessorTime */
                    index = ScanResCheckName ( aVar );
                    scanRes [index].v5 = ( aVar->type == ASN_GAUGE );
                case 6:  /* cpuPercentUserTime */
                    index = ScanResCheckName ( aVar );
                    scanRes [index].v6 = ( aVar->type == ASN_GAUGE );
                case 10: /* cpuDPCsQueuedPerSec */
                    index = ScanResCheckName ( aVar );
                    scanRes [index].v10 = ( aVar->type == ASN_GAUGE );
                case 11: /* cpuInterruptsPerSec */
                    index = ScanResCheckName ( aVar );
                    scanRes [index].v11 = ( aVar->type == ASN_GAUGE );
                default:
                    /* don't care */
                break;
            }
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int j;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v2  && scanRes [i].v3  && scanRes [i].v4  &&
                scanRes [i].v5  && scanRes [i].v6  && scanRes [i].v10 &&
                scanRes [i].v11 )
        {
            retVal = STATE_OK;
            printf ( "OK name: " );
            for ( j = 0; j < scanRes [i].len; j++ )
            {
                printf ( "%c", (char) scanRes [i].name[j] );
            }
            printf ( "\n" );
        }
        free ( scanRes [i].name );
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

oid* PrepOid ( long aNum, size_t aNLen, size_t aOidLen )
{
    oid* ret;
    size_t jLen = 0; /* junk length */
    int i; /* index into name */
    int j = 13; /* index into aOid */

    ret = malloc ( sizeof ( oid ) * aOidLen );
    if ( ret == NULL ) ErrorMalloc ();
    OidCpy ( ret, &jLen, baseOID, &baseOIDLen );
    (ret)[11] = aNum;
    (ret)[12] = aNLen;
    for ( i = 0; i < aNLen; i++ ) (ret)[j++] = name [i];
    return ( ret );
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    size_t nLen = strlen ( name );
    size_t oidLen = 11 + 1 + 1 + nLen;
    oid* o_pDPCT  = PrepOid (  2, nLen, oidLen );
    oid* o_pIntT  = PrepOid (  3, nLen, oidLen );
    oid* o_pPrivT = PrepOid (  4, nLen, oidLen );
    oid* o_pProcT = PrepOid (  5, nLen, oidLen );
    oid* o_pUserT = PrepOid (  6, nLen, oidLen );
    oid* o_dpcQpS = PrepOid ( 10, nLen, oidLen );
    oid* o_intpS  = PrepOid ( 11, nLen, oidLen );

    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.pDPCT  = ReadGauge32 ( sess2, o_pDPCT,  oidLen );
    v.pIntT  = ReadGauge32 ( sess2, o_pIntT,  oidLen );
    v.pPrivT = ReadGauge32 ( sess2, o_pPrivT, oidLen );
    v.pProcT = ReadGauge32 ( sess2, o_pProcT, oidLen );
    v.pUserT = ReadGauge32 ( sess2, o_pUserT, oidLen );
    v.dpcQpS = ReadGauge32 ( sess2, o_dpcQpS, oidLen );
    v.intpS  = ReadGauge32 ( sess2, o_intpS,  oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
    free ( o_intpS );
    free ( o_dpcQpS );
    free ( o_pUserT );
    free ( o_pProcT );
    free ( o_pPrivT );
    free ( o_pIntT );
    free ( o_pDPCT );
}

#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "pDPCT",  GAUGE, rrdHBeat, "0", "100" );
    params [i++] = CreateDSStr ( "pIntT",  GAUGE, rrdHBeat, "0", "100" );
    params [i++] = CreateDSStr ( "pPrivT", GAUGE, rrdHBeat, "0", "100" );
    params [i++] = CreateDSStr ( "pProcT", GAUGE, rrdHBeat, "0", "100" );
    params [i++] = CreateDSStr ( "pUserT", GAUGE, rrdHBeat, "0", "100" );
    params [i++] = CreateDSStr ( "dpcQpS", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "intpS",  GAUGE, rrdHBeat, "0", "U" );
    params [i++] = RRA_MAX_7H_1M;
    params [i++] = RRA_AVG_7H_1M;
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;
    char buf[1024];

    chk = sscanf ( aData, "%1023s;%u;%u;%u;%u;%u;%u;%u", buf,
            &(v.pDPCT), &(v.pIntT), &(v.pPrivT), &(v.pProcT), &(v.pUserT),
            &(v.dpcQpS), &(v.intpS)
            );
    if ( chk != 10 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
    addInf = StrDup ( buf );
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* pDPCT  = UIntToStr   ( v.pDPCT );
    char* pIntT  = UIntToStr   ( v.pIntT );
    char* pPrivT = UIntToStr   ( v.pPrivT );
    char* pProcT = UIntToStr   ( v.pProcT );
    char* pUserT = UIntToStr   ( v.pUserT );
    char* dpcQpS = UIntToStr   ( v.dpcQpS );
    char* intpS  = UIntToStr   ( v.intpS );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup (
            "pDPCT:pIntT:pPrivT:pProcT:pUserT:dpcQpS:intpS" );
    params [i++] = StrConcat ( 15, ts, ":", pDPCT, ":", pIntT, ":", pPrivT,
            ":", pProcT, ":", pUserT, ":", dpcQpS, ":", intpS );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( intpS );
    free ( dpcQpS );
    free ( pUserT );
    free ( pProcT );
    free ( pPrivT );
    free ( pIntT );
    free ( pDPCT );
    free ( ts );
}

void PlotDataUsages ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[%]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "usg", aTiming );
    PPAddPlot ( pp, "time in user mode",       "%", FALSE );
    PPAddPlot ( pp, "time in privileged mode", "%", TRUE );
    title = StrConcat ( 4, hostName, " cpu ", addInf, " usages" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 100, FALSE );
    params [i++] = CreateDEF ( "pPrivTMa", aFN, "pPrivT", MAX );
    params [i++] = CreateDEF ( "pPrivTAv", aFN, "pPrivT", AVG );
    params [i++] = CreateDEF ( "pUserTMa", aFN, "pUserT", MAX );
    params [i++] = CreateDEF ( "pUserTAv", aFN, "pUserT", AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "pUserTMa", C_LOAD3_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "pPrivTMa", C_LOAD1_MAX, pp, 1, MAX, TRUE );
    params [i++] = CreatePlot ( PLOT_LINE2, "pUserTMa", C_LOAD3_AVG, pp, 0, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_STACK, "pPrivTAv", C_LOAD1_AVG, pp, 1, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "pUserTAv", "pUserTAv", "pUserTMa", "pUserTMa", "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "pPrivTAv", "pPrivTAv", "pPrivTMa", "pPrivTMa", "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataDpcIntTime ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[%]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "dit", aTiming );
    PPAddPlot ( pp, "time servicing DPCs",      "%", FALSE );
    PPAddPlot ( pp, "time servicing interrups", "%", FALSE );
    title = StrConcat ( 4, hostName, " cpu ", addInf, " servicing DPCs and interrups" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 100, FALSE );
    params [i++] = CreateDEF ( "pDPCTM", aFN, "pDPCT", MAX );
    params [i++] = CreateDEF ( "pDPCTA", aFN, "pDPCT", AVG );
    params [i++] = CreateDEF ( "pIntTM", aFN, "pIntT", MAX );
    params [i++] = CreateDEF ( "pIntTA", aFN, "pIntT", AVG );
    params [i++] = CreatePlot ( PLOT_LINE3, "pDPCTM", C_USAGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "pDPCTA", C_USAGE1_AVG, pp, 0, AVG, TRUE );
    params [i++] = CreatePlot ( PLOT_LINE3, "pIntTM", C_USAGE2_MAX, pp, 1, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "pIntTA", C_USAGE2_AVG, pp, 1, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "pDPCTA", "pDPCTA", "pDPCTM", "pDPCTM", "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "pIntTA", "pIntTA", "pIntTM", "pIntTM", "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataDpcIntCnt ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[1/s]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "die", aTiming );
    PPAddPlot ( pp, "DPCs queued", "1/s", FALSE );
    PPAddPlot ( pp, "Interrups",   "1/s", FALSE );
    title = StrConcat ( 4, hostName, " cpu ", addInf, " DPCs and interrups" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_NONE, 0, 0, FALSE );
    params [i++] = CreateDEF ( "dpcQpSM", aFN, "dpcQpS", MAX );
    params [i++] = CreateDEF ( "dpcQpSA", aFN, "dpcQpS", AVG );
    params [i++] = CreateDEF ( "intpSM",  aFN, "intpS",  MAX );
    params [i++] = CreateDEF ( "intpSA",  aFN, "intpS",  AVG );
    params [i++] = CreatePlot ( PLOT_LINE3, "dpcQpSM", C_USAGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "dpcQpSA", C_USAGE1_AVG, pp, 0, AVG, TRUE );
    params [i++] = CreatePlot ( PLOT_LINE3, "intpSM",  C_USAGE2_MAX, pp, 1, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "intpSA",  C_USAGE2_AVG, pp, 1, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "dpcQpSA", "dpcQpSA", "dpcQpSM", "dpcQpSM", "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "intpSA",  "intpSA",  "intpSM",  "intpSM",  "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataUsages     ( fn, PLOT_6H_1M,   now );
    PlotDataDpcIntTime ( fn, PLOT_6H_1M,   now );
    PlotDataDpcIntCnt  ( fn, PLOT_6H_1M,   now );
    PlotDataUsages     ( fn, PLOT_24H_5M,  now );
    PlotDataDpcIntTime ( fn, PLOT_24H_5M,  now );
    PlotDataDpcIntCnt  ( fn, PLOT_24H_5M,  now );
    PlotDataUsages     ( fn, PLOT_7D_15M,  now );
    PlotDataDpcIntTime ( fn, PLOT_7D_15M,  now );
    PlotDataDpcIntCnt  ( fn, PLOT_7D_15M,  now );
    PlotDataUsages     ( fn, PLOT_31D_60M, now );
    PlotDataDpcIntTime ( fn, PLOT_31D_60M, now );
    PlotDataDpcIntCnt  ( fn, PLOT_31D_60M, now );
    PlotDataUsages     ( fn, PLOT_366D_4H, now );
    PlotDataDpcIntTime ( fn, PLOT_366D_4H, now );
    PlotDataDpcIntCnt  ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    int retVal = STATE_OK;

    ReadVals ();
    PrintExit ( retVal );
    printf ( " - cpu %s usage: %u%%", addInf, v.pProcT );
    if ( npPerfData )
    {
        printf ( "|'DPCs receiving and servicing'=%u%%", v.pDPCT );
        printf ( " 'interrupts receiving and servicing'=%u%%", v.pIntT );
        printf ( " 'privileged mode non-idle time'=%u%%", v.pPrivT );
        printf ( " 'not idle'=%u%%", v.pProcT );
        printf ( " 'user mode non-idle time'=%u%%", v.pUserT );
        printf ( " 'DPCs queued per second'=%u", v.dpcQpS );
        printf ( " 'Interrupts per second'=%u\n", v.intpS );
    }
    else
    {
        printf ( "|%s;%u;%u;%u;%u;%u;%u;%u\n",
                addInf, v.pDPCT, v.pIntT, v.pPrivT, v.pProcT, v.pUserT,
                v.dpcQpS, v.intpS );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( checkData )
    {
        if ( !isNameSet )  Usage ( "Missing 'name' parameters" );
    }
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ( NULL );
        if ( plotPerfData ) PlotPerfData ( NULL );
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
