/*
 * File:      check_winf_mem.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

/* see also: http://www.microsoft.com/resources/documentation/windowsnt/4/workstation/reskit/en-us/04memory.mspx */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include "globals.h"
#include "nagiosif.h"
#if(WITH_RRD)
#include "rrdif.h"
#endif /* WITH_RRD */
#include "snmpif.h"
#include "syslogif.h"
#include "strutils.h"
#include "utilities.h"


/* data type for storing scan results */
typedef struct
{
    int v2;      /* found memoryAvailableKBytes */
    int v4;      /* found memoryCommittedBytes */
    int v5;      /* found memoryCacheBytes */
    int v6;      /* found memoryCacheBytesPeak */
    int v7;      /* found memoryPageFaultsPerSec */
    int v8;      /* found memoryPagesInputPerSec */
    int v9;      /* found memoryPagesOutputPerSec */
    int v10;     /* found memoryPagesPerSec */
    int v11;     /* found memoryPoolNonpagedBytes */
    int v12;     /* found memoryPoolPagedBytes */
    int v13;     /* found memoryPoolPagedResidentBytes */
    int v14;     /* found memorySystemCacheResidentBytes */
    int v15;     /* found memorySystemCodeResidentBytes */
    int v16;     /* found memorySystemCodeTotalBytes */
    int v17;     /* found memorySystemDriverResidentBytes */
    int v18;     /* found memorySystemDriverTotalBytes */
} tScanRes;

/* data type for storing data */
typedef struct
{
    uint   rAvail; /* physical memory available to processes */
    uint   commit; /* committed virtual memory, in bytes */
    uint   cached; /* cached bytes */
    uint   cachPk; /* maximum of cached bytes since boot time */
    uint   pFpS;   /* page faults per second */
    uint   pIps;   /* pages input per second */
    uint   pOps;   /* pages output per second */
    uint   pps;    /* pages in/out per second */
    uint   pNB;    /* pool nonpaged bytes */
    uint   pPB;    /* pool paged bytes */
    uint   pPRB;   /* pool paged resident bytes */
    uint   sCaRB;  /* system cache resident bytes */
    uint   sCoRB;  /* system code resident bytes */
    uint   sCoTB;  /* system code total bytes */
    uint   sDRB;   /* system driver resident bytes */
    uint   sDTB;   /* system driver total bytes */
} tDataSet;


/* global input variables */
uint  warn = 5;   /* pages per second */
uint  crit = 10;  /* pages per second */

/*
 * SNMP base OID
 * INFORMANT-STD::memoryAvailableBytes.0
 * = 1.3.6.1.4.1.9600.1.1.2.1.0
 * INFORMANT-STD::memoryAvailableKBytes.0
 * = 1.3.6.1.4.1.9600.1.1.2.2.0
 * INFORMANT-STD::memoryAvailableMBytes.0
 * = 1.3.6.1.4.1.9600.1.1.2.3.0
 * INFORMANT-STD::memoryCommittedBytes.0
 * = 1.3.6.1.4.1.9600.1.1.2.4.0
 * INFORMANT-STD::memoryCacheBytes.0
 * = 1.3.6.1.4.1.9600.1.1.2.5.0
 * INFORMANT-STD::memoryCacheBytesPeak.0
 * = 1.3.6.1.4.1.9600.1.1.2.6.0
 * INFORMANT-STD::memoryPageFaultsPerSec.0
 * = 1.3.6.1.4.1.9600.1.1.2.7.0
 * INFORMANT-STD::memoryPagesInputPerSec.0
 * = 1.3.6.1.4.1.9600.1.1.2.8.0
 * INFORMANT-STD::memoryPagesOutputPerSec.0
 * = 1.3.6.1.4.1.9600.1.1.2.9.0
 * INFORMANT-STD::memoryPagesPerSec.0
 * = 1.3.6.1.4.1.9600.1.1.2.10.0
 * INFORMANT-STD::memoryPoolNonpagedBytes.0
 * = 1.3.6.1.4.1.9600.1.1.2.11.0
 * INFORMANT-STD::memoryPoolPagedBytes.0
 * = 1.3.6.1.4.1.9600.1.1.2.12.0
 * INFORMANT-STD::memoryPoolPagedResidentBytes.0
 * = 1.3.6.1.4.1.9600.1.1.2.13.0
 * INFORMANT-STD::memorySystemCacheResidentBytes.0
 * = 1.3.6.1.4.1.9600.1.1.2.14.0
 * INFORMANT-STD::memorySystemCodeResidentBytes.0
 * = 1.3.6.1.4.1.9600.1.1.2.15.0
 * INFORMANT-STD::memorySystemCodeTotalBytes.0
 * = 1.3.6.1.4.1.9600.1.1.2.16.0
 * INFORMANT-STD::memorySystemDriverResidentBytes.0
 * = 1.3.6.1.4.1.9600.1.1.2.17.0
 * INFORMANT-STD::memorySystemDriverTotalBytes.0
 * = 1.3.6.1.4.1.9600.1.1.2.18.0
 */
oid baseOID[] = { 1, 3, 6, 1, 4, 1, 9600, 1, 1, 2 };
size_t baseOIDLen = 10;

/* scan results */
tScanRes* scanRes = NULL;
int scanResLen = 0;

/* results of the SNMP queries */
tDataSet v;


void SetGlobals ()
{
    /* progName is set by ReadArgs() */
    srvName = "check_winf_mem";
    progVersion = "0.1";
    helpIndexP = NULL;
    helpThresP = "[-w WARN] [-c CRIT]";
    helpPurpose = "Checks the memory usage of a Windows(R) computer.";
    /* community is set by ReadArgs() */
    /* commFile is set by ReadArgs() */
#if(WITH_RRD)
    rrdStep = 1 * 60;
    rrdHBeat = 3 * 60;
#endif /* WITH_RRD */
    supportsPerfLog = TRUE;
    hasIndexParam = FALSE;
    hasAddInfParam = FALSE;
}

void HelpSpecificParams ()
{
    printf (
    "-w, --warning UINT\n"
    "    Pages in/out per second at which a warning is generated.\n"
    "    Default is %u.\n", warn
    );
    printf (
    "\n"
    "-c, --critical UINT\n"
    "    Pages in/out per second at which a critical is generated.\n"
    "    Default is %u.\n", crit
    );
}

void HelpExplain ()
{
    printf (
    "The plugin checks if paging exceeds certain thresholds. It also reports\n"
    "different types of memory usages.\n"
    );
}

void HelpPerformanceData ()
{
    printf (
    "Native performance data is provided in the format (all in one line):\n"
    "<rAvail>;<commit>;<cached>;<cachPk>;<pFpS>;<pIps>;<pOps>;<pps>;\n"
    "<pNB>;<pPB>;<pPRB>;<sCaRB>;<sCoRB>;<sCoTB>;<sDRB>;<sDTB>\n"
    "where\n"
    "    rAvail ... physical memory available to processes, in kBytes.\n"
    "    commit ... committed virtual memory, in bytes.\n"
    "    cached ... cached bytes.\n"
    "    cachPk ... maximum of cached bytes since boot time.\n"
    "    pFpS   ... page faults per second.\n"
    "    pIps   ... pages input per second.\n"
    "    pOps   ... pages output per second.\n"
    "    pps    ... pages in/out per second.\n"
    "    pNB    ... pool nonpaged bytes.\n"
    "    pPB    ... pool paged bytes.\n"
    "    pPRB   ... pool paged resident bytes.\n"
    "    sCaRB  ... system cache resident bytes.\n"
    "    sCoRB  ... system code resident bytes.\n"
    "    sCoTB  ... system code total bytes.\n"
    "    sDRB   ... system driver resident bytes.\n"
    "    sDTB   ... system driver total bytes.\n"
    "\n"
    );
    printf (
    "Nagios Plugins compatible performance data is provided in the format:\n"
    "'physical memory available'=<rAvail>KB "
    "'committed virtual memory'=<commit>B "
    "'cached'=<cached>B "
    "'maximum cached since boot time'=<cachPk>B "
    "'page faults per second.'=<pFpS> "
    "'pages input per second.'=<pIps> "
    "'pages output per second.'=<pOps> "
    "'pages in/out per second.'=<pps>;<warn>;<crit> "
    "'pool nonpaged'=<pNB>B "
    "'pool paged'=<pPB>B "
    "'pool paged resident'=<pPRB>B "
    "'system cache resident'=<sCaRB>B "
    "'system code resident'=<sCoRB>B "
    "'system code total'=<sCoTB>B "
    "'system driver resident'=<sDRB>B "
    "'system driver total'=<sDTB>B\n"
    "\n"
    );
}

int ChkOptSpecificParams ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-w" )
            ||  !strcmp ( argv [(*i)], "--warning" ) )
    {
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            warn = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || (*tail) ) Usage ( "Invalid 'warn' parameter" );
        } else Usage ( "Missing 'warn' parameter(s)" );
    } else if ( !strcmp ( argv [(*i)], "-c" )
            ||  !strcmp ( argv [(*i)], "--critical" ) )
    {
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            crit = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || (*tail) ) Usage ( "Invalid 'crit' parameter" );
        } else Usage ( "Missing 'crit' parameter(s)" );
    } else ret = FALSE;
    return ( ret );
}

void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar )
{
    if ( aVar->name_length == 12 )
    {
        if ( OidCmp ( aVar->name, baseOID, baseOIDLen ) == 0 )
        {
            if ( aVar->name [11] == 0 )
            {
                switch ( aVar->name [10] )
                {
                    case 2: /* memoryAvailableKBytes */
                        scanResLen = 1;
                        scanRes = realloc ( scanRes,
                                sizeof ( tScanRes ) * scanResLen );
                        if ( scanRes == NULL ) ErrorMalloc ();
                        scanRes [0].v2 = ( aVar->type == ASN_GAUGE );
                        scanRes [0].v4 = FALSE;
                        scanRes [0].v5 = FALSE;
                        scanRes [0].v6 = FALSE;
                        scanRes [0].v7 = FALSE;
                        scanRes [0].v8 = FALSE;
                        scanRes [0].v9 = FALSE;
                        scanRes [0].v10 = FALSE;
                        scanRes [0].v11 = FALSE;
                        scanRes [0].v12 = FALSE;
                        scanRes [0].v13 = FALSE;
                        scanRes [0].v14 = FALSE;
                        scanRes [0].v15 = FALSE;
                        scanRes [0].v16 = FALSE;
                        scanRes [0].v17 = FALSE;
                        scanRes [0].v18 = FALSE;
                    break;
                    case 4: /* memoryCommittedBytes */
                        scanRes [0].v4 = ( aVar->type == ASN_GAUGE );
                    break;
                    case 5: /* memoryCacheBytes */
                        scanRes [0].v5 = ( aVar->type == ASN_GAUGE );
                    break;
                    case 6: /* memoryCacheBytesPeak */
                        scanRes [0].v6 = ( aVar->type == ASN_GAUGE );
                    break;
                    case 7: /* memoryPageFaultsPerSec */
                        scanRes [0].v7 = ( aVar->type == ASN_GAUGE );
                    break;
                    case 8: /* memoryPagesInputPerSec */
                        scanRes [0].v8 = ( aVar->type == ASN_GAUGE );
                    break;
                    case 9: /* memoryPagesOutputPerSec */
                        scanRes [0].v9 = ( aVar->type == ASN_GAUGE );
                    break;
                    case 10: /* memoryPagesPerSec */
                        scanRes [0].v10 = ( aVar->type == ASN_GAUGE );
                    break;
                    case 11: /* memoryPoolNonpagedBytes */
                        scanRes [0].v11 = ( aVar->type == ASN_GAUGE );
                    break;
                    case 12: /* memoryPoolPagedBytes */
                        scanRes [0].v12 = ( aVar->type == ASN_GAUGE );
                    break;
                    case 13: /* memoryPoolPagedResidentBytes */
                        scanRes [0].v13 = ( aVar->type == ASN_GAUGE );
                    break;
                    case 14: /* memorySystemCacheResidentBytes */
                        scanRes [0].v14 = ( aVar->type == ASN_GAUGE );
                    break;
                    case 15: /* memorySystemCodeResidentBytes */
                        scanRes [0].v15 = ( aVar->type == ASN_GAUGE );
                    break;
                    case 16: /* memorySystemCodeTotalBytes */
                        scanRes [0].v16 = ( aVar->type == ASN_GAUGE );
                    break;
                    case 17: /* memorySystemDriverResidentBytes */
                        scanRes [0].v17 = ( aVar->type == ASN_GAUGE );
                    break;
                    case 18: /* memorySystemDriverTotalBytes */
                        scanRes [0].v18 = ( aVar->type == ASN_GAUGE );
                    break;
                    default:
                        /* don't care */
                    break;
                }
            }
            else Log ( LOG_CRIT, "Unexpected index (!= 0) while scanning" );
        } else Log ( LOG_CRIT, "Unexpected OID while scanning services" );
    } else Log ( LOG_CRIT, "Wrong length while scanning services" );
}

int ScanServices ()
{
    int i;
    int retVal = STATE_UNKNOWN;

    SnmpWalk ( baseOID, baseOIDLen );
    for ( i = 0; i < scanResLen; i++ )
    {
        if (    scanRes [i].v2  && scanRes [i].v4  && scanRes [i].v5  &&
                scanRes [i].v6  && scanRes [i].v7  && scanRes [i].v8  &&
                scanRes [i].v9  && scanRes [i].v10 && scanRes [i].v11 &&
                scanRes [i].v12 && scanRes [i].v13 && scanRes [i].v14 &&
                scanRes [i].v15 && scanRes [i].v16 && scanRes [i].v17 &&
                scanRes [i].v18 )
        {
            retVal = STATE_OK;
            printf ( "OK\n" );
        }
    }
    if ( retVal == STATE_UNKNOWN ) printf ( "FAILED\n" );
    return retVal;
}

void ReadVals ()
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    oid o_rAvail [12];
    oid o_commit [12];
    oid o_cached [12];
    oid o_cachPk [12];

    oid o_pFpS   [12];
    oid o_pIps   [12];
    oid o_pOps   [12];
    oid o_pps    [12];

    oid o_pNB    [12];
    oid o_pPB    [12];
    oid o_pPRB   [12];
    oid o_sCaRB  [12];

    oid o_sCoRB  [12];
    oid o_sCoTB  [12];
    oid o_sDRB   [12];
    oid o_sDTB   [12];
    size_t oidLen;

    OidCpy ( o_rAvail, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_commit, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_cached, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_cachPk, &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_pFpS,   &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_pIps,   &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_pOps,   &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_pps,    &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_pNB,    &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_pPB,    &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_pPRB,   &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_sCaRB,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_sCoRB,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_sCoTB,  &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_sDRB,   &oidLen, baseOID, &baseOIDLen );
    OidCpy ( o_sDTB,   &oidLen, baseOID, &baseOIDLen );
    o_rAvail [10] =  2; o_rAvail [11] = 0;
    o_commit [10] =  4; o_commit [11] = 0;
    o_cached [10] =  5; o_cached [11] = 0;
    o_cachPk [10] =  6; o_cachPk [11] = 0;
    o_pFpS   [10] =  7; o_pFpS   [11] = 0;
    o_pIps   [10] =  8; o_pIps   [11] = 0;
    o_pOps   [10] =  9; o_pOps   [11] = 0;
    o_pps    [10] = 10; o_pps    [11] = 0;
    o_pNB    [10] = 11; o_pNB    [11] = 0;
    o_pPB    [10] = 12; o_pPB    [11] = 0;
    o_pPRB   [10] = 13; o_pPRB   [11] = 0;
    o_sCaRB  [10] = 14; o_sCaRB  [11] = 0;
    o_sCoRB  [10] = 15; o_sCoRB  [11] = 0;
    o_sCoTB  [10] = 16; o_sCoTB  [11] = 0;
    o_sDRB   [10] = 17; o_sDRB   [11] = 0;
    o_sDTB   [10] = 18; o_sDTB   [11] = 0;
    oidLen = 12;
    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    v.rAvail = ReadGauge32 ( sess2, o_rAvail, oidLen );
    v.commit = ReadGauge32 ( sess2, o_commit, oidLen );
    v.cached = ReadGauge32 ( sess2, o_cached, oidLen );
    v.cachPk = ReadGauge32 ( sess2, o_cachPk, oidLen );
    v.pFpS   = ReadGauge32 ( sess2, o_pFpS,   oidLen );
    v.pIps   = ReadGauge32 ( sess2, o_pIps,   oidLen );
    v.pOps   = ReadGauge32 ( sess2, o_pOps,   oidLen );
    v.pps    = ReadGauge32 ( sess2, o_pps,    oidLen );
    v.pNB    = ReadGauge32 ( sess2, o_pNB,    oidLen );
    v.pPB    = ReadGauge32 ( sess2, o_pPB,    oidLen );
    v.pPRB   = ReadGauge32 ( sess2, o_pPRB,   oidLen );
    v.sCaRB  = ReadGauge32 ( sess2, o_sCaRB,  oidLen );
    v.sCoRB  = ReadGauge32 ( sess2, o_sCoRB,  oidLen );
    v.sCoTB  = ReadGauge32 ( sess2, o_sCoTB,  oidLen );
    v.sDRB   = ReadGauge32 ( sess2, o_sDRB,   oidLen );
    v.sDTB   = ReadGauge32 ( sess2, o_sDTB,   oidLen );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}


#if(WITH_RRD)
void CreateDataFile ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-s" );
    params [i++] = IntToStr ( rrdStep );
    params [i++] = CreateDSStr ( "rAvail", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "commit", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "cached", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "cachPk", GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "pFpS",   GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "pIps",   GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "pOps",   GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "pps",    GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "pNB",    GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "pPB",    GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "pPRB",   GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "sCaRB",  GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "sCoRB",  GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "sCoTB",  GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "sDRB",   GAUGE, rrdHBeat, "0", "U" );
    params [i++] = CreateDSStr ( "sDTB",   GAUGE, rrdHBeat, "0", "U" );
    params [i++] = RRA_MAX_25H_5M;
    params [i++] = RRA_AVG_25H_5M;
    params [i++] = RRA_MAX_8D_15M;
    params [i++] = RRA_AVG_8D_15M;
    params [i++] = RRA_MAX_32D_1H;
    params [i++] = RRA_AVG_32D_1H;
    params [i++] = RRA_MAX_367D_4H;
    params [i++] = RRA_AVG_367D_4H;
    RRDCreate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
}

void ParseData ( const char* aData )
{
    int chk;

    chk = sscanf ( aData, "%u;%u;%u;%u;%u;%u;%u;%u;%u;%u;%u;%u;%u;%u;%u;%u",
            &(v.rAvail), &(v.commit), &(v.cached), &(v.cachPk),
            &(v.pFpS),   &(v.pIps),   &(v.pOps),   &(v.pps),
            &(v.pNB),    &(v.pPB),    &(v.pPRB),   &(v.sCaRB),
            &(v.sCoRB),  &(v.sCoTB),  &(v.sDRB),   &(v.sDTB) );
    if ( chk != 16 )
    {
        char msg[1024];
        snprintf ( msg, 1024,
                "Could not parse performance data: %s.\n", aData );
        Error ( LOG_ERR, msg );
    }
}

void WriteData ( const char* aRRDFPath )
{
    int i = 0;
    char* params[1024]; /* Now that should be enough! */
    char* ts     = UIntToStr   ( GetTimestamp () );
    char* rAvail = UIntToStr   ( v.rAvail );
    char* commit = UIntToStr   ( v.commit );
    char* cached = UIntToStr   ( v.cached );
    char* cachPk = UIntToStr   ( v.cachPk );
    char* pFpS   = UIntToStr   ( v.pFpS );
    char* pIps   = UIntToStr   ( v.pIps );
    char* pOps   = UIntToStr   ( v.pOps );
    char* pps    = UIntToStr   ( v.pps );
    char* pNB    = UIntToStr   ( v.pNB );
    char* pPB    = UIntToStr   ( v.pPB );
    char* pPRB   = UIntToStr   ( v.pPRB );
    char* sCaRB  = UIntToStr   ( v.sCaRB );
    char* sCoRB  = UIntToStr   ( v.sCoRB );
    char* sCoTB  = UIntToStr   ( v.sCoTB );
    char* sDRB   = UIntToStr   ( v.sDRB );
    char* sDTB   = UIntToStr   ( v.sDTB );

    params [i++] = StrDup ( "JUNK" );
    params [i++] = StrDup ( aRRDFPath );
    params [i++] = StrDup ( "-t" );
    params [i++] = StrDup (
            "rAvail:commit:cached:cachPk:"
            "pFpS:pIps:pOps:pps:"
            "pNB:pPB:pPRB:sCaRB:"
            "sCoRB:sCoTB:sDRB:sDTB" );
    params [i++] = StrConcat ( 33, ts, ":",
            rAvail, ":", commit, ":", cached, ":", cachPk, ":",
            pFpS,   ":", pIps,   ":", pOps,   ":", pps,    ":",
            pNB,    ":", pPB,    ":", pPRB,   ":", sCaRB,  ":",
            sCoRB,  ":", sCoTB,  ":", sDRB,   ":", sDTB    );
    RRDUpdate ( i, params, aRRDFPath );
    while ( i > 0 ) free ( params [--i] );
    free ( sDTB );
    free ( sDRB );
    free ( sCoTB );
    free ( sCoRB );
    free ( sCaRB );
    free ( pPRB );
    free ( pPB );
    free ( pNB );
    free ( pps );
    free ( pOps );
    free ( pIps );
    free ( pFpS );
    free ( cachPk );
    free ( cached );
    free ( commit );
    free ( rAvail );
    free ( ts );
}

void PlotDataPageFs ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[1/s]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "fs", aTiming );
    PPAddPlot ( pp, "page faults",  "[1/s]", FALSE );
    PPAddPlot ( pp, "pages in/out", "[1/s]", FALSE );
    title = StrConcat ( 2, hostName, " pages faults" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_AUTO, 0, 0, FALSE );
    params [i++] = CreateDEF ( "pFpSM", aFN, "pFpS", MAX );
    params [i++] = CreateDEF ( "pFpSA", aFN, "pFpS", AVG );
    params [i++] = CreateDEF ( "ppsM",  aFN, "pps",  MAX );
    params [i++] = CreateDEF ( "ppsA",  aFN, "pps",  AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "pFpSM", C_ERR1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "pFpSA", C_ERR1_AVG, pp, 0, AVG, TRUE );
    params [i++] = CreatePlot ( PLOT_LINE3, "ppsM",  C_THRU_MAX, pp, 1, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "ppsA",  C_THRU_AVG, pp, 1, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "pFpSA", "pFpSA", "pFpSM", "pFpSM", "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "ppsA",  "ppsA",  "ppsM",  "ppsM",  "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataPageIO ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[1/s]  <-- out | in -->";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "io", aTiming );
    PPAddPlot ( pp, "pages input",  "[1/s]", FALSE );
    PPAddPlot ( pp, "pages output", "[1/s]", FALSE );
    title = StrConcat ( 2, hostName, " pages in and out" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1000, PLOT_SI_AUTO, 0, 0, TRUE );
    params [i++] = CreateDEF ( "pIpsM",  aFN, "pIps",  MAX );
    params [i++] = CreateDEF ( "pIpsA",  aFN, "pIps",  AVG );
    params [i++] = CreateDEF ( "pOpsMR", aFN, "pOps",  MAX );
    params [i++] = CreateDEF ( "pOpsAR", aFN, "pOps",  AVG );
    params [i++] = StrDup ( "CDEF:pOpsM=pOpsMR,-1,*" );
    params [i++] = StrDup ( "CDEF:pOpsA=pOpsAR,-1,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "pIpsM",  C_THRU_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "pIpsA",  C_THRU_AVG, pp, 0, AVG, TRUE );
    params [i++] = CreatePlot ( PLOT_AREA,  "pOpsMR", C_THRU_MAX, pp, 1, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "pOpsAR", C_THRU_AVG, pp, 1, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "pIpsA",  "pIpsA",  "pIpsM",  "pIpsM",  "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 1, "pOpsAR", "pOpsAR", "pOpsMR", "pOpsMR", "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataPageRAvail ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[B]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "ravail", aTiming );
    PPAddPlot ( pp, "available memory",  "[B]", FALSE );
    title = StrConcat ( 2, hostName, " available memory" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1024, PLOT_SI_AUTO, 0, 0, FALSE );
    params [i++] = CreateDEF ( "rAvailMR", aFN, "rAvail",  MAX );
    params [i++] = CreateDEF ( "rAvailAR", aFN, "rAvail",  AVG );
    params [i++] = StrDup ( "CDEF:rAvailM=rAvailMR,1024,*" );
    params [i++] = StrDup ( "CDEF:rAvailA=rAvailAR,1024,*" );
    params [i++] = CreatePlot ( PLOT_AREA,  "rAvailM",  C_FREE_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "rAvailA",  C_FREE_AVG, pp, 0, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "rAvailA", "rAvailA", "rAvailM", "rAvailM",  "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataCommit ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[B]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "commit", aTiming );
    PPAddPlot ( pp, "committed memory",  "[B]", FALSE );
    title = StrConcat ( 2, hostName, " committed memory" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1024, PLOT_SI_AUTO, 0, 0, FALSE );
    params [i++] = CreateDEF ( "commitM", aFN, "commit",  MAX );
    params [i++] = CreateDEF ( "commitA", aFN, "commit",  AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "commitM",  C_FREE_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "commitA",  C_FREE_AVG, pp, 0, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "commitA", "commitA", "commitM", "commitM",  "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataCache ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[B]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "cache", aTiming );
    PPAddPlot ( pp, "cached",  "[B]", FALSE );
    title = StrConcat ( 2, hostName, " cached" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1024, PLOT_SI_AUTO, 0, 0, FALSE );
    params [i++] = CreateDEF ( "cachedM", aFN, "cached",  MAX );
    params [i++] = CreateDEF ( "cachedA", aFN, "cached",  AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "cachedM",  C_FREE_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "cachedA",  C_FREE_AVG, pp, 0, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "cachedA", "cachedA", "cachedM", "cachedM",  "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataNPB ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[B]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "npb", aTiming );
    PPAddPlot ( pp, "pool nonpaged",  "[B]", FALSE );
    title = StrConcat ( 2, hostName, " pool nonpaged" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1024, PLOT_SI_AUTO, 0, 0, FALSE );
    params [i++] = CreateDEF ( "pNBM", aFN, "pNB",  MAX );
    params [i++] = CreateDEF ( "pNBA", aFN, "pNB",  AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "pNBM",  C_LOAD1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "pNBA",  C_LOAD1_AVG, pp, 0, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "pNBA", "pNBA", "pNBM", "pNBM",  "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataPPB ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[B]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "ppb", aTiming );
    PPAddPlot ( pp, "pool paged",          "[B]", FALSE );
    PPAddPlot ( pp, "pool paged resident", "[B]", FALSE );
    title = StrConcat ( 2, hostName, " pool paged" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1024, PLOT_SI_AUTO, 0, 0, FALSE );
    params [i++] = CreateDEF ( "pPBM",  aFN, "pPB", MAX );
    params [i++] = CreateDEF ( "pPBA",  aFN, "pPB", AVG );
    params [i++] = CreateDEF ( "pPRBM", aFN, "pPB", MAX );
    params [i++] = CreateDEF ( "pPRBA", aFN, "pPB", AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "pPBM",  C_GAUGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_AREA,  "pPRBM", C_FREE_MAX,   pp, 1, MAX, TRUE );
    params [i++] = CreatePlot ( PLOT_LINE2, "pPBA",  C_GAUGE1_AVG, pp, 0, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "pPRBA", C_FREE_AVG,   pp, 1, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "pPBA",  "pPBA",  "pPBM",  "pPBM",  "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 0, "pPRBA", "pPRBA", "pPRBM", "pPRBM", "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataCode ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[B]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "code", aTiming );
    PPAddPlot ( pp, "system code total",    "[B]", FALSE );
    PPAddPlot ( pp, "system code resident", "[B]", FALSE );
    title = StrConcat ( 2, hostName, " system code" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1024, PLOT_SI_AUTO, 0, 0, FALSE );
    params [i++] = CreateDEF ( "sCoTBM", aFN, "sCoTB", MAX );
    params [i++] = CreateDEF ( "sCoTBA", aFN, "sCoTB", AVG );
    params [i++] = CreateDEF ( "sCoRBM", aFN, "sCoRB", MAX );
    params [i++] = CreateDEF ( "sCoRBA", aFN, "sCoRB", AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "sCoTBM", C_GAUGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_AREA,  "sCoRBM", C_FREE_MAX,   pp, 1, MAX, TRUE );
    params [i++] = CreatePlot ( PLOT_LINE2, "sCoTBA", C_GAUGE1_AVG, pp, 0, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "sCoRBA", C_FREE_AVG,   pp, 1, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "sCoTBA", "sCoTBA", "sCoTBM", "sCoTBM", "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 0, "sCoRBA", "sCoRBA", "sCoRBM", "sCoRBM", "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotDataDriver ( tFN* aFN, int aTiming, int aNow )
{
    tPP* pp = PPInit ();
    int i = 0;
    char* title = NULL;
    char* vLabel = "[B]";
    char* params[1024]; /* Now that should be enough! */

    FNSetPNGData ( aFN, "drv", aTiming );
    PPAddPlot ( pp, "device drives total",    "[B]", FALSE );
    PPAddPlot ( pp, "device drives resident", "[B]", FALSE );
    title = StrConcat ( 2, hostName, " device drives" );
    i = PlotStdParams ( params, aFN, title, vLabel,
            aTiming, aNow - rrdHBeat, 1024, PLOT_SI_AUTO, 0, 0, FALSE );
    params [i++] = CreateDEF ( "sDTBM", aFN, "sDTB", MAX );
    params [i++] = CreateDEF ( "sDTBA", aFN, "sDTB", AVG );
    params [i++] = CreateDEF ( "sDRBM", aFN, "sDRB", MAX );
    params [i++] = CreateDEF ( "sDRBA", aFN, "sDRB", AVG );
    params [i++] = CreatePlot ( PLOT_AREA,  "sDTBM", C_GAUGE1_MAX, pp, 0, MAX, FALSE );
    params [i++] = CreatePlot ( PLOT_AREA,  "sDRBM", C_FREE_MAX,   pp, 1, MAX, TRUE );
    params [i++] = CreatePlot ( PLOT_LINE2, "sDTBA", C_GAUGE1_AVG, pp, 0, AVG, FALSE );
    params [i++] = CreatePlot ( PLOT_LINE2, "sDRBA", C_FREE_AVG,   pp, 1, AVG, TRUE );
    i = PlotPrintParams ( i, params, pp, 0, "sDTBA", "sDTBA", "sDTBM", "sDTBM", "%5.1lf" );
    i = PlotPrintParams ( i, params, pp, 0, "sDRBA", "sDRBA", "sDRBM", "sDRBM", "%5.1lf" );
    RRDGraph ( i, params, aFN, pp );
    while ( i > 0 ) free ( params [--i] );
    free ( title );
    PPDone ( pp );
    FNUnsetPNGData ( aFN );
}

void PlotPerfData ()
{
    int now = time ( NULL );
    tFN* fn = FNInit ();

    PlotDataPageFs     ( fn, PLOT_6H_1M,   now );
    PlotDataPageIO     ( fn, PLOT_6H_1M,   now );
    PlotDataPageRAvail ( fn, PLOT_6H_1M,   now );
    PlotDataCommit     ( fn, PLOT_6H_1M,   now );
    PlotDataCache      ( fn, PLOT_6H_1M,   now );
    PlotDataNPB        ( fn, PLOT_6H_1M,   now );
    PlotDataPPB        ( fn, PLOT_6H_1M,   now );
    PlotDataCode       ( fn, PLOT_6H_1M,   now );
    PlotDataDriver     ( fn, PLOT_6H_1M,   now );
    PlotDataPageFs     ( fn, PLOT_24H_5M,  now );
    PlotDataPageIO     ( fn, PLOT_24H_5M,  now );
    PlotDataPageRAvail ( fn, PLOT_24H_5M,  now );
    PlotDataCommit     ( fn, PLOT_24H_5M,  now );
    PlotDataCache      ( fn, PLOT_24H_5M,  now );
    PlotDataNPB        ( fn, PLOT_24H_5M,  now );
    PlotDataPPB        ( fn, PLOT_24H_5M,  now );
    PlotDataCode       ( fn, PLOT_24H_5M,  now );
    PlotDataDriver     ( fn, PLOT_24H_5M,  now );
    PlotDataPageFs     ( fn, PLOT_7D_15M,  now );
    PlotDataPageIO     ( fn, PLOT_7D_15M,  now );
    PlotDataPageRAvail ( fn, PLOT_7D_15M,  now );
    PlotDataCommit     ( fn, PLOT_7D_15M,  now );
    PlotDataCache      ( fn, PLOT_7D_15M,  now );
    PlotDataNPB        ( fn, PLOT_7D_15M,  now );
    PlotDataPPB        ( fn, PLOT_7D_15M,  now );
    PlotDataCode       ( fn, PLOT_7D_15M,  now );
    PlotDataDriver     ( fn, PLOT_7D_15M,  now );
    PlotDataPageFs     ( fn, PLOT_31D_60M, now );
    PlotDataPageIO     ( fn, PLOT_31D_60M, now );
    PlotDataPageRAvail ( fn, PLOT_31D_60M, now );
    PlotDataCommit     ( fn, PLOT_31D_60M, now );
    PlotDataCache      ( fn, PLOT_31D_60M, now );
    PlotDataNPB        ( fn, PLOT_31D_60M, now );
    PlotDataPPB        ( fn, PLOT_31D_60M, now );
    PlotDataCode       ( fn, PLOT_31D_60M, now );
    PlotDataDriver     ( fn, PLOT_31D_60M, now );
    PlotDataPageFs     ( fn, PLOT_366D_4H, now );
    PlotDataPageIO     ( fn, PLOT_366D_4H, now );
    PlotDataPageRAvail ( fn, PLOT_366D_4H, now );
    PlotDataCommit     ( fn, PLOT_366D_4H, now );
    PlotDataCache      ( fn, PLOT_366D_4H, now );
    PlotDataNPB        ( fn, PLOT_366D_4H, now );
    PlotDataPPB        ( fn, PLOT_366D_4H, now );
    PlotDataCode       ( fn, PLOT_366D_4H, now );
    PlotDataDriver     ( fn, PLOT_366D_4H, now );
    FNDone ( fn );
}
#endif /* WITH_RRD */

int CheckData ()
{
    int retVal = STATE_OK;

    ReadVals ();
    if ( warn < v.pps ) retVal = MAXIMUM ( retVal, STATE_WARNING );
    if ( crit < v.pps ) retVal = MAXIMUM ( retVal, STATE_CRITICAL );
    PrintExit ( retVal );
    printf ( " - paging: %i page faults / s; %i/%i pages i/o / s",
            v.pFpS, v.pIps, v.pOps );
    if ( npPerfData )
    {
        printf ( "|'physical memory available.'=%uKB",    v.rAvail );
        printf ( " 'committed virtual memory.'=%uB",      v.commit );
        printf ( " 'cached'=%uB",                         v.cached );
        printf ( " 'maximum cached since boot time'=%uB", v.cachPk );
        printf ( " 'page faults per second.'=%u",         v.pFpS );
        printf ( " 'pages input per second.'=%u",         v.pIps );
        printf ( " 'pages output per second.'=%u",        v.pOps );
        printf ( " 'pages in/out per second.'=%u;%u;%u",  v.pps, warn, crit );
        printf ( " 'pool nonpaged'=%uB",                  v.pNB );
        printf ( " 'pool paged'=%uB",                     v.pPB );
        printf ( " 'pool paged resident'=%uB",            v.pPRB );
        printf ( " 'system cache resident'=%uB",          v.sCaRB );
        printf ( " 'system code resident'=%uB",           v.sCoRB );
        printf ( " 'system code total'=%uB",              v.sCoTB );
        printf ( " 'system driver resident'=%uB",         v.sDRB );
        printf ( " 'system driver total'=%uB\n",          v.sDTB );
    }
    else
    {
        printf ( "|%u;%u;%u;%u;%u;%u;%u;%u;%u;%u;%u;%u;%u;%u;%u;%u\n",
                v.rAvail, v.commit, v.cached, v.cachPk,
                v.pFpS, v.pIps, v.pOps, v.pps,
                v.pNB, v.pPB, v.pPRB, v.sCaRB,
                v.sCoRB, v.sCoTB, v.sDRB, v.sDTB );
    }
    return ( retVal );
}

int main ( int argc, char** argv )
{
    int retVal = STATE_OK;

    SetGlobals ();
    ReadArgs ( argc, argv );
    if ( scanServices ) retVal = ScanServices ();
    else
    {
        if ( checkData ) retVal = CheckData ();
#if(WITH_RRD)
        if ( logPerfData ) LogPerfData ();
        if ( plotPerfData ) PlotPerfData ();
#endif /* WITH_RRD */
    }
    if ( hostName != NULL ) free ( hostName );
    if ( hostAddr != NULL ) free ( hostAddr );
    if ( progName != NULL ) free ( progName );
    return ( retVal );
}
