/*
 * File:      globals.h
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#ifndef GLOBALS_H
#define GLOBALS_H


#ifndef NETSNMP_NO_INLINE
#define NETSNMP_NO_INLINE /* Turn off inlining */
#endif


#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>


/**********************/
/* Exported constants */

#if(WITH_RRD)
/*
 * Directory where the host directories are created which in turn contain the
 * rrd files.
 */
#ifndef RRDDIR
#define RRDDIR "/usr/local/nagios/perflog/rrd"
#endif

/*
 * Directory where the host directories are created which in turn contain the
 * rrd files.
 */
#ifndef PNGDIR
#define PNGDIR "/usr/local/nagios/perflog/img"
#endif
#endif /* WITH_RRD */

/*
 * Facility where messages are logged.
 */
#ifndef LOGFACILITY
#define LOGFACILITY LOG_DAEMON
#endif


/*********************/
/* Just to make sure */

#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE  !FALSE
#endif

#ifndef ulong
#define ulong unsigned long int
#endif


/***********************/
/* Little helper macro */

#define MAXIMUM(a,b) ( (a) > (b) ) ? (a) : (b)

/********************/
/* Global variables */

/*
 * basename of the program when running. (Default: NULL)
 */
extern char* progName;

/*
 * service name, often the same as progName; used to build the file names
 * of the rrd archives and plots. (Default: NULL)
 */
extern char* srvName;

/*
 * version of the program. (Default: NULL)
 */
extern char* progVersion;

/*
 * logging level. (Default: LOG_ERR)
 */
extern int loggingLevel;

/*
 * TRUE if the host should be scanned for services of this type.
 * (Default: FALSE)
 */
extern int scanServices;

/*
 * TRUE if the actual service check should be made. (Default: FALSE)
 */
extern int checkData;

#if(WITH_RRD)
/*
 * TRUE if the performance data should be checked. (Default: FALSE)
 */
extern int logPerfData;

/*
 * TRUE if the performance data should be plotted. (Default: FALSE)
 */
extern int plotPerfData;
#endif /* WITH_RRD */

/*
 * TRUE if the Nagios Plugins compatible performance data format should be
 * used. (Default: FALSE)
 */
extern int npPerfData;

/*
 * String containing the synopsis of the program specific index parameters
 * e.g. "-s SYSTEM -i INDEX". (Default: NULL)
 */
extern char* helpIndexP;

/*
 * String containing the synopsis of the program specific threshold parameters
 * e.g. "-w WARN -c CRIT". (Default: NULL)
 */
extern char* helpThresP;

/*
 * String containing a short description of the program.
 */
extern char* helpPurpose;

/*
 * The variable where the host name will be stored. (Default: NULL)
 */
extern char* hostName;

/*
 * Set to TRUE if host name has been set. (Default: FALSE)
 */
extern int isHostNameSet;

/*
 * The variable where the host address will be stored. (Default: NULL)
 */
extern char* hostAddr;

/*
 * Set to TRUE if host address has been set. (Default: FALSE)
 */
extern int isHostAddrSet;

/*
 * Does the plugin have a '--index' parameter? This is special because it is
 * common enough to handle that globally. (Default: FALSE)
 */
extern int hasIndexParam;

/*
 * The variable where the index passed by an command line argument will be
 * stored. (Default: -1)
 */
extern int idx;

/*
 * Set to TRUE if idx has been set by the command line. (Default: FALSE)
 */
extern int isIdxSet;

/*
 * Does the plugin have a '--additional-info' parameter? This is special
 * because it is common enough to handle that globally. (Default: FALSE)
 */
extern int hasAddInfParam;

/*
 * The variable where the additional info passed by an command line argument
 * will be stored. This is only used (usable) if supportsPerfLog is TRUE.
 * (Default: NULL)
 */
extern char* addInf;

/*
 * Set to TRUE if addInf has been set by the command line. (Default: FALSE)
 */
extern int isAddInfSet;

/*
 * The community string if provided on the command line, else NULL.
 * (Default: NULL)
 */
extern char* community;

/*
 * The community string file name if provided on the command line,
 * else NULL. (Default: NULL)
 */
extern char* commFile;

/*
 * Defaults to SNMP_VERSION_2c, can be set to SNMP_VERSION_1 by a command
 * line switch.
 */
extern long snmpVersion;

/*
 * Number of times to retry a SNMP request. This can be set by the --retries
 * command line option. Defaults to 5.
 */
extern int snmpRetries;

/*
 * Timeout for the first SNMP request in microseconds. This is just a global
 * variable in case I want to make it configurable some day. Defaults to
 * 1000000 (1 Second).
 */
extern int snmpTimeOut;

/*
 * Does the plugin support performance logging and plotting? (Default: FALSE)
 */
extern int supportsPerfLog;

#if(WITH_RRD)
/*
 * Base(!) directory for the rrd archives. (Default: RRDDIR)
 */
extern char* rrdDir;

/*
 * RRD Step (c.f. man rrdcreate). (Default: -1)
 */
extern int rrdStep;

/*
 * RRD Heartbeat (c.f. man rrdcreate). (Default: -1)
 */
extern int rrdHBeat;

/*
 * Base(!) directory for the png files. (Default: PNGDIR)
 */
extern char* pngDir;
#endif /* WITH_RRD */


/***********************************************************/
/* Function which have to be provided by the using program */

/*
 * Writes the plugin specific parameters to stdout. An example
 * would be:
 *
 * void HelpSpecificParams ()
 * {
 *     printf (
 *     "-w, --warning INTEGER\n"
 *     "    Level at which a warning is generated.\n"
 *     "\n"
 *     "-c, --critical INTEGER\n"
 *     "    Level at which a critical is generated.\n"
 *     "\n"
 *     );
 * }
 */
extern void HelpSpecificParams ();

/*
 * This function writes a help text about the function of this program to
 * stdout. Example:
 *
 * void HelpExplain ()
 * {
 *     printf (
 *     "This plugin monitors the function of the foobar device. If the\n"
 *     "device reports a failing foo entity, then a warning is generated.\n"
 *     "If two or more entities fail, then a critical is generated.\n"
 *     "\n"
 *     );
 * }
 */
extern void HelpExplain ();

/*
 * This function writes a help text about the performance data provided by
 * this program to stdout. Example:
 *
 * void HelpPerformanceData ()
 * {
 *     printf (
 *     "Native performance data is provided in the format:\n"
 *     "<foo>;<bar>\n"
 *     "where\n"
 *     "    foo ... foo counter\n"
 *     "    bar ... current bar value\n"
 *     "\n"
 *     );
 *     printf (
 *     "Nagios Plugins compatible performance data is provided in the format:\n"
 *     "'foo counter'=<foo>c "
 *     "'current bar value'=<bar>\n"
 *     "\n"
 *     );
 * }
 */
extern void HelpPerformanceData ();

/*
 * This function should try to interpret the argument argv[(*i)]
 * as a program specific option. If that is a known argument, appropriate
 * measures should be taken, *i should be incremented and TRUE should be
 * returned. If the function can not handle this parameter, then FALSE
 * should be returned.
 *
 * The function can also interpret additional parameters, each time
 * incrementing *i.
 *
 * Example:
 *
 * int ChkOptSpecificParams ( int argc, char** argv, int* i )
 * {
 *     int ret = TRUE;
 *     if (        !strcmp ( argv [(*i)], "-w" )
 *             ||  !strcmp ( argv [(*i)], "--warning" ) )
 *     {
 *         isWarnSet = TRUE;
 *         (*i)++;
 *         if ( (*i) < argc )
 *         {
 *             char *tail = argv [(*i)];
 *             errno = 0;
 *             warn = strtod ( argv [(*i)++], &tail );
 *             if ( errno || *tail ) Usage ( "Invalid 'warn' parameter" );
 *         } else
 *         {
 *             Usage ( "Missing 'warn' parameter(s)" );
 *         }
 *     } else ret = FALSE;
 *     return ( ret );
 * }
 */
int ChkOptSpecificParams ( int argc, char** argv, int* i );

#if(WITH_RRD)
/*
 * Creates a rrd file suitable for the data which is to be handled.
 */
extern void CreateDataFile ( const char* dataFile );

/*
 * Parses performance data as given by aData into the global data set.
 */
void ParseData ( const char* aData );

/*
 * Writes the performance data from the global data set to the RRD archive.
 */
void WriteData ( const char* aRRDFPath );
#endif /* WITH_RRD */

/*
 * This function is called for each valid variable which is found during
 * a SnmpWalk().
 *
 * It could/should be used to figure out if this object has some meaning
 * while scanning for a service.
 */
extern void SnmpWalkHandleVariable ( netsnmp_variable_list* aVar );

#endif /* GLOBALS_H */
