/*
 * File:      rrdif.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the rets of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#if(WITH_RRD)

#include <stdio.h> /* rrd.h needs stdio.h to be included first (type FILE) */
#include <rrd.h>
#include <dirent.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include "globals.h"
#include "nagiosif.h"
#include "rrdif.h"
#include "syslogif.h"
#include "strutils.h"


int TimingGetRange ( int aTiming )
{
    int ret = 0;

    switch ( aTiming )
    {
        case PLOT_1H_10S:  ret = 3600;            break;
        case PLOT_6H_1M:   ret = 3600 *  6;       break;
        case PLOT_24H_5M:  ret = 3600 * 24;       break;
        case PLOT_7D_15M:  ret = 3600 * 24 *   7; break;
        case PLOT_31D_60M: ret = 3600 * 24 *  31; break;
        case PLOT_366D_4H: ret = 3600 * 24 * 366; break;
        default: Error ( LOG_ERR, "Invalid timing" ); break;
    }
    return ret;
}

int TimingGetInterval ( int aTiming )
{
    int ret = 0;

    switch ( aTiming )
    {
        case PLOT_1H_10S:  ret = 10;          break;
        case PLOT_6H_1M:   ret = 60;          break;
        case PLOT_24H_5M:  ret = 60 *  5;     break;
        case PLOT_7D_15M:  ret = 60 * 15;     break;
        case PLOT_31D_60M: ret = 60 * 60;     break;
        case PLOT_366D_4H: ret = 60 * 40 * 4; break;
        default: Error ( LOG_ERR, "Invalid timing" ); break;
    }
    return ret;
}

float TimingGetXScale ( int aTiming )
{
    float ret = 1.00;

    switch ( aTiming )
    {
        case PLOT_1H_10S:  ret = 1.00; break;
        case PLOT_6H_1M:   ret = 1.00; break;
        case PLOT_24H_5M:  ret = 2.00; break;
        case PLOT_7D_15M:  ret = 1.00; break;
        case PLOT_31D_60M: ret = 1.00; break;
        case PLOT_366D_4H: ret = 0.25; break;
        default: Error ( LOG_ERR, "Invalid timing" ); break;
    }
    return ret;
}

char* TimingGetDescr ( int aTiming )
{
    char* ret = NULL;

    switch ( aTiming )
    {
        case PLOT_1H_10S:  ret = StrDup ( "1h_10s" );   break;
        case PLOT_6H_1M:   ret = StrDup ( "6h_1m" );   break;
        case PLOT_24H_5M:  ret = StrDup ( "24h_5m" );  break;
        case PLOT_7D_15M:  ret = StrDup ( "7d_15m" );  break;
        case PLOT_31D_60M: ret = StrDup ( "31d_1h" ); break;
        case PLOT_366D_4H: ret = StrDup ( "366d_4h" ); break;
        default: Error ( LOG_ERR, "Invalid timing" ); break;
    }
    return ret;
}

void CheckDir ( const char* aDir )
{
    DIR* dir = opendir ( aDir );

    if ( dir )
    {
        closedir ( dir );
    } else
    {
        mode_t mode = S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
        /* mkdir returns 0(!) on success, -1 on failure */
        if ( mkdir ( aDir, mode ) )
        {
            char msg[1024];
            snprintf ( msg, 1024,
                    "Could not make directory %s.\n",
                    aDir );
            Error ( LOG_ERR, msg );
        }
    }
}

char* BuildRRDFilePath ( const char* aRRDHostDirPath )
{
    char* ret;
    char* part;

    if ( hasAddInfParam ) part = StrConcat ( 4, srvName, "_", addInf, ".rrd" );
    else                  part = StrConcat ( 2, srvName, ".rrd" );
    ret = StrConcat ( 3, aRRDHostDirPath, "/", part );
    free ( part );
    return ret;
}


void CheckRRDDataFile ( const char* dataFile )
{
    FILE* file = fopen ( dataFile, "r+" );

    if ( file ) fclose ( file );
    else        CreateDataFile ( dataFile );
}

char* CreateDSStr ( const char* aName, int aType, int aHBeat, char* aMin,
        char* aMax )
{
    char* ret;
    char* type = NULL;
    char* hBeat = IntToStr ( aHBeat );

    switch ( aType )
    {
        case GAUGE:    type = "GAUGE";    break;
        case COUNTER:  type = "COUNTER";  break;
        case DERIVE:   type = "DERIVE";   break;
        case ABSOLUTE: type = "ABSOLUTE"; break;
        default:
        {
            char msg[1024];

            snprintf ( msg, 1024,
                    "Invalid type of DS: %i.\n", aType );
            Error ( LOG_ERR, msg );
        }
        break;
    }
    ret = StrConcat ( 10, "DS:", aName, ":", type, ":", hBeat, ":",
            aMin, ":", aMax );
    return ( ret );
}

char* CreateRRAStr ( int aCF, int aRangeHours, int aCDPIntMinutes )
{
    char* ret;
    char* cf = NULL;
    int secondsPerCDP = 60 * aCDPIntMinutes;
    int secondsRange = 60 * 60 * aRangeHours;
    char* cdpPerPDP = IntToStr ( secondsPerCDP / rrdStep );
    char* rows = IntToStr ( secondsRange / secondsPerCDP );

    switch ( aCF )
    {
        case MIN: cf = "MIN";     break;
        case MAX: cf = "MAX";     break;
        case AVG: cf = "AVERAGE"; break;
        case LST: cf = "LAST";    break;
        default:
        {
            char msg[1024];

            snprintf ( msg, 1024,
                    "Invalid consolidating function for RRA: %i.\n", aCF );
            Error ( LOG_ERR, msg );
        }
        break;
    }
    ret = StrConcat ( 6, "RRA:", cf, ":0.5:", cdpPerPDP, ":", rows );
    free ( rows );
    free ( cdpPerPDP );
    return ret;
}

void RRDCreate ( int aParamCnt, char** aParams, const char* aRRDFilePath )
{
    int chk;

    LogParams ( aParamCnt, aParams );
    optind = 1; /* rrd_*'s weired interface */
    rrd_clear_error ();
    /* rrd_* functions return 0 on success. */
    chk = rrd_create ( aParamCnt, aParams );
    CheckRRDError ( chk, "create", aRRDFilePath );
}

void RRDUpdate ( int aParamCnt, char** aParams, const char* aRRDFilePath )
{
    int chk;

    LogParams ( aParamCnt, aParams );
    optind = 1; /* rrd_*'s weired interface */
    rrd_clear_error ();
    /* rrd_* functions return 0 on success. */
    chk = rrd_update ( aParamCnt, aParams );
    CheckRRDError ( chk, "update", aRRDFilePath );
}

char* BuildPNGFileName ( const char* aPltId, int aTiming )
{
    char* ret;
    char* ai;
    char* pi;
    char* ti;

    if ( hasAddInfParam ) ai = StrConcat ( 2, "_", addInf );
    else                  ai = StrDup ( "" );
    if ( NULL != aPltId ) pi = StrConcat ( 2, "_", aPltId );
    else                  pi = StrDup ( "" );
    ti = TimingGetDescr ( aTiming );
    ret = StrConcat ( 6, srvName, ai, pi, "_", ti, ".png" );
    /* ti is not a malloced string */
    free ( ai );
    free ( pi );
    return ret;
}

char* BuildPNGFilePath ( const char* aPNGHostDir, const char* aPNGFileName )
{
    char* ret;

    ret = StrConcat ( 3, aPNGHostDir, "/", aPNGFileName );
    return ret;
}

char* BuildTmpPNGFilePath ( const char* aPNGFilePath )
{
    char* ret;
    char* pid;

    pid = IntToStr ( getpid() );
    ret = StrConcat ( 3, aPNGFilePath, ".", pid );
    free ( pid );
    return ret;
}

tFN* FNInit ()
{
    tFN* fn;

    fn = malloc ( sizeof ( tFN ) );
    if ( fn == NULL ) ErrorMalloc ();
    fn->rrdHostDirPath = GetRRDHostDirPath ();
    fn->rrdFilePath = BuildRRDFilePath ( fn->rrdHostDirPath );
    fn->pngHostDirName = GetPNGHostDirName ();
    fn->pngFileName = NULL;
    fn->pngFilePath = NULL;
    fn->tmpPNGFilePath = NULL;
    CheckDir ( fn->pngHostDirName );
    return fn;
}

void FNDone ( tFN* aFN )
{
    FNUnsetPNGData ( aFN );
    free ( aFN->pngHostDirName );
    free ( aFN->rrdFilePath );
    free ( aFN->rrdHostDirPath );
    free ( aFN );
}

void FNSetPNGData ( tFN* aFN, const char* aPlotId, int aTiming )
{
    FNUnsetPNGData ( aFN );
    aFN->pngFileName = BuildPNGFileName ( aPlotId, aTiming );
    aFN->pngFilePath = BuildPNGFilePath ( aFN->pngHostDirName,
            aFN->pngFileName );
    aFN->tmpPNGFilePath = BuildTmpPNGFilePath ( aFN->pngFilePath );
}

void FNUnsetPNGData ( tFN* aFN )
{
    if ( aFN->tmpPNGFilePath != NULL ) free ( aFN->tmpPNGFilePath );
    if ( aFN->pngFilePath    != NULL ) free ( aFN->pngFilePath );
    if ( aFN->pngFileName    != NULL ) free ( aFN->pngFileName );
    aFN->tmpPNGFilePath = NULL;
    aFN->pngFilePath    = NULL;
    aFN->pngFileName    = NULL;
}

tPP* PPInit ()
{
    tPP* pp;
    pp = malloc ( sizeof ( tPP ) );
    if ( pp == NULL ) ErrorMalloc ();
    pp->dim = NULL;
    pp->unit = NULL;
    pp->stack = NULL;
    pp->count = 0;
    pp->dimLen = 0;
    pp->unitLen = 0;
    return pp;
}

void PPAddPlot ( tPP* aPP, const char* aDim, const char* aUnit, int aStack )
{
    int index = aPP->count;
    int dimLen = strlen ( aDim );
    int unitLen = strlen ( aUnit );

    aPP->count++;
    if ( aPP->count == 1 )
    {
        aPP->dim = malloc ( sizeof ( char* ) );
        aPP->unit = malloc ( sizeof ( char* ) );
        aPP->stack = malloc ( sizeof ( int ) );
    } else
    {
        aPP->dim = realloc ( aPP->dim, sizeof ( char* ) * aPP->count );
        aPP->unit = realloc ( aPP->unit, sizeof ( char* ) * aPP->count );
        aPP->stack = realloc ( aPP->stack, sizeof ( int ) * aPP->count );
    }
    if ( aPP->dim == NULL ) ErrorMalloc ();
    if ( aPP->unit == NULL ) ErrorMalloc ();
    if ( aPP->stack == NULL ) ErrorMalloc ();
    aPP->dim[index] = StrDup ( aDim );
    aPP->unit[index] = StrDup ( aUnit );
    aPP->stack[index] = aStack;
    if ( aStack ) dimLen += strlen ( "[+] " );
    if ( dimLen > aPP->dimLen ) aPP->dimLen = dimLen;
    if ( unitLen > aPP->unitLen ) aPP->unitLen = unitLen;
}

void PPDone ( tPP* aPP )
{
    while ( aPP->count > 0 )
    {
        aPP->count--;
        free ( aPP->unit[aPP->count] );
        free ( aPP->dim[aPP->count] );
    }
    free ( aPP->stack );
    free ( aPP->unit );
    free ( aPP->dim );
    free ( aPP );
}

int PlotStdParams ( char** aParams, tFN* aFN, const char* aTitle,
        const char* aVLabel, int aTiming, int aEnd, int aSIBase, int aSIExp,
        int aLower, int aUpper, int aUsingBidir )
{
    int range = TimingGetRange ( aTiming );
    int interval = TimingGetInterval ( aTiming );
    float xScale = TimingGetXScale ( aTiming );
    int start = aEnd - range;
    int w1 = (int) ( ( (float) ( range / interval ) ) * xScale );
    int h1 = ( aUsingBidir ) ? PLOT_H_BIDIR : PLOT_H_UNIDIR;
    int i = 0;
    int usingExp = TRUE;
    int exp = 0;

    printf ( "TITLE %s\n",  aTitle );
    printf ( "START %i\n",  start );
    printf ( "END %i\n",    aEnd );
    switch ( aSIExp )
    {
        case PLOT_SI_AUTO:  usingExp = FALSE; break;
        case PLOT_SI_FEMTO: exp = -15; break;
        case PLOT_SI_PICO:  exp = -12; break;
        case PLOT_SI_NANO:  exp =  -9; break;
        case PLOT_SI_MICRO: exp =  -6; break;
        case PLOT_SI_MILLI: exp =  -3; break;
        case PLOT_SI_NONE:  exp =   0; break;
        case PLOT_SI_KILO:  exp =   3; break;
        case PLOT_SI_MEGA:  exp =   6; break;
        case PLOT_SI_GIGA:  exp =   9; break;
        case PLOT_SI_TERA:  exp =  12; break;
        case PLOT_SI_PETA:  exp =  15; break;
        default: Error ( LOG_ERR, "Invalid SI Exponent Name given" );
    }
    aParams [i++] = StrDup ( "JUNK" );
    aParams [i++] = StrDup ( aFN->tmpPNGFilePath );
    aParams [i++] = StrDup ( "-t" );
    aParams [i++] = StrDup ( aTitle );
    aParams [i++] = StrDup ( "-v" );
    aParams [i++] = StrDup ( aVLabel );
    aParams [i++] = StrDup ( "-b" );
    aParams [i++] = IntToStr ( aSIBase );
    if ( usingExp )
    {
        aParams [i++] = StrDup ( "-X" );
        aParams [i++] = IntToStr ( exp );
    }
    if ( aLower < aUpper )
    {
        aParams [i++] = StrDup ( "-l" );
        aParams [i++] = IntToStr ( aLower );
        aParams [i++] = StrDup ( "-u" );
        aParams [i++] = IntToStr ( aUpper );
    }
    aParams [i++] = StrDup ( "-s" );
    aParams [i++] = IntToStr ( start );
    aParams [i++] = StrDup ( "-e" );
    aParams [i++] = IntToStr ( aEnd );
    aParams [i++] = StrDup ( "-w" );
    aParams [i++] = IntToStr ( w1 );
    aParams [i++] = StrDup ( "-h" );
    aParams [i++] = IntToStr ( h1 );
    aParams [i++] = StrDup ( "-a" );
    aParams [i++] = StrDup ( "PNG" );
    return i;
}

char* CreateDEF ( const char* aVName, tFN* aFN, const char* aDS,
        int aCF )
{
    char* ret;
    char* cf = NULL;

    switch ( aCF )
    {
        case MIN: cf = "MIN";     break;
        case MAX: cf = "MAX";     break;
        case AVG: cf = "AVERAGE"; break;
        case LST: cf = "LAST";    break;
        default:
            Error ( LOG_CRIT, "Invalid aCF passed to CreateDEF()." );
        break;
    }
    ret = StrConcat ( 8, "DEF:", aVName, "=", aFN->rrdFilePath, ":", aDS,
            ":", cf );
    return ret;
}

char* CreatePlot ( int aType, const char* aVName, const char* aColor,
        tPP* aPP, int aPPI, int aCF, int aNL )
{
    char* ret;
    char* type = NULL;
    char* cf = NULL;
    char* stack = aPP->stack[aPPI] ? "[+] " : "";

    switch ( aType )
    {
        case PLOT_AREA:  type = "AREA"; break;
        case PLOT_LINE1: type = "LINE1"; break;
        case PLOT_LINE2: type = "LINE2"; break;
        case PLOT_LINE3: type = "LINE3"; break;
        case PLOT_STACK: type = "STACK"; break;
        default:
            Error ( LOG_CRIT, "Invalid aType passed to CreatePlot()." );
        break;
    }
    switch ( aCF )
    {
        case MIN: cf = "min"; break;
        case MAX: cf = "max"; break;
        case AVG: cf = "avg"; break;
        case LST: cf = "lst"; break;
        default:
            Error ( LOG_CRIT, "Invalid aCF passed to CreatePlot()." );
        break;
    }
    if ( aNL )
    {
        ret = StrConcat ( 10, type, ":", aVName, aColor, ":", stack,
                aPP->dim[aPPI], ", ", cf, "\\n" );
    } else
    {
        char* padStr;
        int i;
        /* Start with the length of the longest dimension string */
        int padLen = aPP->dimLen;

        /* add two blanks (inter-column space) */
        padLen += 2;
        /* add the length of the cf string TODO: Why? */
        /* TODO padLen += 3; */
        /* subtract the length of this dimension string */
        padLen -= strlen ( aPP->dim[aPPI] );
        /* subtract the length of the stack string (may be 0) */
        padLen -= strlen ( stack );
        /* create the padding string */
        padStr = malloc ( padLen + 1 );
        for ( i = 0; i < padLen; i++ ) padStr [i] = ' ';
        padStr [padLen] = '\000';
        ret = StrConcat ( 10 , type, ":", aVName, aColor, ":", stack,
                aPP->dim[aPPI], ", ",  cf, padStr );
        free ( padStr );
    }
    return ret;
}

int PlotPrintParams (int aI, char** aParams, tPP* aPP, int aPPI,
        const char* aVNAvg, const char* aVNMin, const char* aVNMax,
        const char* aVNLst, const char* aFormat )
{
    int i = aI;

    aParams [i++] = StrConcat ( 4, "PRINT:", aVNAvg, ":AVERAGE:AVG ", aFormat );
    aParams [i++] = StrConcat ( 4, "PRINT:", aVNMin, ":MIN:MIN ",     aFormat );
    aParams [i++] = StrConcat ( 4, "PRINT:", aVNMax, ":MAX:MAX ",     aFormat );
    aParams [i++] = StrConcat ( 4, "PRINT:", aVNLst, ":LAST:LST ",    aFormat );
    return i;
}

void RRDGraph ( int aParamCnt, char** aParams, tFN* aFN, tPP* aPP )
{
    int chk = -1;
    int w;
    int h;
    char** results;
    int j = 0;
#if(NEW_RRD_GRAPH)
    double j1;
    double j2;
#endif

    LogParams ( aParamCnt, aParams );
    optind = 1; /* rrd_*'s weired interface */
    rrd_clear_error ();
    /* rrd_* functions return 0 on success. */
    chk = rrd_graph ( aParamCnt, aParams, &results, &w, &h
#if(NEW_RRD_GRAPH)
            , NULL, &j1, &j2
#endif
            );
    CheckRRDError ( chk, "graph", aFN->rrdFilePath );
    if ( rename ( aFN->tmpPNGFilePath, aFN->pngFilePath ) == -1 )
            Error ( LOG_ERR, "Could not rename file" );
    printf ( "IMAGE %s\n",  aFN->pngFileName );
    printf ( "WIDTH %i\n",  w );
    printf ( "HEIGHT %i\n", h );
    while ( results [j] )
    {
        if ( ( j % 4 ) == 0 )
        {
            int i = j / 4;

            if ( i < aPP->count )
            {
                printf ( "DIMENSION %s\n", aPP->dim[i] );
                printf ( "UNIT %s\n", aPP->unit[i] );
            }
        }
        printf ( "%s\n", results [j++] );
    }
    printf ( "END\n" );
    while ( j > 0 ) free ( results [--j] );
    free ( results );
}

void LogPerfData ()
{
    char* hostDirPath = GetRRDHostDirPath ();
    char* rrdFilePath = BuildRRDFilePath ( hostDirPath );

    CheckDir ( hostDirPath );
    CheckRRDDataFile ( rrdFilePath );
    if ( !checkData )
    {
        char* data = GetData ();
        ParseData ( data );
        free ( data );
    } /* else the global variable v should already contain the parsed data. */
    WriteData ( rrdFilePath );
    free ( rrdFilePath );
    free ( hostDirPath );
}

#endif /* WITH_RRD */
