/*
 * File:      snmpif.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */


#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <pwd.h>
#include <stdlib.h>
#include <unistd.h>
#include "nagiosif.h"
#include "snmpif.h"
#include "strutils.h"
#include "syslogif.h"


void ReadLine ( char** aBuff, int* aBuffSize, FILE* aFile )
{
    int i = 0;
    int eof = FALSE;
    int newLine = FALSE;
    while ( !eof && !newLine )
    {
        int ch;
        while ( i >= *aBuffSize )
        {
            *aBuffSize += 128;
            *aBuff = realloc ( *aBuff, *aBuffSize );
            if ( !(*aBuff) ) ErrorMalloc ();
        }
        ch = getc ( aFile );
        if ( ch == EOF )
        {
            eof = TRUE;
        } else
        {
            if ( ch == '\n' )
            {
                newLine = TRUE;
                (*aBuff) [i++] = '\0';
            } else
            {
                (*aBuff) [i++] = (char) ch;
            }
        }
    }
}

int ReadFile ( char** aBuff, int* aBuffSize, char* aFileName )
{
    int ret = FALSE;
    FILE* file = fopen ( aFileName, "r" );
    if ( file )
    {
        ReadLine ( aBuff, aBuffSize, file );
        ret = TRUE;
        fclose ( file );
    }
    return ( ret );
}

/*
 * Note: This will malloc a buffer which is "returned" in aSess->community.
 * Somebody should free it ... that is: Call "free ( aSess->community );"
 * when you are done.
 */
void ConfigureCommunity ( struct snmp_session* aSess )
{
    int strLen = 512; /* common length for allocated strings */
    char* comBuf = malloc ( strLen );
    char* comRet = NULL;
    if ( !comBuf ) ErrorMalloc ();
    if ( community )
    {
        strncpy ( comBuf, community, 511 );
        Log ( LOG_DEBUG, "Using community from command line." );
    } else
    {
        int readIt = FALSE;

        if ( commFile )
        {
            Log ( LOG_DEBUG,
                    "Trying to read community from user supplied file." );
            readIt = ReadFile ( &comBuf, &strLen, commFile );
            if ( readIt )
            {
                Log ( LOG_DEBUG, "Using community from user supplied file." );
            }
        } /* else never mind, try next */
        if ( !readIt )
        {
            char* home;
            Log ( LOG_DEBUG,
                    "Failed to read community from user supplied file. "
                    "Trying to read community from $HOME." );
            home = getenv ( "HOME" );
            if ( home )
            {
                char* fName = "/.s4nag/community";
                char* userCfg = malloc ( strLen );
                if ( !userCfg ) ErrorMalloc ();
                strncpy ( userCfg, home, 128 );
                strncat ( userCfg, fName, 128 );
                readIt = ReadFile ( &comBuf, &strLen, userCfg );
                free ( userCfg );
                if ( readIt )
                        Log ( LOG_DEBUG, "Using community from $HOME." );
            } else
            {
                Log ( LOG_DEBUG, "Could not read $HOME." );
            }
        } /* else we've already got our community string */
        if ( !readIt )
        {
            struct passwd* pwd;
            Log ( LOG_DEBUG, "Failed to read community from $HOME."
                    "Trying to read community from passwd home." );
            pwd = getpwuid ( geteuid () );
            if ( pwd )
            {
                char* home = pwd->pw_dir;
                if ( home )
                {
                    char* fName = "/.s4nag/community";
                    char* userCfg = malloc ( strLen );
                    if ( !userCfg ) ErrorMalloc ();
                    strncpy ( userCfg, home, 128 );
                    strncat ( userCfg, fName, 128 );
                    readIt = ReadFile ( &comBuf, &strLen, userCfg );
                    free ( userCfg );
                } /* else never mind, try next */
            } /* else never mind, try next */
            if ( readIt )
                    Log ( LOG_DEBUG, "Using community from passwd home." );
        } /* else we've already got our community string */
        if ( !readIt )
        {
            Log ( LOG_DEBUG, "Failed to read community from passwd home. "
                    "Trying to read community from /etc/s4nag/community." );
            readIt = ReadFile ( &comBuf, &strLen, "/etc/s4nag/community" );
            if ( readIt )
            {
                Log ( LOG_DEBUG,
                        "Using community from /etc/s4nag/community." );
            }
        } /* else we've already got our community string */
        if ( !readIt )
        {
            Log ( LOG_DEBUG,
                    "Failed to read community from /etc/s4nag/community. "
                    "Reverting to default community 'public'." );
            strncpy ( comBuf, "public", strlen ( "public" ) + 1 );
        } /* else we've already got our community string */
    }
    comRet = StrDup ( comBuf );
    aSess->community = (u_char*) comRet;
    aSess->community_len = strlen ( comRet );
    free ( comBuf );
}

void ConfigureSession1 ( struct snmp_session* aSess )
{
    snmp_sess_init ( aSess );
    if ( isHostAddrSet ) aSess->peername = StrDup ( hostAddr );
    else                 aSess->peername = StrDup ( hostName );
    aSess->version = snmpVersion;
    aSess->retries = snmpRetries;
    aSess->timeout = snmpTimeOut;
    ConfigureCommunity ( aSess );
}

void OIDToString ( char* aBuffer, int aBufLen, oid aOid[], size_t aOidLen)
{
    int i; /* counter for the aOid array */
    int j = 0; /* counter for the destination buffer */

    for ( i = 0 ; i < aOidLen ; i++ )
    {
        char num[32];
        int k = 0; /* counter for the num buffer */
        sprintf ( num, "%li", aOid [i] );
        if ( i != 0 ) aBuffer [j++] = '.';
        while ( num [k] && ( j < aBufLen - 1 ) ) aBuffer [j++] = num [k++];
    }
    aBuffer [j++] = '\0';
}

void SNMPError ( const char* aMsg, const char* aPeer, oid aOid[],
        size_t aOidLen )
{
    char msg [1024];
    char oid [1024] = "";
    OIDToString ( oid, 1024, aOid, aOidLen );
    snprintf ( msg, 1024, "%s Peer = %s, OID = %s", aMsg, aPeer, oid );
    Error ( LOG_ERR, msg );
}

void CheckType ( uint aFoundType, uint aExpectedType, char* aExpectedTypeName,
        struct snmp_session* aSess, oid aOid[], size_t aOidLen )
{
    /* cf. /usr/include/net-snmp/library/asn1.h */
    if ( aFoundType != aExpectedType )
    {
        char msg [1024] = "";

        snprintf ( msg, 1024,
                "Expected type 0x%x (%s), found type 0x%x.",
                aExpectedType, aExpectedTypeName, aFoundType );
        SNMPError ( msg, aSess->peername, aOid, aOidLen );
    }
}

int OidCmp ( oid aOid1[], oid aOid2[], size_t aLen )
{
    int i = 0;
    int diff;

    if ( aLen > MAX_OID_LEN )
            Error ( LOG_ERR, "Length of OID exceeds MAX_OID_LEN" );
    do
    {
        diff = aOid1[i] - aOid2[i];
        i++;
    }
    while ( ( diff == 0 ) && ( i < aLen ) );
    return diff;
}

void OidCpy ( oid aDstOid[], size_t* aDstLen, oid aSrcOid[], size_t* aSrcLen )
{
    int i;

    if ( *aSrcLen > MAX_OID_LEN )
            Error ( LOG_ERR, "Length of Source OID exceeds MAX_OID_LEN" );
    for ( i = 0; i < *aSrcLen; i++ ) aDstOid[i] = aSrcOid[i];
    *aDstLen = *aSrcLen;
}

int SnmpWalkHandleResult ( struct snmp_session* aSess, struct snmp_pdu* aRcvd,
        oid aCurrOid[], size_t* aCurrLen, oid aRootOid[], size_t* aRootLen )
{
    netsnmp_variable_list* var = aRcvd->variables;
    int cont = TRUE;

    if ( var )
    {
        while ( ( var != NULL ) && cont )
        {
            /* check if we're still inside the subtree */
            if ( var->name_length < *aRootLen )
                    cont = FALSE;
            else if ( OidCmp ( aRootOid, var->name, *aRootLen ) )
                    cont = FALSE;
            else
            {
                SnmpWalkHandleVariable ( var );
                if      ( var->type == SNMP_ENDOFMIBVIEW )   cont = FALSE;
                else if ( var->type == SNMP_NOSUCHOBJECT )   cont = FALSE;
                else if ( var->type == SNMP_NOSUCHINSTANCE ) cont = FALSE;
                else
                {
                    OidCpy ( aCurrOid, aCurrLen,
                            var->name, &(var->name_length) );
                }
            }
            var = var->next_variable;
        }
    }
    else SNMPError ( "Got STAT_SUCCESS and SNMP_ERR_NOERROR but no var?!?",
            aSess->peername, aCurrOid, *aCurrLen );
    return cont;
}

void SnmpWalk ( oid aRootOid[], size_t aRootLen )
{
    struct snmp_session sess1;
    struct snmp_session* sess2;
    struct snmp_pdu* send;
    struct snmp_pdu* rcvd;
    oid currOid[MAX_OID_LEN];
    size_t currLen;
    int cont = TRUE;
    int found = FALSE; /* to catch an end of mib, I'll need that */
    int status;

    ConfigureSession1 ( &sess1 );
    sess2 = snmp_open ( &sess1 );
    if ( sess2 == NULL ) Error ( LOG_ERR, "Could not open session!" );
    OidCpy ( currOid, &currLen, aRootOid, &aRootLen );
    do
    {
        send = snmp_pdu_create ( SNMP_MSG_GETNEXT );
        snmp_add_null_var ( send, currOid, currLen );
        /* actually send the request */
        status = snmp_synch_response ( sess2, send, &rcvd );
        if ( status == STAT_SUCCESS )
        {
            if ( rcvd )
            {
                if ( rcvd->errstat == SNMP_ERR_NOERROR )
                {
                    found = TRUE;
                    cont = SnmpWalkHandleResult ( sess2, rcvd,
                            currOid, &currLen, aRootOid, &aRootLen );
                }
                else
                {
                    if ( ( rcvd->errstat != SNMP_ERR_NOSUCHNAME ) || !found )
                    {
                        char msg [1024];

                        snprintf ( msg, 1024, "Error in packet: %s",
                                snmp_errstring ( rcvd->errstat ) );
                        SNMPError ( msg, sess2->peername, currOid, currLen );
                    }
                    else
                    {
                        /* I've already found something and this is a "no such
                         * name error". So this is just an end-of-mib.
                         */
                        cont = FALSE;
                    }
                }
            }
            else SNMPError ( "SNMP response is NULL.", sess2->peername,
                    currOid, currLen );
        }
        else if ( status == STAT_TIMEOUT )
                SNMPError ( "Timeout!", sess2->peername, currOid, currLen );
        else SNMPError ( "SNMP request returned an error.", sess2->peername,
                currOid, currLen );
        snmp_free_pdu ( rcvd );
    }
    while ( cont );
    free ( sess1.peername );
    free ( sess1.community );
    snmp_close_sessions ();
}

struct snmp_pdu* SnmpGet ( struct snmp_session* aSess,
        oid aOid[], size_t aOidLen, uint aType, char* aTypeName )
{
    struct snmp_pdu* send;
    struct snmp_pdu* rcvd;
    int status;

    send = snmp_pdu_create ( SNMP_MSG_GET );
    snmp_add_null_var ( send, aOid, aOidLen );
    /* actually send the request */
    status = snmp_synch_response ( aSess, send, &rcvd );
    if ( status == STAT_SUCCESS )
    {
        if ( rcvd )
        {
            if ( rcvd->errstat == SNMP_ERR_NOERROR )
            {
                if ( rcvd->variables )
                {
                    CheckType ( rcvd->variables->type, aType, aTypeName,
                            aSess, rcvd->variables->name,
                            rcvd->variables->name_length );
                }
                else SNMPError (
                        "Got STAT_SUCCESS and SNMP_ERR_NOERROR but no var!",
                        aSess->peername, aOid, aOidLen );
            }
            else
            {
                char msg [1024];
                snprintf ( msg, 1024, "Error in packet: %s",
                        snmp_errstring ( rcvd->errstat ) );
                SNMPError ( msg, aSess->peername, aOid, aOidLen );
            }
        }
        else SNMPError ( "SNMP response is NULL.", aSess->peername, aOid,
                aOidLen );
    }
    else if ( status == STAT_TIMEOUT )
            SNMPError ( "Timeout!", aSess->peername, aOid, aOidLen );
    else SNMPError ( "SNMP request returned an error.", aSess->peername,
            aOid, aOidLen );
    return ( rcvd );
}

int ReadInteger ( struct snmp_session* aSess, oid aOid[], size_t aOidLen )
{
    struct snmp_pdu* rcvd;
    int ret;

    rcvd = SnmpGet ( aSess, aOid, aOidLen, ASN_INTEGER, "an integer" );
    ret = *(rcvd->variables->val.integer);
    snmp_free_pdu ( rcvd );
    return ( ret );
}

uint ReadCounter32 ( struct snmp_session* aSess, oid aOid[], size_t aOidLen )
{
    struct snmp_pdu* rcvd;
    uint ret;

    rcvd = SnmpGet ( aSess, aOid, aOidLen, ASN_COUNTER, "a counter32" );
    ret = (uint) *(rcvd->variables->val.integer);
    snmp_free_pdu ( rcvd );
    return ( ret );
}

uint ReadGauge32 ( struct snmp_session* aSess, oid aOid[], size_t aOidLen )
{
    struct snmp_pdu* rcvd;
    uint ret;

    rcvd = SnmpGet ( aSess, aOid, aOidLen, ASN_GAUGE, "a gauge32" );
    ret = (uint) *(rcvd->variables->val.integer);
    snmp_free_pdu ( rcvd );
    return ( ret );
}

/* Returns a newly allocated string. Free it! */
char* VarListStrDup ( netsnmp_variable_list* aVar )
{
    char* ret = NULL;

    if ( aVar->type == ASN_OCTET_STR )
    {
        ret = malloc ( aVar->val_len + 1 );
        if ( ret == NULL ) ErrorMalloc ();
        strncpy ( ret, (char*) (aVar->val.string), aVar->val_len );
        ret [aVar->val_len] = '\0';
    }
    else
    {
        Log ( LOG_CRIT,
                "VarListStrDup() called with aVar->type != ASN_OCTET_STR" );
    }
    return ret;
}

/* Returns a newly allocated string. Free it! */
char* ReadString ( struct snmp_session* aSess, oid aOid[], size_t aOidLen )
{
    struct snmp_pdu* rcvd;
    char* ret;

    rcvd = SnmpGet ( aSess, aOid, aOidLen, ASN_OCTET_STR, "a string" );
    ret = VarListStrDup ( rcvd->variables );
    snmp_free_pdu ( rcvd );
    return ( ret );
}

