/*
 * File:      strutils.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include "globals.h"
#include "strutils.h"
#include "syslogif.h"


char* StrDup ( const char* aStr )
{
    char* ret = strdup ( aStr );

    if ( NULL == ret ) ErrorMalloc ();
    return ( ret );
}

char* IntToStr ( long aInt )
{
    char buffer[32]; /* an 64(!) bit integer can have a sign plus 19 digits */
    int len = snprintf ( buffer, 32, "%li", aInt );

    if ( 32 <= len )
    {
        Error ( LOG_ERR, "Got a really big integer here ... too big." );
    }
    /* TODOLO: Okay, I could do this inline ... */
    return ( StrDup ( buffer ) );
}

char* UIntToStr ( unsigned long aULong )
{
    char buffer[32]; /* an 64(!) bit integer can have a sign plus 19 digits */
    int len = snprintf ( buffer, 32, "%lu", aULong );

    if ( 32 <= len )
    {
        Error ( LOG_ERR, "Got a really big unsigned here ... too big." );
    }
    /* TODOLO: Okay, I could do this inline ... */
    return ( StrDup ( buffer ) );
}

char* DoubleToStr ( double aDouble )
{
    char buffer[64]; /* no way to really tell, is there? */
    int len = snprintf ( buffer, 64, "%lf", aDouble );

    if ( 64 <= len )
    {
        Error ( LOG_ERR, "Got a really big double here ... too big." );
    }
    /* TODOLO: Okay, I could do this inline ... */
    return ( StrDup ( buffer ) );
}

char* StrConcat ( uint aCnt, ... )
{
    size_t* lens = malloc ( sizeof ( size_t ) * aCnt );
    size_t len = 1;
    int ofs = 0;
    int i;
    char* ret;
    va_list str;

    /* first loop: get lengths */
    if ( NULL == lens ) ErrorMalloc ();
    va_start ( str, aCnt );
    for ( i = 0; i < aCnt; i++ )
    {
        lens [i] = strlen ( va_arg ( str, char* ) );
        len += lens [i];
    }
    va_end ( str );
    /* second loop: allocate memory and fill it */
    ret = malloc ( len );
    va_start ( str, aCnt );
    if ( NULL == ret ) ErrorMalloc ();
    for ( i = 0; i < aCnt; i++ )
    {
        strncpy ( &(ret[ofs]), va_arg ( str, char* ), lens [i] );
        ofs += lens [i];
    }
    ret[ofs] = '\000';
    va_end ( str );
    free ( lens );
    return ( ret );
}

int CheckIdent ( const char* aIdent )
{
    int ret = TRUE;
    int i = 0;

    while ( ( aIdent [i] != '\0' ) && ( ret ) )
    {
        int ucl = FALSE;
        int lcl = FALSE;
        int dig = FALSE;
        int usc = FALSE;

        if ( ( 'A' <= aIdent [i] ) && ( aIdent [i] <= 'Z' ) ) ucl = TRUE;
        if ( ( 'a' <= aIdent [i] ) && ( aIdent [i] <= 'z' ) ) lcl = TRUE;
        if ( ( '0' <= aIdent [i] ) && ( aIdent [i] <= '9' ) ) dig = TRUE;
        if ( '_' == aIdent [i] ) usc = TRUE;
        if ( !( ucl || lcl || dig || usc ) ) ret = FALSE;
        i++;
    }
    if ( i == 0 ) ret = FALSE;
    return ( ret );
}

char* Basename ( const char *aFilename )
{
    char* ret = NULL;

    if ( NULL == aFilename ) ret = StrDup ( "" );
    else
    {
        int len = strlen ( aFilename );

        if ( 0 == len ) ret = StrDup ( "" );
        else
        {
            int onlySlashes = TRUE; /* so far only slashes were found */
            int slash = TRUE; /* last char was a slash / at begin of string */
            int i;
            int p1 = 0; /* ~ begin of last component */
            int p2 = -1; /* ~ end of last component */
            int dLen; /* position of last slash sequence */

            for ( i = 0; i < len; i++ )
            {
                if ( '/' == aFilename [i] )
                {
                    if ( !slash )
                    {
                        slash = TRUE;
                        p2 = i;
                    } /* no else needed */
                }
                else
                {
                    if ( slash )
                    {
                        onlySlashes = FALSE;
                        slash = FALSE;
                        p1 = i;
                    } /* no else needed */
                }
            }
            if ( p2 < p1 )
            {
                if ( onlySlashes ) ret = StrDup ( "" );
                else
                {
                    p2 = len - 1;
                    dLen = p2 - p1 + 1;
                    ret = malloc ( dLen + 1 );
                    strncpy ( ret, &(aFilename [p1]), dLen );
                    ret [dLen] = '\000';
                }
            }
            else
            {
                ret = StrDup ( "" );
            }
        }
    }
    return ( ret );
}
