/*
 * File:      utilities.c
 * Copyright: (c) 2006 by Peter Gritsch
 * Email:     s4nag@no-where.at
 *
 * This file is part of SNMP4Nagios.
 *
 * SNMP4Nagios is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * SNMP4Nagios is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SNMP4Nagios; if not, write to the
 *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "globals.h"
#include "nagiosif.h"
#include "snmpif.h"
#include "strutils.h"


void HelpCopyright ()
{
    printf ( "%s %s (SNMP4Nagios %s)\n", progName, progVersion, VERSION );
    printf (
    "\n"
    "Copyright (c) 2006 Peter Gritsch\n"
    "SNMP4Nagios is free software; you can redistribute it and/or modify\n"
    "it under the terms of the GNU General Public License. See the file\n"
    "COPYING for more details.\n"
    "\n"
    );
}

void HelpUsage ()
{
    if ( helpIndexP == NULL )
            printf ( "Usage: %s\n", progName );
    else    printf ( "Usage: %s %s\n", progName, helpIndexP );
    if ( helpThresP == NULL ) printf ( "       [-S] [-T]" );
    else                      printf ( "       [-S] [-T %s ]", helpThresP );
#if(WITH_RRD)
    if ( supportsPerfLog ) printf ( " [-L] [-P]" );
#endif /* WITH_RRD */
    printf ( " [-V] [-h]\n" );
    printf ( "       [-H HOST] [-A ADDRESS] [-C COMMUNITY|-F COMMFILE]\n" );
    printf ( "       [-v] [-1|-2] [-r RETRIES] [-t TIMEOUT ]\n" );
#if(WITH_RRD)
    if ( supportsPerfLog ) printf ( "       [-R RRDDIR] [-G GRAPHDIR]\n" );
#endif /* WITH_RRD */
}

void HelpModesStd ()
{
    printf (
    "Modes:\n"
    "\n"
    "-S, --scan\n"
    "    Scan the host for services. This disables all other modes.\n"
    "\n"
    "-T, --test\n"
    "    Check the service.\n"
    "\n"
    );
}

#if(WITH_RRD)
void HelpModesPerflog ()
{
    printf (
    "-L, --log\n"
    "    Log performance data. If -T is given, then the data is passed\n"
    "    internally from the check otherwise it is read from the\n"
    "    environment.\n"
    "\n"
    "-P, --plot\n"
    "    Create plots.\n"
    "\n"
    );
}
#endif /* WITH_RRD */

void HelpModesOther ()
{
    printf (
    "-V, --version\n"
    "    Print version and exit.\n"
    "\n"
    "-h, --help\n"
    "    This help.\n"
    "\n"
    );
}

void HelpGeneral ()
{
    printf (
    "General options:\n"
    "\n"
    "-H, --host HOSTNAME\n"
    "    Name of the host which should be be checked. This defaults to the\n"
    "    contents of the environment variable $NAGIOS_HOSTNAME and is\n"
    "    needed for logging and plotting.\n"
    "\n"
    "-A, --address BYTE.BYTE.BYTE.BYTE\n"
    "    IP Address of the host which should be checkded. This defaults to\n"
    "    the contents of the environment variable $NAGIOS_HOSTADDRESS and\n"
    "    is preferred for setting up SNMP connections.\n"
    "\n"
    );
    printf (
    "-v, --verbose\n"
    "    Increase verbosity. This can be given several times but there may\n"
    "    simply be nothing more to say. Note also that messages are written\n"
    "    to syslog and that you might need to use three '-v's to see\n"
    "    anything.\n"
    "\n"
    );
    printf (
    "-N, --nagios-plugins-format\n"
    "    Use the alternative 'Nagios Plugins' performance data format.\n"
    "\n"
    );
}

void HelpSNMP ()
{
    printf (
    "SNMP Options:\n"
    "\n"
    "-1, --snmp-v1\n"
    "    Use the SNMP protocol version 1.\n"
    "\n"
    "-2, --snmp-v2c\n"
    "    Use the SNMP protocol version 2c (default).\n"
    "\n"
    "-C, --community STRING\n"
    "    Use the community STRING.\n"
    "\n"
    "-F, --community-file FILENAME\n"
    "    Read the SNMP community string from FILENAME.\n"
    "\n"
    "-r, --retries INTEGER\n"
    "    Number of times to retry a request. Note that net-snmp uses\n"
    "    increasing timeouts for retries.\n"
    "\n"
    "-t, --snmp-timeout INTEGER\n"
    "    Timeout for the first SNMP request in milliseconds.\n"
    "\n"
    );
    printf (
    "If a community string is provided (-C) then it is used. Else if a\n"
    "community file is provided, then the community is read from it.\n"
    "Per default the program tries to read the community string from the\n"
    "files $HOME/.snmp4nagios/community, <home>/.snmp4nagios/community\n"
    "(where <home> is read from /etc/passwd) and /etc/snmp4nagios/community.\n"
    "The first file that can be read is used. If no community is given, then\n"
    "the program will use the community 'public'.\n"
    "\n"
    );
}

#if(WITH_RRD)
void HelpOptionsLogPlot ()
{
    printf (
    "Logging and plotting options:\n"
    "\n"
    "-R, --rrd-directory DIRNAME\n"
    "    RRD archives' base directory. The filename of the archive is build\n"
    "    by appending the hostname, the check's name and -- depending on the\n"
    "    service -- other information to DIRNAME. If this option is not\n"
    "    given, then the compiled in default is used.\n"
    "\n"
    );
    printf (
    "-G, --graph-directory DIRNAME\n"
    "    PNG plots' base directory. The filename of the plots is build by\n"
    "    appending the hostname, the check's name and -- depending on the\n"
    "    service -- other information to DIRNAME. If this option is not\n"
    "    given, then the compiled in default is used.\n"
    "\n"
    );
    if ( hasAddInfParam )
    {
        printf (
        "-I, --additional-info ADDINF\n"
        "    Additional info used to find the RRD archive and image pathes.\n"
        "\n"
        );
    }
}
#endif /* WITH_RRD */

void Version ()
{
    HelpCopyright ();
    exit ( STATE_OK );
}

void Usage ( char* aMsg )
{
    printf ( "Usage error: %s\n", aMsg );
    HelpUsage ();
    exit ( STATE_UNKNOWN );
}

void Help ()
{
    HelpCopyright ();
    printf ( "%s\n", helpPurpose );
    printf ( "\n" );
    HelpUsage ();
    printf ( "\n" );
    HelpModesStd ();
#if(WITH_RRD)
    if ( supportsPerfLog ) HelpModesPerflog ();
#endif /* WITH_RRD */
    HelpModesOther ();
    HelpGeneral ();
    HelpSNMP ();
#if(WITH_RRD)
    if ( supportsPerfLog ) HelpOptionsLogPlot ();
#endif /* WITH_RRD */
    printf ( "Plugin-specific parameters:\n" );
    printf ( "\n" );
    HelpSpecificParams ();
    HelpExplain ();
    if ( supportsPerfLog ) HelpPerformanceData ();
    exit ( STATE_OK );
}

int ChkOptModesStd ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-S" )
            ||  !strcmp ( argv [(*i)], "--scan" ) )
    {
        scanServices = TRUE;
        (*i)++;
    } else if ( !strcmp ( argv [(*i)], "-T" )
            ||  !strcmp ( argv [(*i)], "--test" ) )
    {
        checkData = TRUE;
        (*i)++;
    } else ret = FALSE;
    return ( ret );
}

#if(WITH_RRD)
int ChkOptModesPerflog ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-L" )
            ||  !strcmp ( argv [(*i)], "--log" ) )
    {
        logPerfData = TRUE;
        (*i)++;
    } else if ( !strcmp ( argv [(*i)], "-P" )
            ||  !strcmp ( argv [(*i)], "--plot" ) )
    {
        plotPerfData = TRUE;
        (*i)++;
    } else ret = FALSE;
    return ( ret );
}
#endif /* WITH_RRD */

int ChkOptModesOther ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-V" )
            ||  !strcmp ( argv [(*i)], "--version" ) )
    {
        Version ();
    } else if ( !strcmp ( argv [(*i)], "-h" )
            ||  !strcmp ( argv [(*i)], "--help" ) )
    {
        Help ();
    } else ret = FALSE;
    return ( ret );
}

int ChkOptGeneral ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-H" )
            ||  !strcmp ( argv [(*i)], "--host" ) )
    {
        isHostNameSet = TRUE;
        (*i)++;
        if ( (*i) < argc ) hostName = StrDup ( argv [(*i)++] );
        else Usage ( "Missing 'host' parameter" );
    } else if ( !strcmp ( argv [(*i)], "-A" )
            ||  !strcmp ( argv [(*i)], "--address" ) )
    {
        isHostAddrSet = TRUE;
        (*i)++;
        if ( (*i) < argc ) hostAddr = StrDup ( argv [(*i)++] );
        else Usage ( "Missing 'address' parameter" );
    } else if ( !strcmp ( argv [(*i)], "-v" )
            ||  !strcmp ( argv [(*i)], "--verbose" ) )
    {
        switch ( loggingLevel )
        {
            case LOG_EMERG:
            case LOG_ALERT:
            case LOG_CRIT:
            case LOG_ERR:
                loggingLevel = LOG_WARNING;
            break;
            case LOG_WARNING:
                loggingLevel = LOG_NOTICE;
            break;
            case LOG_NOTICE:
                loggingLevel = LOG_DEBUG;
            break;
            case LOG_INFO:
            case LOG_DEBUG:
            break;
            default:
                loggingLevel = LOG_ERR;
            break;
        }
        (*i)++;
    } else if ( !strcmp ( argv [(*i)], "-N" )
            ||  !strcmp ( argv [(*i)], "--nagios-plugins-format" ) )
    {
        npPerfData = TRUE;
        (*i)++;
    } else ret = FALSE;
    return ( ret );
}

int ChkOptSNMP ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-1" )
            ||  !strcmp ( argv [(*i)], "--snmp-v1" ) )
    {
        snmpVersion = SNMP_VERSION_1;
        (*i)++;
    } else if ( !strcmp ( argv [(*i)], "-2" )
            ||  !strcmp ( argv [(*i)], "--snmp-v2c" ) )
    {
        snmpVersion = SNMP_VERSION_2c;
        (*i)++;
    } else if ( !strcmp ( argv [(*i)], "-C" )
            ||  !strcmp ( argv [(*i)], "--community" ) )
    {
        (*i)++;
        if ( (*i) < argc ) community = argv [(*i)++];
        else Usage ( "Missing 'community' parameter" );
    } else if ( !strcmp ( argv [(*i)], "-F" )
            ||  !strcmp ( argv [(*i)], "--community-file" ) )
    {
        (*i)++;
        if ( (*i) < argc ) commFile = argv [(*i)++];
        else Usage ( "Missing 'community file' parameter" );
    } else if ( !strcmp ( argv [(*i)], "-r" )
            ||  !strcmp ( argv [(*i)], "--retries" ) )
    {
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            snmpRetries = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || *tail )
                    Usage ( "Invalid 'retries' parameter" );
        }
        else Usage ( "Missing 'retries' parameter" );
    } else if ( !strcmp ( argv [(*i)], "-t" )
            ||  !strcmp ( argv [(*i)], "--snmp-timeout" ) )
    {
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];
            errno = 0;
            snmpTimeOut = 1000 * strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || *tail )
                    Usage ( "Invalid 'SNMP timeout' parameter" );
        }
        else Usage ( "Missing 'SNMP timeout' parameter" );
    } else ret = FALSE;
    return ( ret );
}

#if(WITH_RRD)
int ChkOptLogPlot ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-R" )
            ||  !strcmp ( argv [(*i)], "--rrd-directory" ) )
    {
        (*i)++;
        if ( (*i) < argc )
        {
            rrdDir = argv [(*i)++];
        } else
        {
            Usage ( "Missing 'rrd-directory' parameter" );
        }
    } else if ( !strcmp ( argv [(*i)], "-G" )
            ||  !strcmp ( argv [(*i)], "--graph-directory" ) )
    {
        (*i)++;
        if ( (*i) < argc )
        {
            pngDir = argv [(*i)++];
        } else
        {
            Usage ( "Missing 'graph-directory' parameter" );
        }
    } else ret = FALSE;
    return ( ret );
}
#endif /* WITH_RRD */

int ChkOptAddInf ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-I" )
            ||  !strcmp ( argv [(*i)], "--additional-info" ) )
    {
        (*i)++;
        if ( (*i) < argc )
        {
            addInf = argv [(*i)++];
            isAddInfSet = TRUE;
        } else
        {
            Usage ( "Missing 'graph-directory' parameter" );
        }
    } else ret = FALSE;
    return ( ret );
}

int ChkOptIndex ( int argc, char** argv, int* i )
{
    int ret = TRUE;

    if (        !strcmp ( argv [(*i)], "-i" )
            ||  !strcmp ( argv [(*i)], "--index" ) )
    {
        isIdxSet = TRUE;
        (*i)++;
        if ( (*i) < argc )
        {
            char *tail = argv [(*i)];

            errno = 0;
            idx = strtoul ( argv [(*i)++], &tail, 0 );
            if ( errno || *tail ) Usage ( "Invalid 'index' parameter" );
        }
        else Usage ( "Missing 'index' parameter" );
    } else ret = FALSE;
    return ( ret );
}

void ReadArgs ( int argc, char** argv )
{
    int i = 0;

    ReadHostName ();
    ReadHostAddr ();
    progName = Basename ( argv [i++] );
    while ( i < argc )
    {
        /*
         * Note the semicolons at the end of the lines. IOW: Do nothing if the
         * function returns TRUE.
         */
        if ( ChkOptModesStd ( argc, argv, &i ) );
#if(WITH_RRD)
        else if ( supportsPerfLog && ChkOptModesPerflog ( argc, argv, &i ) );
#endif /* WITH_RRD */
        else if ( ChkOptModesOther ( argc, argv, &i ) );
        else if ( ChkOptGeneral ( argc, argv, &i ) );
        else if ( ChkOptSNMP ( argc, argv, &i ) );
#if(WITH_RRD)
        else if ( supportsPerfLog && ChkOptLogPlot ( argc, argv, &i ) );
#endif /* WITH_RRD */
        else if ( hasAddInfParam && ChkOptAddInf ( argc, argv, &i ) );
        else if ( hasIndexParam && ChkOptIndex ( argc, argv, &i ) );
        else if ( ChkOptSpecificParams ( argc, argv, &i ) );
        else
        {
            char msg[512] = "Unknown parameter: ";
            strncat ( msg, argv [i], 256 );
            Usage ( msg );
        }
    }
    if ( checkData && hasIndexParam )
    {
        if ( !isIdxSet ) Usage ( "Missing 'index' parameters" );
    }
    if ( ( checkData
#if(WITH_RRD)
            || logPerfData || plotPerfData
#endif /* WITH_RRD */
            ) && hasAddInfParam )
    {
        if ( !isAddInfSet ) Usage ( "Missing 'addInf' parameters" );
    }
#if(WITH_RRD)
    if ( logPerfData || plotPerfData )
    {
        if ( !isHostNameSet ) Usage ( "'Missing 'host name' parameters" );
    }
#endif /* WITH_RRD */
    if ( !( isHostNameSet || isHostAddrSet ) )
    {
        Usage ( "Missing both 'host name' and 'host address' parameters" );
    }
}
